"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BarcodeConfigurator = void 0;
const typesLib = require("@sage/xtrem-decimal");
const xtrem_client_1 = require("@sage/xtrem-client");
const ui = require("@sage/xtrem-ui");
const barcode_parser_configuration_1 = require("../shared-functions/barcode-parser-configuration");
const screen_management_1 = require("./screen-management");
class BarcodeConfigurator {
    constructor() {
        this._barcodeConfiguration = undefined;
    }
    /**
     *  Initialize the barcode configuration
     *
     * Configurations are returned only if they are all defined:
     * - Barcode configuration must be read and have lines.
     * - Parser configuration must be able to be generated.
     * - Field configuration is only generated if all provided fields were found on the screen.
     * @param pageInstance
     * @param codeFlow
     * @returns instance of barcode configuration or undefined
     */
    async initConfigurator(pageInstance, codeFlow) {
        const _barcodeConfiguration = await this._getBarcodeConfiguration(pageInstance, codeFlow);
        const _parserConfiguration = _barcodeConfiguration
            ? this._computeParserConfiguration(_barcodeConfiguration)
            : undefined;
        const _barcodeManagementConfiguration = _barcodeConfiguration && _parserConfiguration
            ? this._computeManagementConfiguration(pageInstance, _barcodeConfiguration, _parserConfiguration)
            : undefined;
        if (!_barcodeConfiguration || !_parserConfiguration || !_barcodeManagementConfiguration) {
            this._parserConfiguration = undefined;
            this._barcodeConfiguration = undefined;
            this._barcodeManagementConfiguration = undefined;
            return undefined;
        }
        this._barcodeConfiguration = _barcodeConfiguration;
        this._parserConfiguration = _parserConfiguration;
        this._barcodeManagementConfiguration = _barcodeManagementConfiguration;
        ui.console.debug(`The barcode configuration ${_barcodeConfiguration.compositeCode ?? ''} of type ${_barcodeManagementConfiguration.typeOfServiceBarcode} has been loaded with ${_barcodeConfiguration.lines.length} parameter(s).`);
        return _barcodeConfiguration;
    }
    /**
     *  Compute the parser configuration :
     *  No configuration is returned if no rows could be retrieved.
     * @param _barcodeConfiguration
     *  @returns parser configuration or undefined
     */
    _computeParserConfiguration(_barcodeConfiguration) {
        const _lines = [];
        const isFixedLength = typesLib.strictEq(_barcodeConfiguration.compositeFormat, barcode_parser_configuration_1.CompositeFormat.fixedLength);
        const _dateSize = (0, barcode_parser_configuration_1.getDateFormatSize)(_barcodeConfiguration.dateFormat);
        const _decimalSeparator = _barcodeConfiguration?.decimalSeparator;
        const _fieldSeparator = _barcodeConfiguration?.fieldSeparator;
        let _recordSize = 0;
        let _numberOfAlias = 0;
        if (!_dateSize) {
            ui.console.error('Date format is not defined');
            return undefined;
        }
        if (!_decimalSeparator) {
            ui.console.error('Decimal separator is not defined');
            return undefined;
        }
        if (!isFixedLength) {
            if (!_fieldSeparator) {
                ui.console.error('Field separator is not defined');
                return undefined;
            }
            if (typesLib.strictEq(_fieldSeparator, _decimalSeparator)) {
                ui.console.error('Field separator is the same as the decimal separator');
                return undefined;
            }
        }
        for (const _line of _barcodeConfiguration.lines) {
            if (typesLib.strictEq(_line.fieldDataType, 'date') && isFixedLength && typesLib.gt(_dateSize, _line.length)) {
                ui.console.error(`Date field is too short for field : ${_line.fieldName}`);
                return undefined;
            }
            if (isFixedLength) {
                _recordSize = typesLib.add(_recordSize, _line.length);
            }
            if (_line.fieldAlias) {
                var _v;
                _v = _numberOfAlias, _numberOfAlias = typesLib.add(_numberOfAlias, 1), _v;
            }
            _lines.push({
                fieldKey: _line.fieldKey,
                description: _line.fieldName,
                ...(isFixedLength && { startPosition: _line.startPosition }),
                ...(isFixedLength && { length: _line.length }),
                fieldAlias: _line?.fieldAlias,
                fieldDataType: _line.fieldDataType,
            });
        }
        if (!isFixedLength && _lines.length) {
            _recordSize = typesLib.add(_recordSize, typesLib.mul((typesLib.sub(_lines.length, 1)), _fieldSeparator.length));
        }
        return _lines.length
            ? {
                typeOfBarcodeParser: 'composite',
                compositeCode: _barcodeConfiguration.compositeCode,
                localizedDescription: _barcodeConfiguration.localizedDescription,
                isBlankDeletion: _barcodeConfiguration.isBlankDeletion,
                compositeFormat: _barcodeConfiguration.compositeFormat,
                dateFormat: _barcodeConfiguration.dateFormat,
                dateLabel: _barcodeConfiguration.dateLabel,
                ...(!isFixedLength && { fieldSeparator: _fieldSeparator }),
                decimalSeparator: _decimalSeparator,
                recordSize: _recordSize,
                numberOfAlias: _numberOfAlias,
                lines: _lines,
            }
            : undefined;
    }
    /**
     *  Compute the screen management configuration
     * @param _barcodeConfiguration
     */
    _computeManagementConfiguration(pageInstance, _barcodeConfiguration, _parserConfiguration) {
        const _dictionary = {};
        let numberOfAlias = 0;
        for (const _line of _barcodeConfiguration.lines) {
            const _fieldName = _line.fieldName;
            if (_line.fieldKey && _fieldName) {
                const _fieldAlias = _line.fieldAlias;
                const _fieldDataType = _line.fieldDataType;
                // We start with the main field
                const _field = pageInstance[_fieldName];
                if (!_field?.id) {
                    ui.console.error(`Field ${_fieldName} is not defined in the page`);
                    continue;
                }
                if (!(0, screen_management_1.isSupportedField)(_field)) {
                    ui.console.error(`Field ${_fieldName} is not supported by the screen management`);
                    continue;
                }
                const _fieldRightType = this._getSupportedFieldDateType(_field);
                // Check if the field type is known directly
                if (_fieldRightType && typesLib.strictNe(_fieldDataType, _fieldRightType)) {
                    ui.console.error(`The type ${_fieldDataType} of the field ${_fieldName} does not match that of the page ${_fieldRightType}.`);
                    return undefined;
                }
                _dictionary[_line.fieldKey] = {
                    mainField: _field,
                };
                // If an alias is defined, we also check if it is supported
                if (_fieldAlias) {
                    var _v;
                    const _field = pageInstance[_fieldAlias];
                    if (!_field?.id) {
                        ui.console.error(`Field alias ${_fieldAlias} of ${_fieldName} is not defined in the page`);
                        continue;
                    }
                    if (!(0, screen_management_1.isSupportedField)(_field)) {
                        ui.console.error(`Field alias ${_fieldAlias} of ${_fieldName} is not supported by the screen management`);
                        continue;
                    }
                    const _fieldRightType = this._getSupportedFieldDateType(_field);
                    // Check if the field type is known directly
                    if (_fieldRightType && typesLib.strictNe(_fieldDataType, _fieldRightType)) {
                        ui.console.error(`The type ${_fieldDataType} of the field alias ${_fieldAlias} of ${_fieldName}  does not match that of the page ${_fieldRightType}.`);
                        return undefined;
                    }
                    _dictionary[_fieldAlias] = {
                        mainField: _field,
                    };
                    _v = numberOfAlias, numberOfAlias = typesLib.add(numberOfAlias, 1), _v;
                }
            }
        }
        const _isSameSize = typesLib.strictEq(Object.keys(_dictionary).length, typesLib.add(_barcodeConfiguration.lines.length, numberOfAlias));
        if (!_isSameSize) {
            ui.console.error('At least one field could not be handled by the screen management, the configuration is rejected.');
        }
        return _isSameSize
            ? {
                typeOfServiceBarcode: _barcodeConfiguration.typeOfServiceBarcode,
                recordSize: _parserConfiguration.recordSize,
                numberOfAlias: numberOfAlias,
                dictionaryFieldSupported: _dictionary,
                parserConfiguration: _parserConfiguration,
            }
            : undefined;
    }
    /**
     * Get the barcode configuration
     * @returns current instance of barcode configuration or undefined
     */
    get barcodeConfiguration() {
        return this._barcodeConfiguration;
    }
    /**
     * Get the barcode management configuration
     * @returns current instance of barcode management configuration or undefined
     */
    get barcodeManagementConfiguration() {
        return this._barcodeManagementConfiguration;
    }
    /**
     * Get the dictionary field supported
     * @returns current instance of dictionary field supported or undefined
     */
    get dictionaryFieldSupported() {
        return this._barcodeManagementConfiguration?.dictionaryFieldSupported;
    }
    /**
     * Get the parser configuration
     * @returns current instance of parser configuration or undefined
     */
    get parserConfiguration() {
        return this._parserConfiguration;
    }
    /**
     * Get the type of service barcode
     * @returns current instance of type of service barcode or undefined
     */
    get typeOfServiceBarcode() {
        return this._barcodeManagementConfiguration?.typeOfServiceBarcode;
    }
    /**
     *  Get the barcode configuration for the given code flow for fabricator
     * @param pageInstance
     *  @param codeFlow
     *  @returns
     */
    async _getBarcodeConfiguration(pageInstance, codeFlow) {
        const _barcodeSetup = await this._getBarcodeSetup(pageInstance, codeFlow);
        if (_barcodeSetup) {
            const _lines = [];
            const _decimalSeparator = _barcodeSetup.decimalSeparator?.trim() ?? '.';
            // Field name must be alphanumeric only (including underscore).
            const _regExAlphaNumeric = /^_*[A-Za-z]+\w*$/;
            for (const _line of _barcodeSetup.lines) {
                // Remove all spaces in the field name
                const _fieldName = _line.fieldName?.trim().replace(/ /g, '');
                const _aliasName = _line?.fieldAlias?.trim().replace(/ /g, '');
                const _fieldAlias = _aliasName && typesLib.strictNe(_fieldName, _aliasName) ? _aliasName : undefined;
                const _fieldKey = _fieldName;
                if (_fieldKey && _fieldName) {
                    if (!_regExAlphaNumeric.test(_fieldName)) {
                        ui.console.error(`Field name is not valid: ${_fieldName}`);
                        return undefined;
                    }
                    if (_fieldAlias && !_regExAlphaNumeric.test(_fieldAlias)) {
                        ui.console.error(`Field key alias for ${_fieldKey} is not valid: ${_fieldAlias}`);
                        return undefined;
                    }
                    _lines.push({
                        fieldKey: _fieldKey,
                        fieldName: _fieldName,
                        fieldAlias: _fieldAlias,
                        description: _line.fieldName?.trim(),
                        startPosition: _line.startPosition,
                        length: _line.length,
                        fieldDataType: this._getSupportedDataType(_line.fieldDataType),
                    });
                }
            }
            // The service type returned for now is a constant, as it is not (yet) provided by reading the configuration
            return _lines.length
                ? {
                    typeOfServiceBarcode: 'composite',
                    compositeCode: _barcodeSetup.compositeCode?.trim(),
                    localizedDescription: _barcodeSetup.localizedDescription?.trim(),
                    isActive: _barcodeSetup.isActive,
                    isBlankDeletion: _barcodeSetup.isBlankDeletion,
                    compositeFormat: this._getCompositeFormat(_barcodeSetup.compositeCodeType),
                    dateFormat: this._getCompositeDateFormat(_barcodeSetup.dateFormat),
                    dateLabel: _barcodeSetup.dateLabel.trim(),
                    fieldSeparator: _barcodeSetup.fieldSeparator.trim(),
                    decimalSeparator: typesLib.strictNe(_decimalSeparator, '') ? _decimalSeparator : '.',
                    lines: _lines,
                }
                : undefined;
        }
        return undefined;
    }
    /**
     *  Get the composite format for the given composite code type
     *  @param compositeCodeType
     *  @returns composite format type key
     */
    _getCompositeFormat(compositeCodeType) {
        switch (compositeCodeType) {
            case 'delimited':
                return barcode_parser_configuration_1.CompositeFormat.delimited;
            case 'fixedLength':
                return barcode_parser_configuration_1.CompositeFormat.fixedLength;
            default:
                return barcode_parser_configuration_1.CompositeFormat.delimited;
        }
    }
    /**
     * Get the current stock site
     *  @param pageInstance Current page instance
     *  @returns the current stock site
     */
    _getCurrentStockSite(pageInstance) {
        return pageInstance.$.storage.get('mobile-selected-stock-site');
    }
    /**
     *  Get the date format for the given date format
     * @param dateFormat
     * @returns date format type key
     */
    _getCompositeDateFormat(dateFormat) {
        switch (dateFormat) {
            case 1:
                return barcode_parser_configuration_1.CompositeDateFormat.ddmmyy;
            case 2:
                return barcode_parser_configuration_1.CompositeDateFormat.ddmmyyyy;
            case 3:
                return barcode_parser_configuration_1.CompositeDateFormat.ddmmyy;
            case 4:
                return barcode_parser_configuration_1.CompositeDateFormat.yyyymmdd;
            case 5:
                return barcode_parser_configuration_1.CompositeDateFormat.yyyy_mm_dd;
            default:
                return barcode_parser_configuration_1.CompositeDateFormat.yyyymmdd;
        }
    }
    /**
     *  Get the data type for the given field data type
     * @param fieldDataType
     * @returns
     */
    _getSupportedDataType(fieldDataType) {
        switch (fieldDataType) {
            case 'alphanumeric':
                return 'string';
            case 'numeric':
                return 'number';
            case 'date':
                return 'DateValue';
            case 'localMenu':
                return 'string';
            default:
                return 'string';
        }
    }
    /**
     *  Get the data type for the given field data type
     * @param fieldDataType
     * @returns data type key or undefined if type is not supported (not possible to check immediately)
     */
    _getSupportedFieldDateType(fieldDataType) {
        if (!(0, screen_management_1.isSupportedField)(fieldDataType) || (0, screen_management_1.isSoundSupportedField)(fieldDataType)) {
            return undefined;
        }
        if (fieldDataType instanceof ui.fields.Text) {
            return 'string';
        }
        if (fieldDataType instanceof ui.fields.Numeric) {
            return 'number';
        }
        if (fieldDataType instanceof ui.fields.Date) {
            return 'DateValue';
        }
        return undefined;
    }
    /**
     *  Get the barcode configuration for the given code flow
     * @param pageInstance
     * @param codeFlow
     * @returns
     */
    async _getBarcodeSetup(pageInstance, codeFlow) {
        try {
            const _codeSite = this._getCurrentStockSite(pageInstance);
            if (!_codeSite) {
                ui.console.error('Stock site is not defined');
                return undefined;
            }
            if (!codeFlow) {
                ui.console.error('Code flow is not defined');
                return undefined;
            }
            return (0, xtrem_client_1.extractEdges)(await pageInstance.$.graph
                .node('@sage/x3-system/CompositeCodeSetup')
                .query(ui.queryUtils.edgesSelector({
                flow: { code: true },
                stockSite: { code: true },
                isActive: true,
                compositeCode: true,
                localizedDescription: true,
                isBlankDeletion: true,
                compositeCodeType: true,
                dateFormat: true,
                dateLabel: true,
                fieldSeparator: true,
                decimalSeparator: true,
                lines: {
                    query: {
                        edges: {
                            node: {
                                lineNumber: true,
                                fieldName: true,
                                fieldAlias: true,
                                startPosition: true,
                                length: true,
                                fieldDataType: true,
                            },
                        },
                        __args: {
                            first: 200,
                        },
                    },
                },
            }, {
                first: 2,
                filter: {
                    flow: { code: codeFlow },
                    stockSite: { code: { _in: [_codeSite, undefined] } },
                    isActive: true,
                },
                orderBy: { stockSite: { code: -1 } },
            }))
                .execute())[0];
        }
        catch (error) {
            ui.console.error(`Error getting barcode configuration: ${error.message}`);
        }
        return undefined;
    }
}
exports.BarcodeConfigurator = BarcodeConfigurator;
//# sourceMappingURL=barcode-configurator.js.map