import { remove, uniq, uniqBy } from 'lodash';
import { MarkerType } from 'reactflow';
import { triggerFieldEvent } from '../../../utils/events';
import { useFieldValue } from '../../../utils/hooks/effects/use-set-field-value';
export const edgeStyle = {
    style: {
        strokeWidth: 1,
        stroke: '#A6A6A6',
    },
    markerEnd: {
        type: MarkerType.Arrow,
        width: 12,
        height: 12,
        color: '#A6A6A6',
    },
};
export const removeTransientNodeDataProperties = (data) => {
    const cleanData = { ...data };
    delete cleanData.type;
    return cleanData;
};
export const removeTransientNodeDataPropertiesFromNodes = (nodes) => nodes.map(n => ({ ...n, selected: false, dragging: false, data: removeTransientNodeDataProperties(n.data) }));
export const removeTransientEdgeDataPropertiesFromEdges = (nodes) => nodes.map(e => {
    const edgeCopy = { ...e };
    delete edgeCopy.markerEnd;
    delete edgeCopy.data;
    delete edgeCopy.style;
    delete edgeCopy.selected;
    return edgeCopy;
});
export const changeEventHandler = (screenId, elementId) => () => triggerFieldEvent(screenId, elementId, 'onChange');
export const usePreviousWorkflowNodes = (nodeId, screenId, elementId, includeCurrent = false) => {
    const fieldValue = useFieldValue(screenId, elementId);
    if (!fieldValue || !nodeId) {
        return [];
    }
    const currentNode = fieldValue.nodes.find(n => n.id === nodeId);
    let currentNodeId = nodeId;
    const previousNodes = includeCurrent && currentNode ? [currentNode] : [];
    // eslint-disable-next-line no-constant-condition
    while (true) {
        currentNodeId = fieldValue.edges.find(e => e.target === currentNodeId)?.source;
        if (!currentNodeId) {
            break;
        }
        const node = fieldValue.nodes.find(n => n.id === currentNodeId);
        if (!node) {
            break;
        }
        // If the node is already in the list, we have a circular reference
        if (previousNodes.find(n => n.id === node.id)) {
            break;
        }
        previousNodes.push(node);
    }
    return previousNodes.reverse();
};
export const useSubsequentWorkflowNodes = (nodeId, screenId, elementId, includeCurrent = false) => {
    const fieldValue = useFieldValue(screenId, elementId);
    if (!fieldValue || !nodeId) {
        return [];
    }
    const currentNode = fieldValue.nodes.find(n => n.id === nodeId);
    let currentNodeId = nodeId;
    const subsequentNodes = includeCurrent && currentNode ? [currentNode] : [];
    // eslint-disable-next-line no-constant-condition
    while (true) {
        currentNodeId = fieldValue.edges.find(e => e.source === currentNodeId)?.target;
        if (!currentNodeId) {
            break;
        }
        const node = fieldValue.nodes.find(n => n.id === currentNodeId);
        if (!node) {
            break;
        }
        // If the node is already in the list, we have a circular reference
        if (subsequentNodes.find(n => n.id === node.id)) {
            break;
        }
        subsequentNodes.push(node);
    }
    return subsequentNodes.reverse();
};
export const useSourceNode = (edgeId, screenId, elementId) => {
    const fieldValue = useFieldValue(screenId, elementId);
    if (!fieldValue) {
        return null;
    }
    const edge = fieldValue?.edges.find(e => e.id === edgeId);
    if (!edge) {
        return null;
    }
    return fieldValue.nodes.find(n => n.id === edge.source) || null;
};
export const useWorkflowNodeVariables = (nodeId, screenId, elementId, includeCurrent = false) => {
    const previousNodes = usePreviousWorkflowNodes(nodeId, screenId, elementId, includeCurrent);
    const collected = { inputVariables: [], oldRootPaths: [] };
    removeTransientNodeDataPropertiesFromNodes(previousNodes).forEach((node) => {
        collected.inputVariables.push(...(node.data.stepVariables ?? []), ...(node.data.outputVariables ?? []));
        if (node.data.oldRootPaths) {
            collected.oldRootPaths.push(...node.data.oldRootPaths);
        }
    });
    return {
        inputVariables: uniqBy(collected.inputVariables, 'path'),
        oldRootPaths: uniq(collected.oldRootPaths),
    };
};
export function removeChildNodesAndEdgesFromStartPoint(nodeId, tempEdges, tempNodes) {
    remove(tempNodes, n => n.id === nodeId);
    const connectingEdges = tempEdges.filter(e => e.source === nodeId);
    // eslint-disable-next-line no-restricted-syntax
    for (const edge of connectingEdges) {
        const targetNode = tempNodes.find(n => n.id === edge.target);
        const edgesToTarget = tempEdges.filter(e => e.target === edge.target);
        if (edgesToTarget.length === 1 && targetNode) {
            removeChildNodesAndEdgesFromStartPoint(edge.target, tempEdges, tempNodes);
        }
        else {
            remove(tempEdges, e => e.id === edge.id);
        }
    }
}
export function createNewNode({ selectedNodeType, values, nodes, position = { x: 20, y: 20 }, }) {
    return {
        id: allocateId(selectedNodeType, nodes),
        position,
        type: selectedNodeType,
        data: {
            type: selectedNodeType,
            ...values,
        },
    };
}
export function createNewEdge({ sourceNodeId, sourceHandleId, edges, targetNodeId, data, }) {
    return {
        id: allocateId(`${sourceNodeId}--${sourceHandleId}`, edges),
        source: sourceNodeId,
        sourceHandle: sourceHandleId,
        target: targetNodeId,
        data,
        ...edgeStyle,
    };
}
export function allocateId(prefix, items) {
    for (let i = 1; i < 1000; i += 1) {
        const id = `${prefix}-${i}`;
        if (!items.find(n => n.id === id))
            return id;
    }
    throw new Error('Could not find a unique id');
}
export function isSingleEdgeRemoval(change, edges) {
    const removalChange = change.find(c => c.type === 'remove');
    if (!removalChange) {
        return false;
    }
    const edgeToRemove = edges.find(e => e.id === removalChange.id);
    const edgesTargetingNode = edgeToRemove ? edges.filter(e => e.target === edgeToRemove.target) : [];
    return edgesTargetingNode.length < 2;
}
export function roundToNearestTwenty(num) {
    return Math.round(num / 20) * 20;
}
//# sourceMappingURL=workflow-component-utils.js.map