"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.defaultElasticMq = exports.requiredQueues = void 0;
exports.mergeRoutings = mergeRoutings;
exports.sortRoutesByTopic = sortRoutesByTopic;
exports.getRoutingData = getRoutingData;
exports.getRoutingFilePath = getRoutingFilePath;
exports.getAllElasticMqConfigs = getAllElasticMqConfigs;
exports.updateElasticMqConfigFile = updateElasticMqConfigFile;
const xtrem_shared_1 = require("@sage/xtrem-shared");
const fs = require("fs");
const _ = require("lodash");
const fsp = require("path");
const logger = require("./logger");
function mergeRoutings(routing1, routing2) {
    return sortRoutesByTopic(_.mergeWith(routing1, routing2, mergeRoutes));
}
function mergeRoutes(routes1, routes2) {
    return _.unionWith(routes1, routes2, _.isEqual);
}
function sortRoutesByTopic(routing) {
    // Sort each route array by topic
    Object.keys(routing).forEach(packageName => {
        routing[packageName].sort((a, b) => a.topic.localeCompare(b.topic));
    });
    return routing;
}
function getRoutingData(routingFile) {
    logger.info(`Getting data from file ${routingFile}`);
    const rawRoutingData = fs.readFileSync(routingFile, 'utf-8');
    return JSON.parse(rawRoutingData);
}
function getRoutingFilePath(packageDir) {
    return fsp.join(packageDir, 'routing.json');
}
function getQueues(routing) {
    const queueList = new Set();
    Object.values(routing).forEach(routes => {
        const queues = _.uniq(routes.map(route => route.queue));
        queues.forEach(queue => queueList.add(queue));
    });
    return [...queueList];
}
const defaultVisibilityTimeout = '30 seconds';
const delay = '0 seconds';
const receiveMessageWait = '0 seconds';
const fifo = 'true';
const deadLetterQueueName = 'dev-dead-letter-queue';
/**
 * Generates the Elastic Queue configuration for a given queue name.
 * @param queueName - The name of the queue.
 * @returns The Elastic Queue configuration.
 */
function elasticQueueConfig(queueName) {
    const deadLetterQueue = queueName === deadLetterQueueName ? undefined : { name: deadLetterQueueName, maxReceiveCount: 5 };
    return {
        [queueName]: {
            defaultVisibilityTimeout,
            delay,
            receiveMessageWait,
            fifo,
            deadLetterQueue,
        },
    };
}
/**
 * Retrieves all ElasticMQ configurations based on the provided application and routing.
 * Optionally, a test package name can be specified to generate unit test queues.
 *
 * @param application - The application object.
 * @param routing - The routing object.
 * @param testPackageName - The name of the optional test package.
 * @returns A dictionary of queue configurations.
 */
function getAllElasticMqConfigs(dir, routing, testPackageName, additionalQueues) {
    const queues = {
        ...getQueues(routing).reduce((r, k) => {
            // Prefix with test-${testPackageName} to generate the unit test queues
            // This name must match what we generate in application.sqsQueueName when application.applicationType is 'test'
            // we cannot use that method here because application.applicationType is 'dev-tool'.
            const queueConfig = elasticQueueConfig((0, xtrem_shared_1.limitSqsQueueName)(testPackageName ? `test-${testPackageName}-${k}` : k));
            return { ...r, ...queueConfig };
        }, {}),
        ...additionalQueues,
    };
    if (!testPackageName) {
        // Add queue variants prefixed by the app name.
        // We have to generate all the queue names regardless of whether the 'app' key is present in the config file.
        let appName = dir.split('/').at(-3);
        // Special hacks to guess the app name from the directory structure.
        if (appName === 'services')
            appName = 'sdmo';
        else if (appName === 'tools')
            appName = dir.split('/').at(-2);
        Object.keys(queues).forEach(queueName => {
            if (!appName)
                throw new xtrem_shared_1.LogicError('Application directory does not match expected format');
            if (!queueName.startsWith(`${appName}--`)) {
                queues[(0, xtrem_shared_1.limitSqsQueueName)(`${appName}--${queueName}`)] = queues[queueName];
            }
        });
    }
    return queues;
}
/**
 * Sorts the ElasticMQ configurations based on their keys in ascending order.
 *
 * @param queues - The dictionary of queues to be sorted.
 * @returns The sorted dictionary of queues.
 */
function sortElasticMqConfigs(queues) {
    const sortedKeys = Object.keys(queues).sort((a, b) => a.localeCompare(b));
    return _.pick(queues, sortedKeys);
}
/**
 * Writes the ElasticMQ configuration to a file.
 *
 * @param filePath - The path of the file to write the configuration to.
 * @param queues - The dictionary of queue configurations.
 */
function writeElasticMqConfigFile(filePath, queues) {
    const include = 'include classpath("application.conf")';
    const queuesStr = JSON.stringify(sortElasticMqConfigs(queues), null, '\t')
        .replace(/: \{/g, ' {')
        .replace(/:/g, ' =')
        .replace(/[",]/g, '');
    const text = `${include}\n\nqueues ${queuesStr}\n`;
    fs.writeFileSync(filePath, text, 'utf-8');
}
/**
 * Reads the ElasticMQ configuration from the specified root path.
 * @param path The path of the configuration file.
 * @returns A dictionary of queue configurations.
 */
function readElasticMqConfigFile(path) {
    if (!fs.existsSync(path))
        return {};
    const text = fs.readFileSync(path, 'utf-8');
    let json = text
        .replace(/^include.*/, '')
        .replace(/\nqueues\s+{/, '{')
        .replace(/(\n\t+)((?:\w|-)+) \{/g, `$1"$2": {`)
        .replace(/(\n\t+)((?:\w|-)+) = (.*)/g, `$1"$2": "$3",`)
        .replace(/([}"])\n/g, `$1,\n`)
        .replace(/,(\n\t*)\}/g, '$1}')
        .trimEnd();
    if (json.slice(-1) === ',')
        json = json.slice(0, -1);
    return JSON.parse(json);
}
/**
 * Updates the ElasticMQ configuration with the provided queues.
 *
 * @param {string} path - The path of the ElasticMq configuration file.
 * @param {Dict<QueueConfig>} queues - The queues to be added or updated in the configuration.
 * @returns {void}
 */
function updateElasticMqConfigFile(path, queues) {
    const oldQueues = readElasticMqConfigFile(path);
    const mergedQueues = _.merge(oldQueues, queues);
    writeElasticMqConfigFile(path, mergedQueues);
}
exports.requiredQueues = ['infrastructure-event', 'intacctReceive'].reduce((acc, queueName) => ({ ...acc, ...elasticQueueConfig(queueName) }), {});
exports.defaultElasticMq = {
    queues: {
        ...elasticQueueConfig(deadLetterQueueName),
        ...elasticQueueConfig('demo-queue'),
        ...exports.requiredQueues,
        ...elasticQueueConfig('unit-test-message-queue'),
        ...elasticQueueConfig('unit-test-notification-queue'),
        ...elasticQueueConfig('unit-test-visibility-multiple-listeners-queue'),
    },
};
//# sourceMappingURL=routing.js.map