import { CompositeCodeSetup, CompositeCodeType, FieldType, GraphApi } from '@sage/x3-system-api';
import { extractEdges, ExtractEdges } from '@sage/xtrem-client';
import * as ui from '@sage/xtrem-ui';
import {
    CompositeDateFormat,
    CompositeFormat,
    getDateFormatSize,
    ParserConfiguration,
    ParserConfigurationLine,
    TypeOfBarcodeParser,
} from '../shared-functions/barcode-parser-configuration';
import { DataTypeSupported } from '../shared-functions/parsed-generic-element';
import {
    BarcodeManagerSupportedFields,
    DictionaryGenericFieldSupported,
    FieldSupported,
    isSoundSupportedField,
    isSupportedField,
} from './screen-management';
import { TypeOfServiceBarcode } from './type-of-barcode-management';

export interface BarcodeConfigurationLine {
    fieldKey: string;
    fieldName: string;
    fieldAlias?: string;
    description?: string;
    startPosition: number;
    length: number;
    fieldDataType: DataTypeSupported;
}

export interface BarcodeConfiguration {
    typeOfServiceBarcode: TypeOfServiceBarcode;
    compositeCode: string;
    localizedDescription: string;
    isActive: boolean;
    isBlankDeletion: boolean;
    compositeFormat: CompositeFormat;
    dateFormat: CompositeDateFormat;
    dateLabel: string;
    fieldSeparator: string;
    decimalSeparator: string;
    lines: BarcodeConfigurationLine[];
}

export interface BarcodeManagementConfiguration {
    typeOfServiceBarcode: TypeOfServiceBarcode;
    recordSize: number;
    numberOfAlias: number;
    parserConfiguration: ParserConfiguration;
    dictionaryFieldSupported: DictionaryGenericFieldSupported;
}

export class BarcodeConfigurator {
    private _barcodeConfiguration?: BarcodeConfiguration;
    private _parserConfiguration?: ParserConfiguration;
    private _barcodeManagementConfiguration?: BarcodeManagementConfiguration;

    constructor() {
        this._barcodeConfiguration = undefined;
    }

    /**
     *  Initialize the barcode configuration
     *
     * Configurations are returned only if they are all defined:
     * - Barcode configuration must be read and have lines.
     * - Parser configuration must be able to be generated.
     * - Field configuration is only generated if all provided fields were found on the screen.
     * @param pageInstance
     * @param codeFlow
     * @returns instance of barcode configuration or undefined
     */
    public async initConfigurator(
        pageInstance: ui.Page<GraphApi>,
        codeFlow: string,
    ): Promise<BarcodeConfiguration | undefined> {
        const _barcodeConfiguration = await this._getBarcodeConfiguration(pageInstance, codeFlow);
        const _parserConfiguration = _barcodeConfiguration
            ? this._computeParserConfiguration(_barcodeConfiguration)
            : undefined;
        const _barcodeManagementConfiguration =
            _barcodeConfiguration && _parserConfiguration
                ? this._computeManagementConfiguration(pageInstance, _barcodeConfiguration, _parserConfiguration)
                : undefined;
        if (!_barcodeConfiguration || !_parserConfiguration || !_barcodeManagementConfiguration) {
            this._parserConfiguration = undefined;
            this._barcodeConfiguration = undefined;
            this._barcodeManagementConfiguration = undefined;
            return undefined;
        }
        this._barcodeConfiguration = _barcodeConfiguration;
        this._parserConfiguration = _parserConfiguration;
        this._barcodeManagementConfiguration = _barcodeManagementConfiguration;

        ui.console.debug(
            `The barcode configuration ${_barcodeConfiguration.compositeCode ?? ''} of type ${_barcodeManagementConfiguration.typeOfServiceBarcode} has been loaded with ${_barcodeConfiguration.lines.length} parameter(s).`,
        );

        return _barcodeConfiguration;
    }

    /**
     *  Compute the parser configuration :
     *  No configuration is returned if no rows could be retrieved.
     * @param _barcodeConfiguration
     *  @returns parser configuration or undefined
     */
    private _computeParserConfiguration(_barcodeConfiguration: BarcodeConfiguration): ParserConfiguration | undefined {
        const _lines = <ParserConfigurationLine[]>[];
        const isFixedLength = _barcodeConfiguration.compositeFormat === CompositeFormat.fixedLength;
        const _dateSize = getDateFormatSize(_barcodeConfiguration.dateFormat);
        const _decimalSeparator = _barcodeConfiguration?.decimalSeparator;
        const _fieldSeparator = _barcodeConfiguration?.fieldSeparator;
        let _recordSize = 0;
        let _numberOfAlias = 0;
        if (!_dateSize) {
            ui.console.error('Date format is not defined');
            return undefined;
        }
        if (!_decimalSeparator) {
            ui.console.error('Decimal separator is not defined');
            return undefined;
        }

        if (!isFixedLength) {
            if (!_fieldSeparator) {
                ui.console.error('Field separator is not defined');
                return undefined;
            }
            if (_fieldSeparator === _decimalSeparator) {
                ui.console.error('Field separator is the same as the decimal separator');
                return undefined;
            }
        }

        for (const _line of _barcodeConfiguration.lines) {
            if (_line.fieldDataType === <DataTypeSupported>'date' && isFixedLength && _dateSize > _line.length) {
                ui.console.error(`Date field is too short for field : ${_line.fieldName}`);
                return undefined;
            }
            if (isFixedLength) {
                _recordSize += _line.length;
            }
            if (_line.fieldAlias) {
                _numberOfAlias++;
            }
            _lines.push(<ParserConfigurationLine>{
                fieldKey: _line.fieldKey,
                description: _line.fieldName,
                ...(isFixedLength && { startPosition: _line.startPosition }),
                ...(isFixedLength && { length: _line.length }),
                fieldAlias: _line?.fieldAlias,
                fieldDataType: _line.fieldDataType,
            });
        }

        if (!isFixedLength && _lines.length) {
            _recordSize += (_lines.length - 1) * _fieldSeparator.length;
        }

        return _lines.length
            ? <ParserConfiguration>{
                  typeOfBarcodeParser: <TypeOfBarcodeParser>'composite',
                  compositeCode: _barcodeConfiguration.compositeCode,
                  localizedDescription: _barcodeConfiguration.localizedDescription,
                  isBlankDeletion: _barcodeConfiguration.isBlankDeletion,
                  compositeFormat: _barcodeConfiguration.compositeFormat,
                  dateFormat: _barcodeConfiguration.dateFormat,
                  dateLabel: _barcodeConfiguration.dateLabel,
                  ...(!isFixedLength && { fieldSeparator: _fieldSeparator }),
                  decimalSeparator: _decimalSeparator,
                  recordSize: _recordSize,
                  numberOfAlias: _numberOfAlias,
                  lines: _lines,
              }
            : undefined;
    }

    /**
     *  Compute the screen management configuration
     * @param _barcodeConfiguration
     */
    private _computeManagementConfiguration(
        pageInstance: ui.Page<GraphApi>,
        _barcodeConfiguration: BarcodeConfiguration,
        _parserConfiguration: ParserConfiguration,
    ): BarcodeManagementConfiguration | undefined {
        const _dictionary = <DictionaryGenericFieldSupported>{};
        let numberOfAlias = 0;
        for (const _line of _barcodeConfiguration.lines) {
            const _fieldName = _line.fieldName;
            if (_line.fieldKey && _fieldName) {
                const _fieldAlias = _line.fieldAlias;
                const _fieldDataType = _line.fieldDataType;
                // We start with the main field
                const _field = (pageInstance as any)[_fieldName];
                if (!_field?.id) {
                    ui.console.error(`Field ${_fieldName} is not defined in the page`);
                    continue;
                }
                if (!isSupportedField(_field)) {
                    ui.console.error(`Field ${_fieldName} is not supported by the screen management`);
                    continue;
                }

                const _fieldRightType = this._getSupportedFieldDateType(_field);
                // Check if the field type is known directly
                if (_fieldRightType && _fieldDataType !== _fieldRightType) {
                    ui.console.error(
                        `The type ${_fieldDataType} of the field ${_fieldName} does not match that of the page ${_fieldRightType}.`,
                    );
                    return undefined;
                }

                _dictionary[_line.fieldKey] = <FieldSupported>{
                    mainField: _field,
                };

                // If an alias is defined, we also check if it is supported
                if (_fieldAlias) {
                    const _field = (pageInstance as any)[_fieldAlias];
                    if (!_field?.id) {
                        ui.console.error(`Field alias ${_fieldAlias} of ${_fieldName} is not defined in the page`);
                        continue;
                    }
                    if (!isSupportedField(_field)) {
                        ui.console.error(
                            `Field alias ${_fieldAlias} of ${_fieldName} is not supported by the screen management`,
                        );
                        continue;
                    }
                    const _fieldRightType = this._getSupportedFieldDateType(_field);
                    // Check if the field type is known directly
                    if (_fieldRightType && _fieldDataType !== _fieldRightType) {
                        ui.console.error(
                            `The type ${_fieldDataType} of the field alias ${_fieldAlias} of ${_fieldName}  does not match that of the page ${_fieldRightType}.`,
                        );
                        return undefined;
                    }

                    _dictionary[_fieldAlias] = <FieldSupported>{
                        mainField: _field,
                    };
                    numberOfAlias++;
                }
            }
        }

        const _isSameSize = Object.keys(_dictionary).length === _barcodeConfiguration.lines.length + numberOfAlias;
        if (!_isSameSize) {
            ui.console.error(
                'At least one field could not be handled by the screen management, the configuration is rejected.',
            );
        }

        return _isSameSize
            ? <BarcodeManagementConfiguration>{
                  typeOfServiceBarcode: _barcodeConfiguration.typeOfServiceBarcode,
                  recordSize: _parserConfiguration.recordSize,
                  numberOfAlias: numberOfAlias,
                  dictionaryFieldSupported: _dictionary,
                  parserConfiguration: _parserConfiguration,
              }
            : undefined;
    }

    /**
     * Get the barcode configuration
     * @returns current instance of barcode configuration or undefined
     */

    public get barcodeConfiguration(): BarcodeConfiguration | undefined {
        return this._barcodeConfiguration;
    }

    /**
     * Get the barcode management configuration
     * @returns current instance of barcode management configuration or undefined
     */
    public get barcodeManagementConfiguration(): BarcodeManagementConfiguration | undefined {
        return this._barcodeManagementConfiguration;
    }

    /**
     * Get the dictionary field supported
     * @returns current instance of dictionary field supported or undefined
     */
    public get dictionaryFieldSupported(): DictionaryGenericFieldSupported | undefined {
        return this._barcodeManagementConfiguration?.dictionaryFieldSupported;
    }

    /**
     * Get the parser configuration
     * @returns current instance of parser configuration or undefined
     */
    public get parserConfiguration(): ParserConfiguration | undefined {
        return this._parserConfiguration;
    }

    /**
     * Get the type of service barcode
     * @returns current instance of type of service barcode or undefined
     */
    public get typeOfServiceBarcode(): TypeOfServiceBarcode | undefined {
        return this._barcodeManagementConfiguration?.typeOfServiceBarcode;
    }

    /**
     *  Get the barcode configuration for the given code flow for fabricator
     * @param pageInstance
     *  @param codeFlow
     *  @returns
     */
    private async _getBarcodeConfiguration(
        pageInstance: ui.Page<GraphApi>,
        codeFlow: string,
    ): Promise<BarcodeConfiguration | undefined> {
        const _barcodeSetup = await this._getBarcodeSetup(pageInstance, codeFlow);
        if (_barcodeSetup) {
            const _lines = <BarcodeConfigurationLine[]>[];
            const _decimalSeparator = _barcodeSetup.decimalSeparator?.trim() ?? '.';
            // Field name must be alphanumeric only (including underscore).
            const _regExAlphaNumeric = /^_*[A-Za-z]+\w*$/;

            for (const _line of _barcodeSetup.lines) {
                // Remove all spaces in the field name
                const _fieldName = _line.fieldName?.trim().replace(/ /g, '');
                const _aliasName = _line?.fieldAlias?.trim().replace(/ /g, '');
                const _fieldAlias = _aliasName && _fieldName !== _aliasName ? _aliasName : undefined;
                const _fieldKey = _fieldName;
                if (_fieldKey && _fieldName) {
                    if (!_regExAlphaNumeric.test(_fieldName)) {
                        ui.console.error(`Field name is not valid: ${_fieldName}`);
                        return undefined;
                    }

                    if (_fieldAlias && !_regExAlphaNumeric.test(_fieldAlias)) {
                        ui.console.error(`Field key alias for ${_fieldKey} is not valid: ${_fieldAlias}`);
                        return undefined;
                    }

                    _lines.push(<BarcodeConfigurationLine>{
                        fieldKey: _fieldKey,
                        fieldName: _fieldName,
                        fieldAlias: _fieldAlias,
                        description: _line.fieldName?.trim(),
                        startPosition: _line.startPosition,
                        length: _line.length,
                        fieldDataType: this._getSupportedDataType(_line.fieldDataType),
                    });
                }
            }
            // The service type returned for now is a constant, as it is not (yet) provided by reading the configuration
            return _lines.length
                ? <BarcodeConfiguration>{
                      typeOfServiceBarcode: <TypeOfServiceBarcode>'composite',
                      compositeCode: _barcodeSetup.compositeCode?.trim(),
                      localizedDescription: _barcodeSetup.localizedDescription?.trim(),
                      isActive: _barcodeSetup.isActive,
                      isBlankDeletion: _barcodeSetup.isBlankDeletion,
                      compositeFormat: this._getCompositeFormat(_barcodeSetup.compositeCodeType),
                      dateFormat: this._getCompositeDateFormat(_barcodeSetup.dateFormat),
                      dateLabel: _barcodeSetup.dateLabel.trim(),
                      fieldSeparator: _barcodeSetup.fieldSeparator.trim(),
                      decimalSeparator: _decimalSeparator !== '' ? _decimalSeparator : '.',
                      lines: _lines,
                  }
                : undefined;
        }
        return undefined;
    }

    /**
     *  Get the composite format for the given composite code type
     *  @param compositeCodeType
     *  @returns composite format type key
     */
    private _getCompositeFormat(compositeCodeType: CompositeCodeType): CompositeFormat {
        switch (<CompositeCodeType>compositeCodeType) {
            case 'delimited':
                return CompositeFormat.delimited;
            case 'fixedLength':
                return CompositeFormat.fixedLength;
            default:
                return CompositeFormat.delimited;
        }
    }

    /**
     * Get the current stock site
     *  @param pageInstance Current page instance
     *  @returns the current stock site
     */
    private _getCurrentStockSite(pageInstance: ui.Page): string | undefined {
        return pageInstance.$.storage.get('mobile-selected-stock-site') as string | undefined;
    }

    /**
     *  Get the date format for the given date format
     * @param dateFormat
     * @returns date format type key
     */
    private _getCompositeDateFormat(dateFormat: number): CompositeDateFormat {
        switch (dateFormat) {
            case 1:
                return CompositeDateFormat.ddmmyy;
            case 2:
                return CompositeDateFormat.ddmmyyyy;
            case 3:
                return CompositeDateFormat.ddmmyy;
            case 4:
                return CompositeDateFormat.yyyymmdd;
            case 5:
                return CompositeDateFormat.yyyy_mm_dd;
            default:
                return CompositeDateFormat.yyyymmdd;
        }
    }

    /**
     *  Get the data type for the given field data type
     * @param fieldDataType
     * @returns
     */
    private _getSupportedDataType(fieldDataType: FieldType): DataTypeSupported {
        switch (fieldDataType) {
            case 'alphanumeric':
                return 'string';
            case 'numeric':
                return 'number';
            case 'date':
                return 'DateValue';
            case 'localMenu':
                return 'string';
            default:
                return 'string';
        }
    }

    /**
     *  Get the data type for the given field data type
     * @param fieldDataType
     * @returns data type key or undefined if type is not supported (not possible to check immediately)
     */
    private _getSupportedFieldDateType(fieldDataType: BarcodeManagerSupportedFields): DataTypeSupported | undefined {
        if (!isSupportedField(fieldDataType) || isSoundSupportedField(fieldDataType)) {
            return undefined;
        }

        if (fieldDataType instanceof ui.fields.Text) {
            return 'string';
        }

        if (fieldDataType instanceof ui.fields.Numeric) {
            return 'number';
        }

        if (fieldDataType instanceof ui.fields.Date) {
            return 'DateValue';
        }

        return undefined;
    }

    /**
     *  Get the barcode configuration for the given code flow
     * @param pageInstance
     * @param codeFlow
     * @returns
     */
    private async _getBarcodeSetup(
        pageInstance: ui.Page<GraphApi>,
        codeFlow: string,
    ): Promise<ExtractEdges<CompositeCodeSetup> | undefined> {
        try {
            const _codeSite = this._getCurrentStockSite(pageInstance);
            if (!_codeSite) {
                ui.console.error('Stock site is not defined');
                return undefined;
            }
            if (!codeFlow) {
                ui.console.error('Code flow is not defined');
                return undefined;
            }

            return extractEdges<any>(
                await pageInstance.$.graph
                    .node('@sage/x3-system/CompositeCodeSetup')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                flow: { code: true },
                                stockSite: { code: true },
                                isActive: true,
                                compositeCode: true,
                                localizedDescription: true,
                                isBlankDeletion: true,
                                compositeCodeType: true,
                                dateFormat: true,
                                dateLabel: true,
                                fieldSeparator: true,
                                decimalSeparator: true,
                                lines: {
                                    query: {
                                        edges: {
                                            node: {
                                                lineNumber: true,
                                                fieldName: true,
                                                fieldAlias: true,
                                                startPosition: true,
                                                length: true,
                                                fieldDataType: true,
                                            },
                                        },
                                        __args: {
                                            first: 200,
                                        },
                                    },
                                },
                            },
                            {
                                first: 2,
                                filter: {
                                    flow: { code: codeFlow },
                                    stockSite: { code: { _in: [_codeSite, undefined] } },
                                    isActive: true,
                                },
                                orderBy: { stockSite: { code: -1 } },
                            },
                        ),
                    )
                    .execute(),
            )[0];
        } catch (error) {
            ui.console.error(`Error getting barcode configuration: ${error.message}`);
        }
        return undefined;
    }
}
