import { Product, ProductInput, ProductSite } from '@sage/x3-master-data-api';
import { UnitOfMeasure } from '@sage/x3-master-data-api-partial';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import { GraphApi, StockChangeInput, StockChangeLineBinding, StockEntryTransaction } from '@sage/x3-stock-api';
import {
    LicensePlateNumber,
    Location,
    Lot,
    LotsSites,
    MobileSettings,
    SerialNumber,
    Stock,
    StockJournalBinding,
    StockManagementRules,
    StockSearchFilter,
    StockStatus,
} from '@sage/x3-stock-data-api';
import { stockControl } from '@sage/x3-stock-data/build/lib/menu-items/stock-control';
import { getCountSerialNumber } from '@sage/x3-stock-data/lib/client-functions/get-count-serial-number';
import { Site } from '@sage/x3-system-api';
import { BarcodeTransactionFlow } from '@sage/x3-system/lib/client-functions/barcode-transaction-flow';
import {
    AsyncCompositeAllowed,
    AsyncVoidFunction,
    DictionaryDataGeneric,
    DictionaryDataGs1,
    DictionaryFieldSupported,
} from '@sage/x3-system/lib/client-functions/screen-management';
import { SupportManagementService } from '@sage/x3-system/lib/client-functions/support-management-service';
import { TypeOfServiceBarcode } from '@sage/x3-system/lib/client-functions/type-of-barcode-management';
import { DataTitle } from '@sage/x3-system/lib/shared-functions/parsed-element';
import { getRegExp } from '@sage/x3-system/lib/shared-functions/pat-converter';
import {
    AggregateGroupSelector,
    AggregateValuesSelector,
    Dict,
    ExtractEdgesPartial,
    aggregateEdgesSelector,
    extractEdges,
    withoutEdges,
} from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import * as ui from '@sage/xtrem-ui';
import { NotifyAndWait } from '../client-functions/display';
import { getUnitNumberOfDecimalList } from '../client-functions/get-unit-number-decimals';
import { managePages } from '../client-functions/manage-pages';
import { findStockManagementRules } from '../client-functions/stock-management-rules';

// Key to use with Composite Data for this application
export const mobileApplicationKey = 'MobileInventoryStockChangeKey';

type DeepPartial<T> = T extends Object ? { [K in keyof T]?: DeepPartial<T[K]> } : T;
type PartialStockEntryTransaction = DeepPartial<StockEntryTransaction>;

export type inputsStockChange = {
    stockChange: StockChangeInput & {
        id?: string;
    };
    username: string;
    currentLine?: number;
    numberSession?: number;
    started: boolean;
    selectedTransaction?: PartialStockEntryTransaction;
    selectedProduct?: ProductInput;
    destination?: string;
    printingMode?: string;
};

export enum PrintingModeEnum {
    noPrint = 1,
    stockLabel = 2,
    substituteValue3 = 3,
    transfertDocument = 4,
    analysisDocument = 5,
    createdContainerLabel = 6,
    stockLabelAndCreatedContainerLabel = 7,
}

interface StockChangeLineWithQuantityMoved extends StockChangeLineBinding {
    quantityMoved: number;
}

interface StockJournalWithQuantityMoved extends StockJournalBinding {
    quantityMoved: number;
}

enum Section {
    main = 'main',
    stockPanel = 'stockPanel',
    detailPanel = 'detailPanel',
}

/**
 * Composite data management for this application.
 * @enum StockChangeDictionaryFieldKey
 *
 * This enum provides per field the associated access key in the data dictionary.
 * It must be maintained in accordance with the fields of this dictionary.
 * Never translate..
 */

enum StockChangeDictionaryFieldKey {
    identifier1Destination = 'identifier1Destination',
    identifier1Filter = 'identifier1Filter',
    identifier2Destination = 'identifier2Destination',
    identifier2Filter = 'identifier2Filter',
    licensePlateNumberFilter = 'licensePlateNumberFilter',
    locationDestination = 'locationDestination',
    locationFilter = 'locationFilter',
    locationSelected = 'locationSelected',
    lotFilter = 'lotFilter',
    mainIdentifier1Destination = 'mainIdentifier1Destination',
    mainIdentifier2Destination = 'mainIdentifier2Destination',
    mainLocationDestination = 'mainLocationDestination',
    mainStatusDestination = 'mainStatusDestination',
    packingUnitDestination = 'packingUnitDestination',
    productSelected = 'productSelected',
    quantityToMove = 'quantityToMove',
    serialNumberFilter = 'serialNumberFilter',
    startingSerialNumber = 'startingSerialNumber',
    statusDestination = 'statusDestination',
    statusFilter = 'statusFilter',
    stockCustomField1Filter = 'stockCustomField1Filter',
    stockCustomField2Filter = 'stockCustomField2Filter',
    sublotFilter = 'sublotFilter',
}

const hideWhenEmptyValue = (value: any, rowValue?: Dict<Stock>) => {
    return typeof value !== 'number' && !value; // we don't want to hide numeric field with value of 0
};

type packingUnit = {
    node: {
        packingUnit: {
            code: string;
            numberOfDecimals: number;
            _id: string;
        };
        packingUnitToStockUnitConversionFactor: string;
        isPackingFactorEntryAllowed: boolean;
        _id: string;
    };
};

@ui.decorators.page<MobileInternalStockChange>({
    title: 'Stock change',
    module: 'x3-stock',
    mode: 'default',
    node: '@sage/x3-stock/StockChange',
    menuItem: stockControl,
    priority: 100,
    isTitleHidden: true,
    authorizationCode: 'CWSSCS',
    access: { node: '@sage/x3-stock/StockChange' },
    skipDirtyCheck: true,
    navigationPanel: undefined,
    async onLoad() {
        this._numberSession = 0;
        if (!(await this._init())) {
            this._disablePage();
        }
    },
    businessActions() {
        return [this.createButton, this.nextButton, this.CancelButton, this.SelectButton];
    },
    headerCard() {
        if (!this.stockPanelSection.isHidden) {
            return {
                title: this.productCode,
                line2: this.productDescription,
            };
        } else {
            return undefined;
        }
    },
})
export class MobileInternalStockChange extends SupportManagementService<GraphApi> {
    private _transactions: PartialStockEntryTransaction[];
    private _transaction: PartialStockEntryTransaction;
    private _notifier = new NotifyAndWait(this);
    private _mobileSettings: MobileSettings;
    private _stockFieldSettings: StockSearchFilter[] = [];
    private _numberSession: number;
    private _productSite: ProductSite;
    /* @internal */
    private _numberOfDecimalList: ExtractEdgesPartial<UnitOfMeasure>[];
    private _packingUnits: packingUnit[];
    private _selectedStockManagementRules: StockManagementRules;
    private _locationProducts: string[];
    private _previousQuantity: string | undefined;
    private _previousSerialNumber: string | undefined;
    private _previousEndingSerialNumber: string | undefined;
    private _licensePlateNumber: ExtractEdgesPartial<LicensePlateNumber> | undefined;

    /*
     *
     *  Technical properties
     *
     */

    @ui.decorators.pageAction<MobileInternalStockChange>({
        title: 'Create',
        buttonType: 'primary',
        isDisabled: true,
        shortcut: ['f2'],
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-stock/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-stock/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);
                message = `**${ui.localize(
                    '@sage/x3-stock/pages__mobile_stock_change_by_identifier__notification__creation_error',
                    'An error has occurred',
                )}**\n\n`;
                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-stock/pages_creation_error_connexion_webservice_contact_administrator',
                    'An error has occurred (connection or webservice error). Please contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
        async onClick() {
            if (this.stockChangeLines.value?.length) {
                this.$.loader.isHidden = false;
                let result: any;
                result = await this.$standardSaveAction.execute(false);
                await this.$.sound.success();
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-stock/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_stock_change__notification__document_created',
                        'Document no. {{documentId}} created.',
                        { documentId: this.$.recordId },
                    ),
                    {
                        fullScreen: true,
                        acceptButton: { text: ui.localize('@sage/x3-stock/button-accept-ok', 'OK') },
                    },
                );
                await this.$.router.emptyPage();
            } else {
                this._notifier.show(
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_stock_change__notification__no_products_error',
                        `Enter at least one product.`,
                    ),
                    'error',
                );
            }
        },
    })
    createButton: ui.PageAction;

    @ui.decorators.pageAction<MobileInternalStockChange>({
        title: 'Next',
        shortcut: ['f3'],
        buttonType: 'primary',
        isHidden: true,
        async onClick() {
            switch (this._getActiveSection()) {
                case Section.main:
                    await this._setActiveSection(Section.stockPanel);
                    break;

                case Section.stockPanel:
                    if (this.stock.selectedRecords.length > 0) {
                        this.stock.unselectAllRecords();
                        await this._getStock();
                        await this.$.commitValueAndPropertyChanges();
                        this.$.setPageClean();
                        await this._setActiveSection(Section.main);
                    } else {
                        await dialogMessage(
                            this,
                            'error',
                            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                            ui.localize(
                                '@sage/x3-stock/pages__mobile_stock_change_lines__notification__no_stock_error',
                                `Select at least one stock line.`,
                            ),
                        );
                    }
                    break;

                default:
                    await this._setActiveSection(Section.main);
                    break;
            }
        },
    })
    nextButton: ui.PageAction;

    @ui.decorators.pageAction<MobileInternalStockChange>({
        title: 'Cancel',
        buttonType: 'secondary',
        isHidden: true,
        async onClick() {
            const currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
            let i: number = 0;
            while (i < this.stockChangeLines.value?.length) {
                if (
                    Number(this.stockChangeLines.value[i].stockId) === Number(currentRecord.stockId) &&
                    Number(this.stockChangeLines.value[i].numberSession) === Number(this._numberSession)
                ) {
                    this.stockChangeLines.removeRecord(this.stockChangeLines.value[i]._id, 0);
                } else {
                    i += 1;
                }
            }
            (currentRecord as any).quantityInPackingUnitDestination = (
                currentRecord as any
            ).quantityInPackingUnitOrigin;
            this.stock.unselectRecord(this.gridBlock.selectedRecordId);
            this.stock.setRecordValue(currentRecord);

            this.productCode.value = this.productSelected.value?.code;
            this.productDescription.value = this.productSelected.value?.localizedDescription1;

            await this._setActiveSection(Section.stockPanel);
        },
    })
    CancelButton: ui.PageAction;

    @ui.decorators.pageAction<MobileInternalStockChange>({
        title: 'Select',
        buttonType: 'primary',
        isHidden: true,
        onError(error) {
            switch (error.message) {
                case '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumberMandatory': {
                    return ui.localize(
                        '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumberMandatory',
                        'You need to select the serial number and add it first.',
                    );
                }
                default: {
                    return error;
                }
            }
        },
        async onClick() {
            await this.$.commitValueAndPropertyChanges();

            const errors: ui.ValidationResult[] = await this.stock.validateWithDetails();
            if (!errors.length) {
                let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
                if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                    if (!this.serialNumberLines.value.length) {
                        throw new Error(
                            '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumberMandatory',
                        );
                    }
                } else {
                    if (Number(this.quantityToMove.value) <= 0) {
                        await dialogMessage(
                            this,
                            'error',
                            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                            ui.localize(
                                '@sage/x3-stock/pages__mobile_stock_change_lines__quantityInPackingUnitDestination_must_be_greater_than_0',
                                'The quantity to move must be greater than 0.',
                            ),
                        );
                        return;
                    }
                }
                if (Number(this.quantityToMove.value) > Number((currentRecord as any).quantityInPackingUnitRest)) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                        `${ui.localize(
                            '@sage/x3-stock/pages__mobile_stock_change_lines__enter_a_quantity_less_than_or_equal_to_the_stock_quantity',
                            'Enter a quantity less than or equal to the stock quantity.',
                        )}`,
                    );
                    return;
                }
                this.stock.selectRecord(this.gridBlock.selectedRecordId);
                if (this.productSelected.value?.serialNumberManagementMode !== 'globalReceivedIssued') {
                    const line = this.stockChangeLines.value.find(
                        line =>
                            Number(line.stockId) === Number(currentRecord.stockId) &&
                            line.numberSession === this._numberSession,
                    );
                    const lineNumber = line ? line.lineNumber : this.stockChangeLines.value.length;
                    this.stockChangeLines.addOrUpdateRecordValue(
                        {
                            _id: `${currentRecord.stockId}|${lineNumber}`,
                            stockSite: this.stockSite.value?.code,
                            lineNumber: lineNumber,
                            product: {
                                code: currentRecord?.product?.product?.code,
                                _id: currentRecord?.product?.product?._id,
                            },
                            stockId: String(currentRecord.stockId),
                            productDescription: this.productSelected.value.localizedDescription1,
                            packingUnit: {
                                code: currentRecord.packingUnit?.code,
                                _id: currentRecord.packingUnit?._id,
                            },
                            quantityInPackingUnit: currentRecord?.quantityInPackingUnit,
                            quantityMoved: this.quantityToMove.value ?? 0,
                            packingUnitToStockUnitConversionFactor:
                                currentRecord.packingUnitToStockUnitConversionFactor,
                            serialNumber: currentRecord.serialNumber,
                            numberSession: this._numberSession,
                            location: {
                                code: currentRecord.location?.code,
                                _id: currentRecord.location?._id,
                            },
                            status: {
                                code: currentRecord.status?.code,
                                _id: currentRecord.status?._id,
                            },
                            identifier1: currentRecord.identifier1,
                            identifier2: currentRecord.identifier2,
                            ...this.getCustomizationInternalStockChangeLines(),
                        },
                        0,
                        undefined,
                    );
                    this.stockChangeLines.addOrUpdateRecordValue(
                        {
                            _id: `${currentRecord.stockId}|${lineNumber}`,
                            packingUnit: {
                                code: this.packingUnitDestination.value
                                    ? this.packingUnitDestination.value
                                    : currentRecord.packingUnit?.code,
                                _id: this.packingUnitDestination.value
                                    ? this.packingUnitDestination.value
                                    : currentRecord.packingUnit?._id,
                            },
                            product: {
                                code: currentRecord?.product?.product?.code,
                                _id: currentRecord?.product?.product?._id,
                            },
                            packingUnitToStockUnitConversionFactor: this
                                .packingUnitToStockUnitConversionFactorDestination.value
                                ? this.packingUnitToStockUnitConversionFactorDestination.value?.toString()
                                : currentRecord?.packingUnitToStockUnitConversionFactor?.toString(),
                            quantityInPackingUnit: (
                                (Number(this.quantityToMove.value) *
                                    Number(currentRecord.packingUnitToStockUnitConversionFactor)) /
                                Number(
                                    this.packingUnitToStockUnitConversionFactorDestination.value
                                        ? this.packingUnitToStockUnitConversionFactorDestination.value
                                        : currentRecord?.packingUnitToStockUnitConversionFactor,
                                )
                            ).toString(),
                            quantityMoved: this.quantityToMove.value ?? 0,
                            quantityInStockUnit: (
                                Number(this.quantityToMove.value) *
                                Number(currentRecord.packingUnitToStockUnitConversionFactor)
                            ).toString(),
                            ...(this.locationDestination.value && {
                                location: {
                                    code: this.locationDestination.value
                                        ? this.locationDestination.value.code
                                        : currentRecord.location?.code,
                                    _id: this.locationDestination.value
                                        ? this.locationDestination.value._id
                                        : currentRecord.location?._id,
                                },
                            }),
                            ...(currentRecord.licensePlateNumber && {
                                licensePlateNumber: {
                                    code: currentRecord.licensePlateNumber.code,
                                    _id: currentRecord.licensePlateNumber._id,
                                },
                            }),
                            ...(this.statusDestination.value && {
                                status: {
                                    code: this.statusDestination.value,
                                    _id: this.statusDestination.value,
                                },
                            }),
                            ...(!this.statusDestination.value &&
                                currentRecord.status &&
                                this.statusDestination.options?.includes(currentRecord.status.code) && {
                                status: {
                                    code: currentRecord.status.code,
                                    _id: currentRecord.status._id,
                                },
                            }),
                            identifier1: this.identifier1Destination.value
                                ? this.identifier1Destination.value
                                : currentRecord.identifier1,
                            identifier2: this.identifier2Destination.value
                                ? this.identifier2Destination.value
                                : currentRecord.identifier2,
                            serialNumber: currentRecord.serialNumber,
                            ...this.getCustomizationInternalStockChangeDetails(),
                        },
                        1,
                        `${currentRecord.stockId}|${lineNumber}`,
                    );
                    this.transaction.isDisabled = true;
                }
                let qtyTotal = 0;
                this.stockChangeLines.value.forEach(line => {
                    if (
                        Number(line.stockId) === Number(currentRecord.stockId) &&
                        line.numberSession === this._numberSession
                    ) {
                        line.stockDetails?.forEach(stockDetail => {
                            qtyTotal =
                                qtyTotal +
                                (Number(stockDetail.quantityInPackingUnit) *
                                    Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                                Number(line.packingUnitToStockUnitConversionFactor);
                        });
                    }
                });
                (currentRecord as any).quantityToMove = qtyTotal;
                (currentRecord as any).quantityInStockUnitDestination =
                    qtyTotal * Number(currentRecord.packingUnitToStockUnitConversionFactor);
                (currentRecord as any).locationDestination = this.locationDestination.value
                    ? this.locationDestination.value
                    : currentRecord?.location;
                (currentRecord as any).statusDestination = this.statusDestination.value
                    ? this.statusDestination.value
                    : currentRecord.status?.code;
                (currentRecord as any).packingUnitDestination =
                    this.packingUnitDestination.value !== '' ? { code: this.packingUnitDestination.value } : null;
                (currentRecord as any).packingUnitToStockUnitConversionFactorDestination = this
                    .packingUnitToStockUnitConversionFactorDestination.value
                    ? this.packingUnitToStockUnitConversionFactorDestination.value
                    : (currentRecord as any).packingUnitToStockUnitConversionFactor;
                (currentRecord as any).identifier1Destination = this.mainIdentifier1Destination.value
                    ? this.mainIdentifier1Destination.value
                    : currentRecord?.identifier1;
                (currentRecord as any).identifier2Destination = this.mainIdentifier2Destination.value
                    ? this.mainIdentifier2Destination.value
                    : currentRecord?.identifier2;
                this.stock.setRecordValue(currentRecord);
            }

            this.productCode.value = this.productSelected.value?.code;
            this.productDescription.value = this.productSelected.value?.localizedDescription1;

            await this._setActiveSection(Section.stockPanel);
        },
    })
    SelectButton: ui.PageAction;

    @ui.decorators.pageAction<MobileInternalStockChange>({
        icon: 'add',
        title: 'Add...',
        onError(error) {
            switch (error.message) {
                case '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumber': {
                    return ui.localize(
                        '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumber',
                        'The serial number is mandatory',
                    );
                }
                case '@sage/x3-stock/serial-number-range-overlap': {
                    return ui.localize(
                        '@sage/x3-stock/serial-number-range-overlap',
                        'The serial numbers are overlapping. Enter another starting or ending serial number.',
                    );
                }
                case '@sage/x3-stock/pages__mobile_stock_change__same-amount-serial-numbers-in-the-range-to-match-quantity-to-move': {
                    return ui.localize(
                        '@sage/x3-stock/pages__mobile_stock_change__same-amount-serial-numbers-in-the-range-to-match-quantity-to-move',
                        'Select the same amount of serial numbers in the range to match the quantity to move.',
                    );
                }
                case '@sage/x3-stock/serial-number-not-sequential': {
                    return ui.localize(
                        '@sage/x3-stock/serial-number-not-sequential',
                        'The serial numbers are not sequential. Check your entry.',
                    );
                }
                case '@sage/x3-stock/destination-is-missing': {
                    return ui.localize('@sage/x3-stock/destination-is-missing', 'Destination is missing.');
                }
                default: {
                    return error;
                }
            }
        },
        buttonType: 'secondary',
        async onClick() {
            if (!this.startingSerialNumber.value) {
                throw new Error(
                    '@sage/x3-stock/pages__mobile_stock_change_lines__notification__error_startingSerialNumber',
                );
            }
            // check that this will add any duplicates
            const startNumberToAdd = this.startingSerialNumber.value.code.match(/\d+$/);
            const endNumberToAdd = Number(this.endingSerialNumber.value.match(/\d+$/));
            let serialNumberAlreadyUsed: boolean = false;
            if (
                this.serialNumberLines.value.some(row => {
                    const rowStartMatch = row.startingSerialNumber.match(/\d+$/);
                    const rowEndMatch = Number(row.endingSerialNumber.match(/\d+$/));
                    if (
                        row.startingSerialNumber.substring(
                            0,
                            row.startingSerialNumber.length - rowStartMatch.toString().length,
                        ) !==
                        this.startingSerialNumber.value.code.substring(
                            0,
                            this.startingSerialNumber.value.code.length - startNumberToAdd.toString().length,
                        )
                    )
                        return false;

                    return Number(startNumberToAdd) <= rowEndMatch && endNumberToAdd >= Number(rowStartMatch);
                })
            ) {
                serialNumberAlreadyUsed = true;
            }
            this.stockChangeLines.value.forEach(line => {
                if (line.product === this.productSelected.value?.code && line.serialNumber) {
                    const startingSerialNumber = Number(line.serialNumber.match(/\d+$/));
                    const endingSerialNumber = Number(
                        this._calculateEndingSerialNumber(
                            line.serialNumber,
                            Number(line.quantityInPackingUnitDestination),
                        ).match(/\d+$/),
                    );
                    if (
                        Number(startNumberToAdd) <= endingSerialNumber &&
                        Number(endNumberToAdd) >= startingSerialNumber
                    ) {
                        serialNumberAlreadyUsed = true;
                    }
                }
            });
            if (serialNumberAlreadyUsed) {
                throw new Error('@sage/x3-stock/serial-number-range-overlap');
            }
            if (
                this.endingSerialNumber.value !=
                this._calculateEndingSerialNumber(
                    this.startingSerialNumber.value?.code,
                    Number(this.quantityToMove.value),
                )
            ) {
                throw new Error(
                    '@sage/x3-stock/pages__mobile_stock_change__same-amount-serial-numbers-in-the-range-to-match-quantity-to-move',
                );
            }
            if (
                (await getCountSerialNumber(
                    this,
                    this.productSelected.value?.code ?? '',
                    this.stockSite.value?.code,
                    this._stockId.value,
                    this.startingSerialNumber.value?.code ?? '',
                    this.endingSerialNumber.value ?? '',
                )) !== Number(this.quantityToMove.value)
            ) {
                throw new Error('@sage/x3-stock/serial-number-not-sequential');
            }

            let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);

            if (
                (!this.locationDestination.value ||
                    this.locationDestination.value._id === currentRecord.location._id) &&
                (!this.statusDestination.value || this.statusDestination.value === currentRecord.status.code) &&
                (!this.packingUnitDestination.value ||
                    this.packingUnitDestination.value === currentRecord.packingUnit.code) &&
                (!this.mainIdentifier1Destination.value ||
                    this.mainIdentifier1Destination.value === currentRecord.identifier1) &&
                (!this.mainIdentifier2Destination.value ||
                    this.mainIdentifier2Destination.value === currentRecord.identifier2)
            ) {
                throw new Error('@sage/x3-stock/destination-is-missing');
            }

            this.serialNumberLines.addRecord({
                quantity: this.quantityToMove.value,
                startingSerialNumber: this.startingSerialNumber.value.code,
                endingSerialNumber: this.endingSerialNumber.value,
            });
            const _lineNumber = this.stockChangeLines.value.length;
            this.stockChangeLines.addRecord(
                {
                    _id: `${currentRecord.stockId}|${_lineNumber}`,
                    stockSite: this.stockSite.value?.code,
                    lineNumber: _lineNumber,
                    product: {
                        code: currentRecord?.product?.product?.code,
                        _id: currentRecord?.product?.product?._id,
                    },
                    stockId: String(currentRecord.stockId),
                    productDescription: this.productSelected.value.localizedDescription1,
                    packingUnit: {
                        code: this.packingUnitDestination.value
                            ? this.packingUnitDestination.value
                            : currentRecord.packingUnit?.code,
                        _id: this.packingUnitDestination.value
                            ? this.packingUnitDestination.value
                            : currentRecord.packingUnit?._id,
                    },
                    quantityInPackingUnit: this.quantityToMove.value?.toString(),
                    quantityMoved: this.quantityToMove.value ?? 0,
                    packingUnitToStockUnitConversionFactor: currentRecord.packingUnitToStockUnitConversionFactor,
                    serialNumber: this.startingSerialNumber.value.code,
                    endingSerialNumber: this.endingSerialNumber.value,
                    numberSession: this._numberSession,
                },
                0,
                undefined,
            );
            this.stockChangeLines.addRecord(
                {
                    _id: `${currentRecord.stockId}|${_lineNumber}`,
                    packingUnit: {
                        code: this.packingUnitDestination.value
                            ? this.packingUnitDestination.value
                            : currentRecord.packingUnit?.code,
                        _id: this.packingUnitDestination.value
                            ? this.packingUnitDestination.value
                            : currentRecord.packingUnit?._id,
                    },
                    product: {
                        code: currentRecord?.product?.product?.code,
                        _id: currentRecord?.product?.product?._id,
                    },
                    packingUnitToStockUnitConversionFactor: this.packingUnitToStockUnitConversionFactorDestination.value
                        ? this.packingUnitToStockUnitConversionFactorDestination.value?.toString()
                        : currentRecord?.packingUnitToStockUnitConversionFactor?.toString(),
                    quantityMoved: this.quantityToMove.value ?? 0,
                    quantityInPackingUnit: (
                        (Number(this.quantityToMove.value) *
                            Number(currentRecord.packingUnitToStockUnitConversionFactor)) /
                        Number(
                            this.packingUnitToStockUnitConversionFactorDestination.value
                                ? this.packingUnitToStockUnitConversionFactorDestination.value
                                : currentRecord?.packingUnitToStockUnitConversionFactor,
                        )
                    ).toString(),
                    quantityInStockUnit: (
                        Number(this.quantityToMove.value) * Number(currentRecord.packingUnitToStockUnitConversionFactor)
                    ).toString(),
                    ...(this.locationDestination.value && {
                        location: {
                            code: this.locationDestination.value
                                ? this.locationDestination.value.code
                                : currentRecord.location?.code,
                            _id: this.locationDestination.value
                                ? this.locationDestination.value._id
                                : currentRecord.location?._id,
                        },
                    }),
                    ...(this.statusDestination.value && {
                        status: {
                            code: this.statusDestination.value,
                            _id: this.statusDestination.value,
                        },
                    }),
                    ...(!this.statusDestination.value &&
                        currentRecord.status &&
                        this.statusDestination.options?.includes(currentRecord.status.code) && {
                        status: {
                            code: currentRecord.status.code,
                            _id: currentRecord.status._id,
                        },
                    }),
                    identifier1: this.mainIdentifier1Destination.value,
                    identifier2: this.mainIdentifier2Destination.value,
                    serialNumber: this.startingSerialNumber.value.code,
                    endingSerialNumber: this.endingSerialNumber.value,
                },
                1,
                `${currentRecord.stockId}|${_lineNumber}`,
            );
            //            (currentRecord as any).quantityToMove = String(0);
            this.quantityToMove.value = 0;
            this.stock.setRecordValue(currentRecord);
            this.startingSerialNumber.value = null;
            this.endingSerialNumber.value = null;

            await this.$.commitValueAndPropertyChanges();
        },
    })
    addSerialRangeButton: ui.PageAction;

    @ui.decorators.section<MobileInternalStockChange>({
        isTitleHidden: true,
    })
    transactionSection: ui.containers.Section;

    @ui.decorators.block<MobileInternalStockChange>({
        parent() {
            return this.transactionSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    transactionBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        parent() {
            return this.transactionSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationInternalStockChangeHeaderBlock: ui.containers.Block;

    @ui.decorators.dateField<MobileInternalStockChange>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Change date',
        isMandatory: true,
        width: 'small',
        maxDate: DateValue.today().toString(),
        onChange() { },
    })
    effectiveDate: ui.fields.Date;

    @ui.decorators.referenceField<MobileInternalStockChange, Site>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Site',
        node: '@sage/x3-system/Site',
        valueField: 'code',
        isMandatory: true,
        isTransient: false,
        isReadOnly: true,
        isHidden: true,
        bind: 'stockSite',
        width: 'small',
        filter() {
            return {
                code: String(this.$.storage.get('mobile-selected-stock-site')),
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: '_id',
            }),
        ],
    })
    stockSite: ui.fields.Reference;

    @ui.decorators.dropdownListField<MobileInternalStockChange>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Transaction',
        //options: ['ALL'],
        isTransient: true,
        onChange() {
            if (this.transaction.value) {
                const transaction = this._transactions.find(trs => trs.code === this.transaction.value);
                this._setTransaction(transaction, false, false);
            } else {
                this.productSelected.isDisabled = true;
            }
        },
    })
    transaction: ui.fields.DropdownList;

    @ui.decorators.textField<MobileInternalStockChange>({
        isHidden: true,
        isTransientInput: true,
    })
    destinationCode: ui.fields.Text;

    @ui.decorators.textField<MobileInternalStockChange>({
        isHidden: true,
        isTransientInput: true,
    })
    printingMode: ui.fields.Text;

    @ui.decorators.section<MobileInternalStockChange>({
        isTitleHidden: true,
        //    mode: 'accordion',
    })
    fromSection: ui.containers.Section;

    @ui.decorators.block<MobileInternalStockChange>({
        parent() {
            return this.fromSection;
        },
        width: 'extra-large',
        title: 'From',
    })
    fromBlock: ui.containers.Block;

    /* @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.fromBlock;
        },
        placeholder: 'Scan your code',
        title: 'Composite code',
        isMandatory: false,
        isTransient: true,
        width: 'large',
        onChange() {},
    })
    compositeCode: ui.fields.Text; */

    @ui.decorators.referenceField<MobileInternalStockChange, Location>({
        parent() {
            return this.fromBlock;
        },
        title: 'Location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        isMandatory: false,
        isTransient: true,
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        helperTextField: 'type',
        width: 'large',
        canFilter: false,
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite.value?.code ?? undefined },
                category: { _nin: ['subcontract', 'customer'] },
                ...(this.productSelected.value && {
                    stock: { _atLeast: 1, product: { product: { code: this.productSelected.value?.code } } },
                }),
            };
            return locationFilter;
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.locationSelected, rawData);
        },
        async onChange() {
            await this._onChange_locationSelected();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    locationSelected: ui.fields.Reference;

    /*
     * onChange locationSelected event (for decorator and management service)
     */
    private readonly _onChange_locationSelected: AsyncVoidFunction = async () => {
        if (this.locationSelected.value) {
            await this._getLocationProducts();
        }
        if (this.productSelected.value) {
            await this._getStock();
        }
    };

    @ui.decorators.referenceField<MobileInternalStockChange, Product>({
        parent() {
            return this.fromBlock;
        },
        title: 'Product',
        node: '@sage/x3-master-data/Product',
        valueField: 'code',
        helperTextField: 'upc',
        placeholder: 'Scan or select...',
        isMandatory: false,
        isTransient: true,
        canFilter: false,
        isAutoSelectEnabled: true,
        isDisabled: false,
        isFullWidth: true,
        shouldSuggestionsIncludeColumns: true,
        filter() {
            return {
                productStatus: { _ne: 'notUsable' },
                stockManagementMode: { _ne: 'notManaged' },
                productSites: { _atLeast: 1, stockSite: { code: this.stockSite.value?.code ?? undefined } },
                ...((this.locationSelected.value || this.stockChangeLines.value.length !== 0) && {
                    code: {
                        ...(this.locationSelected.value && { _in: this._locationProducts }),
                        ...(this.stockChangeLines.value.length !== 0 && {
                            _nin: this.stockChangeLines.value.map(line => line.product?.code),
                        }),
                    },
                }),
            };
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.productSelected, rawData);
        },
        async onChange() {
            await this._onChange_productSelected();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: 'localizedDescription1',
            }),
            ui.nestedFields.text({
                bind: 'upc',
                prefix: 'UPC:',
            }),
            ui.nestedFields.text({
                bind: 'globalTradeItemNumber',
                title: 'GTIN',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'lotManagementMode',
                optionType: '@sage/x3-stock-data/LotManagementMode',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'serialNumberManagementMode',
                optionType: '@sage/x3-master-data/SerialNumberManagement',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'expirationManagementMode',
                optionType: '@sage/x3-master-data/ExpirationManagement',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'stockVersionMode',
                optionType: '@sage/x3-stock-data/StockVersionMode',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Product, UnitOfMeasure>({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'stockUnit',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Product, UnitOfMeasure>({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'stockUnit',
                valueField: 'numberOfDecimals',
                isHidden: true,
            }),
        ],
    })
    productSelected: ui.fields.Reference;

    /*
     * onChange productSelected event (for decorator and management service)
     */
    private readonly _onChange_productSelected: AsyncVoidFunction = async () => {
        do {
            if (await this.productSelected.value?.code) {
                const result = withoutEdges(
                    await this.$.graph
                        .node('@sage/x3-master-data/ProductSite')
                        .query(
                            ui.queryUtils.edgesSelector(
                                {
                                    isBeingCounted: true,
                                    product: {
                                        globalTradeItemNumber: true,
                                    },
                                },
                                {
                                    filter: {
                                        product: {
                                            code: this.productSelected.value?.code,
                                        },
                                        stockSite: {
                                            code: this.stockSite.value?.code,
                                        },
                                    },
                                },
                            ),
                        )
                        .execute(),
                ) as ExtractEdgesPartial<ProductSite>[];

                if (result.length === 0) {
                    this.productSelected.value = null;
                    break;
                }

                if (result[0].isBeingCounted === true) {
                    if (
                        !(await this.$.dialog
                            .confirmation(
                                'warn',
                                ui.localize('@sage/x3-stock/dialog-warning-title', 'Warning'),
                                ui.localize(
                                    '@sage/x3-stock/product-blocked-by-count-continue',
                                    'Product blocked by count. Do you want to continue?',
                                ),
                                {
                                    acceptButton: {
                                        text: ui.localize('@sage/x3-stock/button-accept-yes', 'Yes'),
                                    },
                                    cancelButton: {
                                        text: ui.localize('@sage/x3-stock/button-cancel-no', 'No'),
                                    },
                                },
                            )
                            .catch(error => false))
                    ) {
                        this.productSelected.value = null;
                        break;
                    }
                }

                this._numberSession = this.stockChangeLines.value.length === 0 ? 0 : this._numberSession + 1;
                this._productSite = await this._getProductSite(this.productSelected.value?.code);
                this._packingUnits = this._productSite.product.packingUnits?.query.edges.map(packingUnit => {
                    return packingUnit;
                });
                this.packingUnitFilter.options = [
                    this._productSite.product.stockUnit.code,
                    ...this._packingUnits.map(packingUnit => {
                        return packingUnit.node.packingUnit.code;
                    }),
                ];
                this.packingUnitDestination.options = [
                    this._productSite.product.stockUnit.code,
                    ...this._packingUnits.map(packingUnit => {
                        return packingUnit.node.packingUnit.code;
                    }),
                ];
                const isGlobalReceivedIssued =
                    this.productSelected.value?.serialNumberManagementMode === 'globalReceivedIssued';
                this.addSerialRangeButton.isHidden = !isGlobalReceivedIssued;
                this.serialNumberBlock.isHidden = !isGlobalReceivedIssued;
                this.listSerialNumberBlock.isHidden = !isGlobalReceivedIssued;
                this.serialNumberLines.isTitleHidden = !isGlobalReceivedIssued;
                this.listSerialNumberBlock.isHidden = !isGlobalReceivedIssued;

                this.serialNumberLines.title = 'Serial number(s) to move    ';
                await this._initPageDetail();
                this.locationSelected.isHidden =
                    !this._productSite.isLocationManaged || !this.getLocationInMobileSettings();
                managePages(
                    this,
                    this.stockSite.value?.code,
                    {
                        product: {
                            code: this._productSite.product.code,
                            lotManagementMode: this._productSite.product.lotManagementMode,
                            serialNumberManagementMode: this._productSite.product.serialNumberManagementMode,
                        },
                        isLocationManaged: this._productSite.isLocationManaged,
                    },
                    '3',
                    ui.localize(
                        '@sage/x3-stock/pages__mobile-settings__mandatory-settings-missing',
                        'Mandatory settings missing.',
                    ),
                    this._stockFieldSettings,
                    this._mobileSettings,
                );

                this.productCode.value = this.productSelected.value?.code;
                this.productDescription.value = this.productSelected.value?.localizedDescription1;
            }
        } while (false);

        this.nextButton.isHidden = !this.productSelected.value;

        await this._getStock();

        if (!this.productSelected.value) {
            this.productSelected.focus();
        }
    };

    @ui.decorators.section<MobileInternalStockChange>({
        isTitleHidden: true,
        //  mode: 'accordion',
    })
    toSection: ui.containers.Section;

    @ui.decorators.block<MobileInternalStockChange>({
        parent() {
            return this.toSection;
        },
        width: 'extra-large',
        title: 'To',
    })
    toBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileInternalStockChange, Location>({
        parent() {
            return this.toBlock;
        },
        title: 'Destination location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        isMandatory: false,
        isTransient: true,
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        width: 'large',
        helperTextField: 'type',
        canFilter: false,
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite.value?.code ?? undefined },
                category: 'internal',
            };
            return locationFilter;
        },
        async onChange() {
            await this._onChange_mainLocationDestination();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    mainLocationDestination: ui.fields.Reference;

    /*
     * onChange mainLocationDestination event (for decorator and management service)
     */
    private readonly _onChange_mainLocationDestination: AsyncVoidFunction = async () => {
        if (this.mainLocationDestination.value) {
            this.mainLocationDestination.getNextField(true)?.focus();
        }
    };

    @ui.decorators.selectField<MobileInternalStockChange>({
        parent() {
            return this.toBlock;
        },
        title: 'Destination status',
        width: 'small',
        isMandatory: false,
        isTransient: true,
        async onChange() {
            await this._onChange_mainStatusDestination();
        },
    })
    mainStatusDestination: ui.fields.Select;

    /*
     * onChange statusDestination event (for decorator and management service)
     */
    private readonly _onChange_mainStatusDestination: AsyncVoidFunction = async () => {
        if (this.mainStatusDestination.value) {
            this.mainStatusDestination.getNextField(true)?.focus();
        }
    };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.toBlock;
        },
        placeholder: 'Scan...',
        title: 'Destination identifier 1',
        isMandatory: false,
        isTransient: true,
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        width: 'large',
        async onChange() {
            await this._onChange_mainIdentifier1Destination();
        },
    })
    mainIdentifier1Destination: ui.fields.Text;
    /*
     * onChange mainIdentifier1Destination event (for decorator and management service)
     */
    private readonly _onChange_mainIdentifier1Destination: AsyncVoidFunction = async () => { };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.toBlock;
        },
        placeholder: 'Scan...',
        title: 'Destination identifier 2',
        isMandatory: false,
        isTransient: true,
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        width: 'large',
        async onChange() {
            await this._onChange_mainIdentifier2Destination();
        },
    })
    mainIdentifier2Destination: ui.fields.Text;

    /*
     * onChange mainIdentifier2Destination event (for decorator and management service)
     */
    private readonly _onChange_mainIdentifier2Destination: AsyncVoidFunction = async () => { };

    @ui.decorators.section<MobileInternalStockChange>({
        isTitleHidden: true,
    })
    stockChangeLinesSection: ui.containers.Section;

    @ui.decorators.block<MobileInternalStockChange>({
        parent() {
            return this.stockChangeLinesSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    stockChangeLinesBlock: ui.containers.Block;

    @ui.decorators.nestedGridField<
        MobileInternalStockChange,
        [StockChangeLineWithQuantityMoved, StockJournalWithQuantityMoved]
    >({
        parent() {
            return this.stockChangeLinesBlock;
        },
        bind: 'stockChangeLines',
        canFilter: true,
        canActivate: true,
        isChangeIndicatorDisabled: true,
        canSelect: false,
        hasLineNumbers: false,
        isTitleHidden: true,
        levels: [
            {
                node: '@sage/x3-stock/StockChangeLine',
                childProperty: 'stockDetails',
                columns: [
                    ui.nestedFields.numeric({
                        bind: 'lineNumber',
                        title: 'Line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileInternalStockChange, StockChangeLineWithQuantityMoved, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: { code: true },
                        isReadOnly: false,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'productDescription',
                        title: 'Description',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityMoved',
                        title: 'Quantity in packing unit',
                        isReadOnly: true,
                        isHiddenDesktop: false,
                        isTransient: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity in packing unit',
                        isReadOnly: true,
                        isHidden: true,
                        prefix(_value, rowValue) {
                            return `${(Number(this.stockChangeLines.getRecordValue(rowValue?._id, 1)?.quantityInPackingUnit) *
                                    Number(
                                        this.stockChangeLines.getRecordValue(rowValue?._id, 1)
                                            ?.packingUnitToStockUnitConversionFactor,
                                    )) /
                                (rowValue?.packingUnitToStockUnitConversionFactor
                                    ? Number(rowValue?.packingUnitToStockUnitConversionFactor)
                                    : 1)
                                } /`;
                        },
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.reference<
                        MobileInternalStockChange,
                        StockChangeLineWithQuantityMoved,
                        UnitOfMeasure
                    >({
                        bind: 'packingUnit',
                        title: 'Packing unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: { code: true },
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'stockId',
                        title: 'stockId',
                        isHidden: true,
                        isTransientInput: true,
                        isReadOnly: true,
                        isHiddenDesktop: false,
                    }),
                    ui.nestedFields.text({
                        bind: 'serialNumber',
                        title: 'Serial number',
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: 'endingSerialNumber',
                        title: 'Ending serial number',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberSession',
                        title: 'Number session',
                        isHidden: true,
                        isTransientInput: true,
                        isReadOnly: true,
                        isHiddenDesktop: false,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        /*  isHidden() {
                            return this._getSidebarInternalStockChangeLine().length === 0;
                        }, */
                        async onClick(rowId) {
                            const _stockChangeLine = this.stockChangeLines.getRecordValue(rowId, 0);
                            if (!_stockChangeLine) {
                                return;
                            }
                            const _stockDetail = this.stockChangeLines.getRecordValue(rowId, 1);
                            if (!_stockDetail) {
                                return;
                            }
                            const _productCode = _stockChangeLine.product?.code;
                            if (!_productCode) {
                                return;
                            }
                            this._productSite = await this._getProductSite(_productCode);
                            this._packingUnits = this._productSite.product.packingUnits?.query.edges.map(
                                packingUnit => {
                                    return packingUnit;
                                },
                            );
                            this.packingUnitFilter.options = [
                                this._productSite.product.stockUnit.code,
                                ...this._packingUnits.map(packingUnit => {
                                    return packingUnit.node.packingUnit.code;
                                }),
                            ];
                            this.packingUnitDestination.options = [
                                this._productSite.product.stockUnit.code,
                                ...this._packingUnits.map(packingUnit => {
                                    return packingUnit.node.packingUnit.code;
                                }),
                            ];
                            this._previousQuantity = _stockDetail.quantityInPackingUnit;
                            this._previousSerialNumber = _stockDetail.serialNumber;
                            this._previousEndingSerialNumber = _stockDetail.endingSerialNumber;
                            this._licensePlateNumber = _stockDetail.licensePlateNumber;
                            if (this._getSidebarInternalStockChangeLine().length === 0) {
                                this.stockChangeLines.openSidebar(rowId, 1);
                            } else {
                                this.stockChangeLines.openSidebar(rowId, 0);
                            }
                        },
                    },
                    {
                        icon: 'delete',
                        title: 'Delete',
                        onClick(rowId, data: any) {
                            const stockLine = this.stock.value.find(
                                line => Number(line.stockId) === Number(data.stockId),
                            );
                            if (stockLine) {
                                this.stock.unselectRecord(stockLine._id);
                            }
                            this.stockChangeLines.removeRecord(rowId, 0);
                            this.transaction.isDisabled = this.stockChangeLines.value.length > 0;
                            this.createButton.isDisabled = this.stockChangeLines.value.length === 0;
                            this.stockChangeLinesBlock.isHidden = this.stockChangeLines.value.length === 0;
                        },
                    },
                ],
                sidebar: {
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock change lines',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarInternalStockChangeLine(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
            {
                node: '@sage/x3-stock-data/StockJournal',
                columns: [
                    /* ui.nestedFields.text({
                        bind: 'productCode' as any,
                        isReadOnly: true,
                        prefix(_value, rowValue) {
                            return rowValue?.product?.code;
                        },
                    }), */
                    /* ui.nestedFields.text({
                        bind: 'movementDescription',
                        isReadOnly: true,
                    }), */
                    ui.nestedFields.numeric({
                        bind: 'quantityMoved',
                        title: 'Quantity in packing unit',
                        isReadOnly: false,
                        isFullWidth: true,
                        isMandatory: true,
                        isTransient: true,
                        postfix(_value, rowValue) {
                            return this.stockChangeLines.getRecordValue(rowValue?._id, 0)?.packingUnit?.code ?? '';
                        },
                        scale(_value, rowValue) {
                            return (
                                this.stockChangeLines.getRecordValue(rowValue?._id, 0)?.packingUnit?.numberOfDecimals ??
                                0
                            );
                        },
                        async onChange(_id, value) {
                            value.quantityInPackingUnit =
                                (Number(value.quantityMoved) *
                                    Number(
                                        this.stockChangeLines.getRecordValue(value?._id, 0)
                                            ?.packingUnitToStockUnitConversionFactor,
                                    )) /
                                Number(
                                    value?.packingUnitToStockUnitConversionFactor
                                        ? value?.packingUnitToStockUnitConversionFactor
                                        : 1,
                                );
                            value.quantityInStockUnit =
                                Number(value.quantityMoved) *
                                Number(
                                    this.stockChangeLines.getRecordValue(value?._id, 0)
                                        ?.packingUnitToStockUnitConversionFactor,
                                );
                            if (
                                value._id &&
                                this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued'
                            ) {
                                let currentQty = Number(value.quantityMoved);
                                if (currentQty > 1) {
                                    value.endingSerialNumber = this._calculateEndingSerialNumber(
                                        value.serialNumber,
                                        currentQty,
                                    );
                                } else {
                                    value.endingSerialNumber = value.serialNumber;
                                }
                                const result = await this._validateSerialNumber(value);
                                if (result) {
                                    this.$.showToast(result, { type: 'warning' });
                                    value.quantityMoved = this._previousQuantity;
                                    value.endingSerialNumber = this._previousEndingSerialNumber;
                                }
                                this._previousQuantity = value.quantityMoved;
                                this._previousEndingSerialNumber = value.endingSerialNumber;
                            }
                            this.stockChangeLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity in packing unit',
                        isHidden: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInStockUnit',
                        title: 'Quantity in stock unit',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileInternalStockChange, StockJournalWithQuantityMoved, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: 'code',
                        isHidden: true,
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            /*  ui.nestedFields.reference<MobileInternalStockChange, ProductSite, UnitOfMeasure>({
                                bind: { stockUnit: { code: true } },
                                title: 'Stock unit',
                                node: '@sage/x3-master-data/UnitOfMeasure',
                                valueField: 'code',
                            }), */
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Destination conversion factor',
                        isHidden(value, rowValue) {
                            return !this._transaction.isUnitChange;
                        },
                        isFullWidth: true,
                        isMandatory: true,
                        isDisabled(value, rowValue) {
                            if (rowValue.packingUnit) {
                                const packingUnitIndex = this._packingUnits
                                    .map(packingUnit => packingUnit.node.packingUnit.code)
                                    .indexOf(rowValue.packingUnit.code);
                                if (packingUnitIndex !== -1) {
                                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                                    return !selectedUnit.isPackingFactorEntryAllowed;
                                } else {
                                    return true;
                                }
                            } else {
                                return true;
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobileInternalStockChange, StockJournalWithQuantityMoved, Location>({
                        bind: 'location',
                        title: 'Destination location',
                        node: '@sage/x3-stock-data/Location',
                        valueField: { code: true },
                        isHidden(value, rowValue) {
                            return !this._transaction.isLocationChange || !this._productSite.isLocationManaged;
                        },
                        isHiddenDesktop: false,
                        isReadOnly: false,
                        isFullWidth: true,
                        // isMandatory: true,
                        filter() {
                            return {
                                stockSite: { code: this.stockSite.value?.code ?? undefined },
                                category: 'internal',
                            };
                        },
                        isDisabled() {
                            return !!this._licensePlateNumber;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileInternalStockChange, StockJournalWithQuantityMoved, StockStatus>({
                        bind: 'status',
                        title: 'Destination status',
                        node: '@sage/x3-stock-data/StockStatus',
                        valueField: { code: true },
                        isHidden(value, rowValue) {
                            return !this._transaction.isStatusChange;
                        },
                        isHiddenDesktop: false,
                        isReadOnly: false,
                        isFullWidth: true,
                        // isMandatory: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileInternalStockChange, StockJournalWithQuantityMoved, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Destination unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isReadOnly: false,
                        isHidden(value, rowValue) {
                            return (
                                !this._transaction.isUnitChange ||
                                this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued'
                            );
                        },
                        isFullWidth: true,
                        // isMandatory: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                                isHidden: true,
                            }),
                        ],
                        filter() {
                            return {
                                code: {
                                    _in: this.packingUnitDestination.options,
                                },
                            };
                        },
                        onChange(_id, value) {
                            if (value.packingUnit) {
                                const packingUnitIndex = this._packingUnits
                                    .map(packingUnit => packingUnit.node.packingUnit.code)
                                    .indexOf(value.packingUnit.code);
                                if (packingUnitIndex !== -1) {
                                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                                    value.packingUnitToStockUnitConversionFactor = Number(
                                        selectedUnit.packingUnitToStockUnitConversionFactor,
                                    );
                                    // (this.stockChangeLines.properties.levels[1].columns[2].properties as any).isHidden =
                                    //    !selectedUnit.isPackingFactorEntryAllowed;
                                } else {
                                    value.packingUnitToStockUnitConversionFactor = 1;
                                }
                                this.stockChangeLines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'serialNumber',
                        node: '@sage/x3-stock-data/SerialNumber',
                        title: 'Starting serial number',
                        valueField: 'code',
                        placeholder: 'Scan or select...',
                        width: 'large',
                        validation: /^$|^[^|]+$/,
                        isHidden() {
                            return ['notManaged', 'issued'].includes(
                                this._productSite.product.serialNumberManagementMode,
                            );
                        },
                        isMandatory() {
                            return (
                                !['notManaged', 'issued'].includes(
                                    this._productSite.product.serialNumberManagementMode,
                                ) && !this._productSite.product.serialSequenceNumber
                            );
                        },
                        isDisabled() {
                            return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                        },
                        filter() {
                            return {
                                _and: [
                                    { product: { code: this._productSite.product.code } },
                                    { stockId: this._stockId.value ?? undefined },
                                ],
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                title: 'Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'stockId',
                                title: 'Stock ID',
                                isReadOnly: true,
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalWithQuantityMoved>) {
                            if (!value._id) {
                                return;
                            }
                            let currentQty = Number(value.quantityInPackingUnit);
                            if (!currentQty || !value.serialNumber) {
                                value.endingSerialNumber = undefined;
                                return;
                            }
                            value.serialNumber = value.serialNumber.toUpperCase();
                            if (currentQty > 1) {
                                value.endingSerialNumber = this._calculateEndingSerialNumber(
                                    value.serialNumber,
                                    currentQty,
                                );
                            } else {
                                value.endingSerialNumber = value.serialNumber;
                            }
                            const result = await this._validateSerialNumber(value);
                            if (result) {
                                this.$.showToast(result, { type: 'warning' });
                                value.serialNumber = this._previousSerialNumber;
                                value.endingSerialNumber = this._previousEndingSerialNumber;
                            }
                            this.stockChangeLines.addOrUpdateRecordValue(value, 1, value._id);
                            this._previousSerialNumber = value.serialNumber;
                            this._previousEndingSerialNumber = value.endingSerialNumber;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'endingSerialNumber',
                        title: 'Ending serial number',
                        isReadOnly: true,
                        isFullWidth: true,
                        isHidden() {
                            return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier1',
                        title: 'Destination identifier 1',
                        isHidden(value, rowValue) {
                            return this._transaction.identifier1Destination === 'hidden';
                        },
                        isReadOnly(value, rowValue) {
                            return this._transaction.identifier1Destination === 'displayed';
                        },
                        isHiddenDesktop: false,
                        isFullWidth: true,
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier2',
                        title: 'Destination identifier 2',
                        isHidden(value, rowValue) {
                            return this._transaction.identifier2Destination === 'hidden';
                        },
                        isReadOnly(value, rowValue) {
                            return this._transaction.identifier2Destination === 'displayed';
                        },
                        isHiddenDesktop: false,
                        isFullWidth: true,
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                    }),
                    ui.nestedFields.reference<
                        MobileInternalStockChange,
                        StockJournalWithQuantityMoved,
                        LicensePlateNumber
                    >({
                        bind: 'licensePlateNumber',
                        title: 'License plate number',
                        valueField: 'code',
                        node: '@sage/x3-stock-data/LicensePlateNumber',
                        isAutoSelectEnabled: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                        isHidden: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId, data: any) {
                            const _stockChangeLine = this.stockChangeLines.getRecordValue(rowId, 0);
                            if (!_stockChangeLine) {
                                return;
                            }
                            const _productCode = _stockChangeLine.product?.code;
                            if (!_productCode) {
                                return;
                            }
                            this._productSite = await this._getProductSite(_productCode);
                            if (!this._productSite) {
                                return;
                            }
                            this._stockId.value = _stockChangeLine.stockId ?? null;
                            this._packingUnits = this._productSite.product.packingUnits?.query.edges.map(
                                packingUnit => {
                                    return packingUnit;
                                },
                            );
                            this.packingUnitFilter.options = [
                                this._productSite.product.stockUnit.code,
                                ...this._packingUnits.map(packingUnit => {
                                    return packingUnit.node.packingUnit.code;
                                }),
                            ];
                            this.packingUnitDestination.options = [
                                this._productSite.product.stockUnit.code,
                                ...this._packingUnits.map(packingUnit => {
                                    return packingUnit.node.packingUnit.code;
                                }),
                            ];
                            this._previousQuantity = data.quantityInPackingUnit;
                            this._previousSerialNumber = data.serialNumber;
                            this._previousEndingSerialNumber = data.endingSerialNumber;
                            this._licensePlateNumber = data.licensePlateNumber;
                            this.stockChangeLines.openSidebar(rowId, 1);
                        },
                    },
                ],
                sidebar: {
                    title: 'Stock change',
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock details',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarInternalStockChangeDetail(),
                                    },
                                },
                            },
                        };
                    },
                    async onRecordConfirmed(_id: string, recordValue: StockJournalWithQuantityMoved | undefined) {
                        const _stockChangeLine = this.stockChangeLines.getRecordValue(_id, 0);
                        if (_stockChangeLine && recordValue) {
                            _stockChangeLine.quantityMoved = recordValue.quantityMoved;
                            /*   (Number(recordValue.quantityInPackingUnit) *
                                    Number(recordValue.packingUnitToStockUnitConversionFactor)) /
                                Number(_stockChangeLine.packingUnitToStockUnitConversionFactor); */
                            _stockChangeLine.serialNumber = recordValue.serialNumber;
                            _stockChangeLine.endingSerialNumber = recordValue.endingSerialNumber;
                            this.stockChangeLines.setRecordValue(_stockChangeLine, 0);
                        }
                    },
                },
            },
        ],
    })
    stockChangeLines: ui.fields.NestedGrid<[StockChangeLineWithQuantityMoved, StockJournalWithQuantityMoved]>;

    /*
     *
     *  Init functions
     *
     */
    private async _init(): Promise<boolean> {
        if (!(await this._initSite())) {
            return false;
        }

        if (this.stockSite.value?.code && !(await this._initControlManager(this.stockSite.value.code))) {
            return false;
        }

        if (this._mobileSettings?.stockField1) {
            this.$.router.emptyPage;
            this.createButton.isDisabled = true;
            this.$standardNewAction;
            this.effectiveDate.value = DateValue.today().toString();
            this.destinationCode.value = this.$.storage.get('mobile-label-destination')?.toString() ?? '';
            //this.expanseDestination.value = true;
            await this._initTransaction();
            this._numberOfDecimalList = await getUnitNumberOfDecimalList(this);
            this.customizationInternalStockChangeHeaderBlock.isHidden =
                this.customizationInternalStockChangeHeaderBlock.layout.$layout?.$items.length === 0;
            this.locationSelected.isHidden = !this.getLocationInMobileSettings();
            return true;
        } else {
            return false;
        }
    }

    private getLocationInMobileSettings(): boolean {
        return (
            this._mobileSettings.stockField1 === 'location' ||
            this._mobileSettings.stockField2 === 'location' ||
            this._mobileSettings.stockField3 === 'location' ||
            this._mobileSettings.stockField4 === 'location' ||
            this._mobileSettings.stockField5 === 'location' ||
            this._mobileSettings.stockField6 === 'location' ||
            this._mobileSettings.stockField7 === 'location' ||
            this._mobileSettings.stockField8 === 'location'
        );
    }

    private _disablePage(): void {
        this.effectiveDate.isDisabled = true;
        this.transaction.isDisabled = true;
        this.productSelected.isDisabled = true;
    }

    private async _initSite(): Promise<boolean> {
        const _stockSite = await getSelectedStockSite(
            this,
            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
            ui.localize(
                '@sage/x3-stock/dialog-error-location-inquiry-set-site',
                'Define a default stock site on the user function profile.',
            ),
        );

        if (!_stockSite) return;

        this.stockSite.value = {
            code: _stockSite,
            _id: _stockSite,
        };

        if (!this.stockSite.value) return;

        if (this.stockSite.value) {
            this._mobileSettings = JSON.parse(this.$.storage.get('mobile-settings-stock-change') as string);

            if (!this._mobileSettings?.stockField1) {
                dialogMessage(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_you_need_to_select_stock_search_parameters_to_set_up_Mobile_Automation_FUNADCSEARCH_function',
                        'You need to select stock search parameters to set up Mobile Automation - FUNADCSEARCH function.',
                    ),
                );
                return false;
            }
            return true;
        }
        return false;
    }

    private async _initPageDetail() {
        this.lotFilter.isHidden = this._productSite.product.lotManagementMode === 'notManaged';
        this.sublotFilter.isHidden = this.lotFilter.isHidden;
        this.sublotFilter.isHidden = this._productSite.product.lotManagementMode !== 'lotAndSublot';
        this._selectedStockManagementRules = await findStockManagementRules(
            this.stockSite.value?.code,
            this._productSite.product.productCategory.code,
            '26',
            this._transaction?.stockMovementCode?.code,
            this,
        );
        this.statusFilter.options = await this._getStockStatus();
        this.mainStatusDestination.options = this.statusFilter.options;
        this.statusDestination.options = this.statusFilter.options;

        this.serialNumberFilter.isHidden = ['notManaged', 'issued'].includes(
            this._productSite.product.serialNumberManagementMode,
        );
        if (['receivedIssued', 'globalReceivedIssued'].includes(this._productSite.product.serialNumberManagementMode)) {
            this.serialNumberFilter.isMandatory = false;
            if (this.lotFilter.isHidden === false) this.lotFilter.isMandatory = false;
            if (this.sublotFilter.isHidden === false) this.sublotFilter.isMandatory = false;
            if (this.statusFilter.isHidden === false) this.statusFilter.isMandatory = false;
            if (this.packingUnitFilter.isHidden === false) this.packingUnitFilter.isMandatory = false;
            if (this.packingUnitDestination.isHidden === false) this.packingUnitDestination.isHidden = true;
            if (this.packingUnitToStockUnitConversionFactorDestination.isHidden === false)
                this.packingUnitToStockUnitConversionFactorDestination.isHidden = true;
        }
        // this.locationDestination.isDisabled = this._productSite.isLicensePlateNumberManaged;
        this.locationDestination.isHidden = !this._productSite.isLocationManaged;
        this.mainLocationDestination.isHidden =
            !this._productSite.isLocationManaged || !this._transaction?.isLocationChange;
        if (!this._productSite.isLocationManaged) {
            this.mainLocationDestination.value = null;
        }

        this.customizationInternalStockChangeLinesBlock.isHidden =
            this.customizationInternalStockChangeLinesBlock.layout.$layout?.$items.length === 0;
        this.customizationInternalStockChangeDetailsBlock.isHidden =
            this.customizationInternalStockChangeDetailsBlock.layout.$layout?.$items.length === 0;
    }

    private async _getStockStatus(): Promise<string[]> {
        const selectedStatus: { _regex: string }[] = [];
        this._selectedStockManagementRules.authorizedSubstatus.split(',').forEach(function (status) {
            selectedStatus.push({ _regex: getRegExp(status).source });
        });
        const response = await this.$.graph
            .node('@sage/x3-stock-data/StockStatus')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        code: true,
                    },
                    {
                        filter: {
                            code: { _or: selectedStatus },
                        },
                    },
                ),
            )
            .execute();

        if (!response.edges || response.edges.length === 0) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_stock_change_details__notification__invalid_stock_status_error',
                    'No stock status',
                ),
            );
        }
        return response.edges.map((stockStatus: any) => stockStatus.node.code);
    }

    private async _initTransaction(): Promise<void | never> {
        try {
            this._transactions = extractEdges(
                await this.$.graph
                    .node('@sage/x3-stock/StockEntryTransaction')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                code: true,
                                isActive: true,
                                stockAutomaticJournal: {
                                    code: true,
                                },
                                localizedDescription: true,
                                isLocationChange: true,
                                isStatusChange: true,
                                isUnitChange: true,
                                transactionType: true,
                                identifier1Destination: true,
                                identifier2Destination: true,
                                identifier1Detail: true,
                                identifier2Detail: true,
                                printingMode: true,
                                defaultStockMovementGroup: {
                                    code: true,
                                },
                                stockMovementCode: {
                                    code: true,
                                },
                                companyOrSiteGroup: {
                                    group: true,
                                },
                            },
                            {
                                filter: {
                                    transactionType: 'stockChange',
                                    isActive: true,
                                    stockChangeDestination: 'internal',
                                    stockChangeAccessMode: { _ne: 'containerNumber' },
                                },
                            },
                        ),
                    )
                    .execute(),
            );
        } catch (err) {
            ui.console.error(err);
        }

        const transaction = this._transactions[0];
        this.transaction.isDisabled = false;
        const hideTransaction = this._transactions.length <= 1;
        hideTransaction ? (this.transactionBlock.isHidden = true) : null;

        this._setTransaction(transaction, hideTransaction, false);
    }

    private _setTransaction(transaction: PartialStockEntryTransaction, hideTransaction = false, disableProduct = true) {
        if (this._transactions) this.transaction.options = this._transactions.map(trs => trs.code);
        this._transaction = transaction;
        this.transaction.value = transaction.code;
        this.transaction.isHidden = hideTransaction;
        this.productSelected.isDisabled = disableProduct;
        this.mainLocationDestination.isHidden = !transaction.isLocationChange;
        this.locationDestination.isHidden = !transaction.isLocationChange;
        this.mainStatusDestination.isHidden = !transaction.isStatusChange;
        this.statusDestination.isHidden = !transaction.isStatusChange;
        this.packingUnitDestination.isHidden = !transaction.isUnitChange;
        this.packingUnitToStockUnitConversionFactorDestination.isHidden = !transaction.isUnitChange;
        this.mainIdentifier1Destination.isHidden = transaction.identifier1Destination === 'hidden';
        this.mainIdentifier1Destination.isReadOnly = transaction.identifier1Destination === 'displayed';
        this.mainIdentifier2Destination.isHidden = transaction.identifier2Destination === 'hidden';
        this.mainIdentifier2Destination.isReadOnly = transaction.identifier2Destination === 'displayed';
        // if (transaction.isStatusChange === true) this.statusFilter.isMandatory = true;
        this.printingMode.value = transaction.printingMode;
    }

    @ui.decorators.section<MobileInternalStockChange>({
        title: 'Stock change',
        isTitleHidden: true,
        isHidden: true,
    })
    stockPanelSection: ui.containers.Section;

    @ui.decorators.section<MobileInternalStockChange>({
        title: 'Stock change',
        isTitleHidden: true,
        isHidden: true,
    })
    detailPanelSection: ui.containers.Section;

    /*This block was defined but never used; it's a ghost block.
    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.stockPanelSection;
        },
    })
    headerStockBlock: ui.containers.Block;*/

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.stockPanelSection;
        },
    })
    stockBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    detailsBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    quantityBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    customizationInternalStockChangeLinesBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    customizationInternalStockChangeDetailsBlock: ui.containers.Block;

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    serialNumberBlock: ui.containers.Block;

    // This technical block should be placed at the end to avoid any unnecessary space.
    @ui.decorators.gridRowBlock<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
        boundTo() {
            return this.stock;
        },
        fieldFilter(columnId: string) {
            return false;
        },
        readOnlyOverride(columnId: string) {
            return undefined;
        },
    })
    gridBlock: ui.containers.GridRowBlock;

    /*@ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    destinationBlock: ui.containers.Block; */

    @ui.decorators.block<MobileInternalStockChange>({
        isTitleHidden: true,
        parent() {
            return this.detailPanelSection;
        },
    })
    listSerialNumberBlock: ui.containers.Block;
    /*
     *
     *  Fields
     *
     */

    @ui.decorators.textField<MobileInternalStockChange>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    productCode: ui.fields.Text;

    @ui.decorators.textField<MobileInternalStockChange>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    productDescription: ui.fields.Text;

    @ui.decorators.referenceField<MobileInternalStockChange, LicensePlateNumber>({
        parent() {
            return this.stockBlock;
        },
        title: 'License plate number',
        valueField: 'code',
        node: '@sage/x3-stock-data/LicensePlateNumber',
        placeholder: 'Scan or select...',
        isMandatory: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        canFilter: false,
        filter() {
            return {
                _and: [{ status: 'inStock' }, { stockSite: { code: this.stockSite.value?.code } }],
            };
        },
        async onChange() {
            await this._onChange_licensePlateNumberFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.reference({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
            }),
        ],
    })
    licensePlateNumberFilter: ui.fields.Reference;

    /*
     * onChange licensePlateNumberFilter event (for decorator and management service)
     */
    private readonly _onChange_licensePlateNumberFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        this.licensePlateNumberFilter.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileInternalStockChange, Location>({
        parent() {
            return this.stockBlock;
        },
        title: 'Location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        isMandatory: false,
        placeholder: 'Scan or select…',
        helperTextField: 'type',
        isAutoSelectEnabled: true,
        isFullWidth: true,
        isTransient: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        canFilter: false,
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite.value?.code },
                category: { _nin: ['subcontract', 'customer'] },
                stock: { _atLeast: 1, product: { product: { code: this.productSelected.value?.code } } },
            };
            return locationFilter;
        },
        isHidden() {
            return !!this.locationSelected.value;
        },
        async onChange() {
            await this._onChange_locationFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
        ],
    })
    locationFilter: ui.fields.Reference;

    /*
     * onChange locationFilter event (for decorator and management service)
     */
    private readonly _onChange_locationFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        if (this.locationFilter.value) {
            this.locationFilter.getNextField(true)?.focus();
        }
    };

    @ui.decorators.referenceField<MobileInternalStockChange, LotsSites>({
        parent() {
            return this.stockBlock;
        },
        title: 'Lot',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/LotsSites',
        valueField: 'lot',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        canFilter: false,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                product: { code: this.productSelected.value?.code },
                storageSite: { code: this.stockSite.value?.code ?? undefined },
            };
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.lotFilter, rawData);
        },
        async onChange() {
            await this._onChange_lotFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
    })
    lotFilter: ui.fields.Reference<LotsSites>;

    /*
     *  onChange lotFilter event (for decorator and management service)
     */
    private readonly _onChange_lotFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        this.lotFilter.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileInternalStockChange, LotsSites>({
        parent() {
            return this.stockBlock;
        },
        title: 'Sublot',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/LotsSites',
        valueField: 'sublot',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        canFilter: false,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                product: { code: this.productSelected.value?.code },
                storageSite: { code: this.stockSite.value?.code ?? undefined },
            };
        },
        async onChange() {
            await this._onChange_sublotFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
    })
    sublotFilter: ui.fields.Reference<LotsSites>;

    /*
     *  onChange serialNumberFilter event (for decorator and management service)
     */
    private readonly _onChange_sublotFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        this.sublotFilter.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileInternalStockChange, SerialNumber>({
        parent() {
            return this.stockBlock;
        },
        title: 'Serial number',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/SerialNumber',
        valueField: 'code',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        canFilter: false,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                product: { code: this.productSelected.value?.code },
                stockSite: { code: this.stockSite.value?.code ?? undefined },
                issueDocumentId: '',
            };
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.serialNumberFilter, rawData);
        },
        async onChange() {
            await this._onChange_serialNumberFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
        ],
    })
    serialNumberFilter: ui.fields.Reference<SerialNumber>;

    /*
     *  onChange serialNumberFilter event (for decorator and management service)
     */
    private readonly _onChange_serialNumberFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        if (this.serialNumberFilter.value) {
            this.stock.value.forEach(stockLine => {
                if (!this.stock.selectedRecords.find(selectedRecord => selectedRecord === stockLine._id)) {
                    this.stock.selectRecord(stockLine._id);
                    this._onRowSelect(stockLine._id, stockLine as any);
                }
            });
        }
    };

    @ui.decorators.selectField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Status',
        placeholder: 'Scan or select...',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._onChange_statusFilter();
        },
    })
    statusFilter: ui.fields.Select;

    /*
     * onChange statusFilter event (for decorator and management service)
     */
    private readonly _onChange_statusFilter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.selectField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Unit',
        placeholder: 'Scan or select...',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._getStock();
        },
    })
    packingUnitFilter: ui.fields.Select;

    @ui.decorators.numericField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Conversion factor',
        placeholder: 'Scan...',
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        min: 0,
        async onChange() {
            await this._getStock();
        },
    })
    packingUnitToStockUnitConversionFactorFilter: ui.fields.Numeric;

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Identifier 1',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._onChange_identifier1Filter();
        },
    })
    identifier1Filter: ui.fields.Text;

    /*
     * onChange identifier1Filter event (for decorator and management service)
     */
    private readonly _onChange_identifier1Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Identifier 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._onChange_identifier2Filter();
        },
    })
    identifier2Filter: ui.fields.Text;

    /*
     * onChange identifier2Filter event (for decorator and management service)
     */
    private readonly _onChange_identifier2Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Stock custom field 1',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._onChange_stockCustomField1Filter();
        },
    })
    stockCustomField1Filter: ui.fields.Text;

    /*
     * onChange stockCustomField1Filter event (for decorator and management service)
     */
    private readonly _onChange_stockCustomField1Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.stockBlock;
        },
        title: 'Stock custom field 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isMandatory: false,
        isFullWidth: true,
        async onChange() {
            await this._onChange_stockCustomField2Filter();
        },
    })
    stockCustomField2Filter: ui.fields.Text;

    /*
     * onChange stockCustomField2Filter event (for decorator and management service)
     */
    private readonly _onChange_stockCustomField2Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.tableField<MobileInternalStockChange, Stock>({
        parent() {
            return this.stockBlock;
        },
        node: '@sage/x3-stock-data/Stock',
        isChangeIndicatorDisabled: false,
        canFilter: false,
        canSelect: true,
        canExport: false,
        canResizeColumns: false,
        canUserHideColumns: false,
        isTitleHidden: false,
        isTransient: true,
        isFullWidth: true,
        isDisabled: false,
        hasSearchBoxMobile: false,
        cardView: true,
        displayMode: ui.fields.TableDisplayMode.compact,
        mobileCard: undefined,
        orderBy: {
            stockSite: 1,
            stockId: 1,
        },
        columns: [
            ui.nestedFields.numeric({
                bind: 'quantityToMove' as any,
                isReadOnly: true,
                isHidden: false,
                isTransient: true,
                postfix(value, rowValue?: Dict<any>) {
                    return `/ ${ui.formatNumberToCurrentLocale(
                        Number(rowValue?.quantityInPackingUnitOrigin),
                        Number(rowValue?.packingUnit.numberOfDecimals),
                    )} ${String(rowValue?.packingUnit?.code)}`;
                },
                title: 'Quantity to move', // this is important to display a title in the grid row block
                isTitleHidden: false,
                isMandatory: false,
                isFullWidth: true,
                max(rowValue: Stock) {
                    return Number((rowValue as any).quantityInPackingUnitOrigin);
                },
                scale(_value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit?.numberOfDecimals;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnit',
                isHidden: true,
                isTitleHidden: true,
                isTransient: false,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnitRest' as any,
                isHidden: true,
                isTitleHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, ProductSite>({
                bind: 'product',
                valueField: { product: { code: true } },
                node: '@sage/x3-master-data/ProductSite',
                isReadOnly: true,
                isHidden: false,
                columns: [
                    ui.nestedFields.text({
                        bind: { product: { code: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { serialNumberManagementMode: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { _id: true } },
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, LicensePlateNumber>({
                bind: 'licensePlateNumber',
                valueField: 'code',
                node: '@sage/x3-stock-data/LicensePlateNumber',
                isReadOnly: true,
                isHidden: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, Location>({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
                isReadOnly: true,
                isHidden: false,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, Location>({
                bind: 'locationDestination' as any,
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.text({
                bind: 'lot',
                isReadOnly: true,
                isHidden() {
                    return this.lotFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                isReadOnly: true,
                isHidden() {
                    return this.sublotFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.link({
                bind: 'globalSerialNumber' as any,
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'serialNumber',
                isReadOnly: true,
                isHidden() {
                    return this.serialNumberFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, StockStatus>({
                bind: 'status',
                valueField: 'code',
                node: '@sage/x3-stock-data/StockStatus',
                isReadOnly: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
                isHidden() {
                    return this.statusFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'statusDestination' as any,
                isReadOnly: true,
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, UnitOfMeasure>({
                bind: 'packingUnit',
                node: '@sage/x3-master-data/UnitOfMeasure',
                isHidden: true,
                valueField: 'code',
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberOfDecimals',
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, UnitOfMeasure>({
                bind: 'packingUnitDestination' as any,
                node: '@sage/x3-master-data/UnitOfMeasure',
                valueField: 'code',
                isHidden: true,
                isTransient: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberOfDecimals',
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactor',
                isReadOnly: true,
                isHidden() {
                    return this.packingUnitToStockUnitConversionFactorFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactorDestination' as any,
                isHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnitOrigin' as any,
                isHidden: true,
                isTitleHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.text({
                bind: 'identifier1',
                isReadOnly: true,
                isHidden() {
                    return this.identifier1Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'identifier1Destination' as any,
                isReadOnly: true,
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'identifier2',
                isReadOnly: true,
                isHidden() {
                    return this.identifier2Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'identifier2Destination' as any,
                isReadOnly: true,
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField1',
                isReadOnly: true,
                isHidden() {
                    return this.stockCustomField1Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField2',
                isReadOnly: true,
                isHidden() {
                    return this.stockCustomField2Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'qualityAnalysisRequestId',
                title: 'Analysis req.',
                isHidden: true,
            }),
            ui.nestedFields.technical({
                bind: 'stockId',
            }),
            ui.nestedFields.technical<MobileInternalStockChange, Stock, Lot>({
                bind: 'lotReference',
                node: '@sage/x3-stock-data/Lot',
                nestedFields: [
                    ui.nestedFields.date({
                        bind: 'expirationDate',
                    }),
                    ui.nestedFields.date({
                        bind: 'useByDate',
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                    }),
                    ui.nestedFields.reference({
                        bind: 'majorVersion',
                        node: '@sage/x3-master-data/MajorVersionStatus',
                        valueField: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                    }),
                ],
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceExpirationDate' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceUseByDate' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lotReferenceMajorVersion' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.technical({
                bind: 'quantityInStockUnit',
            }),
            ui.nestedFields.technical({
                bind: 'allocatedQuantity',
            }),
            ui.nestedFields.technical({
                bind: 'owner',
            }),
        ],
        onError(error) {
            switch (error.message) {
                case '@sage/x3-stock/serial-number-not-sequential': {
                    return ui.localize(
                        '@sage/x3-stock/serial-number-not-sequential',
                        'The serial numbers are not sequential. Check your entry.',
                    );
                }
                default:
                    return error;
            }
        },
        async onRowSelected(recordId: string, rowItem: Stock) {
            if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                await this._onRowClick(recordId, rowItem);
            } else {
                await this._onRowSelect(recordId, rowItem);
            }
        },
        async onRowUnselected(recordId: string, rowItem: Stock) {
            let currentRecord = this.stock.getRecordValue(recordId);
            if (currentRecord) {
                let i: number = 0;
                while (i < this.stockChangeLines.value.length) {
                    if (
                        Number(this.stockChangeLines.value[i].stockId) === Number(currentRecord.stockId) &&
                        Number(this.stockChangeLines.value[i].numberSession) === Number(this._numberSession)
                    ) {
                        this.stockChangeLines.removeRecord(this.stockChangeLines.value[i]._id, 0);
                    } else {
                        i += 1;
                    }
                }
                (currentRecord as any).quantityToMove = (currentRecord as any).quantityInPackingUnitOrigin;
                this.stock.setRecordValue(currentRecord);
            }
        },
        sortColumns(firstColumn, secondColumn) {
            if (firstColumn.bind === secondColumn.bind) return 0;
            if (firstColumn.bind === 'quantityToMove') {
                return secondColumn.bind === (this._stockFieldSettings[0] as string) ? 1 : -1;
            } else if (secondColumn.bind === 'quantityToMove') {
                return firstColumn.bind === (this._stockFieldSettings[0] as string) ? -1 : 1;
            }

            for (const stockFieldSetting of Object.keys(this._stockFieldSettings)) {
                if (!stockFieldSetting || stockFieldSetting === 'none') break;
                if (firstColumn.bind === (stockFieldSetting as string)) return -1;
                if (secondColumn.bind === (stockFieldSetting as string)) return 1;
            }

            return 1;
        },
        mapServerRecord(record: Partial<Stock>) {
            return {
                ...record,
                quantityInPackingUnitOrigin: this._getQuantityInPackingUnitOrigin(record),
                quantityInPackingUnitRest: this._getQuantityInPackingUnitRest(record),
                quantityToMove: this._getQuantityToMove(record).toString(),
                locationDestination: this._getLocationDestination(record),
                statusDestination: this._getStatusDestination(record),
                packingUnitDestination: this._getPackingUnitDestination(record),
                packingUnitToStockUnitConversionFactorDestination:
                    this._getPackingUnitToStockUnitConversionFactorDestination(record),
                identifier1Destination: this._getIdentifier1Destination(record),
                identifier2Destination: this._getIdentifier2Destination(record),
                lotReferenceExpirationDate: record.lotReference?.expirationDate,
                lotReferenceMajorVersion: record.lotReference?.majorVersion?.code,
                lotReferenceUseByDate: record.lotReference?.useByDate,
                globalSerialNumber: ui.localize('@sage/x3-stock/label-view-all', 'View list'),
            };
        },

        async onRowClick(recordId: string, rowItem: Stock) {
            await this._onRowClick(recordId, rowItem);
        },
    })
    stock: ui.fields.Table<Stock>;

    @ui.decorators.detailListField<MobileInternalStockChange, Stock>({
        parent() {
            return this.detailsBlock;
        },
        node: '@sage/x3-stock-data/Stock',
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
        fields: [
            ui.nestedFields.reference<MobileInternalStockChange, Stock, ProductSite>({
                bind: 'product',
                valueField: { product: { code: true } },
                node: '@sage/x3-master-data/ProductSite',
                title: 'Product',
                isReadOnly: true,
                isHidden: false,
                columns: [
                    ui.nestedFields.text({
                        bind: { product: { code: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { serialNumberManagementMode: true } },
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: { product: { _id: true } },
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, LicensePlateNumber>({
                bind: 'licensePlateNumber',
                title: 'License plate number',
                valueField: 'code',
                node: '@sage/x3-stock-data/LicensePlateNumber',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, Location>({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
                title: 'Location',
                isReadOnly: true,
                isHidden() {
                    return !this._productSite.isLocationManaged;
                },
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'stockId',
                title: 'Stock ID',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'lotReferenceMajorVersion' as any,
                title: 'Major version',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceExpirationDate' as any,
                title: 'Expiration date',
                isReadOnly: true,
                isHidden(value: Date) {
                    return !value || this.productSelected.value?.expirationManagementMode === 'notManaged';
                },
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceUseByDate' as any,
                title: 'Use-by date',
                isReadOnly: true,
                isHidden(value: Date) {
                    return !value || this.productSelected.value?.expirationManagementMode === 'notManaged';
                },
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'lotCustomField1',
                title: 'Lot custom field 1',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return !value?.lotCustomField1;
                },
                columns: [
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'lotCustomField2',
                title: 'Lot custom field 2',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return !value?.lotCustomField2;
                },
                columns: [
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.text({
                bind: 'serialNumber',
                title: 'Serial no.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.link({
                bind: 'globalSerialNumber' as any,
                title: 'Serial no.',
                isHidden() {
                    return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                },
                async onClick(_id, rowData: any) {
                    const options: ui.dialogs.PageDialogOptions = {
                        skipDirtyCheck: true,
                    };
                    try {
                        await this.$.dialog.page(
                            '@sage/x3-stock/MobileGlobalSerialDetails',
                            {
                                product: this.productSelected.value?.code ?? '',
                                stockId: rowData.stockId,
                                subtitle: this.productSelected.value?.localizedDescription1.value ?? '',
                            },
                            options,
                        );
                    } catch (error) {
                        if (error) {
                            this.$.showToast(error.message, { timeout: 10000, type: 'error' });
                        }
                    }
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityToMove' as any,
                title: 'Quantity to move',
                isReadOnly: true,
                isTransient: true,
                isHidden: true,
                postfix(value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit.code;
                },
                scale(value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnit',
                title: 'Packing qty.',
                isReadOnly: true,
                isHidden: false,
                postfix() {
                    return this.stock.getRecordValue(this.gridBlock.selectedRecordId).packingUnit?.code;
                },
                scale() {
                    return (
                        this.stock.getRecordValue(this.gridBlock.selectedRecordId)?.packingUnit?.numberOfDecimals ?? 0
                    );
                },
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, UnitOfMeasure>({
                bind: 'packingUnit',
                node: '@sage/x3-master-data/UnitOfMeasure',
                valueField: 'code',
                title: 'Unit',
                isHidden: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberOfDecimals',
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactor',
                title: 'Conversion factor',
                isHidden: false,
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, UnitOfMeasure>({
                bind: 'packingUnitDestination' as any,
                node: '@sage/x3-master-data/UnitOfMeasure',
                valueField: 'code',
                title: 'Destination unit',
                isHidden: true,
                isTransient: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberOfDecimals',
                        isHidden: true,
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactorDestination' as any,
                title: 'Destination conversion factor',
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInStockUnit',
                title: 'Stock qty.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
                postfix() {
                    return this.productSelected.value?.stockUnit?.code ?? '';
                },
                scale() {
                    return this.productSelected.value?.stockUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'allocatedQuantity',
                title: 'Allocated qty.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
                postfix() {
                    return this.productSelected.value?.stockUnit?.code ?? '';
                },
                scale() {
                    return this.productSelected.value?.stockUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.reference<MobileInternalStockChange, Stock, StockStatus>({
                bind: 'status',
                valueField: 'code',
                title: 'Status',
                isReadOnly: true,
                node: '@sage/x3-stock-data/StockStatus',
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.text({
                bind: 'identifier1',
                title: 'Identifier 1',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'identifier2',
                title: 'Identifier 2',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'qualityAnalysisRequestId',
                title: 'Analysis req.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'owner',
                title: 'Owner',
                isReadOnly: true,
                isHidden: true,
            }),
        ],
    })
    stockDetails: ui.fields.DetailList<Stock>;

    @ui.decorators.numericField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        postfix() {
            return `/ ${ui.formatNumberToCurrentLocale(
                this.detailPanelSection.isHidden
                    ? (this.stock.getRecordValue(this.gridBlock.selectedRecordId) as any)?.quantityInPackingUnitOrigin
                    : (this.stock.getRecordValue(this.gridBlock.selectedRecordId) as any)?.quantityInPackingUnitRest,
                this.stock.getRecordValue(this.gridBlock.selectedRecordId)?.packingUnit?.numberOfDecimals,
            )} ${this.stock.getRecordValue(this.gridBlock.selectedRecordId)?.packingUnit?.code}`;
        },
        title: 'Quantity to move',
        isMandatory: false,
        isFullWidth: true,
        isTransient: true,
        max() {
            return (this.stock.getRecordValue(this.gridBlock.selectedRecordId) as any)?.quantityInPackingUnitRest;
        },
        scale() {
            return (
                (<any>(<unknown>this.stock.getRecordValue(this.gridBlock?.selectedRecordId ?? '')))?.packingUnit
                    ?.numberOfDecimals ?? 0
            );
        },
        async onChange() {
            await this._onChange_quantityToMove();
        },
    })
    quantityToMove: ui.fields.Numeric;

    /*
     * onChange quantityToMove event (for decorator and management service)
     */
    private readonly _onChange_quantityToMove: AsyncVoidFunction = async () => {
        let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
        (currentRecord as any).quantityToMove = String(this.quantityToMove.value);
        this.stock.setRecordValue(currentRecord);
        await this.$.commitValueAndPropertyChanges();
    };

    @ui.decorators.referenceField<MobileInternalStockChange, Location>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Destination location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        isMandatory: false,
        isTransient: true,
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        helperTextField: 'type',
        width: 'large',
        canFilter: false,
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite.value?.code ?? undefined },
                category: 'internal',
            };
            return locationFilter;
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.locationDestination, rawData);
        },
        async onChange() {
            await this._onChange_locationDestination();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    locationDestination: ui.fields.Reference;

    /*
     * onChange locationDestination event (for decorator and management service)
     */
    private readonly _onChange_locationDestination: AsyncVoidFunction = async () => {
        this._setSelectButton();
        if (this.locationDestination.value) {
            this.locationDestination.getNextField(true)?.focus();
        }
    };

    @ui.decorators.selectField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Destination status',
        width: 'small',
        isMandatory: false,
        isTransient: true,
        async onChange() {
            await this._onChange_statusDestination();
        },
    })
    statusDestination: ui.fields.Select;

    /*
     * onChange statusDestination event (for decorator and management service)
     */
    private readonly _onChange_statusDestination: AsyncVoidFunction = async () => {
        this._setSelectButton();
        if (this.statusDestination.value) {
            this.statusDestination.getNextField(true)?.focus();
        }
    };

    @ui.decorators.dropdownListField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Destination unit',
        width: 'small',
        isMandatory: true,
        isTransient: true,
        async onChange() {
            await this._onChange_packingUnitDestination();
        },
    })
    packingUnitDestination: ui.fields.DropdownList;

    /*
     * onChange packingUnitDestination event (for decorator and management service)
     */
    private readonly _onChange_packingUnitDestination: AsyncVoidFunction = async () => {
        if (!this.packingUnitDestination.value) return;

        let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId) as any;
        const selectedValue = this.packingUnitDestination.value;
        const packingUnitIndex = this._packingUnits
            .map(packingUnit => packingUnit.node.packingUnit.code)
            .indexOf(selectedValue);
        if (packingUnitIndex !== -1) {
            const selectedUnit = this._packingUnits[packingUnitIndex].node;
            this.packingUnitToStockUnitConversionFactorDestination.value = Number(
                selectedUnit.packingUnitToStockUnitConversionFactor,
            );
            this.packingUnitToStockUnitConversionFactorDestination.isDisabled = true;
            (currentRecord as any).packingUnitDestination = selectedUnit.packingUnit;
            (currentRecord as any).packingUnitToStockUnitConversionFactorDestination = String(
                selectedUnit.packingUnitToStockUnitConversionFactor,
            );
        } else {
            this.packingUnitToStockUnitConversionFactorDestination.value = 1;
            this.packingUnitToStockUnitConversionFactorDestination.isDisabled = true;
            (currentRecord as any).packingUnitToStockUnitConversionFactorDestination = 1;
        }
        this.stock.setRecordValue(currentRecord);
        await this.$.commitValueAndPropertyChanges();
        this._setSelectButton();

        this.packingUnitDestination.getNextField(true)?.focus();
    };

    @ui.decorators.numericField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Destination conversion factor',
        isDisabled: true,
        isMandatory: false,
        isTransient: true,
        scale: 5,
        async onChange() {
            let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
            (currentRecord as any).packingUnitToStockUnitConversionFactorDestination = String(
                this.packingUnitToStockUnitConversionFactorDestination.value,
            );
            this.stock.setRecordValue(currentRecord);
            await this.$.commitValueAndPropertyChanges();
        },
    })
    packingUnitToStockUnitConversionFactorDestination: ui.fields.Numeric;

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        placeholder: 'Scan...',
        title: 'Destination identifier 1',
        isMandatory: false,
        isTransient: true,
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        width: 'large',
        async onChange() {
            await this._onChange_identifier1Destination();
        },
    })
    identifier1Destination: ui.fields.Text;

    /*
     * onChange identifier1Destination event (for decorator and management service)
     */

    private readonly _onChange_identifier1Destination: AsyncVoidFunction = async () => { };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.quantityBlock;
        },
        placeholder: 'Scan...',
        title: 'Destination identifier 2',
        isMandatory: false,
        isTransient: true,
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        width: 'large',
        async onChange() {
            await this._onChange_identifier2Destination();
        },
    })
    identifier2Destination: ui.fields.Text;

    /*
     * onChange identifier2Destination event (for decorator and management service)
     */
    private readonly _onChange_identifier2Destination: AsyncVoidFunction = async () => { };

    @ui.decorators.referenceField<MobileInternalStockChange, SerialNumber>({
        parent() {
            return this.serialNumberBlock;
        },
        title: 'Starting serial number',
        placeholder: 'Scan or select…',
        node: '@sage/x3-stock-data/SerialNumber',
        valueField: 'code',
        isMandatory: false,
        isTransient: true,
        isFullWidth: true,
        canFilter: false,
        isDisabled: false,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                _and: [
                    { product: { code: this.productSelected.value?.code } },
                    { stockId: this._stockId.value ?? undefined },
                ],
            };
        },
        async onChange() {
            await this._onChange_startingSerialNumber();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                title: 'Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'stockId',
                title: 'Stock ID',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    startingSerialNumber: ui.fields.Reference<SerialNumber>;

    /*
     * onChange startingSerialNumber event (for decorator and management service)
     */
    private readonly _onChange_startingSerialNumber: AsyncVoidFunction = async () => {
        await this.onChangeBody();
    };

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.serialNumberBlock;
        },
        title: 'Ending serial number',
        isMandatory: false,
        isTransient: true,
        isReadOnly: true,
        isFullWidth: true,
        async validation(value: string) {
            if (this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued') return;
            let currentQty = Number(this.quantityToMove.value);
            if (
                (await getCountSerialNumber(
                    this,
                    this.productSelected.value?.code ?? '',
                    this.stockSite.value?.code,
                    this._stockId.value,
                    this.startingSerialNumber.value?.code ?? '',
                    this.endingSerialNumber.value ?? '',
                )) !== currentQty
            ) {
                return ui.localize(
                    '@sage/x3-stock/serial-number-not-sequential',
                    'The serial numbers are not sequential. Check your entry.',
                );
            }
        },
        async onChange() {
            await this._onChange_endingSerialNumber();
        },
    })
    endingSerialNumber: ui.fields.Text;

    /*
     * onChange endingSerialNumber event (for decorator and management service)
     */
    private readonly _onChange_endingSerialNumber: AsyncVoidFunction = async () => { };

    private async _onRowClick(recordId: string, rowItem: Stock) {
        this.stockDetails.value = [rowItem]; // populate details list
        this.gridBlock.selectedRecordId = recordId; // populate grid row block
        this._stockId.value = rowItem.stockId;
        const currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId) as any;
        this.quantityToMove.value = Number(currentRecord?.quantityToMove);
        if (currentRecord?.locationDestination) {
            this.locationDestination.value = currentRecord?.locationDestination;
        } else if (this.mainLocationDestination.value) {
            this.locationDestination.value = this.mainLocationDestination.value;
        } else {
            this.locationDestination.value = null;
        }
        this.locationDestination.isDisabled = currentRecord?.licensePlateNumber;
        if (currentRecord?.statusDestination) {
            this.statusDestination.value = currentRecord?.statusDestination;
        } else if (this.mainStatusDestination.value) {
            this.statusDestination.value = this.mainStatusDestination.value;
        } else {
            this.statusDestination.value = null;
        }
        if (currentRecord?.packingUnitDestination) {
            this.packingUnitDestination.value = currentRecord?.packingUnitDestination.code;
        } else {
            this.packingUnitDestination.value = null;
        }
        this.packingUnitToStockUnitConversionFactorDestination.value =
            currentRecord?.packingUnitToStockUnitConversionFactorDestination;
        if (this.mainIdentifier1Destination.value) {
            this.identifier1Destination.value = this.mainIdentifier1Destination.value;
        } else {
            this.identifier1Destination.value = currentRecord?.identifier1Destination;
        }
        if (this.mainIdentifier2Destination.value) {
            this.identifier2Destination.value = this.mainIdentifier2Destination.value;
        } else {
            this.identifier2Destination.value = currentRecord?.identifier2Destination;
        }
        this._setSelectButton();

        if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
            this.serialNumberLines.isHidden = false;
            this.serialNumberLines.value = [];
            this.stockChangeLines.value.forEach(line => {
                if (
                    Number(line.stockId) === Number(rowItem.stockId) &&
                    line.numberSession === this._numberSession &&
                    line.serialNumber
                ) {
                    this.serialNumberLines.addRecord({
                        quantity: Number(line.quantityInPackingUnitDestination),
                        startingSerialNumber: line.serialNumber,
                        endingSerialNumber: line.endingSerialNumber,
                    });
                }
            });
            let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
            this.stock.setRecordValue(currentRecord);
        }
        await this.$.commitValueAndPropertyChanges();
        await this.stock.validateWithDetails();

        await this._setActiveSection(Section.detailPanel);
    }

    private async _onRowSelect(recordId: string, rowItem: Stock) {
        let currentRecord = this.stock.getRecordValue(recordId);
        if (currentRecord) {
            const line = this.stockChangeLines.value.find(
                line =>
                    Number(line.stockId) === Number(currentRecord.stockId) &&
                    line.numberSession === this._numberSession,
            );
            const lineNumber = line ? line.lineNumber : this.stockChangeLines.value.length;

            if (
                (!this.mainLocationDestination.value ||
                    this.mainLocationDestination.value._id === currentRecord.location?._id) &&
                (!this.mainStatusDestination.value ||
                    this.mainStatusDestination.value === currentRecord.status?.code) &&
                (!this.packingUnitDestination.value ||
                    this.packingUnitDestination.value === currentRecord.packingUnit?.code) &&
                (!this.mainIdentifier1Destination.value ||
                    this.mainIdentifier1Destination.value === currentRecord.identifier1) &&
                (!this.mainIdentifier2Destination.value ||
                    this.mainIdentifier2Destination.value === currentRecord.identifier2)
            ) {
                // If you are on the main section or if a dispatch is in progress, no error message is displayed.
                if (!this.controlManager.isDispatchInProgress && this._getActiveSection() !== Section.main) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                        `${ui.localize(
                            '@sage/x3-stock/pages__mobile_stock_change_lines__destination_is_missing',
                            'Destination is missing.',
                        )}`,
                    );
                }

                this.stock.unselectRecord(recordId);
                return;
            }

            this.stockChangeLines.addOrUpdateRecordValue(
                {
                    _id: `${currentRecord.stockId}|${lineNumber}`,
                    stockSite: this.stockSite.value?.code,
                    lineNumber: lineNumber,
                    product: {
                        code: currentRecord.product?.product?.code,
                        _id: currentRecord.product?.product?._id,
                    },
                    stockId: String(currentRecord.stockId),
                    productDescription: this.productSelected.value.localizedDescription1,
                    packingUnit: {
                        code: currentRecord.packingUnit?.code,
                        _id: currentRecord.packingUnit?._id,
                    },
                    quantityInPackingUnit: currentRecord.quantityInPackingUnit,
                    quantityMoved: Number(currentRecord.quantityInPackingUnit),
                    packingUnitToStockUnitConversionFactor: currentRecord.packingUnitToStockUnitConversionFactor,
                    serialNumber: currentRecord.serialNumber,
                    numberSession: this._numberSession,
                    location: {
                        code: currentRecord.location?.code,
                        _id: currentRecord.location?._id,
                    },
                    status: {
                        code: currentRecord.status?.code,
                        _id: currentRecord.status?._id,
                    },
                    identifier1: currentRecord.identifier1,
                    identifier2: currentRecord.identifier2,
                    ...this.getCustomizationInternalStockChangeLines(),
                },
                0,
                undefined,
            );
            this.stockChangeLines.addOrUpdateRecordValue(
                {
                    _id: `${currentRecord.stockId}|${lineNumber}`,
                    packingUnit: {
                        code: currentRecord.packingUnit?.code,
                        _id: currentRecord.packingUnit?._id,
                    },
                    product: {
                        code: currentRecord.product?.product?.code,
                        _id: currentRecord.product?.product?._id,
                    },
                    packingUnitToStockUnitConversionFactor:
                        currentRecord?.packingUnitToStockUnitConversionFactor?.toString(),
                    quantityInPackingUnit: currentRecord.quantityInPackingUnit,
                    quantityInStockUnit: currentRecord.quantityInStockUnit,
                    quantityMoved: Number(currentRecord.quantityInPackingUnit),
                    ...(currentRecord.licensePlateNumber && {
                        licensePlateNumber: {
                            code: currentRecord.licensePlateNumber.code,
                            _id: currentRecord.licensePlateNumber._id,
                        },
                    }),
                    ...(this._productSite.isLocationManaged && {
                        location: {
                            code: this.mainLocationDestination.value
                                ? this.mainLocationDestination.value.code
                                : currentRecord.location?.code,
                            _id: this.mainLocationDestination.value
                                ? this.mainLocationDestination.value._id
                                : currentRecord.location?._id,
                        },
                    }),
                    ...(this.mainStatusDestination.value && {
                        status: {
                            code: this.mainStatusDestination.value,
                            _id: this.mainStatusDestination.value,
                        },
                    }),
                    ...(!this.mainStatusDestination.value &&
                        currentRecord.status &&
                        this.mainStatusDestination.options?.includes(currentRecord.status.code) && {
                        status: {
                            code: currentRecord.status.code,
                            _id: currentRecord.status._id,
                        },
                    }),
                    identifier1: this.mainIdentifier1Destination.value
                        ? this.mainIdentifier1Destination.value
                        : currentRecord.identifier1,
                    identifier2: this.mainIdentifier2Destination.value
                        ? this.mainIdentifier2Destination.value
                        : currentRecord.identifier2,
                    serialNumber: currentRecord.serialNumber,
                    ...this.getCustomizationInternalStockChangeDetails(),
                },
                1,
                `${currentRecord.stockId}|${lineNumber}`,
            );
            this.transaction.isDisabled = true;
        }
        let qtyTotal = 0;
        this.stockChangeLines.value.forEach(line => {
            if (Number(line.stockId) === Number(currentRecord.stockId) && line.numberSession === this._numberSession) {
                line.stockDetails?.forEach(stockDetail => {
                    qtyTotal =
                        qtyTotal +
                        (Number(stockDetail.quantityInPackingUnit) *
                            Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                        Number(line.packingUnitToStockUnitConversionFactor);
                });
            }
        });
        (currentRecord as any).quantityToMove = qtyTotal;
        (currentRecord as any).quantityInStockUnitDestination =
            qtyTotal * Number((currentRecord as any).packingUnitToStockUnitConversionFactor);
        (currentRecord as any).locationDestination = this.mainLocationDestination.value
            ? this.mainLocationDestination.value
            : currentRecord?.location;
        (currentRecord as any).statusDestination = this.mainStatusDestination.value
            ? this.mainStatusDestination.value
            : currentRecord?.status?.code;
        (currentRecord as any).packingUnitDestination =
            this.packingUnitDestination.value !== '' ? { code: this.packingUnitDestination.value } : null;
        (currentRecord as any).packingUnitToStockUnitConversionFactorDestination = this
            .packingUnitToStockUnitConversionFactorDestination.value
            ? this.packingUnitToStockUnitConversionFactorDestination.value
            : (currentRecord as any).packingUnitToStockUnitConversionFactor;
        (currentRecord as any).identifier1Destination = this.mainIdentifier1Destination.value
            ? this.mainIdentifier1Destination.value
            : currentRecord?.identifier1;
        (currentRecord as any).identifier2Destination = this.mainIdentifier2Destination.value
            ? this.mainIdentifier2Destination.value
            : currentRecord?.identifier2;
        this.stock.setRecordValue(currentRecord);
    }

    @ui.decorators.textField<MobileInternalStockChange>({
        parent() {
            return this.serialNumberBlock;
        },
        isTransient: true,
        isHidden: true,
    })
    _stockId: ui.fields.Text;

    @ui.decorators.tableField<MobileInternalStockChange>({
        parent() {
            return this.listSerialNumberBlock;
        },
        canFilter: false,
        canSelect: false,
        canExport: false,
        canResizeColumns: false,
        canUserHideColumns: false,
        isTitleHidden: false,
        isTransient: true,
        isFullWidth: true,
        isDisabled: false,
        node: '@sage/x3-stock/StockCountSerialNumber',
        mobileCard: undefined,
        columns: [
            ui.nestedFields.text({
                bind: 'startingSerialNumber',
                title: 'Starting serial Number',
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantity',
                title: 'Quantity',
                isReadOnly: true,
                postfix() {
                    return this.stock.getRecordValue(this.gridBlock.selectedRecordId)?.packingUnit?.code;
                },
                scale() {
                    return (
                        this.stock.getRecordValue(this.gridBlock.selectedRecordId ?? '')?.packingUnit
                            ?.numberOfDecimals ?? 0
                    );
                },
            }),
            ui.nestedFields.text({
                bind: 'endingSerialNumber',
                title: 'Ending Serial Number',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
        dropdownActions: [
            {
                icon: 'bin',
                title: 'Delete',
                onClick(recordId: string) {
                    const removedRecordSerialNumber = this.serialNumberLines.getRecordValue(recordId);
                    const removedIdSerialNumber = this.stockChangeLines.value.find(
                        number => number.serialNumber === removedRecordSerialNumber.startingSerialNumber,
                    )?._id;
                    this.stockChangeLines.removeRecord(removedIdSerialNumber, 0);

                    const currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
                    if (currentRecord) {
                        (currentRecord as any).quantityToMove = String(0);
                        this.stock.setRecordValue(currentRecord);
                    }

                    //remove from the card
                    this.serialNumberLines.removeRecord(recordId);
                    this.startingSerialNumber.isDisabled = false;
                },
            },
        ],
        fieldActions() {
            return [this.addSerialRangeButton];
        },
    })
    serialNumberLines: ui.fields.Table<any>;

    private _setSelectButton() {
        const currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId ?? '') as any;
        this.SelectButton.isDisabled =
            !currentRecord ||
            ((!this.locationDestination.value || this.locationDestination.value._id === currentRecord.location._id) &&
                (!this.statusDestination.value || this.statusDestination.value === currentRecord.status.code) &&
                (!this.packingUnitDestination.value ||
                    this.packingUnitDestination.value === currentRecord.packingUnit.code) &&
                (!this.mainIdentifier1Destination.value ||
                    this.mainIdentifier1Destination.value === currentRecord.identifier1) &&
                (!this.mainIdentifier2Destination.value ||
                    this.mainIdentifier2Destination.value === currentRecord.identifier2));
    }

    private async _validateSerialNumber(
        stockDetail: StockJournalWithQuantityMoved | ExtractEdgesPartial<StockJournalWithQuantityMoved>,
    ): Promise<string | undefined> {
        if (
            (await getCountSerialNumber(
                this,
                this._productSite.product.code,
                this.stockSite.value?.code,
                this._stockId.value,
                stockDetail.serialNumber ?? '',
                stockDetail.endingSerialNumber ?? '',
            )) !== Number(stockDetail.quantityInPackingUnit)
        ) {
            return ui.localize(
                '@sage/x3-stock/serial-number-not-sequential',
                'The serial numbers are not sequential. Check your entry.',
            );
        }
        if (stockDetail.serialNumber && stockDetail.endingSerialNumber) {
            let serialNumberAlreadyUsed: boolean = false;
            const startNumberToAdd = Number(stockDetail.serialNumber.match(/\d+$/));
            const endNumberToAdd = Number(stockDetail.endingSerialNumber.match(/\d+$/));
            this.stockChangeLines.value.forEach(line => {
                line.stockDetails?.forEach(item => {
                    if (
                        item.product?.code === this._productSite.product.code &&
                        item.serialNumber &&
                        item.endingSerialNumber &&
                        item._id !== stockDetail._id
                    ) {
                        const startingSerialNumber = Number(item.serialNumber.match(/\d+$/));
                        const endingSerialNumber = Number(item.endingSerialNumber.match(/\d+$/));
                        if (startNumberToAdd <= endingSerialNumber && endNumberToAdd >= startingSerialNumber) {
                            serialNumberAlreadyUsed = true;
                            return ui.localize(
                                '@sage/x3-stock/serial-number-range-overlap',
                                'The serial numbers are overlapping. Enter another starting or ending serial number.',
                            );
                        }
                    }
                });
            });
            if (serialNumberAlreadyUsed) {
                return ui.localize(
                    '@sage/x3-stock/serial-number-range-overlap',
                    'The serial numbers are overlapping. Enter another starting or ending serial number.',
                );
            }
        }
        return undefined;
    }

    private _getQuantityInPackingUnitOrigin(record: Partial<Stock>): Number {
        if ((record as any).quantityInPackingUnitOrigin) {
            return (record as any).quantityInPackingUnitOrigin;
        } else {
            let _quantityInPackingUnitOrigin: Number = Number(record.quantityInPackingUnit);
            this.stockChangeLines.value?.forEach(line => {
                if (Number(line.stockId) === Number(record.stockId) && line.numberSession !== this._numberSession) {
                    _quantityInPackingUnitOrigin =
                        Number(_quantityInPackingUnitOrigin) -
                        (Number(line.stockDetails[0].quantityInPackingUnit) *
                            Number(line.stockDetails[0].packingUnitToStockUnitConversionFactor)) /
                        Number(line.packingUnitToStockUnitConversionFactor);
                }
            });
            return _quantityInPackingUnitOrigin;
        }
    }

    private _getQuantityInPackingUnitRest(record: Partial<Stock>): Number {
        let _quantityInPackingUnitRest: Number = this._getQuantityInPackingUnitOrigin(record);
        if (this.productSelected.value?.serialNumberManagementMode === 'globalReceivedIssued') {
            this.stockChangeLines.value?.forEach(line => {
                if (Number(line.stockId) === Number(record.stockId) && line.numberSession === this._numberSession) {
                    line.stockDetails?.forEach(stockDetail => {
                        _quantityInPackingUnitRest = Number(
                            Number(_quantityInPackingUnitRest) -
                            (Number(stockDetail.quantityInPackingUnit) *
                                Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                            Number(line.packingUnitToStockUnitConversionFactor),
                        );
                    });
                }
            });
        }
        return _quantityInPackingUnitRest;
    }

    private _getStockChangeLine(record: Partial<Stock>): any | undefined {
        const _return = this.stockChangeLines.value?.find(
            line => Number(line.stockId) === Number(record.stockId) && line.numberSession === this._numberSession,
        );
        return _return;
    }

    private _getQuantityToMove(record: Partial<Stock>): Number {
        let _quantityToMove: Number;
        _quantityToMove = 0;
        this.stockChangeLines.value?.forEach(line => {
            if (Number(line.stockId) === Number(record.stockId) && line.numberSession === this._numberSession) {
                line.stockDetails?.forEach(stockDetail => {
                    _quantityToMove =
                        Number(_quantityToMove) +
                        (Number(stockDetail.quantityInPackingUnit) *
                            Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                        Number(record.packingUnitToStockUnitConversionFactor);
                });
            }
        });
        if (_quantityToMove === 0) {
            _quantityToMove = Number(record.quantityInPackingUnit);
        }
        const _quantityInPackingUnitRest = this._getQuantityInPackingUnitRest(record);
        /*    if (Number(_quantityToMove) > Number(_quantityInPackingUnitRest)) {
            _quantityToMove = Number(_quantityInPackingUnitRest);
        } */
        return _quantityToMove;
    }

    private _getLocationDestination(record: Partial<Stock>): Partial<Location> | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return line.stockDetails[0].location;
        } else {
            if (!!record.licensePlateNumber) {
                return record.location;
            } else {
                return undefined;
            }
        }
    }

    private _getPackingUnitToStockUnitConversionFactorDestination(record: Partial<Stock>): Number | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return Number(line.stockDetails[0].packingUnitToStockUnitConversionFactor);
        } else {
            return Number(record.packingUnitToStockUnitConversionFactor);
        }
    }

    private _getIdentifier1Destination(record: Partial<Stock>): String | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return line.stockDetails[0].identifier1;
        } else {
            return record.identifier1;
        }
    }

    private _getIdentifier2Destination(record: Partial<Stock>): String | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return line.stockDetails[0].identifier2;
        } else {
            return record.identifier2;
        }
    }

    private _getPackingUnitDestination(record: Partial<Stock>): Partial<UnitOfMeasure> | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return {
                code: line.stockDetails[0].packingUnit.code,
                _id: line.stockDetails[0].packingUnit._id,
                numberOfDecimals: line.stockDetails[0].packingUnit.numberOfDecimals,
            };
        } else {
            return record.packingUnit;
        }
    }

    private _getStatusDestination(record: Partial<Stock>): String | undefined {
        const line = this._getStockChangeLine(record);
        if (line) {
            return line.stockDetails[0].status?.code;
        } else {
            return '';
        }
    }
    private _calculateEndingSerialNumber(startingSerialNumber: string, quantity: number): string {
        return startingSerialNumber.replace(/\d+$/, match => {
            const endingNumber = (Number(match) + quantity - 1).toString();
            const lengthDiff = Math.max(endingNumber.length - match.length, 0);
            return endingNumber.padStart(match.length + lengthDiff, '0');
        });
    }

    private async onChangeBody() {
        let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId);
        let currentQty = Number(this.quantityToMove.value);
        if (!currentQty || !this.startingSerialNumber.value) {
            this.endingSerialNumber.value = null;
            return;
        }

        this.startingSerialNumber.value.code = this.startingSerialNumber.value.code.toUpperCase();
        if (currentQty > 1) {
            this.endingSerialNumber.value = this._calculateEndingSerialNumber(
                this.startingSerialNumber.value.code,
                currentQty,
            );
        } else {
            this.endingSerialNumber.value = this.startingSerialNumber.value.code;
        }
        if (currentQty > (currentRecord as any).quantityInPackingUnitOrigin) this.addSerialRangeButton.isHidden = true;
        else {
            this.addSerialRangeButton.isHidden = false;
        }
        await this.$.commitValueAndPropertyChanges();
        const validationResult = await this.endingSerialNumber.validate();
        if (validationResult) {
            this.$.showToast(validationResult, { type: 'warning' });
        }
    }

    private async _getProductSite(productCode: string): Promise<ProductSite> {
        const response = await this.$.graph
            .node('@sage/x3-master-data/ProductSite')
            .read(
                {
                    isLocationManaged: true,
                    isLicensePlateNumberManaged: true,
                    defaultInternalContainer: {
                        code: true,
                    },
                    _id: true,
                    product: {
                        _id: true,
                        code: true,
                        localizedDescription1: true,
                        lotManagementMode: true,
                        serialNumberManagementMode: true,
                        stockVersionMode: true,
                        expirationManagementMode: true,
                        serialSequenceNumber: true,
                        lotSequenceNumber: true,
                        stockManagementMode: true,
                        defaultPotencyInPercentage: true,
                        expirationLeadTime: true,
                        expirationTimeUnit: true,
                        useByDateCoefficient: true,
                        stockUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        productCategory: {
                            code: true,
                            _id: true,
                        },
                        packingUnits: {
                            query: {
                                edges: {
                                    node: {
                                        packingUnit: {
                                            code: true,
                                            numberOfDecimals: true,
                                            _id: true,
                                        },
                                        packingUnitToStockUnitConversionFactor: true,
                                        isPackingFactorEntryAllowed: true,
                                        _id: true,
                                    },
                                },
                            },
                        },
                    },
                },
                `${productCode}|${this.stockSite.value?.code}`,
            )
            .execute();

        return response as ProductSite;
    }

    private async _getStock() {
        if (!this.productSelected.value) {
            this.stock.value = [];
            return;
        }
        const response = await this.$.graph
            .node('@sage/x3-stock-data/Stock')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        quantityInPackingUnit: true,
                        product: {
                            product: {
                                code: true,
                                serialNumberManagementMode: true,
                                localizedDescription1: true,
                                _id: true,
                            },
                            _id: true,
                        },
                        licensePlateNumber: {
                            code: true,
                            _id: true,
                        },
                        location: {
                            code: true,
                            _id: true,
                        },
                        lot: true,
                        sublot: true,
                        serialNumber: true,
                        status: {
                            code: true,
                            _id: true,
                        },
                        packingUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        packingUnitToStockUnitConversionFactor: true,
                        identifier1: true,
                        identifier2: true,
                        stockCustomField1: true,
                        stockCustomField2: true,
                        stockId: true,
                        lotReference: {
                            expirationDate: true,
                            useByDate: true,
                            lotCustomField1: true,
                            lotCustomField2: true,
                            majorVersion: {
                                code: true,
                                _id: true,
                            },
                            _id: true,
                        },
                        quantityInStockUnit: true,
                        allocatedQuantity: true,
                        qualityAnalysisRequestId: true,
                        owner: true,
                    },
                    {
                        filter: {
                            availableQuantity: { _gt: '0' },
                            isBeingCounted: false,
                            owner: this.stockSite.value?.code,
                            product: { product: { code: this.productSelected.value?.code } },
                            ...(this.locationSelected.value && {
                                location: { code: this.locationSelected.value.code },
                            }),
                            ...(!this.locationSelected.value &&
                                this._productSite.isLocationManaged &&
                                this.getLocationInMobileSettings() &&
                                this.locationFilter.value && { location: { code: this.locationFilter.value.code } }),
                            ...(!this.lotFilter.isHidden && this.lotFilter.value && { lot: this.lotFilter.value.lot }),
                            ...(!this.sublotFilter.isHidden &&
                                this.sublotFilter.value && { sublot: this.sublotFilter.value.sublot }),
                            ...(!this.serialNumberFilter.isHidden &&
                                this.serialNumberFilter.value && { serialNumber: this.serialNumberFilter.value.code }),
                            ...(!this.statusFilter.isHidden &&
                                this.statusFilter.value && { status: { code: this.statusFilter.value } }),
                            ...(!this.packingUnitFilter.isHidden &&
                                this.packingUnitFilter.value && {
                                packingUnit: { code: this.packingUnitFilter.value },
                            }),
                            ...(!this.packingUnitToStockUnitConversionFactorFilter.isHidden &&
                                this.packingUnitToStockUnitConversionFactorFilter.value && {
                                packingUnitToStockUnitConversionFactor:
                                    this.packingUnitToStockUnitConversionFactorFilter.value,
                            }),
                            ...(!this.identifier1Filter.isHidden &&
                                this.identifier1Filter.value && { identifier1: this.identifier1Filter.value }),
                            ...(!this.identifier2Filter.isHidden &&
                                this.identifier2Filter.value && { identifier2: this.identifier2Filter.value }),
                            ...(!this.stockCustomField1Filter.isHidden &&
                                this.stockCustomField1Filter.value && {
                                stockCustomField1: this.stockCustomField1Filter.value,
                            }),
                            ...(!this.stockCustomField2Filter.isHidden &&
                                this.stockCustomField2Filter.value && {
                                stockCustomField2: this.stockCustomField2Filter.value,
                            }),
                            ...(!this.licensePlateNumberFilter.isHidden &&
                                this.licensePlateNumberFilter.value && {
                                licensePlateNumber: this.licensePlateNumberFilter.value.code,
                            }),
                        },
                        first: 1000,
                    },
                ),
            )
            .execute();
        this.stock.value = response.edges.map(stockLine => {
            return {
                _id: stockLine?.node?._id,
                quantityInPackingUnit: stockLine?.node?.quantityInPackingUnit,
                product: {
                    product: stockLine?.node?.product?.product
                        ? {
                            code: stockLine?.node?.product?.product?.code,
                            serialNumberManagementMode: stockLine?.node?.product?.product?.serialNumberManagementMode,
                            _id: stockLine?.node?.product?.product?._id,
                        }
                        : undefined,
                },
                licensePlateNumber: stockLine?.node?.licensePlateNumber
                    ? {
                        code: stockLine?.node?.licensePlateNumber?.code,
                        _id: stockLine?.node?.licensePlateNumber?._id,
                    }
                    : undefined,
                location: stockLine?.node?.location
                    ? {
                        code: stockLine?.node?.location?.code,
                        _id: stockLine?.node?.location?._id,
                    }
                    : undefined,
                lot: stockLine?.node?.lot,
                sublot: stockLine?.node?.sublot,
                serialNumber: stockLine?.node?.serialNumber,
                status: stockLine?.node?.status
                    ? {
                        code: stockLine?.node?.status?.code,
                        _id: stockLine?.node?.status?._id,
                    }
                    : undefined,
                packingUnit: stockLine?.node?.packingUnit
                    ? {
                        code: stockLine?.node?.packingUnit?.code,
                        numberOfDecimals: stockLine?.node?.packingUnit?.numberOfDecimals,
                        _id: stockLine?.node?.packingUnit?._id,
                    }
                    : undefined,
                packingUnitToStockUnitConversionFactor: stockLine?.node?.packingUnitToStockUnitConversionFactor,
                identifier1: stockLine?.node?.identifier1,
                identifier2: stockLine?.node?.identifier2,
                stockCustomField1: stockLine?.node?.stockCustomField1,
                stockCustomField2: stockLine?.node?.stockCustomField2,
                stockId: stockLine?.node?.stockId,

                lotReference: stockLine?.node?.lotReference
                    ? {
                        expirationDate: stockLine?.node?.lotReference?.expirationDate,
                        useByDate: stockLine?.node?.lotReference?.useByDate,
                        lotCustomField1: stockLine?.node?.lotReference?.lotCustomField1,
                        lotCustomField2: stockLine?.node?.lotReference?.lotCustomField2,
                        majorVersion: stockLine?.node?.lotReference?.majorVersion
                            ? {
                                code: stockLine?.node?.lotReference?.majorVersion?.code,
                                _id: stockLine?.node?.lotReference?.majorVersion?._id,
                            }
                            : undefined,
                        _id: stockLine?.node?.lotReference._id,
                    }
                    : undefined,
                quantityInStockUnit: stockLine?.node?.quantityInStockUnit,
                allocatedQuantity: stockLine?.node?.allocatedQuantity,
                qualityAnalysisRequestId: stockLine?.node?.qualityAnalysisRequestId,
                owner: stockLine?.node?.owner,
            };
        });
    }

    private async _getLocationProducts() {
        const response = await this.$.graph
            .node('@sage/x3-stock-data/Stock')
            .aggregate.query(
                aggregateEdgesSelector<Stock, AggregateGroupSelector<Stock>, AggregateValuesSelector<Stock>>(
                    {
                        group: {
                            product: {
                                product: {
                                    code: {
                                        _by: 'value',
                                    },
                                },
                            },
                        },
                        values: {
                            _id: {
                                distinctCount: true,
                            },
                        },
                    },
                    {
                        filter: {
                            location: {
                                code: this.locationSelected.value?.code,
                            },
                            stockSite: {
                                code: this.stockSite.value?.code,
                            },
                        },
                        first: 500,
                    },
                ),
            )
            .execute();
        this._locationProducts = [];
        if (response.edges.length >= 1) {
            response.edges.forEach(item => {
                this._locationProducts.push(item.node.group.product.product.code);
            });
        }
    }
    private _getSidebarInternalStockChangeLine(): ui.SidebarFieldDefinition<StockChangeLineWithQuantityMoved>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockChangeLineWithQuantityMoved>[] = [];
        this.customizationInternalStockChangeLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockChangeLineWithQuantityMoved>);
            }
        });
        return _FieldsSidebar;
    }

    private _getSidebarInternalStockChangeDetail(): ui.SidebarFieldDefinition<StockJournalWithQuantityMoved>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockJournalWithQuantityMoved>[] = [];
        _FieldsSidebar.push('quantityMoved');
        _FieldsSidebar.push('location');
        _FieldsSidebar.push('status');
        _FieldsSidebar.push('packingUnit');
        _FieldsSidebar.push('packingUnitToStockUnitConversionFactor');
        if (!['notManaged', 'issued'].includes(this._productSite.product.serialNumberManagementMode)) {
            _FieldsSidebar.push('serialNumber');
        }
        if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
            _FieldsSidebar.push('endingSerialNumber');
        }
        _FieldsSidebar.push('identifier1');
        _FieldsSidebar.push('identifier2');
        this.customizationInternalStockChangeDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockJournalWithQuantityMoved>);
            }
        });
        return _FieldsSidebar;
    }

    private getCustomizationInternalStockChangeLines() {
        let response = {};
        this.customizationInternalStockChangeLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private getCustomizationInternalStockChangeDetails() {
        let response = {};
        this.customizationInternalStockChangeDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    /**
     *
     * Technical composite data methods
     *
     */

    /**
     * Initialize ControlManager
     * @returns true when ControlManager is usable
     */
    private async _initControlManager(site: string): Promise<boolean> {
        return await this.createAndInitServiceManagement(
            site,
            mobileApplicationKey,
            BarcodeTransactionFlow.stockChange,
            {
                [DataTitle.gtin]: {
                    mainField: this.productSelected,
                },
                [DataTitle.batchLot]: {
                    mainField: this.lotFilter,
                },
                [DataTitle.serial]: {
                    mainField: this.serialNumberFilter,
                },
            } as DictionaryFieldSupported,
            undefined,
            this._checkCompositeDataAllowed,
        );
    }

    /**
     * This asynchronous readonly function check if current composite data is valid or not
     * @param dictionaryDataComposite : dictionary of composite data block.
     * @param scannedFieldId : the scanned field id
     * @returns false when data must be discarded
     */
    /** @internal */
    private readonly _checkCompositeDataAllowed: AsyncCompositeAllowed = async (
        dictionaryDataComposite: DictionaryDataGs1 | DictionaryDataGeneric,
        scannedFieldId?: string,
    ) => {
        if (this._getActiveSection() !== Section.main) {
            const _productCode = <string | undefined>this.productSelected.value?.code;
            if (_productCode) {
                switch (this.typeOfServiceBarcode) {
                    case <TypeOfServiceBarcode>'gs1':
                        const _productGs1 = dictionaryDataComposite[DataTitle.gtin]?.data;
                        const _globalTradeItemNumber = <string | undefined>(
                            this.productSelected.value?.globalTradeItemNumber
                        );

                        if (
                            _globalTradeItemNumber &&
                            ((!_productGs1 && Object.keys(dictionaryDataComposite).length) ||
                                (_productGs1 && _globalTradeItemNumber !== _productGs1))
                        ) {
                            await this.$.sound.error();
                            await dialogMessage(
                                this,
                                'error',
                                ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                                ui.localize(
                                    '@sage/x3-stock/dialog-error-gs-1-data-is-not-related-to-product',
                                    'The GS1 data is not related to the product: {{ product }}.',
                                    { product: _productCode },
                                ),
                                {
                                    fullScreen: false,
                                },
                            );
                            return false;
                        }
                        break;

                    case <TypeOfServiceBarcode>'composite':
                        const _genericDictionary = <DictionaryDataGeneric>(<unknown>dictionaryDataComposite);
                        const _productComposite =
                            _genericDictionary[StockChangeDictionaryFieldKey.productSelected]?.data;

                        // We check that we have not provided a GTIN, UPC or other code to identify the product...
                        if (_productComposite && _productCode !== _productComposite) {
                            try {
                                if (
                                    (await this.productSelected.fetchSuggestions(<string>_productComposite))[0]
                                        ?.code === _productCode
                                ) {
                                    break;
                                }
                            } catch (error) {}
                        }

                        if (
                            (!_productComposite && Object.keys(dictionaryDataComposite).length) ||
                            (_productComposite && _productCode !== _productComposite)
                        ) {
                            await this.$.sound.error();
                            await dialogMessage(
                                this,
                                'error',
                                ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                                ui.localize(
                                    '@sage/x3-stock/dialog-error-composite-data-is-not-related-to-product',
                                    'The composite data is not related to the product: {{ product }}.',
                                    { product: _productCode },
                                ),
                                {
                                    fullScreen: false,
                                },
                            );
                            return false;
                        }
                        break;

                    default:
                        break;
                }
            }
        }
        return true;
    };

    /**
     *
     * Technical screen management methods
     *
     */

    /**
     *  @returns the active section
     */
    private _getActiveSection(): Section {
        if (!this.transactionSection.isHidden) {
            return Section.main;
        } else if (!this.stockPanelSection.isHidden) {
            return Section.stockPanel;
        } else if (!this.detailPanelSection.isHidden) {
            return Section.detailPanel;
        } else {
            return Section.main;
        }
    }

    /**
     *
     * @param newSection
     * @returns  the old active section
     */
    private async _setActiveSection(newSection: Section): Promise<Section> {
        const _previousSection = this._getActiveSection();
        let _newSection = newSection;
        // validate the new section
        switch (_newSection) {
            case Section.main:
                this.productSelected.value = null;
                this.locationSelected.isHidden = !this.getLocationInMobileSettings();
                this.locationSelected.value = null;
                this.packingUnitDestination.value = null;
                this._resetFieldFilters();
                break;
            case Section.stockPanel:
                this.locationFilter.isHidden =
                    !this._productSite.isLocationManaged ||
                    !this.getLocationInMobileSettings() ||
                    !!this.locationSelected.value;
                break;
            case Section.detailPanel:
                break;

            default:
                _newSection = Section.main;
                break;
        }

        // hide/show buttons
        this.createButton.isHidden = _newSection !== Section.main;
        this.nextButton.isHidden =
            (_newSection !== Section.main && _newSection !== Section.stockPanel) ||
            (_newSection === Section.main && !this.productSelected.value);
        this.CancelButton.isHidden = _newSection !== Section.detailPanel;
        this.SelectButton.isHidden = _newSection !== Section.detailPanel;

        this.createButton.isDisabled = this.stockChangeLines.value.length === 0 || _newSection !== Section.main;

        /**
         * To avoid display effects, we prefer to fill in the fields before showing the sections.
         */

        // TODO: fill in the fields for the section selected

        // hide/show sections
        this.transactionSection.isHidden = _newSection !== Section.main;
        this.fromSection.isHidden = _newSection !== Section.main;
        this.toSection.isHidden = _newSection !== Section.main;
        this.stockChangeLinesSection.isHidden = _newSection !== Section.main;
        this.stockPanelSection.isHidden = _newSection !== Section.stockPanel;
        this.detailPanelSection.isHidden = _newSection !== Section.detailPanel;
        this.stockChangeLinesBlock.isHidden = _newSection !== Section.main;

        await this.$.commitValueAndPropertyChanges();

        // Action to do after change
        switch (_newSection) {
            case Section.main:
                if (!this.productSelected.value) {
                    this.productSelected.focus();
                }
                break;
            case Section.stockPanel:
                break;
            case Section.detailPanel:
                break;

            default:
                break;
        }

        return _previousSection;
    }

    private _resetFieldFilters() {
        const filters = [
            this.licensePlateNumberFilter,
            this.locationFilter,
            this.lotFilter,
            this.sublotFilter,
            this.serialNumberFilter,
            this.statusFilter,
            this.packingUnitFilter,
            this.packingUnitToStockUnitConversionFactorFilter,
            this.identifier1Filter,
            this.identifier2Filter,
            this.stockCustomField1Filter,
            this.stockCustomField2Filter,
        ];

        filters.forEach(filter => {
            if (!filter.isHidden) {
                filter.value = null;
            }
        });
    }
}
