import { AutomaticJournal } from '@sage/x3-finance-data-api';
import {
    Container,
    MajorVersionStatus,
    Product,
    ProductSite,
    Supplier,
    SupplierProduct,
    UnitOfMeasure,
} from '@sage/x3-master-data-api';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import { PurchaseEntryTransaction, PurchaseReceiptLineBinding } from '@sage/x3-purchasing-api';
import { SalesDelivery, SalesDeliveryLine } from '@sage/x3-sales-api';
import {
    LicensePlateNumber,
    Location,
    StockJournalBinding,
    StockManagementRules,
    StockStatus,
    Warehouse,
} from '@sage/x3-stock-data-api';
import { receipt } from '@sage/x3-stock-data/build/lib/menu-items/receipt';
import { MiscellaneousTable, Site } from '@sage/x3-system-api';
import { getRegExp } from '@sage/x3-system/lib/shared-functions/pat-converter';
import { ExtractEdgesPartial, Filter, extractEdges } from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import { Dict, MAX_INT_32 } from '@sage/xtrem-shared';
import * as ui from '@sage/xtrem-ui';
import { getUnitNumberOfDecimalList } from '../client-functions/get-unit-number-decimals';
import { readParameterValue } from '../client-functions/read-parameter';
import { findDefaultLocation, findStockManagementRules } from '../client-functions/stock-management-rules';

export type DeepPartial<T> = T extends Object ? { [K in keyof T]?: DeepPartial<T[K]> } : T;
type PartialPurchaseEntryTransaction = DeepPartial<PurchaseEntryTransaction>;

type packingUnit = {
    node: {
        packingUnit: {
            code: string;
            numberOfDecimals: number;
            _id: string;
        };
        packingUnitToStockUnitConversionFactor: string;
        isPackingFactorEntryAllowed: boolean;
        _id: string;
        isPackingUnit: boolean;
    };
};

type SalesDeliveryLineNodeResponse = {
    _id: string;
    salesDelivery: {
        id: string;
        salesSite: {
            code: string;
            _id: string;
        };
        receivingSite: {
            code: string;
            _id: string;
        };
        _id: string;
    };
    lineNumber: number;
    stockUnit: {
        code: string;
        numberOfDecimals: number;
        _id: string;
    };
    salesUnit: {
        code: string;
        numberOfDecimals: number;
        _id: string;
    };
    product: {
        code: string;
        _id: string;
    };
    salesOrderLine: {
        purchaseOrder: {
            orderFromSupplier: {
                code: {
                    code: string;
                    _id: string;
                };
                _id: string;
            };
            _id: string;
        };
        _id: string;
    };
    quantityInStockUnitReceived: string;
    quantityInSalesUnitDelivered: string;
    salesUnitToStockUnitConversionFactor: string;
    majorProductVersion: {
        code: string;
        _id: string;
    };
    minorProductVersion: string;
};

type StockJournalNodeResponse =
    // StockJournalBinding;
    {
        _id: string;
        stockSite: {
            code: string;
            _id: string;
        };
        documentId: string;
        documentLine: number;
        sequence: number;
        product: {
            code: string;
            _id: string;
        };
        quantityInStockUnit: number;
        quantityInPackingUnit: number;
        packingUnit: {
            code: string;
            numberOfDecimals: number;
            _id: string;
        };
        isUpdated: boolean;
        lot: string;
        sublot: string;
        supplierLot: string;
        lotCustomField1: string;
        lotCustomField2: string;
        lotCustomField3: string;
        lotCustomField4: string;
        majorVersion: {
            code: string;
            _id: string;
        };
        minorVersion: string;
        identifier1: string;
        identifier2: string;
        expirationDate: string;
        useByDate: string;
        status: string;
        serialNumber: string;
        startingSerialNumber: string;
        endingSerialNumber: string;
    };

@ui.decorators.page<MobileIntersiteReceipt>({
    title: 'Intersite receipt',
    module: 'x3-purchasing-sales',
    mode: 'default',
    menuItem: receipt,
    priority: 150,
    node: '@sage/x3-purchasing/PurchaseReceipt',
    isTitleHidden: true,
    authorizationCode: 'CWSPTH',
    access: { node: '@sage/x3-purchasing/PurchaseReceipt' },
    navigationPanel: undefined,
    headerCard() {
        if (this.destinationBlock.isHidden) {
            return undefined;
        } else {
            return {
                title: this.headerTitleLeft,
                titleRight: this.headerTitleRight,
                line2: this.headerLine2Left,
                line2Right: this.headerLine2Right,
                line3: this.headerLine3Left,
                line3Right: this.headerLine3Right,
            };
        }
    },

    async onLoad() {
        if (!(await this._init())) {
            this._disablePage();
        }
    },
    businessActions() {
        return [this.createButton, this.detailsNextButton, this.addButton];
    },
})
export class MobileIntersiteReceipt extends ui.Page {
    /*
     *
     *  Technical properties
     *
     */
    private _entryTransactionMap: Map<string, PartialPurchaseEntryTransaction>;
    private _selectedEntryTransaction: PartialPurchaseEntryTransaction | undefined;
    private _numberOfDecimalList: ExtractEdgesPartial<UnitOfMeasure>[];
    private _isLocationPreloaded: boolean;
    private _product: Product;
    private _productSite: ProductSite;
    private _productSalesSite: ProductSite;
    private _supplierProduct: SupplierProduct | null;
    private _packingUnits: packingUnit[];
    private _isOverReceiptAllowed: boolean | undefined;
    private _purchaseStockManagementRules: StockManagementRules;
    private _salesUnitToStockUnitConversionFactor: string;
    private _previousSalesUnitToStockUnitConversionFactor: string;
    private _quantityToReceiveNoRounded: number;
    private _quantityToReceiveRounded: number;
    private _previousQuantityToReceive: number;
    private _remainingQuantityNoRounded: number;
    private _remainingQuantityRounded: number;
    private _remainingQuantityInitial: number;
    private _previousLocationCode: string | undefined;
    private _previousContainerCode: string | undefined;
    private _salesDeliveryLine: SalesDeliveryLineNodeResponse;
    private _stockJournalNodeResponses: StockJournalNodeResponse[];
    private _salesDeliveryLines: string[];
    private _currStockJournal: number;
    private _sequence: number;
    private _sideBarRowId: string;
    private _sideBarLevel: number;

    /*
     *
     *  Page Actions
     *
     */

    @ui.decorators.pageAction<MobileIntersiteReceipt>({
        title: 'Create',
        buttonType: 'primary',
        shortcut: ['f2'],
        isDisabled: true,
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-purchasing-sales/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-purchasing-sales/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);
                message = `**${ui.localize(
                    '@sage/x3-purchasing-sales/dialog-error-purchase-receipt-creation',
                    'An error occurred',
                )}**\n\n`;
                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-purchasing-sales/pages_creation_error_connexion_webservice_contact_administrator',
                    'An error occurred (connection or web service error). Contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-purchasing-sales/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
        async onClick() {
            const validateError = (await this.receiptDate.validate()) || (await this.supplierPackingSlip.validate());
            if (validateError) {
                this.$.removeToasts();
                this.$.showToast(
                    `${ui.localize(
                        '@sage/x3-purchasing-sales/notification-error-intersite-receipt-error',
                        'Error',
                    )}: ${validateError}`,
                    { type: 'error', timeout: 30000 },
                );
            } else if (this.lines.value.length > 0) {
                this.$.loader.isHidden = false;
                await this.$standardSaveAction.execute(false);
                this.$.loader.isHidden = true;
                await this.$.sound.success();
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-purchasing-sales/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-purchasing-sales/dialog-success-purchase-receipt-creation',
                        'Document created: {{receiptId}}.',
                        { receiptId: this.id.value },
                    ),
                );
                await this.$.router.emptyPage();
            } else {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-purchasing-sales/notification-error-intersite-receipt-no-products',
                        `Enter at least one product.`,
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
        },
    })
    createButton: ui.PageAction;

    /*
     *
     *  Sections
     *
     */

    @ui.decorators.section<MobileIntersiteReceipt>({
        isTitleHidden: true,
    })
    mainSection: ui.containers.Section;

    /*
     *
     *  Blocks
     *
     */

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
    })
    productBlock: ui.containers.Block;

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'small',
        isTitleHidden: true,
    })
    customizationIntersiteReceiptHeaderBlock: ui.containers.Block;

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
        isHidden: true,
    })
    destinationBlock: ui.containers.Block;

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    customizationIntersiteReceiptLinesBlock: ui.containers.Block;

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    customizationIntersiteReceiptDetailsBlock: ui.containers.Block;

    @ui.decorators.block<MobileIntersiteReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
        isHidden: true,
    })
    receiptBlock: ui.containers.Block;

    /*
     *
     *  Fields
     *
     */

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isHidden: true,
    })
    id: ui.fields.Text<MobileIntersiteReceipt>;

    @ui.decorators.referenceField<MobileIntersiteReceipt, Site>({
        node: '@sage/x3-system/Site',
        valueField: 'code',
        isHidden: true,
        bind: 'receiptSite',
        helperTextField: 'name',
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: '_id',
            }),
        ],
    })
    receiptSite: ui.fields.Reference<Site, MobileIntersiteReceipt>;

    @ui.decorators.dateField<MobileIntersiteReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Receipt date',
        isMandatory: true,
        maxDate: DateValue.today().toString(), // restricts to current date or earlier
    })
    receiptDate: ui.fields.Date;

    @ui.decorators.dropdownListField<MobileIntersiteReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Transaction',
        isMandatory: true,
        isTransientInput: true,
        onChange() {
            if (this.transaction.value) {
                this._selectedEntryTransaction = this._entryTransactionMap.get(this.transaction.value);
                this.stockMovementCode.value = this._selectedEntryTransaction?.stockMovementCode
                    ? {
                        code: this._selectedEntryTransaction.stockMovementCode.code,
                        _id: this._selectedEntryTransaction.stockMovementCode._id,
                    }
                    : null;
                this.stockAutomaticJournal.value = this._selectedEntryTransaction?.stockAutomaticJournal
                    ? {
                        code: this._selectedEntryTransaction.stockAutomaticJournal.code,
                        _id: this._selectedEntryTransaction.stockAutomaticJournal._id,
                    }
                    : null;
            }
        },
    })
    transaction: ui.fields.DropdownList;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isHidden: true,
        isTransientInput: true,
    })
    destination: ui.fields.Text<MobileIntersiteReceipt>;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isHidden: true,
        isTransientInput: true,
    })
    document: ui.fields.Text<MobileIntersiteReceipt>;

    @ui.decorators.referenceField<MobileIntersiteReceipt, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        canFilter: false,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 14,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementCode: ui.fields.Reference;

    @ui.decorators.referenceField<MobileIntersiteReceipt, AutomaticJournal>({
        node: '@sage/x3-finance-data/AutomaticJournal',
        valueField: 'code',
        isHidden: true,
        canFilter: false,
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockAutomaticJournal: ui.fields.Reference<AutomaticJournal, MobileIntersiteReceipt>;

    @ui.decorators.labelField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden: true,
    })
    legalCompany: ui.fields.Label;

    @ui.decorators.referenceField<MobileIntersiteReceipt, SalesDelivery>({
        parent() {
            return this.productBlock;
        },
        title: 'Delivery number',
        placeholder: 'Scan or select...',
        node: '@sage/x3-sales/SalesDelivery',
        valueField: 'id',
        helperTextField: { salesSite: { code: true } },
        isHelperTextHidden: true,
        canFilter: false,
        isTransient: true,
        filter() {
            let deliveryFilter: any = {
                salesSite: { code: { _ne: this.receiptSite.value.code } },
                isIntersite: true,
                isValidated: true,
                isLoanDelivery: false,
                shipmentDate: { _lte: this.receiptDate.value },
                receivingSite: { code: this.receiptSite.value.code },
                salesDeliveryLines: {
                    _none: true,
                    receiptStatus: { _ne: 0 },
                },
            };
            if (this.lines.value.length > 0) {
                deliveryFilter = {
                    ...deliveryFilter,
                    id: {
                        _nin: this.lines.value.map((line: any) => line.salesDelivery.id),
                    },
                    salesDeliveryLines: {
                        _atLeast: 1,
                        receiptStatus: { _eq: 0 },
                        salesOrderLine: {
                            purchaseOrder: {
                                orderFromSupplier: {
                                    code: {
                                        code: this.supplier.value,
                                    },
                                },
                            },
                        },
                    },
                };
            }
            return deliveryFilter;
        },
        async onChange() {
            if (this.salesDelivery.value) {
                this.supplier.value = await this._getSupplier(this.salesDelivery.value.id);
                if (!this.supplier.value) {
                    this.supplier.value = this.salesDelivery.value.salesSite.code;
                }
                if (this.lines.value.length > 0) {
                    this.supplierPackingSlip.value = null;
                    this.supplierPackingSlipDate.value = null;
                } else {
                    this.supplierPackingSlip.value = this.salesDelivery.value.id;
                    this.supplierPackingSlipDate.value = DateValue.today().toString();
                }
                this.salesDeliveryLine.isDisabled = false;
            } else {
                if (this.lines.value.length > 0) {
                    this.supplierPackingSlip.value = this._getSupplierPackingSlip();
                    this.supplierPackingSlipDate.value = DateValue.today().toString();
                } else {
                    this.supplier.value = null;
                    this.supplierPackingSlip.value = null;
                    this.supplierPackingSlipDate.value = null;
                }
                this.salesDeliveryLine.isDisabled = true;
            }
            this.createButton.isHidden = false;
            this.createButton.isDisabled = this.lines.value.length === 0;
        },
        isFullWidth: true,
        isAutoSelectEnabled: true,
        columns: [
            ui.nestedFields.text({
                bind: 'id',
                title: 'Order no.',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'salesSite',
                valueField: 'code',
                title: 'Sales site',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    salesDelivery: ui.fields.Reference;

    @ui.decorators.referenceField<MobileIntersiteReceipt, Supplier>({
        parent() {
            return this.productBlock;
        },
        title: 'Supplier',
        placeholder: 'Scan or select...',
        node: '@sage/x3-master-data/Supplier',
        valueField: { code: { code: true } },
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        isReadOnly: true,
        validation: /^$|^[^|]+$/,
        columns: [
            ui.nestedFields.reference({
                bind: 'code',
                title: 'Supplier',
                isReadOnly: true,
                node: '@sage/x3-master-data/BusinessPartner',
                valueField: 'code',
            }),
            ui.nestedFields.text({
                bind: 'shortCompanyName',
                title: 'Short description',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'companyName1',
                title: 'Company name',
                isReadOnly: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isActive',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    supplier: ui.fields.Reference;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Supplier packing slip',
        placeholder: 'Scan or enter...',
        isMandatory: false,
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        maxLength: 20,
        isReadOnly: true,
    })
    supplierPackingSlip: ui.fields.Text;

    @ui.decorators.dateField<MobileIntersiteReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Supplier packing slip date',
        isHidden: true,
    })
    supplierPackingSlipDate: ui.fields.Date;

    @ui.decorators.referenceField<MobileIntersiteReceipt, SalesDeliveryLine>({
        parent() {
            return this.productBlock;
        },
        title: 'Product',
        placeholder: 'Scan or select...',
        node: '@sage/x3-sales/SalesDeliveryLine',
        valueField: { product: { code: true } },
        helperTextField: { product: { upc: true } },
        isHelperTextHidden: true,
        canFilter: false,
        isTransient: true,
        filter() {
            let deliveryLineFilter: any = {
                salesDelivery: { id: this.salesDelivery.value?.id },
                receiptStatus: { _eq: 0 },
            };
            if (this.lines.value.length > 0) {
                let deliveryLines: string[] = [];
                this.lines.value.forEach((line: any) => {
                    deliveryLines.push(line.salesDeliveryLine._id);
                });
                if (deliveryLines.length > 0) {
                    deliveryLineFilter = {
                        ...deliveryLineFilter,
                        _id: {
                            _nin: deliveryLines,
                        },
                    };
                }
            }
            return deliveryLineFilter;
        },
        async onChange() {
            if (this.salesDeliveryLine.value) {
                this._product = await this.getProduct(this.salesDeliveryLine.value.product.code);
                this._productSite = await this.getProductSite(
                    this.salesDeliveryLine.value.product.code,
                    this.receiptSite.value?.code ?? '',
                );
                this._productSalesSite = await this.getProductSite(
                    this.salesDeliveryLine.value.product.code,
                    this.salesDelivery.value?.salesSite.code,
                );
                this.supplierPackingSlip.value = this._getSupplierPackingSlip();
                this.supplierPackingSlipDate.value = DateValue.today().toString();
                await this.gotoDestinationBlock();
            }
        },
        isFullWidth: true,
        isAutoSelectEnabled: true,
        columns: [
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
                isReadOnly: true,
                canFilter: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInSalesUnitDelivered' as any,
                canFilter: false,
                isReadOnly: true,
                postfix(_value: any, rowData?: any) {
                    return rowData?.salesUnit.code;
                },
                scale(value, rowValue?: Dict<any>) {
                    return rowValue?.salesUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'localizedDescription1',
                canFilter: true,
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'upc',
                canFilter: true,
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'salesUnit',
                valueField: 'code',
                canFilter: false,
                isHidden: true,
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'salesUnit',
                valueField: 'numberOfDecimals',
                canFilter: false,
                isHidden: true,
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'lineNumber',
                canFilter: false,
                isHidden: true,
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'receiptStatus',
                canFilter: false,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-sales/SalesDelivery',
                bind: 'salesDelivery',
                valueField: 'id',
                isHidden: true,
            }),
        ],
    })
    salesDeliveryLine: ui.fields.Reference;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerTitleLeft: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerTitleRight: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerLine2Left: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerLine2Right: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerLine3Left: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        isTransient: true,
        isHidden() {
            return this.destinationBlock.isHidden ?? true;
        },
    })
    headerLine3Right: ui.fields.Text;

    @ui.decorators.pageAction<MobileIntersiteReceipt>({
        title: 'Add',
        shortcut: ['f3'],
        buttonType: 'primary',
        async onClick() {
            if (await this._validate()) {
                const lineIndex = this.lines.value.length - 1;
                const detailIndex = this.lines.getRecordValue(`${lineIndex}`, 0)?.stockDetails?.length ?? 0;
                const selectedUnit = this._packingUnits.find(
                    unit => unit.node.packingUnit.code === this.destinationReceiptUnit.value,
                );
                this.lines.addOrUpdateRecordValue(
                    {
                        _id: `${lineIndex}|${detailIndex}`,
                        stockSite: this.receiptSite.value
                            ? {
                                code: this.receiptSite.value.code,
                                _id: this.receiptSite.value._id,
                            }
                            : undefined,
                        product: {
                            code: this._product.code,
                            _id: this._product._id,
                        },
                        effectiveDate: this.receiptDate.value ?? undefined,
                        ...(this.destinationLot.value && { lot: this.destinationLot.value }),
                        ...(this.destinationSublot.value && { sublot: this.destinationSublot.value }),
                        ...(this.destinationSupplierLot.value && { supplierLot: this.destinationSupplierLot.value }),
                        ...(this.destinationLocation.value && {
                            location: {
                                code: this.destinationLocation.value.code,
                                _id: this.destinationLocation.value._id,
                            },
                        }),
                        ...(this.destinationSerialNumber.value && { serialNumber: this.destinationSerialNumber.value }),
                        ...(this.destinationLotCustomField1.value && {
                            lotCustomField1: this.destinationLotCustomField1.value,
                        }),
                        ...(this.destinationLotCustomField2.value && {
                            lotCustomField2: this.destinationLotCustomField2.value,
                        }),
                        ...(this.destinationLotCustomField3.value && {
                            lotCustomField3: this.destinationLotCustomField3.value.toString(),
                        }),
                        ...(this.destinationLotCustomField4.value && {
                            lotCustomField4: this.destinationLotCustomField4.value,
                        }),
                        packingUnit: selectedUnit
                            ? {
                                code: selectedUnit.node.packingUnit.code,
                                numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                                _id: selectedUnit.node.packingUnit._id,
                            }
                            : this._product
                                ? this._product.stockUnit
                                : undefined,
                        quantityInPackingUnit: this.destinationQuantityToReceive.value?.toString(),
                        packingUnitToStockUnitConversionFactor:
                            this.destinationPackingUnitToStockUnitConversionFactor.value?.toString(),
                        expirationDate: this.destinationExpirationDate?.value ?? undefined,
                        ...(this.destinationStatus.value && {
                            status: {
                                code: this.destinationStatus.value.toString(),
                                _id: this.destinationStatus.value.toString(),
                            },
                        }),
                        ...(this.destinationIdentifier1.value && {
                            identifier1: this.destinationIdentifier1.value.toString(),
                        }),
                        ...(this.destinationIdentifier2.value && {
                            identifier2: this.destinationIdentifier2.value.toString(),
                        }),
                        ...(this.destinationUseByDate.value && {
                            useByDate: this.destinationUseByDate.value.toString(),
                        }),
                        ...(this.destinationMajorVersion.value && {
                            majorVersion: {
                                code: this.destinationMajorVersion.value.code,
                                _id: this.destinationMajorVersion.value._id,
                            },
                        }),
                        ...(this.destinationMinorVersion.value && { minorVersion: this.destinationMinorVersion.value }),
                        ...(this.destinationLicensePlateNumber.value && {
                            licensePlateNumber: {
                                code: this.destinationLicensePlateNumber.value.code,
                                _id: this.destinationLicensePlateNumber.value._id,
                            },
                        }),
                        sequence: this._sequence,
                        //remainingQuantityInStockUnit: this.destinationRemainingQuantityInStockUnit.value ?? 0,
                        ...this.getcustomizationIntersiteReceiptDetails(),
                    },
                    1,
                    `${lineIndex}`,
                );
                const purchaseReceiptLine = this.lines.getRecordValue(`${lineIndex}`, 0);
                if (purchaseReceiptLine) {
                    purchaseReceiptLine.receiptUnit = selectedUnit
                        ? {
                            code: selectedUnit.node.packingUnit.code,
                            numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                            _id: selectedUnit.node.packingUnit._id,
                        }
                        : this._product
                            ? this._product.stockUnit
                            : undefined;
                    purchaseReceiptLine.receiptUnitToStockUnitConversionFactor =
                        this.destinationPackingUnitToStockUnitConversionFactor.value?.toString();
                    let quantityInReceiptUnitReceived: number;
                    quantityInReceiptUnitReceived = 0;
                    purchaseReceiptLine.stockDetails?.forEach(stockDetail => {
                        quantityInReceiptUnitReceived += Number(stockDetail.quantityInPackingUnit);
                    });
                    purchaseReceiptLine.quantityInReceiptUnitReceived = quantityInReceiptUnitReceived.toString();
                    this.lines.addOrUpdateRecordValue(purchaseReceiptLine, 0, undefined);
                }
                if (this._productSalesSite && this._productSalesSite.stockManagementMode !== 'notManaged') {
                    this.destinationRemainingQuantity.value = this._calculateRemainingQuantityInStockJournalReponse(
                        this._stockJournalNodeResponses[this._currStockJournal],
                    );
                    if (this.destinationRemainingQuantity.value === 0) {
                        if (this._getNextStockJournal()) {
                            this._displayStockJournal(this._stockJournalNodeResponses[this._currStockJournal]);
                            this.destinationReceiptUnit.isDisabled = true;
                            this.destinationContainer.isDisabled = true;
                            this.destinationLicensePlateNumber.isDisabled = true;
                            this.detailsNextButton.isHidden = this._getDetailsNextButtonHidden();
                            this.destinationQuantityToReceive.focus();
                        } else {
                            await this._updatePurchaseReceiptLine();
                        }
                    } else {
                        this._displayStockJournal(this._stockJournalNodeResponses[this._currStockJournal]);
                        this.destinationReceiptUnit.isDisabled = true;
                        this.destinationContainer.isDisabled = true;
                        this.destinationLicensePlateNumber.isDisabled = true;
                        this.destinationQuantityToReceive.focus();
                        this.detailsNextButton.isHidden = this._getDetailsNextButtonHidden();
                    }
                } else {
                    this.destinationRemainingQuantity.value = this._calculateRemainingQuantityInLine(
                        this.destinationReceiptUnit.value ?? '',
                        Number(this.destinationPackingUnitToStockUnitConversionFactor.value),
                    );
                    if (this.destinationRemainingQuantity.value === 0) {
                        await this._updatePurchaseReceiptLine();
                    } else {
                        this._displayLine();
                        this.destinationQuantityToReceive.focus();
                    }
                }
            }
        },
    })
    addButton: ui.PageAction;

    @ui.decorators.pageAction<MobileIntersiteReceipt>({
        title: 'Next',
        shortcut: ['f2'],
        buttonType: 'secondary',
        async onClick() {
            if (
                this._productSalesSite &&
                this._productSalesSite.stockManagementMode !== 'notManaged' &&
                this._getNextStockJournal()
            ) {
                this._displayStockJournal(this._stockJournalNodeResponses[this._currStockJournal]);
                this.detailsNextButton.isHidden = this._getDetailsNextButtonHidden();
                this.destinationQuantityToReceive.focus();
            } else {
                if (
                    (this.lines.value.length > 0 &&
                        this.lines.value[this.lines.value.length - 1] &&
                        this.lines.value[this.lines.value.length - 1].stockDetails?.length) ??
                    0 > 0
                ) {
                    await this._updatePurchaseReceiptLine();
                } else {
                    if (this.lines.value.length > 0 && this.lines.value[this.lines.value.length - 1]) {
                        this.lines.removeRecord(this.lines.value[this.lines.value.length - 1]._id, 0);
                    }
                    this.productBlock.isHidden = false;
                    this.destinationBlock.isHidden = true;
                    this.customizationIntersiteReceiptHeaderBlock.isHidden =
                        this.customizationIntersiteReceiptHeaderBlock.layout.$layout?.$items.length === 0;
                    this.destinationBlock.isHidden = true;
                    this.customizationIntersiteReceiptLinesBlock.isHidden = true;
                    this.customizationIntersiteReceiptDetailsBlock.isHidden = true;
                    this.receiptBlock.isHidden = this.lines.value.length === 0;
                    this.addButton.isHidden = true;
                    this.detailsNextButton.isHidden = true;
                    this.salesDelivery.isDisabled = false;
                    this.salesDeliveryLine.isDisabled = false;
                    this.createButton.isHidden = false;
                    this.createButton.isDisabled = true;
                }
            }
        },
    })
    detailsNextButton: ui.PageAction;

    @ui.decorators.dropdownListField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Unit',
        placeholder: 'Enter unit',
        options: ['UN'],
        isMandatory: true,
        isTransient: true,
        onChange() {
            if (this.destinationReceiptUnit.value) {
                const selectedUnit = this._packingUnits.find(
                    packingUnit => packingUnit.node.packingUnit.code === this.destinationReceiptUnit.value,
                );
                if (selectedUnit) {
                    this.destinationPackingUnitToStockUnitConversionFactor.value = Number(
                        selectedUnit.node.packingUnitToStockUnitConversionFactor,
                    );
                    this.destinationPackingUnitToStockUnitConversionFactor.isDisabled =
                        !selectedUnit.node.isPackingFactorEntryAllowed;
                } else {
                    this.destinationPackingUnitToStockUnitConversionFactor.value = 1;
                    this.destinationPackingUnitToStockUnitConversionFactor.isDisabled = true;
                }
                this.destinationQuantityToReceive.scale = selectedUnit?.node.packingUnit.numberOfDecimals;
                this._salesUnitToStockUnitConversionFactor = String(
                    this.destinationPackingUnitToStockUnitConversionFactor.value,
                );

                const tmp = Math.pow(10, Number(selectedUnit?.node?.packingUnit?.numberOfDecimals));

                this.destinationRemainingQuantity.value =
                    (Number(this.destinationRemainingQuantity.value) *
                        Number(this._previousSalesUnitToStockUnitConversionFactor)) /
                    this.destinationPackingUnitToStockUnitConversionFactor.value;
                if (tmp !== 0) {
                    this.destinationRemainingQuantity.value =
                        Math.round(this.destinationRemainingQuantity.value * tmp) / tmp;
                }

                this._quantityToReceiveNoRounded =
                    (Number(this.destinationQuantityToReceive.value) *
                        Number(this._previousSalesUnitToStockUnitConversionFactor)) /
                    this.destinationPackingUnitToStockUnitConversionFactor.value;
                if (tmp !== 0) {
                    this._quantityToReceiveRounded = Math.round(this._quantityToReceiveNoRounded * tmp) / tmp;
                } else {
                    this._quantityToReceiveRounded = this._quantityToReceiveNoRounded;
                }
                this.destinationQuantityToReceive.value = this._quantityToReceiveRounded;
                this.destinationQuantityToReceive.focus();
                this.destinationPreviousReceiptUnit.value = this.destinationReceiptUnit.value;
                this._previousSalesUnitToStockUnitConversionFactor = this._salesUnitToStockUnitConversionFactor;
            }
            this._showAddButton();
        },
    })
    destinationReceiptUnit: ui.fields.DropdownList;

    @ui.decorators.labelField<MobileIntersiteReceipt>({
        isHidden: true,
        isTransient: true,
    })
    destinationReceiptUnitCode: ui.fields.Label;

    @ui.decorators.labelField<MobileIntersiteReceipt>({
        isHidden: true,
        isTransient: true,
    })
    destinationPreviousReceiptUnit: ui.fields.Label;

    @ui.decorators.numericField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Quantity',
        placeholder: 'Enter quantity',
        isMandatory: true,
        isTransient: true,
        validation(value: number) {
            const regex = /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/; // reg ex for any positive numbers (integers or decimals) excluding 0
            if ((value.toString().match(regex)?.length ?? 0) === 0) {
                return ui.localize('@sage/x3-purchasing-sales/validate-error-invalid-value', 'invalid value');
            }
            if (
                this.salesDelivery.value &&
                this._salesDeliveryLine &&
                (this.destinationQuantityToReceive.value ?? 0) > (this.destinationRemainingQuantity.value ?? 0) &&
                !this._isOverReceiptAllowed
            ) {
                return ui.localize(
                    '@sage/x3-purchasing-sales/validation-error-intersite-receipt-details-quantity-received-is-greater-than-the-ordered-quantity',
                    'The quantity received is greater than the ordered quantity.',
                );
            }
            return undefined;
        },
        scale() {
            const selectedUnit = this._packingUnits.find(
                unit => unit.node.packingUnit.code === this.destinationReceiptUnit.value,
            );
            return selectedUnit?.node.packingUnit.numberOfDecimals ?? 0;
        },
        min: 0,
        async onChange() {
            if (this.salesDelivery.value && this._salesDeliveryLine) {
                if ((this.destinationQuantityToReceive.value ?? 0) > (this.destinationRemainingQuantity.value ?? 0)) {
                    if (!this._isOverReceiptAllowed) {
                        await dialogMessage(
                            this,
                            'error',
                            ui.localize('@sage/x3-purchasing-sales/dialog-error-title', 'Error'),
                            ui.localize(
                                '@sage/x3-purchasing-sales/dialog-error-intersite-receipt-details-quantity-received-is-greater-than-the-ordered-quantity',
                                `The quantity received is greater than the ordered quantity.`,
                            ),
                            {
                                fullScreen: false,
                                acceptButton: {
                                    text: ui.localize('@sage/x3-purchasing-sales/button-accept-ok', 'OK'),
                                },
                            },
                        );
                        this.destinationQuantityToReceive.focus();
                        return;
                    } else {
                        if (
                            !(await dialogConfirmation(
                                this,
                                'warn',
                                ui.localize('@sage/x3-purchasing-sales/dialog-warning-title', 'Warning'),
                                ui.localize(
                                    '@sage/x3-purchasing-sales/dialog-warn-intersite-receipt-details-quantity-message',
                                    'Quantity larger than expected. Validate ?',
                                ),
                                {
                                    fullScreen: false,
                                    acceptButton: {
                                        text: ui.localize('@sage/x3-purchasing-sales/button-accept-yes', 'Yes'),
                                    },
                                    cancelButton: {
                                        text: ui.localize('@sage/x3-purchasing-sales/button-cancel-no', 'No'),
                                    },
                                }, //
                            ))
                        ) {
                            this.destinationQuantityToReceive.value = 0;
                            await this.$.commitValueAndPropertyChanges();
                            this.destinationQuantityToReceive.focus();
                        }
                    }
                }
            }
            if (!this.destinationQuantityToReceive.value) {
                this.destinationQuantityToReceive.value = 0;
            }
            if (this.destinationQuantityToReceive.value !== this._quantityToReceiveRounded) {
                this._quantityToReceiveNoRounded = this.destinationQuantityToReceive.value;
            }
            this._quantityToReceiveRounded = this.destinationQuantityToReceive.value;
            this._showAddButton();
        },
    })
    destinationQuantityToReceive: ui.fields.Numeric;

    @ui.decorators.numericField<MobileIntersiteReceipt>({
        isReadOnly: true,
        isHidden: true,
        isTransient: true,
    })
    destinationRemainingQuantity: ui.fields.Numeric;

    @ui.decorators.numericField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Conversion factor',
        isDisabled: true,
        isHidden: false,
        isTransient: true,
        onChange() {
            if (this.salesDelivery.value && this.lines.value.length > 0) {
                this.lines.value.forEach(line => {
                    if (
                        line.salesDelivery._id === this.salesDelivery.value?._id &&
                        line.salesDeliveryLine._id === this._salesDeliveryLine?._id
                    ) {
                        line.stockDetails?.forEach(stockDetail => {
                            if (
                                Number(this.destinationPackingUnitToStockUnitConversionFactor.value) !==
                                Number(stockDetail.packingUnitToStockUnitConversionFactor)
                            ) {
                                this.destinationPackingUnitToStockUnitConversionFactor.value = Number(
                                    stockDetail.packingUnitToStockUnitConversionFactor,
                                );
                                this.destinationPackingUnitToStockUnitConversionFactor.focus();
                                this.$.showToast(
                                    `${ui.localize(
                                        '@sage/x3-purchasing-sales/notification-The-conversion-factor-can-t-be-different-as-the-one-entered-previously-on-the-same-stock-line',
                                        "The conversion factor can't be different as the one entered previously on the same stock line.",
                                    )}`,
                                    { type: 'error', timeout: 5000 },
                                );
                            }
                        });
                    }
                });
            }

            this._salesUnitToStockUnitConversionFactor = String(
                this.destinationPackingUnitToStockUnitConversionFactor.value,
            );
        },
    })
    destinationPackingUnitToStockUnitConversionFactor: ui.fields.Numeric;

    @ui.decorators.dropdownListField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Status',
        placeholder: 'Enter status',
        isTransient: true,
        onChange() {
            this.destinationStatus.getNextField(true)?.focus();
        },
    })
    destinationStatus: ui.fields.DropdownList;

    @ui.decorators.referenceField<MobileIntersiteReceipt, Container>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Container',
        node: '@sage/x3-master-data/Container',
        valueField: 'code',
        isAutoSelectEnabled: true,
        placeholder: 'Scan or select...',
        width: 'large',
        isTransient: true,
        canFilter: false,
        filter() {
            return {
                isInternal: { _eq: true },
                isActive: { _eq: true },
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.text({
                bind: 'containerType',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: 'description',
                title: 'Description',
            }),
        ],
    })
    destinationContainer: ui.fields.Reference;

    @ui.decorators.referenceField<MobileIntersiteReceipt, LicensePlateNumber>({
        parent() {
            return this.destinationBlock;
        },
        title: 'License plate number',
        placeholder: 'Scan or select...',
        isFullWidth: true,
        node: '@sage/x3-stock-data/LicensePlateNumber',
        isAutoSelectEnabled: true,
        valueField: 'code',
        canFilter: false,
        filter() {
            let licensePlateNumberFilter: Filter<LicensePlateNumber> = {
                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
                isActive: { _eq: true },
                _or: [
                    {
                        isSingleProduct: { _eq: true },
                        stock: {
                            _atLeast: 1,
                            product: { product: { code: this._product.code ?? undefined } },
                        },
                    },
                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                    { isSingleProduct: { _eq: false } },
                ],
            };

            if (this.destinationLocation.value?.code) {
                licensePlateNumberFilter._or = [
                    {
                        location: { code: this.destinationLocation.value.code },
                    },
                    {
                        _and: [{ location: { code: undefined } }, { status: 'free' }],
                    },
                ];
            }

            if (this.destinationContainer.value) {
                licensePlateNumberFilter = {
                    ...licensePlateNumberFilter,
                    container: { code: (this.destinationContainer.value as Container)?.code },
                };
            }

            return licensePlateNumberFilter;
        },
        async onChange() {
            if (!this.destinationLicensePlateNumber.value) {
                this.destinationLocation.isDisabled = false;
                this.destinationLocation.value = null;
                this.destinationContainer.isDisabled = false;
                await this.$.commitValueAndPropertyChanges(); // without this, when you clear out LPN and then, without tabbing out, click Location's lookup button directly, nothing will happen
                return;
            }

            if (!!this.destinationLicensePlateNumber?.value?.location?.code) {
                this.destinationLocation.isDisabled = true;
                this.destinationLocation.value = this.destinationLicensePlateNumber.value.location;
                if ((this.destinationLicensePlateNumber.value as LicensePlateNumber).container) {
                    this.destinationContainer.value = (
                        this.destinationLicensePlateNumber.value as LicensePlateNumber
                    ).container;
                    this.destinationContainer.isDisabled = true;
                } else {
                    this.destinationContainer.isDisabled = false;
                }
            } else {
                if (this.destinationLocation.isDisabled) {
                    this.destinationLocation.value = null;
                    this.destinationLocation.isDisabled = false;
                }
                this.destinationContainer.isDisabled = false;
            }

            await this.$.commitValueAndPropertyChanges();
            this._showAddButton();
            this.destinationLicensePlateNumber.getNextField(true)?.focus();
        },
        isTransient: true,
        isMandatory: false,
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'License plate number',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Location',
                bind: 'location',
                valueField: 'code',
                title: 'Location',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Container',
                bind: 'container',
                valueField: 'code',
                title: 'Container',
                isReadOnly: true,
            }),
            ui.nestedFields.label({
                bind: 'status',
                title: 'Status',
                map(value: any) {
                    switch (value) {
                        case 'free':
                            return 'Free';
                        case 'inStock':
                            return 'In Stock';
                        default:
                            return '';
                    }
                },
                borderColor: ui.tokens.colorsYang100,
                optionType: '@sage/x3-stock-data/ContainerStatus',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isActive',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isSingleProduct',
                isHidden: true,
            }),
        ],
    })
    destinationLicensePlateNumber: ui.fields.Reference;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Suggested location',
        isReadOnly: true,
        width: 'large',
        isTransient: true,
    })
    destinationSuggestedLocation: ui.fields.Text;

    @ui.decorators.referenceField<MobileIntersiteReceipt, Location>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Location',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/Location',
        isAutoSelectEnabled: true,
        valueField: 'code',
        canFilter: false,
        isTransient: true,
        filter() {
            return {
                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
            };
        },
        onChange() {
            this._showAddButton();
            if (this.destinationLocation.value) this.destinationLocation.getNextField(true)?.focus();
            this.destinationWarehouse.value = !this.destinationLocation.value?.warehouse
                ? ''
                : this.destinationLocation.value.warehouse.code;
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Warehouse',
                bind: 'warehouse',
                valueField: 'code',
                isHidden: true,
            }),
        ],
        minLookupCharacters: 1,
        isFullWidth: true,
        isMandatory: true,
    })
    destinationLocation: ui.fields.Reference;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        isHidden: true,
        isReadOnly: true,
        isTransient: true,
    })
    destinationWarehouse: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Serial number',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isDisabled: true,
        isTransient: true,
    })
    destinationSerialNumber: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Supplier lot',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isDisabled: true,
        isTransient: true,
    })
    destinationSupplierLot: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Lot',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isTransient: true,
        isDisabled: true,
    })
    destinationLot: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Sublot',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isDisabled: true,
        isTransient: true,
    })
    destinationSublot: ui.fields.Text;

    @ui.decorators.referenceField<MobileIntersiteReceipt, MajorVersionStatus>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Major version',
        placeholder: 'Scan or select...',
        node: '@sage/x3-master-data/MajorVersionStatus',
        isAutoSelectEnabled: true,
        valueField: 'code',
        isDisabled: true,
        isTransient: true,
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Major version',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'status',
                title: 'Status',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    destinationMajorVersion: ui.fields.Reference;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Minor version',
        placeholder: 'Scan...',
        isDisabled: true,
        isTransient: true,
    })
    destinationMinorVersion: ui.fields.Text;

    @ui.decorators.dateField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Expiration date',
        placeholder: 'Enter an expiration date.',
        isDisabled: true,
        isTransient: true,
    })
    destinationExpirationDate: ui.fields.Date;

    @ui.decorators.dateField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Enter a use-by date.',
        title: 'Use-by date',
        isDisabled: true,
        isTransient: true,
    })
    destinationUseByDate: ui.fields.Date;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom text 1',
        validation: /^$|^[^|]+$/,
        maxLength: 20,
        isDisabled: true,
        isTransient: true,
    })
    destinationLotCustomField1: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom text 2',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isDisabled: true,
        isTransient: true,
    })
    destinationLotCustomField2: ui.fields.Text;

    @ui.decorators.numericField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom number',
        isDisabled: true,
        isTransient: true,
    })
    destinationLotCustomField3: ui.fields.Numeric;

    @ui.decorators.dateField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Enter a date',
        title: 'Lot custom date',
        isDisabled: true,
        isTransient: true,
    })
    destinationLotCustomField4: ui.fields.Date;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Identifier 1',
        placeholder: 'Scan...',
        isMandatory: false,
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isDisabled: true,
        isTransient: true,
    })
    destinationIdentifier1: ui.fields.Text;

    @ui.decorators.textField<MobileIntersiteReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Identifier 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        isMandatory: false,
        maxLength: 10,
        isDisabled: true,
        isTransient: true,
    })
    destinationIdentifier2: ui.fields.Text;

    @ui.decorators.nestedGridField<MobileIntersiteReceipt, [PurchaseReceiptLineBinding, StockJournalBinding]>({
        parent() {
            return this.receiptBlock;
        },
        bind: 'lines',
        canFilter: true,
        canActivate: true,
        isChangeIndicatorDisabled: true,
        canSelect: false,
        isTitleHidden: true,
        levels: [
            {
                node: '@sage/x3-purchasing/PurchaseReceiptLine',
                childProperty: 'stockDetails',
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'ID',
                        isHiddenDesktop: false,
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lineNumber',
                        title: 'Line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: { code: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, SalesDelivery>({
                        bind: 'salesDelivery',
                        node: '@sage/x3-sales/SalesDelivery',
                        valueField: { id: true },
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'id',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInReceiptUnitReceived',
                        title: 'Quantity in receipt unit',
                        isReadOnly: true,
                        postfix(_value, rowValue) {
                            return rowValue?.receiptUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.receiptUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, UnitOfMeasure>({
                        bind: 'receiptUnit',
                        title: 'Receipt unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: { code: true },
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'receiptUnitToStockUnitConversionFactor',
                        title: 'Receipt unit to stock unit conversion factor',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, Site>({
                        bind: 'receiptSite',
                        title: 'Receipt Site',
                        node: '@sage/x3-system/Site',
                        valueField: { code: true },
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, MiscellaneousTable>({
                        bind: 'stockMovementGroup',
                        title: 'Stock movement group',
                        node: '@sage/x3-system/MiscellaneousTable',
                        valueField: { code: true },
                        isHidden: true,
                        filter: {
                            glossaryId: {
                                glossaryId: 9,
                            },
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, PurchaseReceiptLineBinding, Container>({
                        bind: 'container',
                        title: 'Container',
                        node: '@sage/x3-master-data/Container',
                        valueField: 'code',
                        isAutoSelectEnabled: true,
                        placeholder: 'Scan or select...',
                        width: 'large',
                        isTransient: true,
                        canFilter: false,
                        isDisabled(value, rowValue) {
                            return Boolean(rowValue?.licensePlateNumber);
                        },
                        isHidden() {
                            return (
                                !this._productSite ||
                                this._purchaseStockManagementRules?.licensePlateNumberEntry !== 'mandatory' ||
                                !this._productSite?.isLicensePlateNumberManaged
                            );
                        },
                        filter() {
                            return {
                                isInternal: { _eq: true },
                                isActive: { _eq: true },
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.text({
                                bind: 'containerType',
                                title: 'Type',
                            }),
                            ui.nestedFields.text({
                                bind: 'description',
                                title: 'Description',
                            }),
                        ],
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        isHidden(recordId, rowItem, _level, _parentIds) {
                            return (
                                this._getSidebarPurchaseReceiptLine().length === 0 &&
                                this.lines.getRecordValue(recordId, 0)?.stockDetails?.length !== 1
                            );
                        },
                        async onClick(rowId) {
                            const purchaseReceiptLine = this.lines.getRecordValue(rowId, 0);
                            if (purchaseReceiptLine && (await this.getSidebarData(rowId, 0))) {
                                if (purchaseReceiptLine.stockDetails?.length === 1) {
                                    this.lines.openSidebar(purchaseReceiptLine.stockDetails[0]._id, 1);
                                } else {
                                    this.lines.openSidebar(rowId, 0);
                                }
                            }
                        },
                    },
                    {
                        icon: 'delete',
                        title: 'Delete',
                        onClick(rowId) {
                            this.lines.removeRecord(rowId, 0);
                            if (!this.lines.value?.length) {
                                this.$.setPageClean();
                                this._init();
                                this.salesDelivery.value = null;
                                this.salesDelivery.isDisabled = false;
                                this.salesDeliveryLine.value = null;
                                this.salesDeliveryLine.isDisabled = true;
                                this.supplier.value = null;
                                this.supplierPackingSlip.value = null;
                                this.salesDelivery.isDirty = false;
                                this.supplier.isDirty = false;
                                this.supplierPackingSlip.isDirty = false;
                                this.createButton.isHidden = false;
                                this.createButton.isDisabled = true;
                                // this._clearPage();
                                this.receiptBlock.isHidden = true;
                            }
                        },
                    },
                ],
                sidebar: {
                    layout() {
                        return {
                            mainSection: {
                                title: 'Purchase receipt line',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarPurchaseReceiptLine(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
            {
                node: '@sage/x3-stock-data/StockJournal',
                orderBy: { _id: 1 },
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'ID',
                        isHiddenDesktop: false,
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isMandatory: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                                isHidden: true,
                            }),
                        ],
                        filter() {
                            return {
                                code: {
                                    _in: this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.code),
                                },
                            };
                        },
                        isReadOnly: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity',
                        min: 0,
                        isNotZero: true,
                        isMandatory: true,
                        isFullWidth: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                        isReadOnly: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInStockUnit',
                        title: 'Quantity in stock unit',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden() {
                            return (this.destinationReceiptUnit.options ?? []).length <= 1;
                        },
                        isMandatory: true,
                        isFullWidth: true,
                        scale: 6,
                        isReadOnly: true,
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, StockStatus>({
                        bind: 'status',
                        title: 'Status',
                        node: '@sage/x3-stock-data/StockStatus',
                        valueField: { code: true },
                        isHidden() {
                            return !this._productSite || this._productSite.stockManagementMode === 'notManaged';
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, LicensePlateNumber>({
                        bind: 'licensePlateNumber',
                        title: 'License plate number',
                        valueField: 'code',
                        node: '@sage/x3-stock-data/LicensePlateNumber',
                        isFullWidth: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-stock-data/Location',
                                bind: 'location',
                                valueField: 'code',
                                title: 'Location',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Container',
                                bind: 'container',
                                valueField: 'code',
                                title: 'Container',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-system/Site',
                                bind: 'stockSite',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isActive',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isSingleProduct',
                                isHidden: true,
                            }),
                        ],
                        canFilter: false,
                        isHidden() {
                            return !this._productSite || !this._productSite.isLicensePlateNumberManaged;
                        },
                        filter() {
                            let licensePlateNumberFilter: Filter<any> = {
                                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
                                isActive: { _eq: true },
                                _or: [
                                    {
                                        isSingleProduct: { _eq: true },
                                        stock: {
                                            _atLeast: 1,
                                            product: { product: { code: this._product.code } },
                                        },
                                    },
                                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                                    { isSingleProduct: { _in: [false, null] } },
                                ],
                            };
                            if (this._previousLocationCode) {
                                licensePlateNumberFilter._or = [
                                    {
                                        location: { code: this._previousLocationCode },
                                    },
                                    {
                                        _and: [{ location: { code: undefined } }, { status: 'free' }],
                                    },
                                ];
                            }
                            return licensePlateNumberFilter;
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.licensePlateNumber) {
                                if (value.licensePlateNumber.location) {
                                    value.location = value.licensePlateNumber.location;
                                    this._previousLocationCode = value.licensePlateNumber.location.code;
                                }
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, Location>({
                        bind: 'location',
                        title: 'Location',
                        node: '@sage/x3-stock-data/Location',
                        valueField: { code: true },
                        isFullWidth: true,
                        isHidden() {
                            return !this._productSite || !this._productSite.isLocationManaged;
                        },
                        isMandatory() {
                            return (
                                this._productSite &&
                                this._productSite.isLocationManaged &&
                                !this._productSite.isLicensePlateNumberManaged
                            );
                        },
                        isReadOnly(_value, rowValue) {
                            return (
                                !this._productSite ||
                                (this._productSite.isLicensePlateNumberManaged &&
                                    !!rowValue?.licensePlateNumber &&
                                    !!rowValue.licensePlateNumber.location)
                            );
                        },
                        filter() {
                            return {
                                ...(this.receiptSite.value && {
                                    stockSite: { code: this.receiptSite.value.code },
                                }),
                                category: { _nin: ['subcontract', 'customer'] },
                            };
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.location) {
                                value.warehouse = value.location.warehouse;
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                            this._previousLocationCode = value.location?.code;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference<MobileIntersiteReceipt, Location, Warehouse>({
                                node: '@sage/x3-stock-data/Warehouse',
                                bind: 'warehouse',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'serialNumber',
                        title: 'Serial number',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return ['notManaged', 'issued'].includes(this._product.serialNumberManagementMode);
                        },
                        isReadOnly: true,
                    }),
                    ui.nestedFields.text({
                        bind: 'supplierLot',
                        title: 'Supplier lot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return this._product.lotManagementMode === 'notManaged';
                        },
                        isDisabled: true,
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'lot',
                        node: '@sage/x3-stock-data/LotsSites',
                        title: 'Lot',
                        valueField: 'lot',
                        placeholder: 'Scan or select...',
                        isFullWidth: true,
                        isNewEnabled: true,
                        validation: /^$|^[^|]+$/,
                        isDisabled: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: 'lot',
                                title: 'Lot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'sublot',
                                title: 'Sublot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                        isHidden() {
                            return this._product.lotManagementMode === 'notManaged';
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'sublot',
                        title: 'Sublot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return this._product.lotManagementMode !== 'lotAndSublot';
                        },
                        isDisabled: true,
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: 'code',
                        isHiddenDesktop: false,
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, MajorVersionStatus>({
                        bind: 'majorVersion',
                        title: 'Major version',
                        node: '@sage/x3-master-data/MajorVersionStatus',
                        valueField: { code: true },
                        isFullWidth: true,
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                !this._product.stockVersionMode ||
                                this._product.stockVersionMode === 'no'
                            );
                        },
                        isDisabled: true,
                        filter() {
                            return {
                                product: { code: this._product.code },
                                _or: [
                                    { status: 'prototypeVersion' },
                                    { status: 'activeVersion' },
                                    { status: 'stoppedVersion' },
                                ],
                                ...(this._product.lotManagementMode !== 'notManaged' &&
                                    this._product.stockVersionMode === 'majorAndMinor' && {
                                    minorVersions: {
                                        _atLeast: 1,
                                        type: { _eq: 'stock' },
                                        useStatus: { _eq: 'availableToUse' },
                                    },
                                }),
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Major Version',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'minorVersion',
                        title: 'Minor version',
                        node: '@sage/x3-master-data/ProductVersion',
                        valueField: 'minorVersion',
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                this._product.stockVersionMode !== 'majorAndMinor'
                            );
                        },
                        filter(rowValue) {
                            return rowValue.majorVersion
                                ? {
                                    product: { code: this._product.code },
                                    majorVersion: rowValue.majorVersion.code,
                                    type: 'stock',
                                    useStatus: 'availableToUse',
                                }
                                : {};
                        },
                        isDisabled: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: 'minorVersion',
                                title: 'Minor version',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'majorVersion',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'type',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'useStatus',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.date({
                        bind: 'expirationDate',
                        placeholder: 'Enter...',
                        title: 'Expiration date',
                        isHidden() {
                            return (
                                (this._product.lotManagementMode === 'notManaged' ||
                                    this._product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedEntryTransaction?.isLotExpirationDateAllowed &&
                                        (this._product.expirationManagementMode === 'roundingBeginningMonth1' ||
                                            this._product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._product.expirationManagementMode === 'withoutRounding'))) ??
                                false
                            );
                        },
                        isDisabled: true,
                    }),
                    ui.nestedFields.date({
                        bind: 'useByDate',
                        placeholder: 'Enter...',
                        title: 'Use-by date',
                        isDisabled: true,
                        isHidden() {
                            return (
                                (this._product.lotManagementMode === 'notManaged' ||
                                    this._product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedEntryTransaction?.isLotExpirationDateAllowed &&
                                        (this._product.expirationManagementMode === 'roundingBeginningMonth1' ||
                                            this._product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._product.expirationManagementMode === 'withoutRounding'))) ??
                                false
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 1',
                        validation: /^$|^[^|]+$/,
                        isDisabled: true,
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField1Allowed
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 2',
                        validation: /^$|^[^|]+$/,
                        isDisabled: true,
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField2Allowed
                            );
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lotCustomField3',
                        placeholder: 'Enter...',
                        title: 'Lot custom number',
                        isDisabled: true,
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField3Allowed
                            );
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'lotCustomField4',
                        placeholder: 'Enter...',
                        title: 'Lot custom date',
                        isDisabled: true,
                        isHidden() {
                            return (
                                this._product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField4Allowed
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier1',
                        title: 'Identifier 1',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isDisabled() {
                            return this._selectedEntryTransaction?.identifier1Detail !== 'entered';
                        },
                        isHidden() {
                            return (
                                (this._selectedEntryTransaction?.identifier1Detail !== 'entered' &&
                                    this._selectedEntryTransaction?.identifier1Detail !== 'displayed') ||
                                !this._productSite ||
                                !this._productSite.isLocationManaged
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier2',
                        title: 'Identifier 2',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isHiddenDesktop: false,
                        maxLength: 10,
                        isDisabled() {
                            return this._selectedEntryTransaction?.identifier2Detail !== 'entered';
                        },
                        isHidden() {
                            return (
                                (this._selectedEntryTransaction?.identifier2Detail !== 'entered' &&
                                    this._selectedEntryTransaction?.identifier2Detail !== 'displayed') ||
                                !this._productSite ||
                                !this._productSite.isLocationManaged
                            );
                        },
                    }),
                    ui.nestedFields.reference<MobileIntersiteReceipt, StockJournalBinding, Site>({
                        bind: 'stockSite',
                        title: 'Site',
                        node: '@sage/x3-system/Site',
                        valueField: { code: true },
                        isHiddenDesktop: false,
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.date({
                        bind: 'effectiveDate',
                        title: 'Effective date',
                        isHiddenDesktop: false,
                        isReadOnly: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId) {
                            if (await this.getSidebarData(rowId, 1)) {
                                this.lines.openSidebar(rowId, 1);
                            }
                        },
                    },
                ],
                sidebar: {
                    title: 'Intersite receipt',
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock details',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarStockJournal(),
                                    },
                                },
                            },
                        };
                    },
                    async onRecordConfirmed(_id: string, recordValue: StockJournalBinding | undefined) {
                        const _purchaseReceiptLine = this.lines.getRecordValue(_id.substring(0, _id.indexOf('|')), 0);
                        if (_purchaseReceiptLine) {
                            _purchaseReceiptLine.receiptUnit = recordValue?.packingUnit;
                            _purchaseReceiptLine.receiptUnitToPurchaseUnitConversionFactor =
                                recordValue?.packingUnitToStockUnitConversionFactor;
                            _purchaseReceiptLine.receiptUnitToStockUnitConversionFactor =
                                recordValue?.packingUnitToStockUnitConversionFactor;
                            let quantityInReceiptUnitReceived: number;
                            quantityInReceiptUnitReceived = 0;
                            _purchaseReceiptLine.stockDetails?.forEach(stockDetail => {
                                quantityInReceiptUnitReceived += Number(stockDetail.quantityInPackingUnit);
                            });
                            _purchaseReceiptLine.quantityInReceiptUnitReceived =
                                quantityInReceiptUnitReceived.toString();
                            this.lines.setRecordValue(_purchaseReceiptLine, 0);
                        }
                    },
                },
            },
        ],
    })
    lines: ui.fields.NestedGrid<[PurchaseReceiptLineBinding, StockJournalBinding]>;

    private async gotoDestinationBlock() {
        if (this.salesDelivery.value && (await this.$.page.validate()) && this.salesDeliveryLine.value) {
            this._salesDeliveryLines = [this.salesDeliveryLine.value.lineNumber.toString()];
            const salesDeliveryLine = await this._getSalesDeliveryLine(this.salesDeliveryLine.value.lineNumber);
            if (!salesDeliveryLine) {
                return;
            }
            this._salesDeliveryLine = salesDeliveryLine;
            const lineIndex = this.lines.value.length;
            this.lines.addOrUpdateRecordValue(
                {
                    _id: `${lineIndex}`,
                    lineNumber: Number(Math.floor(Math.random() * MAX_INT_32)),
                    product: {
                        code: this._product.code,
                        _id: this._product._id,
                    },
                    receiptSite: this.receiptSite.value ?? undefined,
                    receiptDate: this.receiptDate.value ?? undefined,
                    supplier: this.supplier.value ?? undefined,
                    salesDelivery: this.salesDelivery.value,
                    salesDeliveryLine: {
                        _id: this._salesDeliveryLine._id,
                        lineNumber: this._salesDeliveryLine.lineNumber,
                        salesDelivery: {
                            id: this._salesDeliveryLine.salesDelivery.id,
                            _id: this._salesDeliveryLine.salesDelivery._id,
                        },
                        salesUnitToStockUnitConversionFactor:
                            this._salesDeliveryLine.salesUnitToStockUnitConversionFactor,
                    },
                    ...this.getcustomizationIntersiteReceiptLines(),
                },
                0,
                undefined,
            );

            await this._initializePage();
            if (this._stockJournalNodeResponses.length === 0) {
                if (
                    await this._getStockJournalsNoUpdated(
                        this._salesDeliveryLine.product.code,
                        this.salesDelivery.value.salesSite.code,
                    )
                ) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-purchasing-sales/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-purchasing-sales/dialog-error-intersite-receipt-there-are-movements-in-waiting-for-this-line',
                            `There are movements in waiting for this line.`,
                        ),
                        {
                            acceptButton: {
                                text: ui.localize('@sage/x3-purchasing-sales/button-accept-ok', 'OK'),
                            },
                        },
                    );
                } else {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-purchasing-sales/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-purchasing-sales/dialog-error-intersite-receipt-there-is-no-stock-movement-for-this-line',
                            `There is no stock movement for this line.`,
                        ),
                        {
                            acceptButton: {
                                text: ui.localize('@sage/x3-purchasing-sales/button-accept-ok', 'OK'),
                            },
                        },
                    );
                }
                this.lines.removeRecord(this.lines.value[this.lines.value.length - 1]._id, 0);
                return;
            }
            if (this._productSalesSite && this._productSalesSite.stockManagementMode !== 'notManaged') {
                this._displayStockJournal(this._stockJournalNodeResponses[this._currStockJournal]);
            } else {
                this._displayLine();
            }
            this.productBlock.isHidden = true;
            this.destinationBlock.isHidden = false;
            this.customizationIntersiteReceiptHeaderBlock.isHidden = true;
            this.destinationBlock.isHidden = false;
            this.customizationIntersiteReceiptLinesBlock.isHidden =
                this.customizationIntersiteReceiptLinesBlock.layout.$layout?.$items.length === 0;
            this.customizationIntersiteReceiptDetailsBlock.isHidden = false;
            this.receiptBlock.isHidden = true;
            this.createButton.isHidden = true;
            this.addButton.isHidden = false;
            this.detailsNextButton.isHidden = this._getDetailsNextButtonHidden();
        }
    }

    private async _init(): Promise<boolean> {
        this._numberOfDecimalList = await getUnitNumberOfDecimalList(this);

        this.receiptDate.value = DateValue.today().toString();
        try {
            // Populate Default Site & Legal company & Entry Transaction fields
            const receiptSiteCode = await getSelectedStockSite(
                this,
                ui.localize('@sage/x3-purchasing-sales/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-purchasing-sales/dialog-error-location-inquiry-set-site',
                    'Define a default stock site on the user function profile.',
                ),
            );
            if (!receiptSiteCode) {
                return false;
            }
            this.receiptSite.value = {
                _id: receiptSiteCode,
                code: receiptSiteCode,
            };
            await this._initTransaction();
        } catch (error) {
            this.$.showToast(error.message, { type: 'error' });
            return false;
        }
        this._getIsLocationPreloaded();

        this._isOverReceiptAllowed =
            (await readParameterValue('RCPORD', this.$.userCode ?? '', this.receiptSite.value?.code ?? '', this)) === 2;

        const salesDeliveryId = this.$.queryParameters.salesDeliveryId;
        const saleSiteId = this.$.queryParameters.salesSite;
        if (salesDeliveryId && saleSiteId) {
            this.salesDelivery.isDisabled = true;
            this.salesDelivery.value = {
                id: salesDeliveryId,
                _id: salesDeliveryId,
                salesSite: {
                    code: saleSiteId,
                    _id: saleSiteId,
                },
            };
            this.supplier.value = (await this._getSupplier(salesDeliveryId as string)) ?? null;
            if (!this.supplier.value) {
                this.supplier.value = this.salesDelivery.value.salesSite.code;
            }
            this.salesDeliveryLine.value = null;
            this.salesDeliveryLine.isDisabled = false;
            this.supplierPackingSlip.value = this.salesDelivery.value.id;
            this.supplierPackingSlipDate.value = DateValue.today().toString();
        } else {
            this.salesDeliveryLine.value = null;
            this.salesDeliveryLine.isDisabled = true;
            this.supplier.value = null;
            this.supplierPackingSlip.value = null;
            this.supplier.isDirty = false;
            this.supplierPackingSlip.isDirty = false;
        }
        this.createButton.isHidden = false;
        this.createButton.isDisabled = true;
        this.addButton.isHidden = true;
        this.detailsNextButton.isHidden = true;
        this.legalCompany.value = await this._getLegalCompany(this.receiptSite.value?.code ?? '');
        this.destination.value = this.$.storage.get('mobile-label-destination')?.toString() ?? '';
        this.document.value = this.$.storage.get('mobile-document-destination')?.toString() ?? '';
        return true;
    }

    private async getProductSite(productCode: string, siteCode: string) {
        const productSiteToReceive = await this.$.graph
            .node('@sage/x3-master-data/ProductSite')
            .read(
                {
                    _id: true,
                    stockSite: { code: true },
                    isLocationManaged: true,
                    isLicensePlateNumberManaged: true,
                    defaultInternalContainer: {
                        code: true,
                        _id: true,
                    },
                    isBeingCounted: true,
                    stockManagementMode: true,
                    defaultLocations: {
                        query: {
                            edges: {
                                node: {
                                    defaultLocation: true,
                                    locationNumber: true,
                                    defaultLocationType: true,
                                    _id: true,
                                },
                            },
                        },
                    },
                    product: {
                        _id: true,
                        code: true,
                        localizedDescription1: true,
                        productCategory: {
                            code: true,
                            _id: true,
                        },
                        lotManagementMode: true,
                        serialNumberManagementMode: true,
                        stockVersionMode: true,
                        expirationManagementMode: true,
                        serialSequenceNumber: true,
                        lotSequenceNumber: true,
                        stockManagementMode: true,
                        defaultPotencyInPercentage: true,
                        expirationLeadTime: true,
                        expirationTimeUnit: true,
                        useByDateCoefficient: true,
                        stockUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        purchaseUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        purchaseUnitToStockUnitConversionFactor: true,
                        packingUnits: {
                            query: {
                                edges: {
                                    node: {
                                        packingUnit: {
                                            code: true,
                                            numberOfDecimals: true,
                                            _id: true,
                                        },
                                        packingUnitToStockUnitConversionFactor: true,
                                        isPackingFactorEntryAllowed: true,
                                        _id: true,
                                    },
                                },
                            },
                        },
                    },
                },
                `${productCode}|${siteCode}`,
            )
            .execute();

        if (!productSiteToReceive) {
            return null;
        }

        this._purchaseStockManagementRules = await findStockManagementRules(
            this.receiptSite.value?.code ?? '',
            productSiteToReceive.product.productCategory.code,
            '3',
            this._selectedEntryTransaction?.stockMovementCode?.code ?? undefined,
            this,
        );

        return productSiteToReceive as any;
    }

    private async getProduct(productCode: string) {
        const productToReceive = await this.$.graph
            .node('@sage/x3-master-data/Product')
            .read(
                {
                    _id: true,
                    code: true,
                    localizedDescription1: true,
                    productCategory: {
                        code: true,
                        _id: true,
                    },
                    lotManagementMode: true,
                    serialNumberManagementMode: true,
                    stockVersionMode: true,
                    expirationManagementMode: true,
                    serialSequenceNumber: true,
                    lotSequenceNumber: true,
                    stockManagementMode: true,
                    defaultPotencyInPercentage: true,
                    expirationLeadTime: true,
                    expirationTimeUnit: true,
                    useByDateCoefficient: true,
                    stockUnit: {
                        code: true,
                        numberOfDecimals: true,
                        _id: true,
                    },
                    purchaseUnit: {
                        code: true,
                        numberOfDecimals: true,
                        _id: true,
                    },
                    purchaseUnitToStockUnitConversionFactor: true,
                    packingUnits: {
                        query: {
                            edges: {
                                node: {
                                    packingUnit: {
                                        code: true,
                                        numberOfDecimals: true,
                                        _id: true,
                                    },
                                    packingUnitToStockUnitConversionFactor: true,
                                    isPackingFactorEntryAllowed: true,
                                    _id: true,
                                },
                            },
                        },
                    },
                },
                `${productCode}`,
            )
            .execute();

        if (!productToReceive) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing-sales/pages__intersite_receipt__notification__invalid_product_error',
                    `Could not retrieve your product: {{ productCode }}.`,
                    {
                        productCode: productCode,
                    },
                ),
                {
                    type: 'error',
                    timeout: 5000,
                },
            );
            return null;
        }

        this._purchaseStockManagementRules = await findStockManagementRules(
            this.receiptSite.value?.code ?? '',
            productToReceive.productCategory.code,
            '3',
            this._selectedEntryTransaction?.stockMovementCode?.code ?? undefined,
            this,
        );

        return productToReceive as any;
    }

    private async _getLegalCompany(defaultSite: string): Promise<string> {
        const response = await this.$.graph
            .node('@sage/x3-system/Site')
            .read(
                {
                    _id: true,
                    legalCompany: { code: true },
                },
                defaultSite,
            )
            .execute();

        if (!response) {
            // This should never occur as long as the user function profile has a valid site set up
            throw new Error(
                ui.localize(
                    '@sage/x3-purchasing-sales/error-intersite-receipt-no-legal-company',
                    'The invalid site {{site}} has no legal company.',
                    {
                        site: defaultSite,
                    },
                ),
            );
        }

        return response.legalCompany.code;
    }

    private _getSupplierPackingSlip(): string | null {
        const salesDeliverys: string[] = [];
        if (this.lines.value.length > 0) {
            this.lines.value.forEach(line => {
                if (!salesDeliverys.find(id => id === line.salesDelivery.id)) {
                    salesDeliverys.push(line.salesDelivery.id);
                }
            });
        }
        if (this.salesDelivery.value && !salesDeliverys.find(id => id === this.salesDelivery?.value?.id)) {
            salesDeliverys.push(this.salesDelivery?.value?.id);
        }
        if (salesDeliverys.length === 1) {
            return salesDeliverys[0];
        } else if (this.supplierPackingSlip.value) {
            return this.supplierPackingSlip.value;
        } else {
            return null;
        }
    }

    private async _getEntryTransactionOptions(): Promise<string[]> {
        const response = extractEdges(
            await this.$.graph
                .node('@sage/x3-purchasing/PurchaseEntryTransaction')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            code: true,
                            stockMovementCode: {
                                code: true,
                                _id: true,
                            },
                            defaultStockMovementGroup: {
                                code: true,
                                _id: true,
                            },
                            stockAutomaticJournal: {
                                code: true,
                                _id: true,
                            },
                            isLotCustomField1Allowed: true,
                            isLotCustomField2Allowed: true,
                            isLotCustomField3Allowed: true,
                            isLotCustomField4Allowed: true,
                            identifier1Entry: true,
                            identifier1Detail: true,
                            identifier2Entry: true,
                            identifier2Detail: true,
                            isLotPotencyAllowed: true,
                            isLotExpirationDateAllowed: true,
                        },
                        {
                            filter: {
                                entryTransactionType: 'receipt',
                                isActive: true,
                            },
                        },
                    ),
                )
                .execute(),
        ) as PurchaseEntryTransaction[];

        this._entryTransactionMap = new Map<string, PurchaseEntryTransaction>();

        // transform Entry Transaction response into a string array
        return response.map((entryTransaction: PurchaseEntryTransaction) => {
            this._entryTransactionMap.set(entryTransaction.code, {
                code: entryTransaction.code,
                _id: entryTransaction._id,
                stockAutomaticJournal: entryTransaction.stockAutomaticJournal,
                stockMovementCode: entryTransaction.stockMovementCode,
                defaultStockMovementGroup: entryTransaction.defaultStockMovementGroup,
                isLotCustomField1Allowed: entryTransaction.isLotCustomField1Allowed,
                isLotCustomField2Allowed: entryTransaction.isLotCustomField2Allowed,
                isLotCustomField3Allowed: entryTransaction.isLotCustomField3Allowed,
                isLotCustomField4Allowed: entryTransaction.isLotCustomField4Allowed,
                identifier1Entry: entryTransaction.identifier1Entry,
                identifier1Detail: entryTransaction.identifier1Detail,
                identifier2Entry: entryTransaction.identifier2Entry,
                identifier2Detail: entryTransaction.identifier2Detail,
                isLotPotencyAllowed: entryTransaction.isLotPotencyAllowed,
                isLotExpirationDateAllowed: entryTransaction.isLotExpirationDateAllowed,
            });
            return entryTransaction.code;
        });
    }

    private async _getIsLocationPreloaded() {
        const responseSite = await this.$.graph
            .node('@sage/x3-master-data/MobileAutomationSetup')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        isLocationPreloaded: true,
                    },
                    {
                        filter: {
                            ...(this.receiptSite.value && { site: { code: this.receiptSite.value.code } }),
                        },
                    },
                ),
            )
            .execute();
        if (responseSite.edges.length !== 0) {
            responseSite.edges.some((edge: { node: { isLocationPreloaded: boolean } }) => {
                this._isLocationPreloaded = edge.node.isLocationPreloaded;
            });
        } else {
            const response = await this.$.graph
                .node('@sage/x3-master-data/MobileAutomationSetup')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            isLocationPreloaded: true,
                        },
                        {
                            filter: {
                                site: null,
                            },
                        },
                    ),
                )
                .execute();
            if (response.edges.length !== 0) {
                response.edges.some((edge: { node: { isLocationPreloaded: boolean } }) => {
                    this._isLocationPreloaded = edge.node.isLocationPreloaded;
                });
            } else {
                this._isLocationPreloaded = true;
            }
        }
    }

    private _disablePage(): void {
        this.salesDelivery.isDisabled = true;
        this.supplier.isDisabled = true;
        this.receiptDate.isDisabled = true;
        this.transaction.isDisabled = true;
        this.supplierPackingSlip.isDisabled = true;
    }

    private async _initTransaction() {
        this.transaction.options = await this._getEntryTransactionOptions();
        switch (this.transaction.options.length) {
            case 0:
                this._disablePage();
                throw new Error('No transaction, cannot continue');
            case 1:
                this.transaction.value = this.transaction.options[0];
                this.transaction.isHidden = true;
                break;
            default:
                this.transaction.value = this.transaction.options[0];
                if (this.lines.value.length > 0) {
                    this.transaction.isDisabled = true;
                }
                break;
        }
        if (this.transaction.value) {
            this._selectedEntryTransaction = this._entryTransactionMap.get(this.transaction.value);
            this.stockMovementCode.value = this._selectedEntryTransaction?.stockMovementCode
                ? {
                    code: this._selectedEntryTransaction.stockMovementCode.code,
                    _id: this._selectedEntryTransaction.stockMovementCode._id,
                }
                : null;
            this.stockAutomaticJournal.value = this._selectedEntryTransaction?.stockAutomaticJournal
                ? {
                    code: this._selectedEntryTransaction.stockAutomaticJournal.code,
                    _id: this._selectedEntryTransaction.stockAutomaticJournal._id,
                }
                : null;
        }
    }

    private async _getSupplier(salesDeliveryId: string): Promise<Supplier | null> {
        const response = extractEdges(
            await this.$.graph
                .node('@sage/x3-sales/SalesDeliveryLine')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            salesDelivery: {
                                id: true,
                                _id: true,
                            },
                            salesOrderLine: {
                                purchaseOrder: {
                                    orderFromSupplier: {
                                        code: {
                                            code: true,
                                            _id: true,
                                        },
                                        shortCompanyName: true,
                                        companyName1: true,
                                        isActive: true,
                                        _id: true,
                                    },
                                },
                                _id: true,
                            },
                            _id: true,
                        },
                        {
                            filter: {
                                salesDelivery: {
                                    id: salesDeliveryId,
                                },
                            },
                        },
                    ),
                )
                .execute(),
        ) as any[];
        let supplier: Supplier | null;
        supplier = null;
        response.forEach(salesDeliveryLine => {
            if (salesDeliveryLine.salesOrderLine?.purchaseOrder) {
                supplier = salesDeliveryLine.salesOrderLine.purchaseOrder.orderFromSupplier;
            }
        });
        return supplier;
    }

    private async _getSupplierByCode(code: string): Promise<Supplier | null> {
        const response = extractEdges(
            await this.$.graph
                .node('@sage/x3-sales/Supplier')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            code: {
                                code: true,
                                _id: true,
                            },
                            shortCompanyName: true,
                            companyName1: true,
                            isActive: true,
                            _id: true,
                        },
                        {
                            filter: {
                                code: {
                                    code: code,
                                },
                            },
                        },
                    ),
                )
                .execute(),
        ) as any[];
        if (response.length > 0) {
            return response[0];
        } else {
            return null;
        }
    }

    async _initializePage(): Promise<boolean> {
        this.destinationQuantityToReceive.value = null;
        this.destinationLicensePlateNumber.value = null;
        this.destinationLocation.value = null;
        this.destinationSerialNumber.value = null;
        this.destinationSupplierLot.value = null;
        this.destinationLot.value = null;
        this.destinationSublot.value = null;
        this.destinationMajorVersion.value = null;
        this.destinationMinorVersion.value = null;
        this.destinationLotCustomField1.value = null;
        this.destinationLotCustomField2.value = null;
        this.destinationLotCustomField3.value = null;
        this.destinationLotCustomField4.value = null;
        this.destinationIdentifier1.value = null;
        this.destinationIdentifier2.value = null;

        if (!this._salesDeliveryLine || !this.salesDelivery.value) {
            return false;
        }

        if (this._productSalesSite && this._productSalesSite.stockManagementMode !== 'notManaged') {
            await this._getStockJournals(this._salesDeliveryLine.product.code, this.salesDelivery.value.salesSite.code);
            if (this._stockJournalNodeResponses.length === 0) {
                return false;
            }
        }

        if (this._salesDeliveryLine.salesOrderLine?.purchaseOrder?.orderFromSupplier) {
            this.supplier.value = this._salesDeliveryLine.salesOrderLine.purchaseOrder.orderFromSupplier;
        } else {
            this.supplier.value = await this._getSupplierByCode(this.salesDelivery.value.salesSite.code);
        }

        this._supplierProduct = await this._getSupplierProduct();

        this._initPackingUnitFields();
        this.destinationReceiptUnit.value = this._salesDeliveryLine.salesUnit.code;
        this.destinationPreviousReceiptUnit.value = this.destinationReceiptUnit.value;

        this._purchaseStockManagementRules = await findStockManagementRules(
            this.receiptSite.value?.code ?? '',
            this._product.productCategory.code,
            '3',
            this.stockMovementCode.value?.code,
            this,
        );

        this.destinationStatus.options = await this._getStatuses();
        this.destinationStatus.value = this._purchaseStockManagementRules?.defaultStatus ?? '';

        const lpnManagementRule = this._purchaseStockManagementRules?.licensePlateNumberEntry;
        const lotManagement = this._product.lotManagementMode;
        const lotNotManaged = lotManagement === 'notManaged';
        this.destinationLot.isHidden = lotNotManaged;
        this.destinationSupplierLot.isHidden = lotNotManaged;
        this.destinationSublot.isHidden = lotManagement !== 'lotAndSublot';
        this.destinationLotCustomField1.isHidden =
            !this._selectedEntryTransaction?.isLotCustomField1Allowed || lotNotManaged;
        this.destinationLotCustomField2.isHidden =
            !this._selectedEntryTransaction?.isLotCustomField2Allowed || lotNotManaged;
        this.destinationLotCustomField3.isHidden =
            !this._selectedEntryTransaction?.isLotCustomField3Allowed || lotNotManaged;
        this.destinationLotCustomField4.isHidden =
            !this._selectedEntryTransaction?.isLotCustomField4Allowed || lotNotManaged;

        const serialNumberManagement = this._product.serialNumberManagementMode;
        this.destinationSerialNumber.isHidden =
            serialNumberManagement === 'notManaged' || serialNumberManagement === 'issued';

        this.destinationLicensePlateNumber.isHidden =
            !this._productSite || !this._productSite.isLicensePlateNumberManaged;
        this.destinationLocation.isHidden = !this._productSite || !this._productSite.isLocationManaged;
        this.destinationLocation.isMandatory = !this.destinationLocation.isHidden;
        this.destinationLocation.isDisabled = false;
        this.destinationReceiptUnit.isDisabled = false;
        this.destinationContainer.isDisabled = false;
        this.destinationLicensePlateNumber.isDisabled = false;
        this.destinationStatus.isHidden = !this._productSite || this._productSite.stockManagementMode === 'notManaged';

        const expirationManagement = this._product.expirationManagementMode;
        const isExpirationManaged = expirationManagement !== 'notManaged';
        this.destinationExpirationDate.isHidden = this.destinationUseByDate.isHidden = !isExpirationManaged;

        switch (this._selectedEntryTransaction?.identifier1Detail) {
            case 'entered':
            case 'displayed': {
                this.destinationIdentifier1.isHidden = false;
                this.destinationIdentifier1.isDisabled = true;
                break;
            }
            default: {
                this.destinationIdentifier1.isHidden = true;
            }
        }

        switch (this._selectedEntryTransaction?.identifier2Detail) {
            case 'entered':
            case 'displayed': {
                this.destinationIdentifier2.isHidden = false;
                this.destinationIdentifier2.isDisabled = true;
                break;
            }
            default: {
                this.destinationIdentifier2.isHidden = true;
            }
        }

        if (!this._productSite || !this._productSite.isLocationManaged) {
            this.destinationIdentifier1.isHidden = true;
            this.destinationIdentifier2.isHidden = true;
        }

        this.destinationContainer.isHidden =
            !this._productSite || lpnManagementRule !== 'mandatory' || !this._productSite.isLicensePlateNumberManaged;
        if (!this.destinationContainer.isHidden) {
            this.destinationContainer.value = this._productSite.defaultInternalContainer;
        }

        const versionManagementMode = this._product.stockVersionMode;

        this.destinationMajorVersion.isHidden = !['major', 'majorAndMinor'].includes(versionManagementMode);
        this.destinationMinorVersion.isHidden = versionManagementMode !== 'majorAndMinor';

        this.destinationQuantityToReceive.value = this.destinationRemainingQuantity.value; // should auto-round value appropriately based on scale property that was set for this field
        this.destinationQuantityToReceive.isDisabled =
            serialNumberManagement !== 'notManaged' && serialNumberManagement !== 'issued';
        this._quantityToReceiveNoRounded = this.destinationQuantityToReceive.value ?? 0;
        this._quantityToReceiveRounded = this.destinationQuantityToReceive.value ?? 0;
        this._remainingQuantityNoRounded = this.destinationRemainingQuantity.value ?? 0;
        this._remainingQuantityRounded = this.destinationRemainingQuantity.value ?? 0;
        this._remainingQuantityInitial = Number(this._salesDeliveryLine.quantityInStockUnitReceived);

        this._salesUnitToStockUnitConversionFactor = this._salesDeliveryLine.salesUnitToStockUnitConversionFactor;
        this._previousSalesUnitToStockUnitConversionFactor =
            this._salesDeliveryLine.salesUnitToStockUnitConversionFactor;
        const selectedUnit = this._packingUnits.find(
            item => item.node.packingUnit.code === this._salesDeliveryLine.salesUnit.code,
        );
        this.destinationPackingUnitToStockUnitConversionFactor.isDisabled = selectedUnit
            ? !selectedUnit.node.isPackingFactorEntryAllowed
            : true;

        if (this._purchaseStockManagementRules) {
            const suggestedLocation = this._productSite
                ? await findDefaultLocation(this._productSite, this._purchaseStockManagementRules, this)
                : null;
            if (!suggestedLocation) {
                this.destinationSuggestedLocation.value = null;
                this.destinationSuggestedLocation.isHidden = true;
            } else {
                this.destinationSuggestedLocation.value = suggestedLocation;
                this.destinationSuggestedLocation.isHidden = false;
                if (this._isLocationPreloaded === true) {
                    try {
                        this.destinationLocation.value = await this.$.graph
                            .node('@sage/x3-stock-data/Location')
                            .read(
                                {
                                    code: true,
                                    type: true,
                                    stockSite: {
                                        code: true,
                                        _id: true,
                                    },
                                    warehouse: {
                                        code: true,
                                        _id: true,
                                    },
                                    _id: true,
                                },
                                `${this.receiptSite.value?.code}|${suggestedLocation}`,
                            )
                            .execute();
                    } catch (e) {
                        await dialogMessage(
                            this,
                            'error',
                            ui.localize('@sage/x3-purchasing-sales/error-loading-location', 'Error loading location'),
                            String(e),
                        );
                        return false;
                    }
                }
            }
        }

        this._initHeader();

        return true;
    }

    private async _getStockJournals(product: string, salesSite: string) {
        try {
            const stockJournals = await this.$.graph
                .node('@sage/x3-stock-data/StockJournal')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            stockSite: {
                                code: true,
                                _id: true,
                            },
                            documentId: true,
                            documentLine: true,
                            sequence: true,
                            product: {
                                code: true,
                                _id: true,
                            },
                            quantityInStockUnit: true,
                            quantityInPackingUnit: true,
                            packingUnit: {
                                code: true,
                                numberOfDecimals: true,
                                _id: true,
                            },
                            isUpdated: true,
                            lot: true,
                            sublot: true,
                            supplierLot: true,
                            lotCustomField1: true,
                            lotCustomField2: true,
                            lotCustomField3: true,
                            lotCustomField4: true,
                            identifier1: true,
                            identifier2: true,
                            majorVersion: {
                                code: true,
                                _id: true,
                            },
                            minorVersion: true,
                            expirationDate: true,
                            useByDate: true,
                            status: {
                                code: true,
                                _id: true,
                            },
                            serialNumber: true,
                            startingSerialNumber: true,
                            endingSerialNumber: true,
                        },
                        {
                            filter: {
                                stockSite: {
                                    code: salesSite,
                                },
                                documentType: 'delivery',
                                ...(this.salesDelivery.value && { documentId: this.salesDelivery.value.id }),
                                documentLine: this._salesDeliveryLine.lineNumber,
                                product: { code: product },
                                quantityInStockUnit: { _lt: 0 },
                                isUpdated: true,
                                isAdjustedMovement: false,
                            },
                            orderBy: {
                                documentId: 1,
                                documentLine: 1,
                                ...(this._product.serialNumberManagementMode === 'receivedIssued' && {
                                    serialNumber: 1,
                                }),
                                sequence: -1,
                            },
                        },
                    ),
                )
                .execute();
            this._stockJournalNodeResponses = [];
            stockJournals.edges.forEach((stockJournal: any) => {
                this._stockJournalNodeResponses.push({
                    stockSite: {
                        code: stockJournal.node.stockSite.code,
                        _id: stockJournal.node.stockSite._id,
                    },
                    documentId: stockJournal.node.documentId,
                    documentLine: stockJournal.node.documentLine,
                    sequence: stockJournal.node.sequence,
                    product: {
                        code: stockJournal.node.product.code,
                        _id: stockJournal.node.product._id,
                    },
                    quantityInStockUnit: Number(stockJournal.node.quantityInStockUnit),
                    quantityInPackingUnit: Number(stockJournal.node.quantityInPackingUnit),
                    packingUnit: {
                        code: stockJournal.node.packingUnit.code,
                        numberOfDecimals: stockJournal.node.packingUnit.numberOfDecimals,
                        _id: stockJournal.node.packingUnit._id,
                    },
                    isUpdated: stockJournal.node.isUpdated,
                    lot: stockJournal.node.lot,
                    sublot: stockJournal.node.sublot,
                    supplierLot: stockJournal.node.supplierLot,
                    lotCustomField1: stockJournal.node.lotCustomField1,
                    lotCustomField2: stockJournal.node.lotCustomField2,
                    lotCustomField3: stockJournal.node.lotCustomField3,
                    lotCustomField4: stockJournal.node.lotCustomField4,
                    identifier1: stockJournal.node.identifier1,
                    identifier2: stockJournal.node.identifier2,
                    ...(stockJournal.node.majorVersion && {
                        majorVersion: {
                            code: stockJournal.node.majorVersion?.code,
                            _id: stockJournal.node.majorVersion._id,
                        },
                    }),
                    minorVersion: stockJournal.node.minorVersion,
                    expirationDate: stockJournal.node.expirationDate,
                    useByDate: stockJournal.node.useByDate,
                    status: stockJournal.node.status?.code,
                    serialNumber: stockJournal.node.serialNumber,
                    startingSerialNumber: stockJournal.node.startingSerialNumber,
                    endingSerialNumber: stockJournal.node.endingSerialNumber,
                    _id: stockJournal.node._id,
                });
            });
            this._currStockJournal = 0;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing-sales/error-loading-stock-journals', 'Error loading stock journals'),
                String(e),
            );
        }
    }

    //this function found the stock movement in awaiting status (isUpdated=false) for the delivery
    private async _getStockJournalsNoUpdated(product: string, salesSite: string): Promise<boolean> {
        try {
            const stockJournals = await this.$.graph
                .node('@sage/x3-stock-data/StockJournal')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                        },
                        {
                            filter: {
                                stockSite: {
                                    code: salesSite,
                                },
                                documentType: 'delivery',
                                ...(this.salesDelivery.value && { documentId: this.salesDelivery.value.id }),
                                documentLine: this._salesDeliveryLine.lineNumber,
                                product: { code: product },
                                quantityInStockUnit: { _lt: 0 },
                                isUpdated: false,
                                isAdjustedMovement: false,
                            },
                        },
                    ),
                )
                .execute();
            return stockJournals.edges.length > 0;
        } catch (e) {
            return false;
        }
    }

    private async _getStatuses(): Promise<string[]> {
        const selectedStatus: { _regex: string }[] = [];
        this._purchaseStockManagementRules.authorizedSubstatus.split(',').forEach(function (status) {
            selectedStatus.push({ _regex: getRegExp(status).source });
        });
        try {
            return extractEdges<{ code: string }>(
                await this.$.graph
                    .node('@sage/x3-stock-data/StockStatus')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                code: true,
                                _id: true,
                            },
                            {
                                filter: {
                                    code: { _or: selectedStatus },
                                },
                            },
                        ),
                    )
                    .execute(),
            ).map(status => status.code);
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing-sales/error-loading-stock-statuses', 'Error loading stock statuses'),
                String(e),
            );
            return [];
        }
    }

    private async _getSupplierProduct(): Promise<SupplierProduct | null> {
        let supplierProduct: SupplierProduct;
        try {
            supplierProduct = (await this.$.graph
                .node('@sage/x3-master-data/SupplierProduct')
                .read(
                    {
                        _id: true,
                        purchaseUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        purchaseUnitToStockUnitConversionFactor: true,
                        packingUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        packingUnitToPurchaseUnitConversionFactor: true,
                    },
                    `${this._salesDeliveryLine.product.code}|${this._salesDeliveryLine.salesOrderLine?.purchaseOrder?.orderFromSupplier?.code?.code}`,
                )
                .execute()) as SupplierProduct;
            return supplierProduct;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize(
                    '@sage/x3-purchasing-sales/error-loading-supplier-product',
                    'Error loading supplier product',
                ),
                String(e),
            );
            return null;
        }
    }

    private _initHeader() {
        if (this.salesDelivery.value && this._salesDeliveryLine) {
            this.headerTitleLeft.value = this.salesDelivery.value.id;
            this.headerTitleRight.value = this._salesDeliveryLine.lineNumber.toString();
            this.headerTitleRight.prefix = ui.localize(
                '@sage/x3-purchasing-sales/pages__mobile_intersite_receipt_header_detail__line_no',
                'Line No.',
            );
            this.headerLine2Left.value = this._product.code;
            this.headerLine2Right.value = this._product.localizedDescription1;
            this.headerLine3Left.value = ui.localize(
                '@sage/x3-purchasing-sales/pages__mobile_intersite_receipt_header_detail__expected_quantity',
                'Expected quantity:',
            );
            const expectedQuantity = Number(this._salesDeliveryLine.quantityInSalesUnitDelivered);
            this.headerLine3Right.value = `${expectedQuantity} ${this._salesDeliveryLine.salesUnit.code}`;
        } else {
            this.headerTitleLeft.value = this._product.code;
            this.headerTitleRight.value = '';
            this.headerLine2Left.value = this._product.localizedDescription1;
            this.headerLine2Right.value = '';
            this.headerLine3Left.value = this.supplier.value?.code;
            this.headerLine3Right.value = '';
        }
    }

    private _initPackingUnitFields() {
        this._packingUnits = [];

        this._packingUnits.push({
            node: {
                packingUnit: {
                    code: this._product.stockUnit.code,
                    numberOfDecimals: this._product.stockUnit.numberOfDecimals,
                    _id: this._product.stockUnit._id,
                },
                packingUnitToStockUnitConversionFactor: '1',
                isPackingFactorEntryAllowed: false,
                _id: this._product.stockUnit._id,
                isPackingUnit: false,
            },
        });

        if (this._product.stockUnit.code !== this._product.purchaseUnit.code) {
            this._packingUnits.push({
                node: {
                    packingUnit: {
                        code: this._product.purchaseUnit.code,
                        numberOfDecimals: this._product.purchaseUnit.numberOfDecimals,
                        _id: this._product.purchaseUnit._id,
                    },
                    packingUnitToStockUnitConversionFactor:
                        this._product.purchaseUnitToStockUnitConversionFactor.toString(),
                    isPackingFactorEntryAllowed: false,
                    _id: this._product.purchaseUnit._id,
                    isPackingUnit: false,
                },
            });
        }

        let productPackingList = extractEdges(this._product.packingUnits.query).filter(productPacking => {
            return !!productPacking.packingUnit?.code;
        });
        let duplicateUnit = false;
        if (productPackingList.length > 0) {
            for (const productPacking of productPackingList) {
                duplicateUnit = false;
                for (const receiptUnit of this._packingUnits) {
                    if (
                        productPacking.packingUnit.code === receiptUnit.node.packingUnit.code &&
                        productPacking.packingUnitToStockUnitConversionFactor ===
                        receiptUnit.node.packingUnitToStockUnitConversionFactor
                    ) {
                        duplicateUnit = true;
                        receiptUnit.node.isPackingUnit = true;
                        break;
                    }
                }
                if (!duplicateUnit) {
                    this._packingUnits.push({
                        node: {
                            packingUnit: {
                                code: productPacking.packingUnit?.code,
                                numberOfDecimals: productPacking.packingUnit?.numberOfDecimals,
                                _id: productPacking.packingUnit?._id,
                            },
                            packingUnitToStockUnitConversionFactor:
                                productPacking.packingUnitToStockUnitConversionFactor,
                            isPackingFactorEntryAllowed: productPacking.isPackingFactorEntryAllowed,
                            _id: productPacking._id,
                            isPackingUnit: true,
                        },
                    });
                }
            }
        }

        if (this._supplierProduct) {
            const receiptUnitOptions = this.destinationReceiptUnit.options;
            if (
                this._supplierProduct.purchaseUnit &&
                !this._packingUnits.find(item => item.node.packingUnit.code === this._supplierProduct.purchaseUnit.code)
            ) {
                this._packingUnits.push({
                    node: {
                        packingUnit: {
                            code: this._supplierProduct.purchaseUnit.code,
                            numberOfDecimals: this._supplierProduct.purchaseUnit.numberOfDecimals,
                            _id: this._supplierProduct.purchaseUnit._id,
                        },
                        packingUnitToStockUnitConversionFactor: '1',
                        isPackingFactorEntryAllowed: false,
                        _id: this._product.stockUnit._id,
                        isPackingUnit: false,
                    },
                });
            }
            if (
                this._supplierProduct.packingUnit &&
                !this._packingUnits.find(item => item.node.packingUnit.code === this._supplierProduct.packingUnit.code)
            ) {
                this._packingUnits.push({
                    node: {
                        packingUnit: {
                            code: this._supplierProduct.packingUnit.code,
                            numberOfDecimals: this._supplierProduct.packingUnit.numberOfDecimals,
                            _id: this._supplierProduct.packingUnit._id,
                        },
                        packingUnitToStockUnitConversionFactor: '1',
                        isPackingFactorEntryAllowed: false,
                        _id: this._product.purchaseUnit._id,
                        isPackingUnit: false,
                    },
                });
            }
        }

        this.destinationPackingUnitToStockUnitConversionFactor.isHidden = false;
        this.destinationReceiptUnit.isHidden = false;
        this.destinationReceiptUnit.options = this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.code);
        this.destinationReceiptUnit.isReadOnly = this.destinationReceiptUnit.options.length <= 1;
        this.destinationReceiptUnit.isDisabled = false;
    }

    private async _validate(): Promise<boolean> {
        const fieldErrors = new Array<string>();
        if (await this._isFieldInvalid(this.destinationReceiptUnit))
            fieldErrors.push(this.destinationReceiptUnit.title ?? '');
        if (await this._isFieldInvalid(this.destinationQuantityToReceive))
            fieldErrors.push(this.destinationQuantityToReceive.title ?? '');
        if (await this._isFieldInvalid(this.destinationStatus)) fieldErrors.push(this.destinationStatus.title ?? '');
        if (await this._isFieldInvalid(this.destinationLicensePlateNumber))
            fieldErrors.push(this.destinationLicensePlateNumber.title ?? '');
        if (await this._isFieldInvalid(this.destinationLocation))
            fieldErrors.push(this.destinationLocation.title ?? '');
        if (await this._isFieldInvalid(this.destinationSerialNumber))
            fieldErrors.push(this.destinationSerialNumber.title ?? '');
        if (await this._isFieldInvalid(this.destinationSupplierLot))
            fieldErrors.push(this.destinationSupplierLot.title ?? '');
        if (await this._isFieldInvalid(this.destinationLot)) fieldErrors.push(this.destinationLot.title ?? '');
        if (await this._isFieldInvalid(this.destinationSublot)) fieldErrors.push(this.destinationSublot.title ?? '');
        if (await this._isFieldInvalid(this.destinationMajorVersion))
            fieldErrors.push(this.destinationMajorVersion.title ?? '');
        if (await this._isFieldInvalid(this.destinationMinorVersion))
            fieldErrors.push(this.destinationMinorVersion.title ?? '');
        if (await this._isFieldInvalid(this.destinationExpirationDate))
            fieldErrors.push(this.destinationExpirationDate.title ?? '');
        if (await this._isFieldInvalid(this.destinationUseByDate))
            fieldErrors.push(this.destinationUseByDate.title ?? '');
        if (await this._isFieldInvalid(this.destinationLotCustomField1))
            fieldErrors.push(this.destinationLotCustomField1.title ?? '');
        if (await this._isFieldInvalid(this.destinationLotCustomField2))
            fieldErrors.push(this.destinationLotCustomField2.title ?? '');
        if (await this._isFieldInvalid(this.destinationLotCustomField3))
            fieldErrors.push(this.destinationLotCustomField3.title ?? '');
        if (await this._isFieldInvalid(this.destinationLotCustomField4))
            fieldErrors.push(this.destinationLotCustomField4.title ?? '');
        if (await this._isFieldInvalid(this.destinationIdentifier1))
            fieldErrors.push(this.destinationIdentifier1.title ?? '');
        if (await this._isFieldInvalid(this.destinationIdentifier2))
            fieldErrors.push(this.destinationIdentifier2.title ?? '');
        if (await this._isFieldInvalid(this.destinationContainer))
            fieldErrors.push(this.destinationContainer.title ?? '');
        if (fieldErrors.length > 0) {
            this.$.removeToasts();
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing-sales/pages__utils__notification__invalid_entry_error',
                    `Check your entry for: {{fieldNames}}`,
                    { fieldNames: fieldErrors },
                ),
                { type: 'error', timeout: 5000 },
            );
            await this.$.page.validate();
        }
        return fieldErrors.length === 0;
    }

    private async _isFieldInvalid(
        field:
            | ui.fields.FilterSelect
            | ui.fields.DropdownList
            | ui.fields.Text
            | ui.fields.Reference
            | ui.fields.Numeric
            | ui.fields.Select
            | ui.fields.Date,
    ): Promise<boolean> {
        return !field.isHidden && ((field.isMandatory && !field.value) || !!(await field.validate()));
    }

    private _calculateRemainingQuantity(stockJournal: ExtractEdgesPartial<StockJournalBinding>): number {
        let pendingQuantityInReceiptUnitReceived = 0;
        if (this.salesDelivery.value && this.lines.value.length > 0) {
            this.lines.value.forEach(line => {
                if (
                    line.salesDelivery.id === this.salesDelivery.value.id &&
                    line.salesDeliveryLine.lineNumber === this._salesDeliveryLine.lineNumber
                ) {
                    line.stockDetails?.forEach(stockDetail => {
                        pendingQuantityInReceiptUnitReceived += Number(stockDetail.quantityInPackingUnit);
                    });
                }
            });
        }
        let _remainingQuantity: number;
        _remainingQuantity = Math.max(
            0,
            (Number(this._salesDeliveryLine.salesUnitToStockUnitConversionFactor) !== 0
                ? (Number(this._salesDeliveryLine.quantityInSalesUnitDelivered) *
                    Number(this._salesDeliveryLine.salesUnitToStockUnitConversionFactor)) /
                Number(stockJournal.packingUnitToStockUnitConversionFactor)
                : 0) - pendingQuantityInReceiptUnitReceived,
        );
        const tmp = Math.pow(10, Number(stockJournal.packingUnit?.numberOfDecimals));
        if (tmp !== 0) {
            _remainingQuantity = Math.round(_remainingQuantity * tmp) / tmp;
        }
        return _remainingQuantity;
    }

    private _calculateRemainingQuantityInStockJournalReponse(stockJournal: StockJournalNodeResponse): number {
        let pendingQuantityInReceiptUnitReceived = 0;
        if (this.salesDelivery.value && this.lines.value.length > 0) {
            this.lines.value.forEach(line => {
                if (
                    line.salesDelivery.id === this.salesDelivery.value.id &&
                    line.salesDeliveryLine.lineNumber === this._salesDeliveryLine.lineNumber
                ) {
                    line.stockDetails?.forEach(stockDetail => {
                        if (stockDetail.sequence === stockJournal.sequence) {
                            pendingQuantityInReceiptUnitReceived += Number(stockDetail.quantityInPackingUnit);
                        }
                    });
                }
            });
        }
        let _remainingQuantity: number;
        _remainingQuantity = Math.max(
            0,
            -(Number(this._salesUnitToStockUnitConversionFactor) !== 0
                ? Number(stockJournal.quantityInStockUnit) / Number(this._salesUnitToStockUnitConversionFactor)
                : 0) - pendingQuantityInReceiptUnitReceived,
        );
        const selectedUnit = this._packingUnits.find(
            line => line.node.packingUnit.code === this.destinationReceiptUnit.value,
        );
        const tmp = Math.pow(10, Number(selectedUnit?.node?.packingUnit?.numberOfDecimals));
        if (tmp !== 0) {
            _remainingQuantity = Math.round(_remainingQuantity * tmp) / tmp;
        }
        return _remainingQuantity;
    }

    private _calculateRemainingQuantityInLine(
        packingUnitCode: string,
        packingUnitToStockUnitConversionFactor: number,
    ): number {
        let pendingQuantityInReceiptUnitReceived = 0;
        if (this.salesDelivery.value && this.lines.value.length > 0) {
            this.lines.value.forEach(line => {
                if (
                    line.salesDelivery.id === this.salesDelivery.value.id &&
                    line.salesDeliveryLine.lineNumber === this._salesDeliveryLine.lineNumber
                ) {
                    line.stockDetails?.forEach(stockDetail => {
                        pendingQuantityInReceiptUnitReceived += Number(stockDetail.quantityInPackingUnit);
                    });
                }
            });
        }
        let _remainingQuantity: number;
        _remainingQuantity = Math.max(
            0,
            (Number(packingUnitToStockUnitConversionFactor) !== 0
                ? (Number(this._salesDeliveryLine.quantityInSalesUnitDelivered) *
                    Number(this._salesUnitToStockUnitConversionFactor)) /
                packingUnitToStockUnitConversionFactor
                : 0) - pendingQuantityInReceiptUnitReceived,
        );
        const selectedUnit = this._packingUnits.find(line => line.node.packingUnit.code === packingUnitCode);
        const tmp = Math.pow(10, Number(selectedUnit?.node?.packingUnit?.numberOfDecimals));
        if (tmp !== 0) {
            _remainingQuantity = Math.round(_remainingQuantity * tmp) / tmp;
        }
        return _remainingQuantity;
    }

    private _getNextStockJournal(): StockJournalNodeResponse | null {
        const startingCurrStockJournal = this._currStockJournal;
        let lookForStockJournals: boolean;
        lookForStockJournals = true;
        let nextStockJournal: StockJournalNodeResponse | null;
        nextStockJournal = null;
        while (lookForStockJournals) {
            if (this._currStockJournal < this._stockJournalNodeResponses.length - 1) {
                this._currStockJournal += 1;
            } else {
                this._currStockJournal = 0;
            }
            if (
                this._calculateRemainingQuantityInStockJournalReponse(
                    this._stockJournalNodeResponses[this._currStockJournal],
                ) !== 0
            ) {
                nextStockJournal = this._stockJournalNodeResponses[this._currStockJournal];
                lookForStockJournals = false;
            }
            if (this._currStockJournal === startingCurrStockJournal) {
                lookForStockJournals = false;
            }
        }
        if (!nextStockJournal) {
            this._currStockJournal = startingCurrStockJournal;
        }
        return nextStockJournal;
    }

    private _getDetailsNextButtonHidden(): boolean {
        if (this._productSalesSite && this._productSalesSite.stockManagementMode !== 'notManaged') {
            let stockJournalWithRemainingQuantityNotNullCount: number;
            stockJournalWithRemainingQuantityNotNullCount = 0;
            this._stockJournalNodeResponses.forEach(stockJournal => {
                if (this._calculateRemainingQuantityInStockJournalReponse(stockJournal) !== 0) {
                    stockJournalWithRemainingQuantityNotNullCount += 1;
                }
            });
            return stockJournalWithRemainingQuantityNotNullCount === 0;
        } else {
            return true;
        }
    }

    private async _updatePurchaseReceiptLine() {
        if (this.lines.value.length > 0) {
            const purchaseReceiptLine = this.lines.getRecordValue((this.lines.value.length - 1).toString(), 0);
            const selectedUnit = this._packingUnits.find(
                line => line.node.packingUnit.code === this.destinationReceiptUnit.value,
            );
            if (purchaseReceiptLine && selectedUnit) {
                purchaseReceiptLine.lineNumber = Number(this.lines.value.length * 1000);
                purchaseReceiptLine.receiptUnit = selectedUnit.node.packingUnit;
                purchaseReceiptLine.receiptUnitToStockUnitConversionFactor = this._salesUnitToStockUnitConversionFactor;
                let quantityInReceiveUnit: number;
                quantityInReceiveUnit = 0;
                purchaseReceiptLine.stockDetails?.forEach(item => {
                    quantityInReceiveUnit += Number(item.quantityInPackingUnit);
                });
                purchaseReceiptLine.quantityInReceiptUnitReceived = quantityInReceiveUnit.toString();
            }

            const lineIndex = this._salesDeliveryLines.findIndex(
                lineNumber => Number(lineNumber) === Number(this._salesDeliveryLine.lineNumber),
            );
            if (lineIndex >= 0) {
                this._salesDeliveryLines.splice(lineIndex, 1);
            }
            this.productBlock.isHidden = false;
            this.destinationBlock.isHidden = true;
            this.customizationIntersiteReceiptHeaderBlock.isHidden =
                this.customizationIntersiteReceiptHeaderBlock.layout.$layout?.$items.length === 0;
            this.destinationBlock.isHidden = true;
            this.customizationIntersiteReceiptLinesBlock.isHidden = true;
            this.customizationIntersiteReceiptDetailsBlock.isHidden = true;
            this.receiptBlock.isHidden = this.lines.value.length === 0;
            this.addButton.isHidden = true;
            this.detailsNextButton.isHidden = true;
            if (this.lines.value.length === 0) {
                this.salesDelivery.value = null;
                this.salesDelivery.isDisabled = false;
                this.salesDeliveryLine.isDisabled = true;
                this.salesDeliveryLine.value = null;
                this.createButton.isHidden = false;
                this.createButton.isDisabled = true;
            } else {
                this.salesDelivery.isDisabled = true;
                this.salesDeliveryLine.isDisabled = false;
                this.salesDeliveryLine.value = null;
                this.createButton.isHidden = false;
                this.createButton.isDisabled = false;
            }
        }
    }

    private _displayStockJournal(stockJournal: StockJournalNodeResponse) {
        this._sequence = stockJournal.sequence;
        this.destinationRemainingQuantity.value = this._calculateRemainingQuantityInStockJournalReponse(stockJournal);
        this.destinationQuantityToReceive.value = this.destinationRemainingQuantity.value;
        this.destinationPackingUnitToStockUnitConversionFactor.value = Number(
            this._salesUnitToStockUnitConversionFactor,
        );
        if (stockJournal.startingSerialNumber) {
            this.destinationSerialNumber.value = stockJournal.startingSerialNumber;
        } else {
            this.destinationSerialNumber.value = stockJournal.serialNumber;
        }
        this.destinationSupplierLot.value = stockJournal.supplierLot;
        this.destinationLot.value = stockJournal.lot;
        this.destinationSublot.value = stockJournal.sublot;
        this.destinationMajorVersion.value = stockJournal.majorVersion
            ? { code: stockJournal.majorVersion.code, _id: stockJournal.majorVersion._id }
            : null;
        this.destinationMinorVersion.value = stockJournal.minorVersion;
        this.destinationLotCustomField1.value = stockJournal.lotCustomField1;
        this.destinationLotCustomField2.value = stockJournal.lotCustomField2;
        this.destinationLotCustomField3.value = Number(stockJournal.lotCustomField3);
        this.destinationLotCustomField4.value = stockJournal.lotCustomField4?.startsWith('1599')
            ? null
            : stockJournal.lotCustomField4;
        this.destinationIdentifier1.value = stockJournal.identifier1;
        this.destinationIdentifier2.value = stockJournal.identifier2;
        this.destinationExpirationDate.value = stockJournal.expirationDate;
        this.destinationUseByDate.value = stockJournal.useByDate;
        this._showAddButton();
    }

    private _displayLine() {
        this.destinationRemainingQuantity.value = this._calculateRemainingQuantityInLine(
            this.destinationReceiptUnit.value ?? '',
            Number(this._salesUnitToStockUnitConversionFactor),
        );
        this.destinationQuantityToReceive.value = this.destinationRemainingQuantity.value;
        this.destinationPackingUnitToStockUnitConversionFactor.value = Number(
            this._salesUnitToStockUnitConversionFactor,
        );
        this._showAddButton();
    }

    private _showAddButton() {
        this.addButton.isDisabled =
            !this.destinationReceiptUnit.value ||
            Number(this.destinationQuantityToReceive.value) <= 0 ||
            (!this.destinationContainer.isHidden &&
                !this.destinationLicensePlateNumber.isHidden &&
                (!this.destinationLicensePlateNumber.value || this.destinationLicensePlateNumber.value?.code === '') &&
                (!this.destinationContainer.value || this.destinationContainer.value?.code === ''));
    }

    private async getSidebarData(rowId: string, level: number): Promise<boolean> {
        this._sideBarRowId = rowId;
        this._sideBarLevel = level;
        const purchaseReceiptLine = this.lines.getRecordValue(
            level === 0 ? rowId : rowId.substring(0, rowId.indexOf('|')),
            0,
        );
        if (
            !purchaseReceiptLine ||
            !purchaseReceiptLine.lineNumber ||
            !purchaseReceiptLine.stockDetails ||
            purchaseReceiptLine.stockDetails.length === 0 ||
            !purchaseReceiptLine.stockDetails[0]._id
        ) {
            return false;
        }
        const stockJournal = this.lines.getRecordValue(
            level === 0 ? purchaseReceiptLine.stockDetails[0]._id : rowId,
            1,
        );
        if (!stockJournal || !stockJournal._id) {
            return false;
        }
        const productCode = purchaseReceiptLine.product?.code;
        if (productCode) {
            this._product = await this.getProduct(productCode);
        }
        if (productCode && this.receiptSite.value) {
            this._productSite = await this.getProductSite(productCode, this.receiptSite.value?.code ?? '');
        }
        if (productCode && this.salesDelivery.value) {
            this._productSalesSite = await this.getProductSite(productCode, this.salesDelivery.value.salesSite.code);
        }
        const salesDeliveryLine = await this._getSalesDeliveryLine(purchaseReceiptLine.salesDeliveryLine.lineNumber);
        if (salesDeliveryLine) {
            this._salesDeliveryLine = salesDeliveryLine;
        }
        this._purchaseStockManagementRules = await findStockManagementRules(
            this.receiptSite.value?.code ?? '',
            this._product.productCategory.code,
            '3',
            this._selectedEntryTransaction?.stockMovementCode?.code ?? undefined,
            this,
        );
        this._initPackingUnitFields();
        this.destinationPreviousReceiptUnit.value =
            stockJournal.packingUnit && stockJournal.packingUnit.code
                ? stockJournal.packingUnit.code
                : this._product.stockUnit.code;
        this._previousSalesUnitToStockUnitConversionFactor =
            this._salesDeliveryLine.salesUnitToStockUnitConversionFactor;
        this._quantityToReceiveNoRounded = Number(stockJournal.quantityInPackingUnit);
        this._quantityToReceiveRounded = Number(stockJournal.quantityInPackingUnit);
        this._previousQuantityToReceive = Number(stockJournal.quantityInPackingUnit);
        //    this._previousContainerCode = stockJournal.container?.code;
        this._previousLocationCode = stockJournal.location?.code;
        return true;
    }

    private _getSidebarPurchaseReceiptLine(): ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>[] = [];
        this.customizationIntersiteReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>);
            }
        });
        return _FieldsSidebar;
    }

    private _getSidebarStockJournal(): ui.SidebarFieldDefinition<StockJournalBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockJournalBinding>[] = [];
        _FieldsSidebar.push('packingUnit');
        _FieldsSidebar.push('quantityInPackingUnit');
        if ((this.destinationReceiptUnit.options ?? []).length > 1) {
            _FieldsSidebar.push('packingUnitToStockUnitConversionFactor');
        }
        if (this._productSite && this._productSite.stockManagementMode !== 'notManaged') {
            _FieldsSidebar.push('status');
        }
        if (this._productSite && this._productSite.isLicensePlateNumberManaged) {
            _FieldsSidebar.push('licensePlateNumber');
        }
        if (this._productSite && this._productSite.isLocationManaged) {
            _FieldsSidebar.push('location');
        }
        if (!['notManaged', 'issued'].includes(this._product.serialNumberManagementMode)) {
            _FieldsSidebar.push('serialNumber');
        }
        if (this._product.lotManagementMode !== 'notManaged') {
            _FieldsSidebar.push('supplierLot');
        }
        if (this._product.lotManagementMode !== 'notManaged') {
            _FieldsSidebar.push('lot');
        }
        if (this._product.lotManagementMode === 'lotAndSublot') {
            _FieldsSidebar.push('sublot');
        }
        if (this._product.lotManagementMode !== 'notManaged' && this._product.stockVersionMode !== 'no') {
            _FieldsSidebar.push('majorVersion');
        }
        if (this._product.lotManagementMode !== 'notManaged' && this._product.stockVersionMode === 'majorAndMinor') {
            _FieldsSidebar.push('minorVersion');
        }
        if (
            this._product.lotManagementMode !== 'notManaged' &&
            (this._selectedEntryTransaction?.isLotExpirationDateAllowed ||
                this._product.expirationManagementMode === 'manualEntry' ||
                this._product.expirationManagementMode === 'mandatoryEntry')
        ) {
            _FieldsSidebar.push('expirationDate');
            _FieldsSidebar.push('useByDate');
        }
        if (
            this._product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField1Allowed
        ) {
            _FieldsSidebar.push('lotCustomField1');
        }
        if (
            this._product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField2Allowed
        ) {
            _FieldsSidebar.push('lotCustomField2');
        }
        if (
            this._product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField3Allowed
        ) {
            _FieldsSidebar.push('lotCustomField3');
        }
        if (
            this._product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField4Allowed
        ) {
            _FieldsSidebar.push('lotCustomField4');
        }
        if (
            this._selectedEntryTransaction?.identifier1Detail === 'entered' ||
            this._selectedEntryTransaction?.identifier1Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier1');
        }
        if (
            this._selectedEntryTransaction?.identifier2Detail === 'entered' ||
            this._selectedEntryTransaction?.identifier2Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier2');
        }
        this.customizationIntersiteReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockJournalBinding>);
            }
        });

        return _FieldsSidebar;
    }

    private getcustomizationIntersiteReceiptLines() {
        let response = {};
        this.customizationIntersiteReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private getcustomizationIntersiteReceiptDetails() {
        let response = {};
        this.customizationIntersiteReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private async _getSalesDeliveryLine(
        salesDeliveryLineNumber: number,
    ): Promise<SalesDeliveryLineNodeResponse | null> {
        if (!this.salesDelivery.value) {
            return null;
        }
        try {
            return await this.$.graph
                .node('@sage/x3-sales/SalesDeliveryLine')
                .read(
                    {
                        _id: true,
                        salesDelivery: {
                            id: true,
                            _id: true,
                            salesSite: {
                                code: true,
                                _id: true,
                            },
                            receivingSite: {
                                code: true,
                                _id: true,
                            },
                        },
                        lineNumber: true,
                        product: {
                            code: true,
                            _id: true,
                        },
                        stockUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        salesUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        salesOrderLine: {
                            purchaseOrder: {
                                orderFromSupplier: {
                                    code: {
                                        code: true,
                                        _id: true,
                                    },
                                    _id: true,
                                },
                                _id: true,
                            },
                            _id: true,
                        },
                        quantityInStockUnitReceived: true,
                        quantityInSalesUnitDelivered: true,
                        salesUnitToStockUnitConversionFactor: true,
                        majorProductVersion: { code: true },
                        minorProductVersion: true,
                    },
                    `${this.salesDelivery.value.id}|${salesDeliveryLineNumber}`,
                )
                .execute();
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing-sales/error-loading-sales-delivery', 'Error loading sales delivery'),
                String(e),
            );
            return null;
        }
    }
}
