import { Dict, ExtractEdges, extractEdges, ExtractEdgesPartial, withoutEdges } from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import * as ui from '@sage/xtrem-ui';

import { AutomaticJournal } from '@sage/x3-finance-data-api';
import { MiscellaneousTable, Site } from '@sage/x3-system-api';
import { AsyncVoidFunction } from '@sage/x3-system/lib/client-functions/screen-management';
import { getRegExp } from '@sage/x3-system/lib/shared-functions/pat-converter';

import { MajorVersionStatus, Product, ProductSite } from '@sage/x3-master-data-api';
import { UnitOfMeasure } from '@sage/x3-master-data-api-partial';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import { GraphApi, MiscellaneousIssueLineBinding, StockEntryTransaction } from '@sage/x3-stock-api';
import {
    LicensePlateNumber,
    Location,
    Lot,
    LotsSites,
    MobileSettings,
    SerialNumber,
    Stock,
    StockJournal,
    StockJournalBinding,
    StockManagementRules,
    StockSearchFilter,
    StockStatus,
} from '@sage/x3-stock-data-api';
import { other } from '@sage/x3-stock-data/build/lib/menu-items/other';
import { getCountSerialNumber } from '@sage/x3-stock-data/lib/client-functions/get-count-serial-number';
import { MAX_INT_32 } from '@sage/xtrem-shared';
import { NotifyAndWait } from '../client-functions/display';
import { getUnitNumberOfDecimalList } from '../client-functions/get-unit-number-decimals';
import { findSetting, managePages } from '../client-functions/manage-pages';
import { findStockManagementRules } from '../client-functions/stock-management-rules';
import { DeepPartial, packingUnit } from '../client-functions/stock-types';

// Key to use with Composite Data for this application
export const mobileApplicationKey = 'MobileMiscellaneousIssueKey';

/**
 * Composite data management for this application.
 * @enum MiscellaneousIssueDictionaryFieldKey
 *
 * This enum provides per field the associated access key in the data dictionary.
 * It must be maintained in accordance with the fields of this dictionary.
 * Never translate..
 */

// TODO: Assign the field names of the dictionary to the corresponding keys of the enum when the values are available

enum MiscellaneousIssueDictionaryFieldKey {
    identifier1Filter = 'identifier1Filter',
    identifier2Filter = 'identifier2Filter',
    licensePlateNumberFilter = 'licensePlateNumberFilter',
    locationFilter = 'locationFilter',
    locationSelected = 'locationSelected',
    lotFilter = 'lotFilter',
    packingUnitFilter = 'packingUnitFilter',
    packingUnitToIssue = 'packingUnitToIssue',
    packingUnitToStockUnitConversionFactorToIssue = 'packingUnitToStockUnitConversionFactorToIssue',
    productSelected = 'productSelected',
    quantityToIssue = 'quantityToIssue',
    serialNumberFilter = 'serialNumberFilter',
    statusFilter = 'statusFilter',
    stockCustomField1Filter = 'stockCustomField1Filter',
    stockCustomField2Filter = 'stockCustomField2Filter',
    sublotFilter = 'sublotFilter',
}

type PartialStockEntryTransaction = DeepPartial<StockEntryTransaction>;

const hideWhenEmptyValue = (value: any, rowValue?: Dict<Stock>) => {
    return typeof value !== 'number' && !value; // we don't want to hide numeric field with value of 0
};
@ui.decorators.page<MobileMiscellaneousIssue>({
    title: 'Miscellaneous issue',
    module: 'x3-stock',
    mode: 'default',
    menuItem: other,
    priority: 200,
    isTitleHidden: true,
    authorizationCode: 'CWSSMO',
    node: '@sage/x3-stock/MiscellaneousIssue',
    access: { node: '@sage/x3-stock/MiscellaneousIssue' },
    skipDirtyCheck: true,
    navigationPanel: undefined,
    async onLoad() {
        if (!(await this._init())) {
            this._disablePage();
        }
    },
    businessActions() {
        return [this.createButton, this.nextButton, this.CancelButton, this.SelectButton];
    },
    headerCard() {
        if (!this.stockLineSection?.isHidden) {
            return {
                title: this.productCode,
                line2: this.productDescription,
            };
        }
        return undefined;
    },
})
export class MobileMiscellaneousIssue extends ui.Page<GraphApi> {
    private _transactions: PartialStockEntryTransaction[];
    private _transaction: PartialStockEntryTransaction;
    private _notifier = new NotifyAndWait(this);
    private _mobileSettings: MobileSettings;
    private _stockFieldSettings: StockSearchFilter[] = [];
    private _numberSession: number;
    private _productSite: ProductSite;
    private _packingUnits: packingUnit[];
    private _selectedStockManagementRules: StockManagementRules;
    private _originalStockLines: any[];
    private _productAtLocation: string[];
    private _numberOfDecimalList: ExtractEdgesPartial<UnitOfMeasure>[];
    private _lots: string[] = [];

    // =================================================================================================================
    // 🟢 FIRST SECTION DISPLAYED
    // Responsible for initializing and rendering the main view.
    // =================================================================================================================

    // Header card
    @ui.decorators.textField<MobileMiscellaneousIssue>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    productCode: ui.fields.Text;

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    productDescription: ui.fields.Text;

    @ui.decorators.section<MobileMiscellaneousIssue>({
        isTitleHidden: true,
    })
    headerSection: ui.containers.Section;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.headerSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    headerBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.headerSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationMiscellaneousIssueHeaderBlock: ui.containers.Block;

    // Mandatory fields for mutation
    @ui.decorators.textField<MobileMiscellaneousIssue>({
        placeholder: 'Scan...',
        isHidden: true,
    })
    id: ui.fields.Text<MobileMiscellaneousIssue>;

    @ui.decorators.selectField<MobileMiscellaneousIssue>({
        isHidden: true,
    })
    entryType: ui.fields.Select<'miscellaneousIssue', MobileMiscellaneousIssue>;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        placeholder: 'Scan or select...',
        canFilter: false,
        isAutoSelectEnabled: true,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 9,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementGroup: ui.fields.Reference;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        placeholder: 'Scan or select...',
        canFilter: false,
        isAutoSelectEnabled: true,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 14,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementCode: ui.fields.Reference;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, AutomaticJournal>({
        node: '@sage/x3-finance-data/AutomaticJournal',
        valueField: 'code',
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        isHidden: true,
        canFilter: false,
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockAutomaticJournal: ui.fields.Reference;

    @ui.decorators.dateField<MobileMiscellaneousIssue>({
        parent() {
            return this.headerBlock;
        },
        title: 'Issue date',
        isMandatory: true,
        width: 'small',
        maxDate: DateValue.today().toString(),
        onChange() { },
    })
    effectiveDate: ui.fields.Date;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, Site>({
        parent() {
            return this.headerBlock;
        },
        title: 'Site',
        node: '@sage/x3-system/Site',
        valueField: 'code',
        placeholder: 'Scan or select...',
        helperTextField: 'name',
        canFilter: false,
        isAutoSelectEnabled: true,
        isMandatory: true,
        isReadOnly: true,
        isHidden: true,
        bind: 'stockSite',
        width: 'small',
        filter() {
            return {
                code: String(this.$.storage.get('mobile-selected-stock-site')),
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: 'name',
            }),
            ui.nestedFields.text({
                bind: '_id',
            }),
        ],
    })
    stockSite: ui.fields.Reference;

    @ui.decorators.dropdownListField<MobileMiscellaneousIssue>({
        parent() {
            return this.headerBlock;
        },
        title: 'Transaction',
        isTransientInput: true,
        async onChange() {
            await this._onChange_transaction();
        },
    })
    transaction: ui.fields.DropdownList;

    /**
     * OnChange transaction readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_transaction: AsyncVoidFunction = async () => {
        if (this.transaction.value) {
            const transaction = this._transactions.find(trs => trs.code === this.transaction.value);
            await this._transactionChange(transaction as any, false);
        } else {
            this.productSelected.isDisabled = true;
        }
    };

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.headerSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    productBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, Location>({
        parent() {
            return this.productBlock;
        },
        title: 'Location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        placeholder: 'Scan or select...',
        helperTextField: 'type',
        canFilter: false,
        isAutoSelectEnabled: true,
        isTransient: true,
        width: 'large',
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite?.value?.code },
                category: { _nin: ['subcontract', 'customer'] },
            };
            return locationFilter;
        },
        async onChange() {
            await this._onChange_locationSelected();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
            ui.nestedFields.checkbox({
                bind: 'isSingleProduct',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    locationSelected: ui.fields.Reference;

    /*
     * onChange locationSelected event (for decorator and management service)
     */
    private readonly _onChange_locationSelected: AsyncVoidFunction = async () => {
        if (!this.locationSelected.value) {
            this.productSelected.value = null;
            this._productAtLocation = [];
            return;
        }

        await this._getProductAtLocation();

        if (this._productAtLocation?.length > 0 && this.locationSelected.value.isSingleProduct) {
            this.productSelected.value = { code: this._productAtLocation[0] };
            this.locationFilter.value = this.locationSelected.value;
            await this.$.commitValueAndPropertyChanges();
            await this._onChange_productSelected();
        }

        this.locationSelected.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, Product>({
        parent() {
            return this.productBlock;
        },
        title: 'Product',
        node: '@sage/x3-master-data/Product',
        valueField: 'code',
        helperTextField: 'upc',
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        isTransient: true,
        isFullWidth: true,
        shouldSuggestionsIncludeColumns: true,
        canFilter: false,
        filter() {
            const { stockSite, _productAtLocation, miscellaneousIssueLines } = this;
            const stockSiteCode = stockSite.value?.code;
            const productCodesToExclude = miscellaneousIssueLines.value.map(line => line.product?.code);

            function splitArrayIntoChunks<T>(array: T[], chunkSize: number): T[][] {
                const chunks: T[][] = [];
                for (let i = 0; i < array.length; i += chunkSize) {
                    chunks.push(array.slice(i, i + chunkSize));
                }
                return chunks;
            }
            const chunkSize = 1000;
            const inClauses = _productAtLocation?.length
                ? splitArrayIntoChunks(_productAtLocation, chunkSize).map(chunk => ({
                    code: { _in: chunk },
                }))
                : [{ code: { _nin: [] as string[] } }];

            return {
                productStatus: { _ne: 'notUsable' },
                _and: [
                    { stockManagementMode: { _ne: 'notManaged' } },
                    {
                        productSites: {
                            _atLeast: 1,
                            ...(stockSiteCode && { stockSite: { code: stockSiteCode } }),
                            stockManagementMode: { _ne: 'notManaged' },
                        },
                    },
                    {
                        code: {
                            ...(productCodesToExclude.length && { _nin: productCodesToExclude }),
                        },
                    },
                ],
                _or: inClauses,
            };
        },
        async onChange() {
            await this._onChange_productSelected();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: 'localizedDescription1',
            }),
            ui.nestedFields.text({
                bind: 'upc',
            }),
            ui.nestedFields.text({
                bind: 'globalTradeItemNumber',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'lotManagementMode',
                optionType: '@sage/x3-stock-data/LotManagementMode',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'serialNumberManagementMode',
                optionType: '@sage/x3-master-data/SerialNumberManagement',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'expirationManagementMode',
                optionType: '@sage/x3-master-data/ExpirationManagement',
                isHidden: true,
            }),
            ui.nestedFields.select({
                bind: 'stockVersionMode',
                optionType: '@sage/x3-stock-data/StockVersionMode',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Product, UnitOfMeasure>({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'stockUnit',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Product, UnitOfMeasure>({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'stockUnit',
                valueField: 'numberOfDecimals',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isVersionManaged',
                isHidden: true,
            }),
        ],
    })
    productSelected: ui.fields.Reference;

    @ui.decorators.section<MobileMiscellaneousIssue>({
        isTitleHidden: true,
    })
    miscellaneousIssueLinesSection: ui.containers.Section;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.miscellaneousIssueLinesSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    miscellaneousIssueLinesBlock: ui.containers.Block;

    @ui.decorators.nestedGridField<MobileMiscellaneousIssue, [MiscellaneousIssueLineBinding, StockJournalBinding]>({
        parent() {
            return this.miscellaneousIssueLinesBlock;
        },
        bind: 'miscellaneousIssueLines',
        title: 'Miscellaneous issue lines',
        isTitleHidden: true,
        canFilter: false,
        canActivate: true,
        isChangeIndicatorDisabled: true,
        canSelect: false,
        hasLineNumbers: false,
        levels: [
            {
                node: '@sage/x3-stock/MiscellaneousIssueLine',
                childProperty: 'stockDetails',
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'ID',
                        isHiddenDesktop: false,
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lineNumber',
                        title: 'Line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.select({
                        bind: 'entryType',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, MiscellaneousIssueLineBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: { code: true },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'productDescription',
                        title: 'Description',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity in packing unit',
                        isReadOnly: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, MiscellaneousIssueLineBinding, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Packing unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: { code: true },
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimal places',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'stockId',
                        title: 'Stock ID',
                        isTransientInput: true,
                        isHidden: true,
                        isReadOnly: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberSession',
                        title: 'Number session',
                        isTransientInput: true,
                        isHidden: true,
                        isReadOnly: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isHidden(recordId, _rowItem, _level, _parentIds) {
                            return this.miscellaneousIssueLines.getRecordValue(recordId, 0)?.stockDetails?.length !== 1;
                        },
                        async onClick(recordId) {
                            const miscellaneousIssueLine = this.miscellaneousIssueLines.getRecordValue(recordId, 0);
                            if (
                                miscellaneousIssueLine &&
                                miscellaneousIssueLine.stockId &&
                                miscellaneousIssueLine.stockDetails?.length === 1
                            ) {
                                const stockJournal = miscellaneousIssueLine.stockDetails[0];
                                this._stockId.value = miscellaneousIssueLine.stockId;
                                const _productCode = miscellaneousIssueLine.product?.code;
                                if (_productCode) {
                                    this._productSite = await this._getProductSite(_productCode);
                                }
                                const packingUnits =
                                    this._productSite.product.packingUnits?.query.edges.map(
                                        packingUnit => packingUnit.node.packingUnit.code,
                                    ) || [];
                                this.packingUnitFilter.options = [
                                    this._productSite.product.stockUnit.code,
                                    ...packingUnits,
                                ];
                                await this._getStockByStockId(miscellaneousIssueLine.stockId);
                                this.miscellaneousIssueLines.openSidebar(stockJournal._id, 1);
                            }
                        },
                    },
                    {
                        icon: 'delete',
                        title: 'Delete',
                        onClick(rowId) {
                            this.miscellaneousIssueLines.removeRecord(rowId, 0);
                            this.transaction.isDisabled = false;
                            if (this.miscellaneousIssueLines.value.length === 0) {
                                this.createButton.isDisabled = true;
                                this.miscellaneousIssueLines.value = [];
                                this.miscellaneousIssueLinesBlock.isHidden = true;
                                this._productAtLocation = [];
                            }
                        },
                    },
                ],
                sidebar: {
                    layout() {
                        return {
                            mainSection: {
                                title: 'Miscellaneous issue lines',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarMiscellaneousIssueLine(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
            {
                node: '@sage/x3-stock-data/StockJournal',
                columns: [
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity in packing unit',
                        isFullWidth: true,
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                        validation(value, rowValue) {
                            return this._validateEntry(value, rowValue);
                        },
                        onChange(_id, value) {
                            if (value.quantityInPackingUnit) {
                                if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                                    value.startingSerialNumber = {
                                        code: value.serialNumber,
                                        _id: value.serialNumber,
                                    };
                                }
                                this._updateRecordValue(_id, value);
                            }
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInStockUnit',
                        title: 'Quantity in stock unit',
                        isDisabled: true,
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, StockJournal, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: 'code',
                        isHidden: true,
                        isDisabled: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Conversion factor',
                        isFullWidth: false,
                        isDisabled: true,
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, StockJournal, Location>({
                        bind: 'location',
                        title: 'Location',
                        node: '@sage/x3-stock-data/Location',
                        valueField: { code: true },
                        isFullWidth: true,
                        filter() {
                            return {
                                stockSite: { code: this.stockSite?.value?.code },
                                category: { _nin: ['subcontract', 'customer'] },
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, StockJournal, StockStatus>({
                        bind: 'status',
                        title: 'Status',
                        node: '@sage/x3-stock-data/StockStatus',
                        valueField: { code: true },
                        isFullWidth: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, StockJournal, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isDisabled: true,
                        isFullWidth: false,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimal places',
                                isHidden: true,
                            }),
                        ],
                        filter() {
                            return {
                                code: {
                                    _in: this.packingUnitToIssue?.options,
                                },
                            };
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousIssue, StockJournal, UnitOfMeasure>({
                        bind: 'stockUnit',
                        title: 'Stock unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isDisabled: true,
                        isFullWidth: false,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimal places',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'serialNumber',
                        title: 'Serial number',
                        isDisabled: true,
                        isFullWidth: true,
                        validation: /^$|^[^|]+$/,
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier1',
                        title: 'Identifier 1',
                        isHidden: true,
                        isDisabled: true,
                        isFullWidth: true,
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier2',
                        title: 'Identifier 2',
                        isHidden: true,
                        isDisabled: true,
                        isFullWidth: true,
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                    }),
                    ui.nestedFields.filterSelect<MobileMiscellaneousIssue, StockJournal, SerialNumber>({
                        bind: 'startingSerialNumber',
                        title: 'Starting serial number',
                        placeholder: 'Scan or select...',
                        node: '@sage/x3-stock-data/SerialNumber',
                        valueField: 'code',
                        canFilter: false,
                        isTransient: true,
                        isFullWidth: true,
                        filter() {
                            return {
                                product: { code: this._productSite?.product?.code },
                                stockId: this._stockId?.value ?? undefined,
                            };
                        },
                        isHidden() {
                            return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                        },
                        async validation(value, rowValue) {
                            if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                                const regex = /^$|^[^|]+$/;
                                if (!(value.toString().match(regex)?.length ?? 0)) {
                                    return ui.localize(
                                        '@sage/x3-stock/validate-error-invalid-value',
                                        'Incorrect value',
                                    );
                                }
                                return await this._validateSerialNumber(value, rowValue);
                            }
                            return undefined;
                        },
                        async onChange(_id, value) {
                            await this._validateSerialNumber(_id, value);
                            this._updateRecordValue(_id, value);
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                                isReadOnly: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'endingSerialNumber',
                        title: 'Ending serial number',
                        placeholder: 'Scan...',
                        isDisabled: true,
                        isFullWidth: true,
                        isTransientInput: true,
                        isHidden() {
                            return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                        },
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        async onClick(recordId) {
                            const [stockId, lineNumber] = recordId.split('|');
                            this._stockId.value = stockId;
                            this._numberSession = Number(
                                this.miscellaneousIssueLines.getRecordValue(lineNumber, 0)?.numberSession,
                            );
                            const _productCode = this.miscellaneousIssueLines.getRecordValue(lineNumber, 0)?.product
                                ?.code;
                            if (_productCode) {
                                this._productSite = await this._getProductSite(_productCode);
                            }

                            const packingUnits =
                                this._productSite.product.packingUnits?.query.edges.map(
                                    packingUnit => packingUnit.node.packingUnit.code,
                                ) || [];
                            this.packingUnitFilter.options = [
                                this._productSite.product.stockUnit.code,
                                ...packingUnits,
                            ];

                            if (this.miscellaneousIssueLines.value.length > 0) {
                                await this._getStockByStockId(stockId);
                            }

                            this.miscellaneousIssueLines.openSidebar(recordId, 1);
                        },
                    },
                ],
                sidebar: {
                    title: 'Miscellaneous issue',
                    async onRecordConfirmed(_id) {
                        const [, lineNumber] = _id.split('|');
                        this._updateMiscellaneousIssueLine(lineNumber);
                    },
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock details',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarStockJournal(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
        ],
    })
    miscellaneousIssueLines: ui.fields.NestedGrid<[MiscellaneousIssueLineBinding, StockJournalBinding]>;

    @ui.decorators.pageAction<MobileMiscellaneousIssue>({
        title: 'Create',
        buttonType: 'primary',
        isDisabled: true,
        shortcut: ['f2'],
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-stock/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-stock/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);
                message = `**${ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__creation_error',
                    'An error has occurred',
                )}**\n\n`;
                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-stock/pages_creation_error_connexion_webservice_contact_administrator',
                    'An error has occurred (connection or webservice error). Please contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
        async onClick() {
            if (this.miscellaneousIssueLines.value?.length) {
                this.$.loader.isHidden = false;
                await this.$standardSaveAction.execute(false);
                await this.$.sound.success();

                let _resultOutput: string | null | undefined = this.$.recordId;
                if (_resultOutput != null && _resultOutput.includes('|')) {
                    _resultOutput = _resultOutput.split('|')[1];
                }
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-stock/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__document_created',
                        'Document no. {{documentId}} created.',
                        { documentId: _resultOutput },
                    ),
                    {
                        fullScreen: true,
                        acceptButton: { text: ui.localize('@sage/x3-stock/button-accept-ok', 'OK') },
                    },
                );
                await this.$.router.emptyPage();
            } else {
                this._notifier.show(
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__no_products_error',
                        `Enter at least one product.`,
                    ),
                    'error',
                );
            }
        },
    })
    createButton: ui.PageAction;

    @ui.decorators.pageAction<MobileMiscellaneousIssue>({
        title: 'Next',
        shortcut: ['f3'],
        buttonType: 'primary',
        isHidden: true,
        async onClick() {
            if (this.stock.selectedRecords.length > 0) {
                await this._getStock();
                await this.$.commitValueAndPropertyChanges();
                this.$.setPageClean();
                this.stock.value = [];
                this.stock.selectedRecords = [];
                this.headerSection.isHidden = false;
                this.miscellaneousIssueLinesSection.isHidden = false;
                this.createButton.isDisabled = this.miscellaneousIssueLines.value.length === 0;
                this.stockLineSection.isHidden = true;
                this.detailStockLineSection.isHidden = true;
                this.miscellaneousIssueLinesBlock.isHidden = false;
                this.createButton.isHidden = false;
                this.nextButton.isHidden = true;
                this.CancelButton.isHidden = true;
                this.SelectButton.isHidden = true;
                this.transaction.isDisabled = true;
                this.productSelected.value = null;
                this.locationSelected.value = null;
                this.locationSelected.focus();
                this.miscellaneousIssueLines.isHidden = false;
                this._productAtLocation = [];
                this._resetFieldFilters();
            } else {
                await dialogMessage(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__no_stock_error',
                        `Select at least one stock line.`,
                    ),
                );
            }
        },
    })
    nextButton: ui.PageAction;

    // =================================================================================================================
    // 🟢 SECOND SECTION DISPLAYED
    // Display stock lines for selected product
    // =================================================================================================================

    @ui.decorators.section<MobileMiscellaneousIssue>({
        title: 'Miscellaneous issue',
        isTitleHidden: true,
        isHidden: true,
    })
    stockLineSection: ui.containers.Section;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        isTitleHidden: true,
        parent() {
            return this.stockLineSection;
        },
    })
    stockLineBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileMiscellaneousIssue, LicensePlateNumber>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'License plate number',
        valueField: 'code',
        node: '@sage/x3-stock-data/LicensePlateNumber',
        placeholder: 'Scan or select...',
        helperTextField: { container: { code: true } },
        canFilter: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        filter() {
            return {
                _and: [{ status: 'inStock' }, { stockSite: { code: this.stockSite?.value?.code } }],
            };
        },
        async onChange() {
            await this._onChange_licensePlateNumberFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.reference({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
            }),
            ui.nestedFields.reference({
                bind: 'container',
                valueField: 'code',
                node: '@sage/x3-master-data/Container',
            }),
        ],
    })
    licensePlateNumberFilter: ui.fields.Reference;

    /**
     * OnChange licensePlateNumberFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_licensePlateNumberFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        if (this.licensePlateNumberFilter.value) {
            this.licensePlateNumberFilter.getNextField(true)?.focus();
        }
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, Location>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        placeholder: 'Scan or select...',
        helperTextField: 'type',
        canFilter: false,
        isAutoSelectEnabled: true,
        isFullWidth: true,
        isTransient: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        filter() {
            const locationFilter: any = {
                stockSite: { code: this.stockSite?.value?.code },
                category: { _nin: ['subcontract', 'customer'] },
                stock: { _atLeast: 1, product: { product: { code: this.productSelected?.value?.code } } },
            };
            return locationFilter;
        },
        async onChange() {
            await this._onChange_locationFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
            }),
        ],
    })
    locationFilter: ui.fields.Reference;

    /**
     * OnChange locationFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_locationFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        if (this.locationFilter.value) {
            this.locationFilter.getNextField(true)?.focus();
        }
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, LotsSites>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Lot',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/LotsSites',
        valueField: 'lot',
        canFilter: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                product: { code: this.productSelected?.value?.code },
                storageSite: { code: this.stockSite?.value?.code },
                _and: [
                    {
                        lot: {
                            ...(this._lots?.length ? { _in: this._lots } : { _nin: [] }),
                        },
                    },
                ],
            };
        },
        async onChange() {
            await this._onChange_lotFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
    })
    lotFilter: ui.fields.Reference<LotsSites>;

    /**
     * OnChange lotFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_lotFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        this.lotFilter.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, LotsSites>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Sublot',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/LotsSites',
        valueField: 'sublot',
        canFilter: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                product: { code: this.productSelected?.value?.code },
                storageSite: { code: this.stockSite?.value?.code },
            };
        },
        async onChange() {
            await this._onChange_sublotFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
    })
    sublotFilter: ui.fields.Reference<LotsSites>;

    /**
     * OnChange sublotFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_sublotFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        this.sublotFilter.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, SerialNumber>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Serial number',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/SerialNumber',
        valueField: 'code',
        canFilter: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        filter() {
            return {
                product: { code: this.productSelected?.value?.code },
                stockSite: { code: this.stockSite?.value?.code },
                issueDocumentId: '',
            };
        },
        async onChange() {
            await this._onChange_serialNumberFilter();
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
        ],
    })
    serialNumberFilter: ui.fields.Reference<SerialNumber>;

    /**
     * OnChange serialNumberFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_serialNumberFilter: AsyncVoidFunction = async () => {
        await this._getStock();
        if (this.serialNumberFilter.value) {
            this.stock.value.forEach(stockLine => {
                if (!this.stock.selectedRecords.find(selectedRecord => selectedRecord === stockLine._id)) {
                    this.stock.selectRecord(stockLine._id);
                    this._onRowSelect(stockLine._id, stockLine as any);
                }
            });
        }
    };

    @ui.decorators.dropdownListField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Status',
        placeholder: 'Scan or select...',
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_statusFilter();
        },
    })
    statusFilter: ui.fields.DropdownList;

    /**
     * OnChange statusFilter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_statusFilter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.dropdownListField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Unit',
        placeholder: 'Scan or select...',
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_packingUnitFilter();
        },
    })
    packingUnitFilter: ui.fields.DropdownList;

    /*
     * onChange packingUnitFilter event (for decorator and management service)
     */
    private readonly _onChange_packingUnitFilter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.numericField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Conversion factor',
        placeholder: 'Enter...',
        isTransient: true,
        isFullWidth: true,
        min: 0,
        async onChange() {
            await this._getStock();
        },
    })
    packingUnitToStockUnitConversionFactorFilter: ui.fields.Numeric;

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Identifier 1',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_identifier1Filter();
        },
    })
    identifier1Filter: ui.fields.Text;

    /**
     * OnChange identifier1Filter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_identifier1Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Identifier 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_identifier2Filter();
        },
    })
    identifier2Filter: ui.fields.Text;

    /**
     * OnChange identifier2Filter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_identifier2Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Stock custom field 1',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_stockCustomField1Filter();
        },
    })
    stockCustomField1Filter: ui.fields.Text;

    /**
     * OnChange stockCustomField1Filter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_stockCustomField1Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        title: 'Stock custom field 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        isFullWidth: true,
        async onChange() {
            await this._onChange_stockCustomField2Filter();
        },
    })
    stockCustomField2Filter: ui.fields.Text;

    /**
     * OnChange stockCustomField2Filter readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_stockCustomField2Filter: AsyncVoidFunction = async () => {
        await this._getStock();
    };

    @ui.decorators.tableField<MobileMiscellaneousIssue>({
        parent() {
            return this.stockLineBlock;
        },
        node: '@sage/x3-stock-data/Stock',
        isChangeIndicatorDisabled: false,
        canFilter: false,
        canSelect: true,
        canExport: false,
        canUserHideColumns: false,
        isTransient: true,
        isFullWidth: true,
        hasSearchBoxMobile: false,
        cardView: true,
        displayMode: ui.fields.TableDisplayMode.compact,
        mobileCard: undefined,
        orderBy: {
            stockSite: 1,
            stockId: 1,
        },
        columns: [
            ui.nestedFields.numeric({
                bind: 'quantityToIssue' as any,
                isReadOnly: true,
                isTransient: true,
                postfix(_value, rowValue?: Dict<any>) {
                    const selectedRecordId = this.gridBlock?.selectedRecordId;
                    const currentRecord = selectedRecordId ? this.stock.getRecordValue(selectedRecordId) : undefined;
                    const packingUnitCode = this.packingUnitToIssue.value
                        ? String(this.packingUnitToIssue.value)
                        : String(currentRecord?.packingUnit?.code ?? rowValue?.packingUnit?.code ?? '');

                    if (currentRecord && rowValue?.stockId === currentRecord.stockId) {
                        return ` ${packingUnitCode}`;
                    } else {
                        return ` ${String(rowValue?.packingUnit?.code)}`;
                    }
                },
                title: 'Quantity to issue',
                isFullWidth: true,
                max(rowValue?: Stock) {
                    return (<any>(<unknown>rowValue))?.quantityInPackingUnitOrigin ?? 0;
                },
                scale(_value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit?.numberOfDecimals;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnit',
                isHidden: true,
                isTitleHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnitRest' as any,
                isHidden: true,
                isTitleHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, ProductSite>({
                bind: 'product',
                valueField: { product: { code: true } },
                node: '@sage/x3-master-data/ProductSite',
                isReadOnly: true,
                columns: [
                    ui.nestedFields.text({
                        bind: { product: { code: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { serialNumberManagementMode: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { localizedDescription1: true } },
                    }),
                    ui.nestedFields.text({
                        bind: { product: { _id: true } },
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, LicensePlateNumber>({
                bind: 'licensePlateNumber',
                valueField: 'code',
                node: '@sage/x3-stock-data/LicensePlateNumber',
                isReadOnly: true,
                isHidden: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Location>({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
                isReadOnly: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
            }),
            ui.nestedFields.text({
                bind: 'lot',
                isReadOnly: true,
                isHidden() {
                    return this.lotFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                isReadOnly: true,
                isHidden() {
                    return this.sublotFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.link({
                bind: 'globalSerialNumber' as any,
                isTransient: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'serialNumber',
                isReadOnly: true,
                isHidden() {
                    return this.serialNumberFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, StockStatus>({
                bind: 'status',
                valueField: 'code',
                node: '@sage/x3-stock-data/StockStatus',
                isReadOnly: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                        isHidden: true,
                    }),
                ],
                isHidden() {
                    return this.statusFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.technical<MobileMiscellaneousIssue, Stock, UnitOfMeasure>({
                bind: 'packingUnit',
                node: '@sage/x3-master-data/UnitOfMeasure',
                nestedFields: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                    ui.nestedFields.numeric({
                        bind: 'numberOfDecimals',
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactor',
                isReadOnly: true,
                isHidden() {
                    return this.packingUnitToStockUnitConversionFactorFilter.isHidden ?? false;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnitOrigin' as any,
                isHidden: true,
                isTitleHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.text({
                bind: 'identifier1',
                isReadOnly: true,
                isHidden() {
                    return this.identifier1Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'identifier2',
                isReadOnly: true,
                isHidden() {
                    return this.identifier2Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField1',
                isReadOnly: true,
                isHidden() {
                    return this.stockCustomField1Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField2',
                isReadOnly: true,
                isHidden() {
                    return this.stockCustomField2Filter.isHidden ?? false;
                },
            }),
            ui.nestedFields.technical({
                bind: 'stockId',
            }),
            ui.nestedFields.technical<MobileMiscellaneousIssue, Stock, Lot>({
                bind: 'lotReference',
                node: '@sage/x3-stock-data/Lot',
                nestedFields: [
                    ui.nestedFields.date({
                        bind: 'expirationDate',
                    }),
                    ui.nestedFields.date({
                        bind: 'useByDate',
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                    }),
                    ui.nestedFields.reference({
                        bind: 'majorVersion',
                        node: '@sage/x3-master-data/MajorVersionStatus',
                        valueField: 'code',
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isSubjectToException',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isOnHold',
                                isHidden: true,
                            }),
                            ui.nestedFields.date({
                                bind: 'exceptionDate',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: '_id',
                    }),
                ],
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceExpirationDate' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.date({
                bind: 'lotReferenceUseByDate' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lotReferenceMajorVersion' as any,
                isTransient: true,
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.technical({
                bind: 'quantityInStockUnit',
            }),
            ui.nestedFields.technical({
                bind: 'allocatedQuantity',
            }),
            ui.nestedFields.technical({
                bind: 'qualityAnalysisRequestId',
            }),
            ui.nestedFields.technical({
                bind: 'owner',
            }),
        ],
        onError(error) {
            if (error.message === '@sage/x3-stock/serial-number-not-sequential') {
                return ui.localize(
                    '@sage/x3-stock/serial-number-not-sequential',
                    'The serial numbers are not sequential. Check your entry.',
                );
            }
            return error;
        },
        async onRowSelected(recordId: string, rowItem: Stock) {
            if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                await this._onRowClick(recordId, rowItem);
            } else {
                await this._onRowSelect(recordId, rowItem);
            }
        },
        async onRowUnselected(recordId: string) {
            let currentRecord = this.stock.getRecordValue(recordId);
            const originalStockLine = this._originalStockLines?.find(line => recordId === line._id);
            if (currentRecord) {
                this.miscellaneousIssueLines?.value?.forEach(line => {
                    if (String(line.numberSession) === String(this._numberSession)) {
                        line.stockDetails?.forEach(stock => {
                            if (String(stock._id?.split('|')[0]) === String(currentRecord.stockId)) {
                                this.miscellaneousIssueLines.removeRecord(String(stock._id), 1);
                            }
                        });
                        this._updateMiscellaneousIssueLine(String(line.numberSession));
                    }
                });
                this.packingUnitToIssue.value = null;
                currentRecord = originalStockLine;
                this.stock.setRecordValue(currentRecord);
                await this.$.commitValueAndPropertyChanges();
            }
        },
        sortColumns(firstColumn, secondColumn) {
            if (firstColumn.bind === secondColumn.bind) return 0;
            if (firstColumn.bind === 'quantityToIssue') {
                return secondColumn.bind === (this._stockFieldSettings[0] as string) ? 1 : -1;
            } else if (secondColumn.bind === 'quantityToIssue') {
                return firstColumn.bind === (this._stockFieldSettings[0] as string) ? -1 : 1;
            }

            for (const stockFieldSetting of Object.keys(this._stockFieldSettings)) {
                if (!stockFieldSetting || stockFieldSetting === 'none') break;
                if (firstColumn.bind === (stockFieldSetting as string)) return -1;
                if (secondColumn.bind === (stockFieldSetting as string)) return 1;
            }

            return 1;
        },
        mapServerRecord(record: any) {
            return {
                ...record,
                quantityInPackingUnitOrigin: this._getQuantityInPackingUnitOrigin(record),
                quantityInPackingUnitRest: this._getQuantityInPackingUnitRest(record),
                quantityToIssue: this._getQuantityToIssue(record).toString(),
                lotReferenceExpirationDate: record.lotReference?.expirationDate,
                lotReferenceMajorVersion: record.lotReference?.majorVersion?.code,
                lotReferenceUseByDate: record.lotReference?.useByDate,
                stockUnit: record.product?.product?.stockUnit?.code,
                globalSerialNumber: ui.localize('@sage/x3-stock/label-view-all', 'View list'),
            };
        },

        async onRowClick(recordId: string, rowItem: Stock) {
            if (this.stock.selectedRecords.includes(recordId)) {
                this.miscellaneousIssueLines?.value?.forEach(line => {
                    if (String(line.numberSession) === String(this._numberSession)) {
                        line.stockDetails?.forEach(stock => {
                            if (String(stock._id?.split('|')[0]) === String(rowItem.stockId)) {
                                this.miscellaneousIssueLines.removeRecord(String(stock._id), 1);
                            }
                        });
                    }
                });
            }

            this.stock.unselectRecord(recordId);
            await this._onRowClick(recordId, rowItem);
        },
    })
    stock: ui.fields.Table<any>;

    // =================================================================================================================
    // 🟢 THIRD SECTION DISPLAYED
    // Display details of the selected stock line.
    // =================================================================================================================

    @ui.decorators.pageAction<MobileMiscellaneousIssue>({
        title: 'Select',
        shortcut: ['f3'],
        buttonType: 'primary',
        isHidden: true,
        onError(error) {
            if (
                error.message ===
                '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumberMandatory'
            ) {
                return ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumberMandatory',
                    'You need to select the serial number and add it first.',
                );
            }
            return error;
        },
        async onClick() {
            await this.$.commitValueAndPropertyChanges();

            const errors: ui.ValidationResult[] = await this.stock.validateWithDetails();
            if (!errors.length) {
                let currentRecord = this.stock.getRecordValue(String(this.gridBlock.selectedRecordId)) as any;
                if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
                    if (!this.serialNumberLines.value.length) {
                        throw new Error(
                            '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumberMandatory',
                        );
                    }
                } else {
                    if (Number(this.quantityToIssue.value) <= 0) {
                        await dialogMessage(
                            this,
                            'error',
                            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                            ui.localize(
                                '@sage/x3-stock/pages__mobile_miscellaneous_issue__quantityInPackingUnitDestination_must_be_greater_than_0',
                                'The quantity to issue must be greater than 0.',
                            ),
                        );
                        return;
                    }
                }
                if (currentRecord.qualityAnalysisRequestId) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-stock/pages__mobile_miscellaneous_issue__miscellaneous_issue_is_impossible_on_a_line_in_quality_control',
                            'Stock line subject to an analysis request.',
                        ),
                    );
                    return;
                }

                const conversionFactorOrigin = Number(currentRecord?.packingUnitToStockUnitConversionFactor);
                const conversionFactorToIssue = Number(this.packingUnitToStockUnitConversionFactorToIssue.value);
                let quantityInPackingUnitRest = Number(currentRecord?.quantityInPackingUnitRest);

                if (conversionFactorOrigin !== conversionFactorToIssue) {
                    const factor =
                        conversionFactorOrigin > conversionFactorToIssue
                            ? conversionFactorOrigin
                            : conversionFactorOrigin / conversionFactorToIssue;
                    quantityInPackingUnitRest = quantityInPackingUnitRest * factor;
                }

                if (Number(this.quantityToIssue.value) > Number(quantityInPackingUnitRest)) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-stock/pages__mobile_miscellaneous_issue__enter_a_quantity_less_than_or_equal_to_the_stock_quantity',
                            'Enter a quantity less than or equal to the stock quantity.',
                        ),
                    );
                    return;
                }

                if (
                    (this.quantityToIssue.value ? Number(this.quantityToIssue.value) : 0) *
                    (this.packingUnitToStockUnitConversionFactorToIssue.value
                        ? Number(this.packingUnitToStockUnitConversionFactorToIssue.value)
                        : 1) >
                    Number(currentRecord.quantityInStockUnit - currentRecord.allocatedQuantity)
                ) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-stock/pages__mobile_miscellaneous_issue__enter_a_quantity_less_than_or_equal_to_the_stock_quantity_minus_allocated_quantity',
                            'Enter a quantity less than or equal to the stock quantity minus the allocated quantity.',
                        ),
                    );
                    this.quantityToIssue.value = null;
                    return;
                }
                this.stock.selectRecord(String(this.gridBlock.selectedRecordId));

                const line = this.miscellaneousIssueLines.value.find(
                    line => Number(line.numberSession) === this._numberSession,
                );
                const lineNumber = line ? line.lineNumber : this._numberSession;
                const numberOfDecimals =
                    (this._numberOfDecimalList?.find(uom => uom.code === this.packingUnitToIssue?.value) || {})
                        .numberOfDecimals ?? 0;
                if (this.productSelected.value?.serialNumberManagementMode !== 'globalReceivedIssued') {
                    this.miscellaneousIssueLines.addOrUpdateRecordValue(
                        {
                            _id: `${lineNumber}`,
                            entryType: 'miscellaneousIssue',
                            lineNumber: lineNumber,
                            product: {
                                code: currentRecord?.product?.product?.code,
                                _id: currentRecord?.product?.product?._id,
                            },
                            stockId: String(currentRecord.stockId),
                            productDescription: this.productSelected?.value?.localizedDescription1,
                            numberSession: String(this._numberSession),
                            packingUnit: {
                                code: this.packingUnitToIssue.value
                                    ? this.packingUnitToIssue.value
                                    : currentRecord.packingUnit?.code,
                                numberOfDecimals: this.packingUnitToIssue.value
                                    ? numberOfDecimals
                                    : currentRecord.packingUnit?.numberOfDecimals,
                                _id: this.packingUnitToIssue.value
                                    ? this.packingUnitToIssue.value
                                    : currentRecord.packingUnit?._id,
                            },
                            packingUnitToStockUnitConversionFactor: String(
                                this.packingUnitToStockUnitConversionFactorToIssue.value,
                            ),
                            quantityInPackingUnit: String(this.quantityToIssue.value),
                        },
                        0,
                        undefined,
                    );
                    this.miscellaneousIssueLines.addOrUpdateRecordValue(
                        {
                            _id: `${currentRecord.stockId}|${lineNumber}`,
                            packingUnit: {
                                code: currentRecord.packingUnit?.code,
                                numberOfDecimals: currentRecord.packingUnit?.numberOfDecimals,
                                _id: currentRecord.packingUnit?._id,
                            },
                            product: {
                                code: currentRecord?.product?.product?.code,
                                _id: currentRecord?.product?.product?._id,
                            },
                            packingUnitToStockUnitConversionFactor: String(
                                currentRecord?.packingUnitToStockUnitConversionFactor,
                            ),
                            quantityInPackingUnit: String(
                                (Number(this.quantityToIssue.value) *
                                    (this.packingUnitToStockUnitConversionFactorToIssue.value || 0)) /
                                (currentRecord.packingUnitToStockUnitConversionFactor || 1),
                            ),
                            quantityInStockUnit: String(
                                Number(this.quantityToIssue.value) *
                                Number(this.packingUnitToStockUnitConversionFactorToIssue.value),
                            ),
                            location: {
                                code: currentRecord.location?.code,
                                _id: currentRecord.location?._id,
                            },
                            status: {
                                code: currentRecord.status?.code,
                                _id: currentRecord.status?._id,
                            },
                            lot: currentRecord.lot ?? undefined,
                            sublot: currentRecord.sublot ?? undefined,
                            serialNumber: currentRecord.serialNumber ?? undefined,
                            licensePlateNumber: currentRecord.licensePlateNumber?.code ?? undefined,
                            stockUnit: {
                                code: this._productSite.product?.stockUnit?.code,
                                numberOfDecimals: this._productSite.product.stockUnit.numberOfDecimals,
                                _id: this._productSite.product.stockUnit._id,
                            },
                            identifier1: currentRecord.identifier1 ?? undefined,
                            identifier2: currentRecord.identifier2 ?? undefined,
                            startingSerialNumber: this.startingSerialNumber?.value?.code ?? undefined,
                            endingSerialNumber: this.endingSerialNumber?.value ?? undefined,
                            stockCustomField1: currentRecord.stockCustomField1 ?? undefined,
                            stockCustomField2: currentRecord.stockCustomField2 ?? undefined,
                        },
                        1,
                        `${lineNumber}`,
                    );
                }

                let qtyTotalInPackingUnit = 0;
                let qtyTotalInStockUnit = 0;

                this.miscellaneousIssueLines.value.forEach(line => {
                    if (Number(line.numberSession) === this._numberSession) {
                        const totals = line.stockDetails?.reduce(
                            (acc, stockDetail) => {
                                if (String(stockDetail._id?.split('|')[0]) === String(currentRecord.stockId)) {
                                    acc.qtyTotalInStockUnit += Number(stockDetail.quantityInStockUnit);
                                    acc.qtyTotalInPackingUnit +=
                                        Number(stockDetail.quantityInPackingUnit) *
                                        Number(stockDetail.packingUnitToStockUnitConversionFactor);
                                }
                                return acc;
                            },
                            { qtyTotalInStockUnit: 0, qtyTotalInPackingUnit: 0 },
                        );

                        if (totals) {
                            qtyTotalInStockUnit += totals.qtyTotalInStockUnit;
                            qtyTotalInPackingUnit += totals.qtyTotalInPackingUnit;
                        }
                    }
                });

                currentRecord.quantityInStockUnit = qtyTotalInStockUnit;
                if (isNaN(Number(this.quantityToIssue.value)) || Number(this.quantityToIssue.value) === 0) {
                    this.quantityToIssue.value = qtyTotalInPackingUnit;
                }
                currentRecord.quantityToIssue = this.quantityToIssue.value;

                const packingUnitIndex = this._packingUnits
                    .map(packingUnit => packingUnit.node.packingUnit.code)
                    .indexOf(this.packingUnitToIssue.value ?? '');
                if (packingUnitIndex !== -1) {
                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                    currentRecord.packingUnit = selectedUnit.packingUnit;
                    currentRecord.quantityToIssue = Number(
                        qtyTotalInStockUnit / Number(selectedUnit.packingUnitToStockUnitConversionFactor),
                    );
                } else {
                    currentRecord.packingUnit = {
                        code: this.packingUnitToIssue.value
                            ? this.packingUnitToIssue.value
                            : currentRecord.packingUnit?.code,
                        numberOfDecimals: this.packingUnitToIssue.value
                            ? numberOfDecimals
                            : currentRecord.packingUnit?.numberOfDecimals,
                        _id: this.packingUnitToIssue.value
                            ? this.packingUnitToIssue.value
                            : currentRecord.packingUnit?._id,
                    };
                    currentRecord.quantityInPackingUnitOrigin =
                        Number(currentRecord.quantityInPackingUnitOrigin) *
                        Number(currentRecord.packingUnitToStockUnitConversionFactor);
                    currentRecord.quantityInPackingUnit =
                        Number(currentRecord.quantityInPackingUnit) *
                        Number(currentRecord.packingUnitToStockUnitConversionFactor);
                    currentRecord.quantityInPackingUnitRest =
                        Number(currentRecord.quantityInPackingUnitRest) *
                        Number(currentRecord.packingUnitToStockUnitConversionFactor);
                    currentRecord.packingUnitToStockUnitConversionFactor = 1;
                }
                this.stock.setRecordValue(currentRecord);

                this._updateMiscellaneousIssueLine(String(this._numberSession));
                await this.$.commitValueAndPropertyChanges();
            }
            this.headerSection.isHidden = true;
            this.miscellaneousIssueLinesSection.isHidden = true;
            this.stockLineSection.isHidden = false;
            this.detailStockLineSection.isHidden = true;
            this.miscellaneousIssueLinesBlock.isHidden = true;
            this.createButton.isHidden = true;
            this.nextButton.isHidden = false;
            this.CancelButton.isHidden = true;
            this.SelectButton.isHidden = true;
            this.packingUnitToIssue.value = null;
            this.productCode.value = this.productSelected.value?.code;
            this.productDescription.value = this.productSelected.value?.localizedDescription1;
        },
    })
    SelectButton: ui.PageAction;

    @ui.decorators.pageAction<MobileMiscellaneousIssue>({
        title: 'Cancel',
        buttonType: 'secondary',
        isHidden: true,
        async onClick() {
            const currentRecord = this.stock.getRecordValue(String(this.gridBlock.selectedRecordId)) as any;
            const originalStockLine = this._originalStockLines?.find(
                line => String(this.gridBlock.selectedRecordId) === line._id,
            );
            this.miscellaneousIssueLines?.value?.forEach(line => {
                if (String(line.numberSession) === String(this._numberSession)) {
                    line.stockDetails?.forEach(stock => {
                        if (String(stock._id?.split('|')[0]) === String(currentRecord.stockId)) {
                            this.miscellaneousIssueLines.removeRecord(String(stock._id), 1);
                        }
                    });

                    this._updateMiscellaneousIssueLine(String(line.numberSession));
                }
            });

            this.packingUnitToIssue.value = null;
            currentRecord.quantityInStockUnit = originalStockLine?.quantityInStockUnit;
            currentRecord.packingUnit = originalStockLine?.packingUnit;
            currentRecord.quantityInPackingUnit = originalStockLine?.quantityInPackingUnit;
            currentRecord.quantityToIssue = currentRecord.quantityInPackingUnitOrigin;
            currentRecord.quantityInPackingUnitDestination = currentRecord.quantityInPackingUnitOrigin;

            this.stock.unselectRecord(String(this.gridBlock.selectedRecordId));
            this.stock.setRecordValue(currentRecord);

            this.headerSection.isHidden = true;
            this.miscellaneousIssueLinesSection.isHidden = true;
            this.stockLineSection.isHidden = false;
            this.detailStockLineSection.isHidden = true;
            this.miscellaneousIssueLinesBlock.isHidden = true;
            this.createButton.isHidden = true;
            this.nextButton.isHidden = false;
            this.CancelButton.isHidden = true;
            this.SelectButton.isHidden = true;

            this.productCode.value = this.productSelected.value?.code;
            this.productDescription.value = this.productSelected.value?.localizedDescription1;
        },
    })
    CancelButton: ui.PageAction;

    @ui.decorators.pageAction<MobileMiscellaneousIssue>({
        icon: 'add',
        title: 'Add...',
        onError(error) {
            switch (error.message) {
                case '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumber': {
                    return ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumber',
                        'The serial number is mandatory',
                    );
                }
                case '@sage/x3-stock/serial-number-range-overlap': {
                    return ui.localize(
                        '@sage/x3-stock/serial-number-range-overlap',
                        'The serial numbers are overlapping. Enter another starting or ending serial number.',
                    );
                }
                case '@sage/x3-stock/pages__mobile_miscellaneous_issue__same-amount-serial-numbers-in-the-range-to-match-quantity-to-issue': {
                    return ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_issue__same-amount-serial-numbers-in-the-range-to-match-quantity-to-issue',
                        'Select the same amount of serial numbers in the range to match the quantity to issue.',
                    );
                }
                case '@sage/x3-stock/serial-number-not-sequential': {
                    return ui.localize(
                        '@sage/x3-stock/serial-number-not-sequential',
                        'The serial numbers are not sequential. Check your entry.',
                    );
                }
                default: {
                    return error;
                }
            }
        },
        buttonType: 'secondary',
        async onClick() {
            if (!this.startingSerialNumber.value) {
                throw new Error(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumber',
                );
            }
            // check that this will add any duplicates
            const startNumberToAdd = this.startingSerialNumber?.value?.code?.match(/\d+$/);
            const endNumberToAdd = Number(this.endingSerialNumber?.value?.match(/\d+$/));
            let serialNumberAlreadyUsed: boolean = false;
            if (
                this.serialNumberLines.value.some(row => {
                    const rowStartMatch = row.startingSerialNumber.match(/\d+$/);
                    const rowEndMatch = Number(row.endingSerialNumber.match(/\d+$/));

                    // check if the 'beginning part' of the serial matches
                    if (
                        !startNumberToAdd ||
                        row.startingSerialNumber.substring(
                            0,
                            row.startingSerialNumber.length - rowStartMatch.toString().length,
                        ) !==
                        this.startingSerialNumber?.value?.code?.substring(
                            0,
                            this.startingSerialNumber.value.code.length - startNumberToAdd.toString().length,
                        )
                    )
                        return false;

                    return Number(startNumberToAdd) <= rowEndMatch && endNumberToAdd >= Number(rowStartMatch);
                })
            ) {
                serialNumberAlreadyUsed = true;
            }
            this.miscellaneousIssueLines.value.forEach(line => {
                const serialNumberPrefix1 = this.startingSerialNumber.value?.code?.match(/^(.*?)(?=\d+$)/);
                if (
                    line.product === this.productSelected.value?.code &&
                    line.stockDetails?.length &&
                    line.stockDetails[0].serialNumber
                ) {
                    const serialNumberPrefix2 = line?.stockDetails[0]?.serialNumber?.match(/^(.*?)(?=\d+$)/);
                    if (serialNumberPrefix1 === serialNumberPrefix2) {
                        const startingSerialNumber = Number(line.stockDetails[0].serialNumber.match(/\d+$/));
                        const endingSerialNumber = Number(
                            this._calculateEndingSerialNumber(
                                line.stockDetails[0].serialNumber,
                                Number(line.quantityInPackingUnit),
                            ).match(/\d+$/),
                        );
                        if (
                            Number(startNumberToAdd) <= endingSerialNumber &&
                            Number(endNumberToAdd) >= startingSerialNumber
                        ) {
                            serialNumberAlreadyUsed = true;
                        }
                    }
                }
            });
            if (serialNumberAlreadyUsed) {
                throw new Error('@sage/x3-stock/serial-number-range-overlap');
            }
            if (
                this.endingSerialNumber.value !=
                this._calculateEndingSerialNumber(
                    this.startingSerialNumber?.value?.code ?? '',
                    Number(this.quantityToIssue.value) *
                    Number(this.packingUnitToStockUnitConversionFactorToIssue.value),
                )
            ) {
                throw new Error(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__same-amount-serial-numbers-in-the-range-to-match-quantity-to-issue',
                );
            }
            if (
                (await getCountSerialNumber(
                    this,
                    this.productSelected.value?.code ?? '',
                    this.stockSite.value?.code,
                    this._stockId.value,
                    this.startingSerialNumber.value?.code ?? '',
                    this.endingSerialNumber.value ?? '',
                )) !==
                Number(this.quantityToIssue.value) * Number(this.packingUnitToStockUnitConversionFactorToIssue.value)
            ) {
                throw new Error('@sage/x3-stock/serial-number-not-sequential');
            }

            let currentRecord = this.stock.getRecordValue(this.gridBlock.selectedRecordId ?? '') as any;
            const serialNumberId = Number(Math.floor(Math.random() * MAX_INT_32));
            this.serialNumberLines.addRecord({
                quantity:
                    Number(this.quantityToIssue.value) *
                    Number(this.packingUnitToStockUnitConversionFactorToIssue.value),
                startingSerialNumber: this.startingSerialNumber.value.code,
                endingSerialNumber: this.endingSerialNumber.value,
                id: serialNumberId,
            });
            // const lineNumber = Number(Math.floor(Math.random() * MAX_INT_32));
            const lineNumber = this._numberSession;
            const numberOfDecimals =
                (this._numberOfDecimalList?.find(uom => uom.code === this.packingUnitToIssue?.value) || {})
                    .numberOfDecimals ?? 0;
            this.miscellaneousIssueLines.addOrUpdateRecordValue(
                {
                    _id: `${lineNumber}`,
                    lineNumber: lineNumber,
                    product: {
                        code: currentRecord?.product?.product?.code,
                        _id: currentRecord?.product?.product?._id,
                    },
                    stockId: String(currentRecord.stockId),
                    productDescription: this.productSelected?.value?.localizedDescription1,
                    numberSession: String(this._numberSession),
                    packingUnit: {
                        code: this.packingUnitToIssue.value
                            ? this.packingUnitToIssue.value
                            : currentRecord.packingUnit?.code,
                        numberOfDecimals: this.packingUnitToIssue.value
                            ? numberOfDecimals
                            : currentRecord.packingUnit?.numberOfDecimals,
                        _id: this.packingUnitToIssue.value
                            ? this.packingUnitToIssue.value
                            : currentRecord.packingUnit?._id,
                    },
                    packingUnitToStockUnitConversionFactor: String(
                        this.packingUnitToStockUnitConversionFactorToIssue.value,
                    ),
                    quantityInPackingUnit: String(this.quantityToIssue.value),
                    ...this._getCustomizationMiscellaneousIssueLines(),
                },
                0,
                undefined,
            );
            this.miscellaneousIssueLines.addOrUpdateRecordValue(
                {
                    _id: `${currentRecord.stockId}|${lineNumber}|${serialNumberId}`,
                    packingUnit: {
                        code: currentRecord.packingUnit?.code,
                        numberOfDecimals: currentRecord.packingUnit?.numberOfDecimals ?? numberOfDecimals,
                        _id: currentRecord.packingUnit?._id,
                    },
                    product: {
                        code: currentRecord?.product?.product?.code,
                        _id: currentRecord?.product?.product?._id,
                    },
                    packingUnitToStockUnitConversionFactor: String(
                        currentRecord?.packingUnitToStockUnitConversionFactor,
                    ),
                    quantityInPackingUnit: String(
                        (Number(this.quantityToIssue.value) *
                            (this.packingUnitToStockUnitConversionFactorToIssue.value || 0)) /
                        (currentRecord.packingUnitToStockUnitConversionFactor || 1),
                    ),
                    quantityInStockUnit: String(
                        Number(this.quantityToIssue.value) *
                        Number(this.packingUnitToStockUnitConversionFactorToIssue.value),
                    ),
                    location: {
                        code: currentRecord.location?.code,
                        _id: currentRecord.location?._id,
                    },
                    status: {
                        code: currentRecord.status?.code,
                        _id: currentRecord.status?._id,
                    },
                    lot: currentRecord.lot ?? undefined,
                    sublot: currentRecord.sublot ?? undefined,
                    serialNumber: this.startingSerialNumber.value?.code ?? undefined,
                    licensePlateNumber: currentRecord.licensePlateNumber?.code ?? undefined,
                    stockUnit: {
                        code: this._productSite.product?.stockUnit?.code,
                        numberOfDecimals: this._productSite.product.stockUnit.numberOfDecimals,
                        _id: this._productSite.product.stockUnit._id,
                    },
                    identifier1: currentRecord.identifier1 ?? undefined,
                    identifier2: currentRecord.identifier2 ?? undefined,
                    startingSerialNumber: this.startingSerialNumber?.value?.code ?? undefined,
                    endingSerialNumber: this.endingSerialNumber?.value ?? undefined,
                    stockCustomField1: currentRecord.stockCustomField1 ?? undefined,
                    stockCustomField2: currentRecord.stockCustomField2 ?? undefined,
                    ...this._getCustomizationMiscellaneousIssueDetails(),
                },
                1,
                `${lineNumber}`,
            );
            (currentRecord as any).quantityToIssue = String(0);
            this.quantityToIssue.value = 0;
            this.stock.setRecordValue(currentRecord);
            this.startingSerialNumber.value = null;
            this.endingSerialNumber.value = null;
            this._updateMiscellaneousIssueLine(String(lineNumber));
            await this.$.commitValueAndPropertyChanges();
        },
    })
    addSerialRangeButton: ui.PageAction;

    @ui.decorators.section<MobileMiscellaneousIssue>({
        title: 'Miscellaneous issue',
        isTitleHidden: true,
        isHidden: true,
    })
    detailStockLineSection: ui.containers.Section;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        isTitleHidden: true,
        parent() {
            return this.detailStockLineSection;
        },
    })
    detailStockLineBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        isTitleHidden: true,
        parent() {
            return this.detailStockLineSection;
        },
    })
    quantityBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        isTitleHidden: true,
        parent() {
            return this.detailStockLineSection;
        },
    })
    serialNumberBlock: ui.containers.Block;

    @ui.decorators.gridRowBlock<MobileMiscellaneousIssue>({
        isTitleHidden: true,
        parent() {
            return this.detailStockLineSection;
        },
        boundTo() {
            return this.stock;
        },
        fieldFilter(columnId: string) {
            return false;
        },
        readOnlyOverride(columnId: string) {
            return undefined;
        },
    })
    gridBlock: ui.containers.GridRowBlock;

    @ui.decorators.detailListField<MobileMiscellaneousIssue, Stock>({
        parent() {
            return this.detailStockLineBlock;
        },
        node: '@sage/x3-stock-data/Stock',
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
        fields: [
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, ProductSite>({
                bind: 'product',
                valueField: { product: { code: true } },
                node: '@sage/x3-master-data/ProductSite',
                title: 'Product',
                isReadOnly: true,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, LicensePlateNumber>({
                bind: 'licensePlateNumber',
                valueField: 'code',
                node: '@sage/x3-stock-data/LicensePlateNumber',
                title: 'License plate number',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Location>({
                bind: 'location',
                valueField: 'code',
                node: '@sage/x3-stock-data/Location',
                title: 'Location',
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'stockId',
                title: 'Stock ID',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: { majorVersion: { code: true } },
                title: 'Major version',
                isReadOnly: true,
                isHidden: (value: Lot) => {
                    return !value?.majorVersion;
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'expirationDate',
                title: 'Expiration date',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return (
                        !this.productSelected.value ||
                        this.productSelected.value.expirationManagementMode === 'notManaged' ||
                        !value?.expirationDate ||
                        Date.parse(value.expirationDate) > Date.now()
                    );
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'useByDate',
                title: 'Use-by date',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return (
                        !this.productSelected.value ||
                        this.productSelected.value.expirationManagementMode === 'notManaged' ||
                        !value?.useByDate ||
                        Date.parse(value.useByDate) > Date.now()
                    );
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'lotCustomField1',
                title: 'Lot custom field 1',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return !value?.lotCustomField1;
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, Lot>({
                node: '@sage/x3-stock-data/Lot',
                bind: 'lotReference',
                valueField: 'lotCustomField2',
                title: 'Lot custom field 2',
                isReadOnly: true,
                isHidden(value: Lot) {
                    return !value?.lotCustomField2;
                },
            }),
            ui.nestedFields.text({
                bind: 'serialNumber',
                title: 'Serial no.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.link({
                bind: 'globalSerialNumber' as any,
                title: 'Serial no.',
                isHidden() {
                    return this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued';
                },
                async onClick(_id, rowData: any) {
                    const options: ui.dialogs.PageDialogOptions = {
                        skipDirtyCheck: true,
                    };
                    try {
                        await this.$.dialog.page(
                            '@sage/x3-stock/MobileGlobalSerialDetails',
                            {
                                product: this.productSelected.value?.code ?? '',
                                stockId: rowData.stockId,
                                subtitle: this.productSelected.value?.localizedDescription1.value ?? '',
                            },
                            options,
                        );
                    } catch (error) {
                        if (error) {
                            this.$.showToast(error.message, { timeout: 10000, type: 'error' });
                        }
                    }
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityToIssue' as any,
                title: 'Quantity to issue',
                isReadOnly: true,
                isTransient: true,
                isHidden: true,
                postfix(_value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit.code;
                },
                scale(value, rowValue?: Dict<any>) {
                    return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPackingUnit' as any,
                title: 'Packing qty.',
                isReadOnly: true,
                postfix() {
                    return this.stock.getRecordValue(String(this.gridBlock.selectedRecordId))?.packingUnit?.code;
                },
                scale() {
                    return (
                        this.stock.getRecordValue(String(this.gridBlock.selectedRecordId))?.packingUnit
                            ?.numberOfDecimals ?? 0
                    );
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, UnitOfMeasure>({
                bind: 'packingUnit',
                node: '@sage/x3-master-data/UnitOfMeasure',
                valueField: 'code',
                title: 'Unit',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'packingUnitToStockUnitConversionFactor' as any,
                title: 'Conversion factor',
                isReadOnly: true,
                scale(value, rowValue?: Dict<any>) {
                    const conversionFactor = rowValue?.packingUnitToStockUnitConversionFactor.toString();
                    const numberOfDec = conversionFactor?.includes('.') ? conversionFactor.split('.')[1].length : 0;
                    return numberOfDec;
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInStockUnit',
                title: 'Stock qty.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
                postfix() {
                    return this.productSelected.value?.stockUnit?.code ?? '';
                },
                scale() {
                    return this.productSelected.value?.stockUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.text({
                bind: 'stockUnit' as any,
                title: 'Stock unit',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'allocatedQuantity',
                title: 'Allocated qty.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
                postfix() {
                    return this.productSelected.value?.stockUnit?.code ?? '';
                },
                scale() {
                    return this.productSelected.value?.stockUnit?.numberOfDecimals ?? 0;
                },
            }),
            ui.nestedFields.reference<MobileMiscellaneousIssue, Stock, StockStatus>({
                bind: 'status',
                valueField: 'code',
                node: '@sage/x3-stock-data/StockStatus',
                title: 'Status',
                isReadOnly: true,
                isHidden: (value: StockStatus) => {
                    return !value?.code;
                },
            }),
            ui.nestedFields.text({
                bind: 'identifier1',
                title: 'Identifier 1',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'identifier2',
                title: 'Identifier 2',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField1',
                title: 'Stock custom field 1',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'stockCustomField2',
                title: 'Stock custom field 2',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'qualityAnalysisRequestId',
                title: 'Analysis req.',
                isReadOnly: true,
                isHidden: hideWhenEmptyValue,
            }),
            ui.nestedFields.text({
                bind: 'owner',
                title: 'Owner',
                isReadOnly: true,
                isHidden: true,
            }),
        ],
    })
    stockDetails: ui.fields.DetailList<Stock>;

    @ui.decorators.dropdownListField<MobileMiscellaneousIssue>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Unit',
        width: 'small',
        options: ['UN'],
        placeholder: 'Select...',
        isMandatory: true,
        isTransient: true,
        isDisabled: true,
        async onChange() {
            await this._onChange_packingUnitToIssue();
        },
    })
    packingUnitToIssue: ui.fields.DropdownList;

    /*
     * onChange packingUnitToIssue event (for decorator and management service)
     */
    private readonly _onChange_packingUnitToIssue: AsyncVoidFunction = async () => {
        if (!this.packingUnitToIssue.value) return;
        const selectedValue = this.packingUnitToIssue.value;
        const packingUnitIndex = this._packingUnits
            .map(packingUnit => packingUnit.node.packingUnit.code)
            .indexOf(selectedValue);
        if (packingUnitIndex !== -1) {
            const selectedUnit = this._packingUnits[packingUnitIndex].node;
            this.quantityToIssue.scale = selectedUnit.packingUnit.numberOfDecimals;

            this.packingUnitToStockUnitConversionFactorToIssue.value = Number(
                selectedUnit.packingUnitToStockUnitConversionFactor,
            );
            this.packingUnitToStockUnitConversionFactorToIssue.isDisabled = !selectedUnit.isPackingFactorEntryAllowed;

            const conversionFactor = this.packingUnitToStockUnitConversionFactorToIssue.value.toString();
            const numberOfDec = conversionFactor.includes('.') ? conversionFactor.split('.')[1].length : 0;
            this.packingUnitToStockUnitConversionFactorToIssue.scale = numberOfDec;
        } else {
            this.packingUnitToStockUnitConversionFactorToIssue.value = 1;
            this.packingUnitToStockUnitConversionFactorToIssue.isDisabled = true;
            this.quantityToIssue.scale =
                (this._numberOfDecimalList?.find(uom => uom.code === this.packingUnitToIssue?.value) || {})
                    .numberOfDecimals ?? 0;
        }
        this.quantityToIssue.value = null;
    };

    @ui.decorators.numericField<MobileMiscellaneousIssue>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Conversion factor',
        isMandatory: true,
        placeholder: 'Enter...',
        isTransient: true,
        async onChange() {
            await this._onChange_packingUnitToStockUnitConversionFactorToIssue();
        },
    })
    packingUnitToStockUnitConversionFactorToIssue: ui.fields.Numeric;

    /*
     * onChange packingUnitToStockUnitConversionFactorToIssue event (for decorator and management service)
     */
    private readonly _onChange_packingUnitToStockUnitConversionFactorToIssue: AsyncVoidFunction = async () => {
        if (!this.packingUnitToStockUnitConversionFactorToIssue.value) return;
        const selectedValue = this.packingUnitToIssue.value ?? '';
        const packingUnitIndex = this._packingUnits
            .map(packingUnit => packingUnit.node.packingUnit.code)
            .indexOf(selectedValue);
        if (packingUnitIndex !== -1) {
            const selectedUnit = this._packingUnits[packingUnitIndex].node;
            this.quantityToIssue.scale = selectedUnit.packingUnit.numberOfDecimals;
        } else {
            this.packingUnitToStockUnitConversionFactorToIssue.value = 1;
            this.packingUnitToStockUnitConversionFactorToIssue.isDisabled = true;
            this.quantityToIssue.scale =
                (this._numberOfDecimalList?.find(uom => uom.code === this.packingUnitToIssue.value) || {})
                    .numberOfDecimals ?? 0;
        }
        this.quantityToIssue.value = null;
    };

    @ui.decorators.numericField<MobileMiscellaneousIssue>({
        parent() {
            return this.quantityBlock;
        },
        title: 'Quantity to issue',
        placeholder: 'Enter...',
        isFullWidth: true,
        isTransient: true,
        postfix() {
            return ` ${this.packingUnitToIssue.value}`;
        },
        max() {
            const selectedRecord = this.stock.getRecordValue(this.gridBlock?.selectedRecordId ?? '');
            const conversionFactorOrigin = Number(selectedRecord?.packingUnitToStockUnitConversionFactor);
            const conversionFactorToIssue = Number(this.packingUnitToStockUnitConversionFactorToIssue.value);
            const quantityInPackingUnitRest = Number(selectedRecord?.quantityInPackingUnitRest);

            if (conversionFactorOrigin === conversionFactorToIssue) {
                return quantityInPackingUnitRest;
            }

            const factor =
                conversionFactorOrigin > conversionFactorToIssue
                    ? conversionFactorOrigin
                    : conversionFactorOrigin / conversionFactorToIssue;
            return quantityInPackingUnitRest * factor;
        },
        scale() {
            return (
                (<any>(<unknown>this.stock.getRecordValue(this.gridBlock?.selectedRecordId ?? '')))?.packingUnit
                    ?.numberOfDecimals ?? 0
            );
        },
        async onChange() {
            await this._onChange_quantityToIssue();
        },
    })
    quantityToIssue: ui.fields.Numeric;

    /*
     * onChange quantityToIssue event (for decorator and management service)
     */
    private readonly _onChange_quantityToIssue: AsyncVoidFunction = async () => {
        const currentRecord = this.stock.getRecordValue(this.gridBlock?.selectedRecordId ?? '') as any;
        currentRecord.quantityToIssue = String(this.quantityToIssue.value);
        this.stock.setRecordValue(currentRecord);
        await this.$.commitValueAndPropertyChanges();
    };

    @ui.decorators.referenceField<MobileMiscellaneousIssue, SerialNumber>({
        parent() {
            return this.serialNumberBlock;
        },
        title: 'Starting serial number',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/SerialNumber',
        valueField: 'code',
        canFilter: false,
        isTransient: true,
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        minLookupCharacters: 1,
        validation: /^$|^[^|]+$/,
        filter() {
            return {
                _and: [
                    { product: { code: this.productSelected?.value?.code } },
                    { stockId: this._stockId?.value ?? undefined },
                ],
            };
        },
        async onChange() {
            await this._onChange_startingSerialNumber();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                title: 'Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'stockId',
                title: 'Stock ID',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    startingSerialNumber: ui.fields.Reference<SerialNumber>;

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.serialNumberBlock;
        },
        title: 'Ending serial number',
        placeholder: 'Scan...',
        isTransient: true,
        isReadOnly: true,
        isFullWidth: true,
        async validation(value: string) {
            if (this._productSite.product.serialNumberManagementMode !== 'globalReceivedIssued') return;
            const currentQty =
                Number(this.quantityToIssue.value) * Number(this.packingUnitToStockUnitConversionFactorToIssue.value);
            if (
                (await getCountSerialNumber(
                    this,
                    this.productSelected.value?.code ?? '',
                    this.stockSite.value?.code,
                    this._stockId.value,
                    this.startingSerialNumber.value?.code ?? '',
                    this.endingSerialNumber.value ?? '',
                )) !== currentQty
            ) {
                return ui.localize(
                    '@sage/x3-stock/serial-number-not-sequential',
                    'The serial numbers are not sequential. Check your entry.',
                );
            }
        },
    })
    endingSerialNumber: ui.fields.Text;

    @ui.decorators.textField<MobileMiscellaneousIssue>({
        parent() {
            return this.serialNumberBlock;
        },
        placeholder: 'Scan...',
        isTransient: true,
        isHidden: true,
    })
    _stockId: ui.fields.Text;

    @ui.decorators.tableField<MobileMiscellaneousIssue>({
        parent() {
            return this.serialNumberBlock;
        },
        canFilter: false,
        canSelect: false,
        canExport: false,
        canUserHideColumns: false,
        isTransient: true,
        isFullWidth: true,
        node: '@sage/x3-stock/StockCountSerialNumber',
        mobileCard: undefined,
        columns: [
            ui.nestedFields.text({
                bind: 'startingSerialNumber',
                title: 'Starting serial number',
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantity',
                title: 'Quantity',
                isReadOnly: true,
                postfix() {
                    return this.stock.getRecordValue(String(this.gridBlock.selectedRecordId))?.packingUnit?.code;
                },
                scale() {
                    return (
                        this.stock.getRecordValue(this.gridBlock.selectedRecordId ?? '')?.packingUnit
                            ?.numberOfDecimals ?? 0
                    );
                },
            }),
            ui.nestedFields.text({
                bind: 'endingSerialNumber',
                title: 'Ending serial number',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
        dropdownActions: [
            {
                icon: 'bin',
                title: 'Remove', // Don't change this title
                onClick(recordId: string) {
                    const removedRecordSerialNumber = this.serialNumberLines.getRecordValue(recordId);
                    this.miscellaneousIssueLines?.value?.forEach((line, index) => {
                        if (String(line.numberSession) === String(this._numberSession)) {
                            line.stockDetails?.forEach((stock, stockIndex) => {
                                if (String(stock._id?.split('|')[2]) === String(removedRecordSerialNumber?.id)) {
                                    this.miscellaneousIssueLines.removeRecord(String(stock._id), 1);
                                }
                            });
                        }
                    });

                    const currentRecord = this.stock.getRecordValue(String(this.gridBlock.selectedRecordId));
                    if (currentRecord) {
                        currentRecord.quantityToIssue = String(0);
                        this.stock.setRecordValue(currentRecord);
                    }

                    //remove from the card
                    this.serialNumberLines.removeRecord(recordId);
                    this.startingSerialNumber.isDisabled = false;
                },
            },
        ],
        fieldActions() {
            return [this.addSerialRangeButton];
        },
    })
    serialNumberLines: ui.fields.Table<any>;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.detailStockLineSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationMiscellaneousIssueLinesBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousIssue>({
        parent() {
            return this.detailStockLineSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationMiscellaneousIssueDetailsBlock: ui.containers.Block;

    // =================================================================================================================
    // 🟢 INIT METHODS
    // =================================================================================================================

    private async _init(): Promise<boolean> {
        await this._initSite();
        this.transaction.isDisabled = false;
        if (this.stockSite.value && this._mobileSettings?.stockField1) {
            this.$.router.emptyPage;
            this.createButton.isDisabled = true;
            this.$standardNewAction;
            this.effectiveDate.value = DateValue.today().toString();
            this.entryType.value = 'miscellaneousIssue';
            await this._initTransaction();
            this._numberOfDecimalList = await getUnitNumberOfDecimalList(this);
            this.customizationMiscellaneousIssueHeaderBlock.isHidden =
                this.customizationMiscellaneousIssueHeaderBlock.layout.$layout?.$items.length === 0;
            return true;
        } else {
            return false;
        }
    }

    private async _initSite(): Promise<void> {
        const _stockSite = await getSelectedStockSite(
            this,
            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
            ui.localize(
                '@sage/x3-stock/dialog-error-location-inquiry-set-site',
                'Define a default stock site on the user function profile.',
            ),
        );

        if (!_stockSite) return;

        this.stockSite.value = {
            code: _stockSite,
            _id: _stockSite,
        };

        if (!this.stockSite.value) return;

        if (this.stockSite.value) {
            this._mobileSettings = JSON.parse(this.$.storage.get('mobile-settings-issue') as string);

            if (!this._mobileSettings?.stockField1) {
                dialogMessage(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_you_need_to_select_stock_search_parameters_to_set_up_Mobile_Automation_FUNADCSEARCH_function',
                        'You need to select stock search parameters to set up Mobile Automation - FUNADCSEARCH function.',
                    ),
                );
            }
        }
    }

    private async _initTransaction(): Promise<void | never> {
        try {
            this._transactions = extractEdges(
                await this.$.graph
                    .node('@sage/x3-stock/StockEntryTransaction')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                _id: true,
                                code: true,
                                isActive: true,
                                stockAutomaticJournal: {
                                    code: true,
                                    _id: true,
                                },
                                localizedDescription: true,
                                isLocationChange: true,
                                isStatusChange: true,
                                isUnitChange: true,
                                transactionType: true,
                                identifier1Destination: true,
                                identifier2Destination: true,
                                identifier1Detail: true,
                                identifier2Detail: true,
                                defaultStockMovementGroup: {
                                    code: true,
                                    _id: true,
                                },
                                stockMovementCode: {
                                    code: true,
                                    _id: true,
                                },
                                companyOrSiteGroup: {
                                    group: true,
                                },
                            },
                            {
                                filter: {
                                    transactionType: 'miscellaneousIssue',
                                    isActive: true,
                                },
                            },
                        ),
                    )
                    .execute(),
            );
        } catch (err) {
            ui.console.error(err);
        }

        const transaction = this._transactions[0];
        this.transaction.isDisabled = false;
        const hideTransaction = this._transactions.length <= 1;
        if (hideTransaction) {
            this.headerBlock.isHidden = true;
        }
        await this._transactionChange(transaction, hideTransaction);
    }

    private _initPackingUnitToTransfer() {
        let productPackingList = extractEdges(this._productSite.product.packingUnits.query).filter(productPacking => {
            return !!productPacking.packingUnit?.code;
        });

        this._packingUnits = productPackingList.map(productPacking => {
            return { node: productPacking };
        });

        let productPackingUnitSelectValues = productPackingList.map(productPacking => {
            return `${productPacking.packingUnit.code}`;
        });

        this.packingUnitToIssue.options = [this._productSite.product.stockUnit.code, ...productPackingUnitSelectValues];
        this.packingUnitToIssue.value = this.packingUnitToIssue.options[0];
        this.packingUnitToStockUnitConversionFactorToIssue.value = 1;
        const numberOfDecimals =
            (this._numberOfDecimalList?.find(uom => uom.code === this.packingUnitToIssue.value) || {})
                .numberOfDecimals ?? 0;
        this.packingUnitToIssue.value
            ? (this.quantityToIssue.scale = numberOfDecimals)
            : (this.quantityToIssue.scale = 0);
    }

    private async _initPageDetail() {
        this.lotFilter.isHidden = this._productSite?.product?.lotManagementMode === 'notManaged';
        if (
            !(this.lotFilter.isHidden || !!this._productSite?.product?.lotSequenceNumber) &&
            ['lotAndSublot', 'mandatoryLot'].includes(this._productSite?.product?.lotManagementMode)
        ) {
            this.lotFilter.isMandatory = true;
        }
        this.sublotFilter.isHidden = this.lotFilter.isHidden;
        this.sublotFilter.isHidden = this._productSite?.product?.lotManagementMode !== 'lotAndSublot';
        this._selectedStockManagementRules = await findStockManagementRules(
            this.stockSite.value?.code,
            this._productSite.product.productCategory.code,
            '2',
            String(this._transaction?.stockMovementCode?.code),
            this,
        );
        this.statusFilter.options = await this._getStockStatus();

        this.serialNumberFilter.isHidden = ['notManaged', 'issued'].includes(
            this._productSite.product.serialNumberManagementMode,
        );
        if (['receivedIssued', 'globalReceivedIssued'].includes(this._productSite.product.serialNumberManagementMode)) {
            if (this.lotFilter.isHidden === false) this.lotFilter.isMandatory = false;
            if (this.sublotFilter.isHidden === false) this.sublotFilter.isMandatory = false;
            if (this.statusFilter.isHidden === false) this.statusFilter.isMandatory = false;
            if (this.packingUnitFilter.isHidden === false) this.packingUnitFilter.isMandatory = false;
        }
        this.customizationMiscellaneousIssueLinesBlock.isHidden =
            this.customizationMiscellaneousIssueLinesBlock.layout.$layout?.$items.length === 0;
        this.customizationMiscellaneousIssueDetailsBlock.isHidden =
            this.customizationMiscellaneousIssueDetailsBlock.layout.$layout?.$items.length === 0;
    }

    // =================================================================================================================
    // 🟢 GET METHODS
    // =================================================================================================================

    private async _getStockStatus(): Promise<string[]> {
        const selectedStatus = this._selectedStockManagementRules.authorizedSubstatus
            .split(',')
            .map(status => ({ _regex: getRegExp(status).source }));

        const response: { edges: { node: { code: string } }[] } = await this.$.graph
            .node('@sage/x3-stock-data/StockStatus')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        code: true,
                    },
                    {
                        filter: {
                            code: { _or: selectedStatus },
                        },
                    },
                ),
            )
            .execute();

        if (!response.edges || response.edges.length === 0) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__invalid_stock_status_error',
                    'No stock status',
                ),
            );
        }

        return response.edges.map(stockStatus => stockStatus.node.code);
    }

    private _getQuantityInPackingUnitOrigin(record: any): number {
        let _quantityInPackingUnitOrigin = Number(record.quantityInPackingUnit);
        this.miscellaneousIssueLines.value?.forEach(line => {
            if (
                Number(line.stockId) === Number(record.stockId) &&
                String(line.numberSession) !== String(this._numberSession)
            ) {
                _quantityInPackingUnitOrigin -= Number(line.quantityInPackingUnit);
                const totalQty = line.stockDetails?.reduce((acc, stockDetail) => {
                    if (String(stockDetail._id?.split('|')[0]) === String(record.stockId)) {
                        acc += Number(stockDetail.quantityInPackingUnit);
                    }
                    return acc;
                }, 0);
                _quantityInPackingUnitOrigin = totalQty || 0;
            }
        });
        return _quantityInPackingUnitOrigin;
    }

    private _getQuantityInPackingUnitRest(record: any): number {
        let _quantityInPackingUnitRest = this._getQuantityInPackingUnitOrigin(record);
        if (this.productSelected.value?.serialNumberManagementMode === 'globalReceivedIssued') {
            this.miscellaneousIssueLines.value.forEach(line => {
                if (String(line.numberSession) === String(this._numberSession)) {
                    const totalQty = line.stockDetails?.reduce((acc, stockDetail) => {
                        if (String(stockDetail._id?.split('|')[0]) === String(record.stockId)) {
                            acc += Number(stockDetail.quantityInPackingUnit);
                        }
                        return acc;
                    }, 0);

                    _quantityInPackingUnitRest -= totalQty || 0;
                }
            });
        }
        return _quantityInPackingUnitRest;
    }

    private _getMiscellaneousIssueLine(record: any): any | undefined {
        const foundLine = this.miscellaneousIssueLines.value?.find(
            line => Number(line.numberSession) === this._numberSession,
        );
        return foundLine;
    }

    private _getQuantityToIssue(record: any): number {
        let _quantityToIssue = 0;
        this.miscellaneousIssueLines.value.forEach(line => {
            if (Number(line.numberSession) === this._numberSession) {
                const totalQty = line.stockDetails?.reduce((acc, stockDetail) => {
                    if (String(stockDetail._id?.split('|')[0]) === String(record.stockId)) {
                        acc +=
                            (Number(stockDetail.quantityInPackingUnit) *
                                Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                            Number(line.packingUnitToStockUnitConversionFactor);
                    }
                    return acc;
                }, 0);

                _quantityToIssue += totalQty || 0;
            }
        });

        if (isNaN(_quantityToIssue) || _quantityToIssue === 0) {
            _quantityToIssue = Number(record.quantityInPackingUnit);
        }
        return _quantityToIssue;
    }

    private _displayDetailFields() {
        this.headerSection.isHidden = true;
        this.miscellaneousIssueLinesSection.isHidden = true;
        this.stockLineSection.isHidden = false;
        this.detailStockLineSection.isHidden = true;
        this.miscellaneousIssueLinesBlock.isHidden = true;
        this.createButton.isHidden = true;
        this.nextButton.isHidden = false;
        this.CancelButton.isHidden = true;
        this.SelectButton.isHidden = true;
    }

    private async _getProductSite(productCode: string): Promise<ProductSite> {
        try {
            const response = await this.$.graph
                .node('@sage/x3-master-data/ProductSite')
                .read(
                    {
                        isLocationManaged: true,
                        isLicensePlateNumberManaged: true,
                        defaultInternalContainer: {
                            code: true,
                        },
                        _id: true,
                        product: {
                            _id: true,
                            code: true,
                            localizedDescription1: true,
                            lotManagementMode: true,
                            serialNumberManagementMode: true,
                            stockVersionMode: true,
                            isVersionManaged: true,
                            expirationManagementMode: true,
                            serialSequenceNumber: true,
                            lotSequenceNumber: true,
                            stockManagementMode: true,
                            defaultPotencyInPercentage: true,
                            expirationLeadTime: true,
                            expirationTimeUnit: true,
                            useByDateCoefficient: true,
                            stockUnit: {
                                code: true,
                                numberOfDecimals: true,
                                _id: true,
                            },
                            productCategory: {
                                code: true,
                                _id: true,
                            },
                            packingUnits: {
                                query: {
                                    edges: {
                                        node: {
                                            packingUnit: {
                                                code: true,
                                                numberOfDecimals: true,
                                                _id: true,
                                            },
                                            packingUnitToStockUnitConversionFactor: true,
                                            isPackingFactorEntryAllowed: true,
                                            _id: true,
                                        },
                                    },
                                },
                            },
                        },
                    },
                    `${productCode}|${this.stockSite.value?.code}`,
                )
                .execute();

            return response as ProductSite;
        } catch (error) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__failed-to-fetch-product-site-data',
                    'There are no sites associated with the product. You need to create a product-site record.',
                ),
            );
        }
    }

    private _buildStockFilter() {
        const lotFilter = this._getLotFilter();
        return {
            availableQuantity: { _gt: '0' },
            isBeingCounted: false,
            owner: this.stockSite.value?.code,
            location: { category: { _nin: ['subcontract', 'customer'] } },
            product: { product: { code: this.productSelected.value?.code } },
            ...(!this.locationSelected.isHidden &&
                this.locationSelected.value && { location: { code: this.locationSelected.value.code } }),
            ...(!this.locationFilter.isHidden &&
                this.locationFilter.value && { location: { code: this.locationFilter.value.code } }),
            ...(!this.lotFilter.isHidden && this.lotFilter.value && { lot: this.lotFilter.value.lot }),
            ...(!this.sublotFilter.isHidden && this.sublotFilter.value && { sublot: this.sublotFilter.value.sublot }),
            ...(!this.licensePlateNumberFilter.isHidden &&
                this.licensePlateNumberFilter.value && {
                licensePlateNumber: this.licensePlateNumberFilter.value.code,
            }),
            ...(!this.serialNumberFilter.isHidden &&
                this.serialNumberFilter.value && { serialNumber: this.serialNumberFilter.value.code }),
            ...(!this.statusFilter.isHidden &&
                this.statusFilter.value && { status: { code: this.statusFilter.value } }),
            ...(!this.packingUnitFilter.isHidden &&
                this.packingUnitFilter.value && { packingUnit: { code: this.packingUnitFilter.value } }),
            ...(!this.packingUnitToStockUnitConversionFactorFilter.isHidden &&
                this.packingUnitToStockUnitConversionFactorFilter.value && {
                packingUnitToStockUnitConversionFactor: this.packingUnitToStockUnitConversionFactorFilter.value,
            }),
            ...(!this.identifier1Filter.isHidden &&
                this.identifier1Filter.value && { identifier1: this.identifier1Filter.value }),
            ...(!this.identifier2Filter.isHidden &&
                this.identifier2Filter.value && { identifier2: this.identifier2Filter.value }),
            ...(!this.stockCustomField1Filter.isHidden &&
                this.stockCustomField1Filter.value && { stockCustomField1: this.stockCustomField1Filter.value }),
            ...(!this.stockCustomField2Filter.isHidden &&
                this.stockCustomField2Filter.value && { stockCustomField2: this.stockCustomField2Filter.value }),
            _and: [
                {
                    ...(!this.lotFilter.isHidden && lotFilter && { lotReference: { ...lotFilter } }),
                },
            ],
        };
    }

    private _mapStockLine(stockLine: any) {
        return {
            _id: stockLine?.node?._id,
            quantityInPackingUnit: Number(stockLine?.node?.quantityInPackingUnit),
            product: {
                product: stockLine?.node?.product?.product
                    ? {
                        code: stockLine?.node?.product?.product?.code,
                        serialNumberManagementMode: stockLine?.node?.product?.product?.serialNumberManagementMode,
                        _id: stockLine?.node?.product?.product?._id,
                    }
                    : undefined,
            },
            licensePlateNumber: stockLine?.node?.licensePlateNumber
                ? {
                    code: stockLine?.node?.licensePlateNumber?.code,
                    _id: stockLine?.node?.licensePlateNumber?._id,
                }
                : undefined,
            location: stockLine?.node?.location
                ? {
                    code: stockLine?.node?.location.code,
                    _id: stockLine?.node?.location._id,
                }
                : undefined,
            lot: stockLine?.node?.lot,
            sublot: stockLine?.node?.sublot,
            serialNumber: stockLine?.node?.serialNumber,
            status: stockLine?.node?.status
                ? {
                    code: stockLine?.node?.status.code,
                    _id: stockLine?.node?.status._id,
                }
                : undefined,
            packingUnit: stockLine?.node?.packingUnit
                ? {
                    code: stockLine?.node?.packingUnit?.code,
                    numberOfDecimals: stockLine?.node?.packingUnit?.numberOfDecimals,
                    _id: stockLine?.node?.packingUnit?._id,
                }
                : undefined,
            packingUnitToStockUnitConversionFactor: stockLine?.node?.packingUnitToStockUnitConversionFactor,
            identifier1: stockLine?.node?.identifier1,
            identifier2: stockLine?.node?.identifier2,
            stockCustomField1: stockLine?.node?.stockCustomField1,
            stockCustomField2: stockLine?.node?.stockCustomField2,
            stockId: stockLine?.node?.stockId,
            lotReference: stockLine?.node?.lotReference
                ? {
                    expirationDate: stockLine?.node?.lotReference?.expirationDate,
                    useByDate: stockLine?.node?.lotReference?.useByDate,
                    lotCustomField1: stockLine?.node?.lotReference?.lotCustomField1,
                    lotCustomField2: stockLine?.node?.lotReference?.lotCustomField2,
                    majorVersion: stockLine?.node?.lotReference?.majorVersion
                        ? {
                            code: stockLine?.node?.lotReference?.majorVersion?.code,
                            status: stockLine?.node?.lotReference?.majorVersion?.status,
                            isSubjectToException: stockLine?.node?.lotReference?.majorVersion?.isSubjectToException,
                            exceptionDate: stockLine?.node?.lotReference?.majorVersion?.exceptionDate,
                            isOnHold: stockLine?.node?.lotReference?.majorVersion?.isOnHold,
                            _id: stockLine?.node?.lotReference?.majorVersion?._id,
                        }
                        : undefined,
                    product: stockLine?.node?.lotReference?.product
                        ? {
                            code: stockLine?.node?.lotReference?.product.code,
                            _id: stockLine?.node?.lotReference?.product._id,
                        }
                        : undefined,
                    _id: stockLine?.node?.lotReference._id,
                }
                : undefined,
            quantityInStockUnit: stockLine?.node?.quantityInStockUnit,
            allocatedQuantity: stockLine?.node?.allocatedQuantity,
            qualityAnalysisRequestId: stockLine?.node?.qualityAnalysisRequestId,
            owner: stockLine?.node?.owner,
        };
    }

    private async _getStock() {
        const filter = this._buildStockFilter();
        try {
            const response = await this.$.graph
                .node('@sage/x3-stock-data/Stock')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            quantityInPackingUnit: true,
                            product: {
                                product: {
                                    code: true,
                                    serialNumberManagementMode: true,
                                    localizedDescription1: true,
                                    _id: true,
                                },
                                _id: true,
                            },
                            licensePlateNumber: {
                                code: true,
                                _id: true,
                            },
                            location: {
                                code: true,
                                _id: true,
                            },
                            lot: true,
                            sublot: true,
                            serialNumber: true,
                            status: {
                                code: true,
                                _id: true,
                            },
                            packingUnit: {
                                code: true,
                                numberOfDecimals: true,
                                _id: true,
                            },
                            packingUnitToStockUnitConversionFactor: true,
                            identifier1: true,
                            identifier2: true,
                            stockCustomField1: true,
                            stockCustomField2: true,
                            stockId: true,
                            lotReference: {
                                expirationDate: true,
                                useByDate: true,
                                lotCustomField1: true,
                                lotCustomField2: true,
                                majorVersion: {
                                    code: true,
                                    status: true,
                                    isSubjectToException: true,
                                    exceptionDate: true,
                                    isOnHold: true,
                                    _id: true,
                                },
                                product: {
                                    code: true,
                                    _id: true,
                                },
                                _id: true,
                            },
                            quantityInStockUnit: true,
                            allocatedQuantity: true,
                            qualityAnalysisRequestId: true,
                            owner: true,
                        },
                        { filter },
                    ),
                )
                .execute();

            this._originalStockLines = [];
            const _stockItems = response.edges.map(this._mapStockLine);
            if (
                this.productSelected?.value?.isVersionManaged &&
                this.productSelected?.value?.stockVersionMode !== 'no'
            ) {
                const _stockItemsValid = await this._filterStockItemsByVersion(_stockItems);
                this.stock.value = _stockItemsValid;
            } else {
                this.stock.value = _stockItems;
            }
            this._originalStockLines = this.stock?.value?.map(item => ({ ...item }));
        } catch (error) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__failed-to-fetch-stock-data',
                    'The item has no stock.',
                ),
            );
        }
    }

    private async _filterStockItemsByVersion(stockItems: any[]): Promise<any[]> {
        const _stockItemsValid: any[] = [];
        this._lots = [];
        for (const item of stockItems) {
            if (item.lotReference?.majorVersion) {
                const isValidVersion = await this.controlVersions(
                    item.lotReference.majorVersion as ExtractEdges<MajorVersionStatus>,
                    this._selectedStockManagementRules.activeVersion,
                    this._selectedStockManagementRules.versionStopped,
                    this._selectedStockManagementRules.prototypeVersion,
                    this.effectiveDate.value ?? '',
                );
                if (isValidVersion) {
                    _stockItemsValid.push(item);
                    this._lots.push(item.lot);
                }
            }
        }
        return _stockItemsValid;
    }

    private async _getStockByStockId(_stockId: string) {
        try {
            const response = await this.$.graph
                .node('@sage/x3-stock-data/Stock')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            quantityInPackingUnit: true,
                            packingUnitToStockUnitConversionFactor: true,
                            stockId: true,
                            quantityInStockUnit: true,
                            allocatedQuantity: true,
                        },
                        {
                            filter: {
                                stockId: _stockId,
                            },
                            first: 500,
                        },
                    ),
                )
                .execute();
            this._originalStockLines = [];
            this._originalStockLines = response.edges.map(stockLine => {
                return {
                    _id: stockLine.node?._id,
                    quantityInPackingUnit: stockLine?.node?.quantityInPackingUnit,
                    packingUnitToStockUnitConversionFactor: stockLine?.node?.packingUnitToStockUnitConversionFactor,
                    stockId: stockLine?.node?.stockId,
                    quantityInStockUnit: stockLine?.node?.quantityInStockUnit,
                    allocatedQuantity: stockLine?.node?.allocatedQuantity,
                };
            });
        } catch (error) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__failed-to-fetch-stock-data',
                    'The item has no stock.',
                ),
            );
        }
    }

    private async _getProductAtLocation() {
        try {
            const response = await this.$.graph
                .node('@sage/x3-stock-data/Stock')
                .aggregate.query(
                    ui.queryUtils.edgesSelector(
                        {
                            group: {
                                product: {
                                    product: {
                                        code: {
                                            _by: 'value',
                                        },
                                    },
                                },
                            },
                        },
                        {
                            filter: {
                                availableQuantity: { _gt: '0' },
                                isBeingCounted: false,
                                owner: this.stockSite.value?.code,
                                location: { code: this.locationSelected.value?.code },
                            },
                            first: 1000,
                        },
                    ),
                )
                .execute();

            this._productAtLocation = response?.edges?.map(item => item.node?.group?.product?.product?.code) || [];
        } catch (error) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__failed-to-fetch-product-at-the-given-location',
                    'The product is not stored at the location.',
                ),
            );
        }
    }

    private _getCustomizationMiscellaneousIssueLines() {
        let response = {};
        this.customizationMiscellaneousIssueLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private _getCustomizationMiscellaneousIssueDetails() {
        let response = {};
        this.customizationMiscellaneousIssueDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private _getSidebarMiscellaneousIssueLine(): ui.SidebarFieldDefinition<MiscellaneousIssueLineBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<MiscellaneousIssueLineBinding>[] = [];
        this.customizationMiscellaneousIssueLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<MiscellaneousIssueLineBinding>);
            }
        });
        return _FieldsSidebar;
    }

    private _getSidebarStockJournal(): ui.SidebarFieldDefinition<StockJournalBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockJournalBinding>[] = [];
        _FieldsSidebar.push('packingUnit');
        _FieldsSidebar.push('packingUnitToStockUnitConversionFactor');
        _FieldsSidebar.push('quantityInPackingUnit');
        _FieldsSidebar.push('stockUnit');
        _FieldsSidebar.push('quantityInStockUnit'), _FieldsSidebar.push('startingSerialNumber');
        _FieldsSidebar.push('endingSerialNumber');
        this.customizationMiscellaneousIssueDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockJournalBinding>);
            }
        });
        return _FieldsSidebar;
    }

    // =================================================================================================================
    // 🟢 ONCHANGE METHODS TO UPDATE THE DATA WHEN THE INPUT VALUE CHANGES.
    // =================================================================================================================

    private async _transactionChange(transaction: PartialStockEntryTransaction, hideTransaction = false) {
        if (this._transactions) {
            this.transaction.options = this._transactions.map(trs => String(trs.code));
        }
        this._transaction = transaction;
        this.transaction.value = String(transaction.code);
        this.transaction.isHidden = hideTransaction;
        this.productSelected.isDisabled = false;

        const _pushSettings = (field: StockSearchFilter): boolean => {
            if (field !== 'none') {
                this._stockFieldSettings?.push(field);
                return true;
            }
            return false;
        };

        const _allPushed =
            _pushSettings(this._mobileSettings?.stockField1) &&
            _pushSettings(this._mobileSettings?.stockField2) &&
            _pushSettings(this._mobileSettings?.stockField3) &&
            _pushSettings(this._mobileSettings?.stockField4) &&
            _pushSettings(this._mobileSettings?.stockField5) &&
            _pushSettings(this._mobileSettings?.stockField6) &&
            _pushSettings(this._mobileSettings?.stockField7) &&
            _pushSettings(this._mobileSettings?.stockField8);

        if (!findSetting('location', this._stockFieldSettings)) {
            this.locationSelected.isHidden = true;
        }

        this.stockMovementCode.value = transaction.stockMovementCode
            ? {
                code: transaction.stockMovementCode.code,
                _id: transaction.stockMovementCode._id,
            }
            : null;
        this.stockMovementGroup.value = transaction.defaultStockMovementGroup
            ? {
                code: transaction.defaultStockMovementGroup.code,
                _id: transaction.defaultStockMovementGroup._id,
            }
            : null;
        this.stockAutomaticJournal.value = transaction.stockAutomaticJournal
            ? {
                code: transaction.stockAutomaticJournal.code,
                _id: transaction.stockAutomaticJournal._id,
            }
            : null;
    }

    private async _onChange_startingSerialNumber() {
        const currentRecord = this.stock.getRecordValue(String(this.gridBlock.selectedRecordId)) as any;
        const currentQty =
            Number(this.quantityToIssue.value) * Number(this.packingUnitToStockUnitConversionFactorToIssue.value);

        if (!currentQty || !this.startingSerialNumber.value) {
            this.endingSerialNumber.value = null;
            return;
        }

        this.startingSerialNumber.value.code = this.startingSerialNumber.value.code?.toUpperCase();

        if (currentQty > 1) {
            this.endingSerialNumber.value = this._calculateEndingSerialNumber(
                String(this.startingSerialNumber.value.code),
                currentQty,
            );
        } else {
            this.endingSerialNumber.value = String(this.startingSerialNumber.value.code);
        }

        this.addSerialRangeButton.isHidden = currentQty > currentRecord.quantityInPackingUnitOrigin;

        await this.$.commitValueAndPropertyChanges();
        const validationResult = await this.endingSerialNumber.validate();
        if (validationResult) {
            this.$.showToast(validationResult, { type: 'warning' });
        }
    }

    /**
     * OnChange product readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */

    private _onChange_productSelected: AsyncVoidFunction = async () => {
        if (await this.productSelected.value?.code) {
            const result = withoutEdges(
                await this.$.graph
                    .node('@sage/x3-master-data/ProductSite')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                isBeingCounted: true,
                                product: {
                                    globalTradeItemNumber: true,
                                },
                            },
                            {
                                filter: {
                                    product: {
                                        code: this.productSelected.value?.code,
                                    },
                                    stockSite: {
                                        code: this.stockSite.value?.code,
                                    },
                                },
                            },
                        ),
                    )
                    .execute(),
            ) as ExtractEdgesPartial<ProductSite>[];

            if (result.length === 0) {
                this.productSelected.value = null;
                this.productSelected.focus();
            }

            if (result[0].isBeingCounted === true) {
                if (
                    !(await this.$.dialog
                        .confirmation(
                            'warn',
                            ui.localize('@sage/x3-stock/dialog-warning-title', 'Warning'),
                            ui.localize(
                                '@sage/x3-stock/product-blocked-by-count-continue',
                                'Product blocked by count. Do you want to continue?',
                            ),
                            {
                                acceptButton: {
                                    text: ui.localize('@sage/x3-stock/button-accept-yes', 'Yes'),
                                },
                                cancelButton: {
                                    text: ui.localize('@sage/x3-stock/button-cancel-no', 'No'),
                                },
                            },
                        )
                        .then(() => true)
                        .catch(() => false))
                ) {
                    this.productSelected.value = null;
                    this.productSelected.focus();
                    return;
                }
            }

            this._numberSession = Math.floor(100000 + Math.random() * 900000);
            this._productSite = await this._getProductSite(this.productSelected.value?.code);
            this._packingUnits = this._productSite.product.packingUnits?.query.edges.map(packingUnit => {
                return packingUnit;
            });
            this.packingUnitFilter.options = [
                this._productSite.product.stockUnit.code,
                ...this._packingUnits.map(packingUnit => {
                    return packingUnit.node.packingUnit.code;
                }),
            ];
            const isGlobalReceivedIssued =
                this.productSelected.value?.serialNumberManagementMode === 'globalReceivedIssued';
            this.addSerialRangeButton.isHidden = !isGlobalReceivedIssued;
            this.serialNumberBlock.isHidden = !isGlobalReceivedIssued;
            this.serialNumberLines.isTitleHidden = !isGlobalReceivedIssued;

            this.serialNumberLines.title = 'Serial number(s) to issue    ';
            await this._initPageDetail();
            managePages(
                this,
                this.stockSite.value?.code,
                {
                    product: {
                        code: this._productSite.product.code,
                        lotManagementMode: this._productSite.product.lotManagementMode,
                        serialNumberManagementMode: this._productSite.product.serialNumberManagementMode,
                    },
                    isLocationManaged: this._productSite.isLocationManaged,
                    isLicensePlateNumberManaged: this._productSite.isLicensePlateNumberManaged,
                },
                '3',
                ui.localize(
                    '@sage/x3-stock/pages__mobile-settings__mandatory-settings-missing',
                    'Mandatory settings missing.',
                ),
                this._stockFieldSettings,
                this._mobileSettings,
            );
            this.productCode.value = this.productSelected.value?.code;
            this.productDescription.value = this.productSelected.value?.localizedDescription1;

            if (!findSetting('location', this._stockFieldSettings) || this._productSite.isLocationManaged !== true) {
                this.locationFilter.isHidden = true;
            }
            if (!this.locationSelected.isHidden && this.locationSelected.value && !this.locationFilter.isHidden) {
                this.locationFilter.isHidden = true;
            }

            this._displayDetailFields();
            await this._getStock();
        } else {
            this.createButton.isDisabled = this.miscellaneousIssueLines.value.length === 0;
        }
    };

    private async _onRowClick(recordId: string, rowItem: Stock) {
        const originalStockLine = this._originalStockLines?.find(line => recordId === line._id);
        const currentRecord = this.stock.getRecordValue(recordId) as any;
        rowItem.quantityInPackingUnit = String(originalStockLine?.quantityInPackingUnit);
        rowItem.quantityInStockUnit = String(originalStockLine?.quantityInStockUnit);
        rowItem.packingUnit = originalStockLine?.packingUnit;
        rowItem.packingUnit.numberOfDecimals = originalStockLine?.packingUnit.numberOfDecimals;

        this.stockDetails.value = [rowItem]; // populate details list
        this.gridBlock.selectedRecordId = recordId; // populate grid row block
        this._stockId.value = rowItem.stockId;

        this._initPackingUnitToTransfer();
        const selectedValue = currentRecord.packingUnit?.code;
        const packingUnitIndex = this._packingUnits
            .map(packingUnit => packingUnit.node.packingUnit.code)
            .indexOf(selectedValue);
        if (packingUnitIndex !== -1) {
            const selectedUnit = this._packingUnits[packingUnitIndex].node;
            this.packingUnitToStockUnitConversionFactorToIssue.isDisabled = !selectedUnit.isPackingFactorEntryAllowed;
            this.quantityToIssue.scale = selectedUnit.packingUnit.numberOfDecimals;
        } else {
            this.packingUnitToStockUnitConversionFactorToIssue.isDisabled = true;
            this.quantityToIssue.scale = currentRecord.packingUnit?.numberOfDecimals;
        }

        this.quantityToIssue.value = null;
        this.packingUnitToStockUnitConversionFactorToIssue.value =
            Number(originalStockLine?.packingUnitToStockUnitConversionFactor) ?? 1;
        this.packingUnitToIssue.value = originalStockLine?.packingUnit?.code;

        const conversionFactor = this.packingUnitToStockUnitConversionFactorToIssue.value.toString();
        const numberOfDec = conversionFactor.includes('.') ? conversionFactor.split('.')[1].length : 0;
        this.packingUnitToStockUnitConversionFactorToIssue.scale = numberOfDec;

        if (this._packingUnits.length) {
            this.packingUnitToIssue.isDisabled = false;
        }

        currentRecord.quantityInStockUnit = originalStockLine?.quantityInStockUnit;
        currentRecord.packingUnit.code = originalStockLine?.packingUnit.code;
        currentRecord.quantityInPackingUnit = originalStockLine?.quantityInPackingUnit;

        this.quantityBlock.layout.$layout?.$items.forEach((item: any) => {
            if (item.$bind != 'packingUnitToIssue' && item.$bind != 'packingUnitToStockUnitConversionFactorToIssue') {
                const line = this._getMiscellaneousIssueLine(
                    this.stock.getRecordValue(String(this.gridBlock.selectedRecordId)) as any,
                );

                if (line && line[item.$bind]) {
                    (this as any)[item.$bind].value = line[item.$bind];
                } else {
                    (this as any)[item.$bind].value = null;
                }
            }
        });

        if (this._productSite.product.serialNumberManagementMode === 'globalReceivedIssued') {
            this.serialNumberLines.isHidden = false;
            this.serialNumberLines.value = [];
        }
        this.stock.setRecordValue(currentRecord);
        await this.$.commitValueAndPropertyChanges();
        await this.stock.validateWithDetails();

        this.headerSection.isHidden = true;
        this.miscellaneousIssueLinesSection.isHidden = true;
        this.stockLineSection.isHidden = true;
        this.detailStockLineSection.isHidden = false;
        this.miscellaneousIssueLinesBlock.isHidden = true;
        this.createButton.isHidden = true;
        this.nextButton.isHidden = true;
        this.CancelButton.isHidden = false;
        this.SelectButton.isHidden = false;
        this.quantityToIssue.focus();
    }

    private async _onRowSelect(recordId: string, rowItem: Stock) {
        let currentRecord = this.stock.getRecordValue(recordId) as any;
        if (currentRecord) {
            const line = this.miscellaneousIssueLines.value.find(
                line => Number(line.numberSession) === this._numberSession,
            );

            const lineNumber = line ? line.lineNumber : this._numberSession;
            const numberOfDecimals =
                (this._numberOfDecimalList?.find(uom => uom.code === currentRecord.packingUnit?.code) || {})
                    .numberOfDecimals ?? 0;
            this.miscellaneousIssueLines.addOrUpdateRecordValue(
                {
                    _id: `${lineNumber}`,
                    entryType: 'miscellaneousIssue',
                    lineNumber: lineNumber,
                    product: {
                        code: currentRecord.product?.product?.code,
                        _id: currentRecord.product?.product?._id,
                    },
                    productDescription: this.productSelected?.value?.localizedDescription1,
                    packingUnit: {
                        code: currentRecord.packingUnit?.code,
                        numberOfDecimals: currentRecord.packingUnit?.numberOfDecimals ?? numberOfDecimals,
                        _id: currentRecord.packingUnit?._id,
                    },
                    quantityInPackingUnit: currentRecord.quantityInPackingUnit,
                    packingUnitToStockUnitConversionFactor: currentRecord.packingUnitToStockUnitConversionFactor,
                    quantityInStockUnit: currentRecord.quantityInPackingUnit,
                    stockId: String(currentRecord.stockId),
                    numberSession: String(this._numberSession),
                    ...this._getCustomizationMiscellaneousIssueLines(),
                },
                0,
                undefined,
            );
            this.miscellaneousIssueLines.addOrUpdateRecordValue(
                {
                    _id: `${currentRecord.stockId}|${lineNumber}`,
                    packingUnit: {
                        code: currentRecord.packingUnit?.code,
                        numberOfDecimals: currentRecord.packingUnit?.numberOfDecimals ?? numberOfDecimals,
                        _id: currentRecord.packingUnit?._id,
                    },
                    product: {
                        code: currentRecord.product?.product?.code,
                        _id: currentRecord.product?.product?._id,
                    },
                    packingUnitToStockUnitConversionFactor:
                        currentRecord?.packingUnitToStockUnitConversionFactor?.toString(),
                    quantityInStockUnit: (
                        Number(currentRecord.quantityInPackingUnit) *
                        Number(currentRecord.packingUnitToStockUnitConversionFactor)
                    ).toString(),
                    quantityInPackingUnit: currentRecord.quantityInPackingUnit,
                    location: {
                        code: currentRecord.location?.code,
                        _id: currentRecord.location?._id,
                    },
                    status: {
                        code: currentRecord.status?.code,
                        _id: currentRecord.status?._id,
                    },
                    lot: currentRecord.lot ?? undefined,
                    sublot: currentRecord.sublot ?? undefined,
                    serialNumber: currentRecord.serialNumber ?? undefined,
                    licensePlateNumber: currentRecord.licensePlateNumber?.code ?? undefined,
                    stockUnit: {
                        code: this._productSite.product?.stockUnit?.code,
                        numberOfDecimals: this._productSite.product.stockUnit.numberOfDecimals,
                        _id: this._productSite.product.stockUnit._id,
                    },
                    identifier1: currentRecord.identifier1 ?? undefined,
                    identifier2: currentRecord.identifier2 ?? undefined,
                    startingSerialNumber: this.startingSerialNumber?.value?.code ?? undefined,
                    endingSerialNumber: this.endingSerialNumber?.value ?? undefined,
                    stockCustomField1: currentRecord.stockCustomField1 ?? undefined,
                    stockCustomField2: currentRecord.stockCustomField2 ?? undefined,
                    ...this._getCustomizationMiscellaneousIssueDetails(),
                },
                1,
                `${lineNumber}`,
            );
        }
        let qtyTotal = 0;
        this.miscellaneousIssueLines.value.forEach(line => {
            if (Number(line.numberSession) === this._numberSession) {
                const totalQty = line.stockDetails?.reduce((acc, stockDetail) => {
                    if (String(stockDetail._id?.split('|')[0]) === String(currentRecord.stockId)) {
                        acc +=
                            (Number(stockDetail.quantityInPackingUnit) *
                                Number(stockDetail.packingUnitToStockUnitConversionFactor)) /
                            Number(line.packingUnitToStockUnitConversionFactor);
                    }
                    return acc;
                }, 0);

                qtyTotal += totalQty || 0;
            }
        });

        (currentRecord as any).quantityToIssue = qtyTotal;
        this.stock.setRecordValue(currentRecord);
        this._updateMiscellaneousIssueLine(String(this._numberSession));
    }

    private _disablePage(): void {
        this.effectiveDate.isDisabled = true;
        this.transaction.isDisabled = true;
        this.productSelected.isDisabled = true;
    }

    private _resetFieldFilters() {
        const filters = [
            this.licensePlateNumberFilter,
            this.locationFilter,
            this.lotFilter,
            this.sublotFilter,
            this.serialNumberFilter,
            this.statusFilter,
            this.packingUnitFilter,
            this.packingUnitToStockUnitConversionFactorFilter,
            this.identifier1Filter,
            this.identifier2Filter,
            this.stockCustomField1Filter,
            this.stockCustomField2Filter,
        ];

        filters.forEach(filter => {
            if (!filter.isHidden) {
                filter.value = null;
            }
        });
    }

    private _calculateEndingSerialNumber(startingSerialNumber: string, quantity: number): string {
        return startingSerialNumber.replace(/\d+$/, match => {
            const endingNumber = (Number(match) + quantity - 1).toString();
            const lengthDiff = Math.max(endingNumber.length - match.length, 0);
            return endingNumber.padStart(match.length + lengthDiff, '0');
        });
    }

    private _validateEntry(value: any, rowValue: any): any {
        const originalStockLine = this._originalStockLines[0];
        const inputQuantityInStockUnit =
            Number(rowValue.quantityInPackingUnit) * Number(rowValue.packingUnitToStockUnitConversionFactor);

        const quantityAvailableInStockUnit =
            Number(originalStockLine.quantityInPackingUnit) *
            Number(originalStockLine.packingUnitToStockUnitConversionFactor);
        if (inputQuantityInStockUnit > quantityAvailableInStockUnit) {
            return ui.localize(
                '@sage/x3-stock/pages__mobile_miscellaneous_issue__entered-quantity-exceeds-the-available-quantity',
                `The entered quantity exceeds the available quantity.`,
            );
        }

        const quantityInStockUnitUnallocated =
            quantityAvailableInStockUnit - Number(originalStockLine.allocatedQuantity);
        if (inputQuantityInStockUnit > quantityInStockUnitUnallocated) {
            return ui.localize(
                '@sage/x3-stock/pages__stock_change_lines__enter_a_quantity_less_than_or_equal_to_the_stock_quantity_minus_allocated_quantity',
                'Enter a quantity less than or equal to the stock quantity minus the allocated quantity.',
            );
        }
        return undefined;
    }

    private _updateRecordValue(_id: any, value: any) {
        value.quantityInStockUnit = String(
            Number(value.quantityInPackingUnit) * Number(value.packingUnitToStockUnitConversionFactor),
        );
        value.serialNumber = value.startingSerialNumber?.code ?? value.serialNumber ?? undefined;
        value.startingSerialNumber = undefined;
        value.startingSerialNumber = value.serialNumber;
        this.miscellaneousIssueLines.addOrUpdateRecordValue(value, 1, value._id);
    }
    private _updateMiscellaneousIssueLine(id: string) {
        const getUniqueValues = <T, K extends keyof T>(arr: T[], key: K): T[K][] => {
            return [...new Set(arr.map(item => item[key]))];
        };

        this.miscellaneousIssueLines?.value?.forEach(line => {
            if (Number(line.numberSession) === this._numberSession) {
                if (line.stockDetails?.length === 0) {
                    this.miscellaneousIssueLines.removeRecord(String(line._id), 0);
                    return;
                }
                const detailsQuantityInstockUnit =
                    line?.stockDetails?.reduce((acc, item) => acc + Number(item.quantityInStockUnit), 0) ?? 0;

                const lineQuantityInPackingUnit =
                    detailsQuantityInstockUnit / Number(line.packingUnitToStockUnitConversionFactor);
                const stockUnitValues = getUniqueValues(line.stockDetails ?? [], 'stockUnit');
                const hasMatchingInnerValues =
                    line.stockDetails?.length <= 1 ||
                    line.stockDetails?.every(innerItem => innerItem.packingUnit === line.packingUnit);

                if (!hasMatchingInnerValues) {
                    this.miscellaneousIssueLines.addOrUpdateRecordValue(
                        {
                            _id: `${id}`,
                            packingUnit: stockUnitValues[0],
                            packingUnitToStockUnitConversionFactor: '1',
                            quantityInPackingUnit: String(detailsQuantityInstockUnit),
                        },
                        0,
                        undefined,
                    );
                } else {
                    this.miscellaneousIssueLines.addOrUpdateRecordValue(
                        {
                            _id: `${id}`,
                            quantityInPackingUnit: String(lineQuantityInPackingUnit),
                        },
                        0,
                        undefined,
                    );
                }
            }
        });
    }

    private async _validateSerialNumber(value: any, rowValue: any) {
        let applyUpdate = false;
        if (typeof rowValue.startingSerialNumber === 'string') {
            if (!rowValue.startingSerialNumber) {
                return ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumber',
                    `The serial number is mandatory`,
                );
            }
            rowValue.startingSerialNumber = {
                code:
                    rowValue.startingSerialNumber?.code?.toUpperCase() ?? rowValue.startingSerialNumber?.toUpperCase(),
            };
            applyUpdate = true;
        } else {
            if (!rowValue.startingSerialNumber?.code) {
                return ui.localize(
                    '@sage/x3-stock/pages__mobile_miscellaneous_issue__notification__error_startingSerialNumber',
                    `The serial number is mandatory`,
                );
            }
        }

        const currentQty =
            Number(rowValue.quantityInPackingUnit) * Number(rowValue.packingUnitToStockUnitConversionFactor);
        if (!currentQty || !rowValue.startingSerialNumber?.code) {
            rowValue.endingSerialNumber = null;
            return;
        }

        rowValue.startingSerialNumber.code = rowValue.startingSerialNumber?.code?.toUpperCase();
        if (currentQty > 1) {
            rowValue.endingSerialNumber = this._calculateEndingSerialNumber(
                String(rowValue.startingSerialNumber?.code),
                currentQty,
            );
        } else {
            rowValue.endingSerialNumber = rowValue.startingSerialNumber?.code;
        }

        const startNumberToAdd = rowValue.startingSerialNumber?.code?.match(/\d+$/);
        const endNumberToAdd = Number(rowValue.endingSerialNumber?.match(/\d+$/));

        const serialPrefix = (serial: string) => serial?.match(/^(.*?)(?=\d+$)/)?.[0];
        const serialNumber = (serial: string) => Number(serial?.match(/\d+$/)?.[0]);

        const isSerialNumberInRange = (start: number, end: number, checkStart: number, checkEnd: number) =>
            checkStart <= end && checkEnd >= start;

        const isPrefixMatching = (prefix1: string, prefix2: string) => prefix1 === prefix2;

        const doesSerialNumberExist = (): boolean => {
            return this.miscellaneousIssueLines.value.some(
                line =>
                    Number(line.numberSession) === this._numberSession &&
                    line.stockDetails?.some(detail => {
                        if (
                            String(detail._id?.split('|')[0]) == String(this._stockId.value) &&
                            String(detail._id?.split('|')[2]) !== String(rowValue._id.split('|')[2])
                        ) {
                            const serialNumberPrefix1 = serialPrefix(rowValue.startingSerialNumber?.code);
                            const serialNumberPrefix2 = serialPrefix(String(detail?.serialNumber));

                            if (
                                serialNumberPrefix1 &&
                                serialNumberPrefix2 &&
                                isPrefixMatching(serialNumberPrefix1, serialNumberPrefix2)
                            ) {
                                const startingSerialNumber = serialNumber(String(detail?.serialNumber));
                                const endingSerialNumber = serialNumber(
                                    this._calculateEndingSerialNumber(
                                        String(detail?.serialNumber),
                                        Number(detail?.quantityInPackingUnit),
                                    ),
                                );

                                if (
                                    isSerialNumberInRange(
                                        Number(startNumberToAdd),
                                        endNumberToAdd,
                                        startingSerialNumber,
                                        endingSerialNumber,
                                    )
                                ) {
                                    return true;
                                }
                            }
                            return false;
                        }
                    }),
            );
        };

        if (doesSerialNumberExist()) {
            return ui.localize(
                '@sage/x3-stock/serial-number-range-overlap',
                `The serial numbers are overlapping. Enter another starting or ending serial number.`,
            );
        }

        const calculatedEndingSerialNumber = this._calculateEndingSerialNumber(
            rowValue.startingSerialNumber?.code ?? '',
            Number(rowValue.quantityInPackingUnit) * Number(rowValue.packingUnitToStockUnitConversionFactor),
        );

        if (rowValue.endingSerialNumber != calculatedEndingSerialNumber) {
            return ui.localize(
                '@sage/x3-stock/pages__mobile_miscellaneous_issue__same-amount-serial-numbers-in-the-range-to-match-quantity-to-issue',
                `Select the same amount of serial numbers in the range to match the quantity to issue.`,
            );
        }

        const countSerialNumber = await getCountSerialNumber(
            this,
            this._productSite?.product?.code ?? '',
            this.stockSite.value?.code,
            this._stockId.value,
            rowValue.startingSerialNumber?.code ?? '',
            rowValue.endingSerialNumber ?? '',
        );

        if (
            countSerialNumber !==
            Number(rowValue.quantityInPackingUnit) * Number(rowValue.packingUnitToStockUnitConversionFactor)
        ) {
            return ui.localize(
                '@sage/x3-stock/serial-number-not-sequential',
                `The serial numbers are not sequential. Check your entry.`,
            );
        }
        if (applyUpdate) {
            this._updateRecordValue(rowValue._id, rowValue);
        }
        return undefined;
    }

    private _getLotFilter(): string | any {
        // No lot management
        if (this._selectedStockManagementRules.outputLot === 'yes') return undefined;

        let lotFilter: any;

        // No use-by date control
        if (this._selectedStockManagementRules.outputLot === 'noUseByDateControl') {
            lotFilter = {
                product: { code: this.productSelected.value?.code },
                _or: [{ useByDate: null }, { useByDate: { _gte: this.effectiveDate.value } }],
            };
        }
        // No expiration date control
        else {
            lotFilter = {
                product: { code: this.productSelected.value?.code },
                _or: [{ expirationDate: null }, { expirationDate: { _gte: this.effectiveDate.value } }],
            };
        }
        return lotFilter;
    }

    private async controlVersions(
        version: ExtractEdges<MajorVersionStatus>,
        activeVersion: string,
        versionStopped: string,
        prototypeVersion: string,
        effectiveDate: string,
    ): Promise<boolean> {
        if (!version || !activeVersion || !versionStopped || !prototypeVersion) {
            throw new Error('Invalid arguments');
        }

        const { status, isOnHold, isSubjectToException, exceptionDate } = version;

        switch (status) {
            case 'prototypeVersion':
                return prototypeVersion !== 'no' && !(prototypeVersion === 'yesExceptOnHold' && isOnHold);
            case 'activeVersion':
                return activeVersion !== 'no' && !(activeVersion === 'yesExceptOnHold' && isOnHold);
            case 'stoppedVersion':
                return (
                    versionStopped !== 'no' &&
                    !(versionStopped === 'noUnlessException' && !isSubjectToException) &&
                    !(
                        versionStopped === 'noUnlessException' &&
                        isSubjectToException &&
                        exceptionDate &&
                        effectiveDate > exceptionDate
                    )
                );
            case 'toActivateVersion':
                return false;
            default:
                return true;
        }
    }
}
