import { AutomaticJournal } from '@sage/x3-finance-data-api';
import {
    Container,
    MajorVersionStatus,
    Product,
    ProductInput,
    ProductSite,
    ProductVersion,
    UnitOfMeasure,
} from '@sage/x3-master-data-api';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import {
    GraphApi,
    MiscellaneousReceiptInput,
    MiscellaneousReceiptLineBinding,
    MiscellaneousReceiptLineInput,
    StockEntryTransaction,
} from '@sage/x3-stock-api';
import {
    LicensePlateNumber,
    Location,
    Lot,
    LotsSites,
    StockJournalBinding,
    StockManagementRules,
    StockStatus,
    Warehouse,
} from '@sage/x3-stock-data-api';
import { other } from '@sage/x3-stock-data/build/lib/menu-items/other';
import { MiscellaneousTable } from '@sage/x3-system-api';
import { Site } from '@sage/x3-system-api-partial';
import { getRegExp } from '@sage/x3-system/lib/shared-functions/pat-converter';
import { ExtractEdges, extractEdges, ExtractEdgesPartial, Filter } from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import { MAX_INT_32 } from '@sage/xtrem-shared';
import * as ui from '@sage/xtrem-ui';
import { controlLot, controlLotReceipt, fieldData, validate } from '../client-functions/control';
import { expirationDateDefaultValue, useByDateDefaultValue } from '../client-functions/defaultValue';
import { NotifyAndWait } from '../client-functions/display';
import { getProductSite } from '../client-functions/get-product-site';
import { getUnitNumberOfDecimalList } from '../client-functions/get-unit-number-decimals';
import { findDefaultLocation, findStockManagementRules } from '../client-functions/stock-management-rules';

// Key to use with Composite Data for this application
export const mobileApplicationKey = 'MobileMiscellaneousReceiptKey';

/**
 * Composite data management for this application.
 * @enum MiscellaneousReceiptDictionaryFieldKey
 *
 * This enum provides per field the associated access key in the data dictionary.
 * It must be maintained in accordance with the fields of this dictionary.
 * Never translate..
 */

enum MiscellaneousReceiptDictionaryFieldKey {
    expirationDate = 'expirationDate',
    identifier1 = 'identifier1',
    identifier2 = 'identifier2',
    licensePlateNumber = 'licensePlateNumber',
    location = 'location',
    lot = 'lot',
    lotCustomField1 = 'lotCustomField1',
    lotCustomField2 = 'lotCustomField2',
    lotCustomField3 = 'lotCustomField3',
    lotCustomField4 = 'lotCustomField4',
    lotSidebar = 'lotSidebar',
    majorVersion = 'majorVersion',
    minorVersion = 'minorVersion',
    packingUnit = 'packingUnit',
    potency = 'potency',
    product = 'product',
    quantityInPackingUnit = 'quantityInPackingUnit',
    serialNumber = 'serialNumber',
    status = 'status',
    sublot = 'sublot',
    supplierLot = 'supplierLot',
    useByDate = 'useByDate',
}

export type DeepPartial<T> = T extends Object ? { [K in keyof T]?: DeepPartial<T[K]> } : T;
type PartialStockEntryTransaction = DeepPartial<StockEntryTransaction>;

export type inputs = {
    miscellaneousReceipt: MiscellaneousReceiptInput & {
        id: string;
    };
    username: string;
    started: boolean;
    selectedTransaction: PartialStockEntryTransaction;
    selectedProduct?: ProductInput;
    destination?: string;
};

export type packingUnit = {
    node: {
        packingUnit: {
            code: string;
            numberOfDecimals: number;
            _id: string;
        };
        packingUnitToStockUnitConversionFactor: string;
        isPackingFactorEntryAllowed: boolean;
        _id: string;
    };
};
@ui.decorators.page<MobileMiscellaneousReceipt>({
    title: 'Miscellaneous receipt',
    module: 'xtrem-x3-stock',
    mode: 'default',
    menuItem: other,
    priority: 100,
    isTitleHidden: true,
    authorizationCode: 'CWSSMR',
    node: '@sage/x3-stock/MiscellaneousReceipt',
    access: { node: '@sage/x3-stock/MiscellaneousReceipt' },
    //    skipDirtyCheck: true,
    navigationPanel: undefined,
    async onLoad() {
        await this._init();
    },
    /*    onDirtyStateUpdated(isDirty: boolean) {
        const isNotMiscLinesCreated = this.miscellaneousReceiptLines.value.length === 0;
        this.createButton.isDisabled = isNotMiscLinesCreated;
    }, */
    businessActions() {
        return [this.createButton, this.addButton];
    },
    headerCard() {
        if (!this.stockHeaderSection.isHidden) {
            return {
                title: this.productHeader,
                line2: this.localizedDescription1Header,
            };
        } else {
            return undefined;
        }
    },
})
export class MobileMiscellaneousReceipt extends ui.Page<GraphApi> {
    private _notifier = new NotifyAndWait(this);
    private _transactions: PartialStockEntryTransaction[];
    private _numberOfDecimalList: ExtractEdgesPartial<UnitOfMeasure>[];
    private _isLocationPreloaded: boolean;
    private _packingUnits: packingUnit[];
    private _productSite: ProductSite;
    private _miscellaneousReceiptLines: MiscellaneousReceiptLineInput[];
    private _selectedTransaction: PartialStockEntryTransaction;
    private _selectedLocation: {
        warehouse: Pick<Location['warehouse'], 'code' | '_id'>;
    };
    private _selectedStockManagementRules: StockManagementRules;
    private _newDetail: boolean = false;
    private _oldSupplierLotValue: string;
    private _majorVersion: string | null;
    private _lots: ExtractEdges<Lot>[];

    private async _fieldsData(validateField: boolean): Promise<fieldData[]> {
        return [
            {
                fieldIsHidden: this.packingUnit.isHidden ?? false,
                fieldValue: this.packingUnit.value,
                fieldName: 'packingUnit',
                validate: validateField ? await this.packingUnit.validate() : undefined,
            },
            {
                fieldIsHidden: this.quantityInPackingUnit.isHidden ?? false,
                fieldValue: this.quantityInPackingUnit.value,
                fieldName: 'quantityInPackingUnit',
                validate: validateField ? await this.quantityInPackingUnit.validate() : undefined,
            },
            {
                fieldIsHidden: this.packingUnitToStockUnitConversionFactorDestination.isHidden ?? false,
                fieldValue: this.packingUnitToStockUnitConversionFactorDestination.value,
                fieldName: 'packingUnitToStockUnitConversionFactorDestination',
                validate: validateField
                    ? await this.packingUnitToStockUnitConversionFactorDestination.validate()
                    : undefined,
            },
            {
                fieldIsHidden: this.status.isHidden ?? false,
                fieldValue: this.status.value,
                fieldName: 'status',
                validate: validateField ? await this.status.validate() : undefined,
            },
            {
                fieldIsHidden: this.licensePlateNumber.isHidden ?? false,
                fieldValue: this.licensePlateNumber.value,
                fieldName: 'licensePlateNumber',
                validate: validateField ? await this.licensePlateNumber.validate() : undefined,
            },
            {
                fieldIsHidden: this.containerDestination.isHidden ?? false,
                fieldValue: this.containerDestination.value,
                fieldName: 'containerDestination',
                validate: validateField ? await this.containerDestination.validate() : undefined,
            },
            {
                fieldIsHidden: this.location.isHidden ?? false,
                fieldValue: this.location.value,
                fieldName: 'location',
                validate: validateField ? await this.location.validate() : undefined,
            },
            {
                fieldIsHidden: this.serialNumber.isHidden ?? false,
                fieldValue: this.serialNumber.value,
                fieldName: 'serialNumber',
                validate: validateField ? await this.serialNumber.validate() : undefined,
            },
            {
                fieldIsHidden: this.supplierLot.isHidden ?? false,
                fieldValue: this.supplierLot.value,
                fieldName: 'supplierLot',
                validate: validateField ? await this.supplierLot.validate() : undefined,
            },
            {
                fieldIsHidden: this.lot.isHidden ?? false,
                fieldValue: this.lot.value,
                fieldName: 'lot',
                validate: validateField ? await this.lot.validate() : undefined,
            },
            {
                fieldIsHidden: this.sublot.isHidden ?? false,
                fieldValue: this.sublot.value,
                fieldName: 'sublot',
                validate: validateField ? await this.sublot.validate() : undefined,
            },
            {
                fieldIsHidden: this.majorVersion.isHidden ?? false,
                fieldValue: this.majorVersion.value,
                fieldName: 'majorVersion',
                validate: validateField ? await this.majorVersion.validate() : undefined,
            },
            {
                fieldIsHidden: this.minorVersion.isHidden ?? false,
                fieldValue: this.minorVersion.value,
                fieldName: 'minorVersion',
                validate: validateField ? await this.minorVersion.validate() : undefined,
            },
            {
                fieldIsHidden: this.potencyDestination.isHidden ?? false,
                fieldValue: this.potencyDestination.value,
                fieldName: 'potencyDestination',
                validate: validateField ? await this.potencyDestination.validate() : undefined,
            },

            {
                fieldIsHidden: this.expirationDate.isHidden ?? false,
                fieldValue: this.expirationDate.value,
                fieldName: 'expirationDate',
                validate: validateField ? await this.expirationDate.validate() : undefined,
            },
            {
                fieldIsHidden: this.lotCustomField1.isHidden ?? false,
                fieldValue: this.lotCustomField1.value,
                fieldName: 'lotCustomField1',
                validate: validateField ? await this.lotCustomField1.validate() : undefined,
            },
            {
                fieldIsHidden: this.lotCustomField2.isHidden ?? false,
                fieldValue: this.lotCustomField2.value,
                fieldName: 'lotCustomField2',
                validate: validateField ? await this.lotCustomField2.validate() : undefined,
            },
            {
                fieldIsHidden: this.lotCustomField3.isHidden ?? false,
                fieldValue: this.lotCustomField3.value,
                fieldName: 'lotCustomField3',
                validate: validateField ? await this.lotCustomField3.validate() : undefined,
            },
            {
                fieldIsHidden: this.lotCustomField4.isHidden ?? false,
                fieldValue: this.lotCustomField4.value,
                fieldName: 'lotCustomField4',
                validate: validateField ? await this.lotCustomField4.validate() : undefined,
            },
            {
                fieldIsHidden: this.identifier1.isHidden ?? false,
                fieldValue: this.identifier1.value,
                fieldName: 'identifier1',
                validate: validateField ? await this.identifier1.validate() : undefined,
            },
            {
                fieldIsHidden: this.identifier2.isHidden ?? false,
                fieldValue: this.identifier2.value,
                fieldName: 'identifier2',
                validate: validateField ? await this.identifier2.validate() : undefined,
            },
            {
                fieldIsHidden: this.useByDate.isHidden ?? false,
                fieldValue: this.useByDate.value,
                fieldName: 'useByDate',
                validate: validateField ? await this.useByDate.validate() : undefined,
            },
        ];
    }

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, Site>({
        node: '@sage/x3-system/Site',
        valueField: 'code',
        isHidden: true,
        bind: 'stockSite',
        helperTextField: 'name',
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: '_id',
            }),
        ],
    })
    stockSite: ui.fields.Reference<Site, MobileMiscellaneousReceipt>;

    /*
     *
     *  Page Actions
     *
     */
    @ui.decorators.pageAction<MobileMiscellaneousReceipt>({
        title: 'Create',
        buttonType: 'primary',
        isDisabled: true,
        shortcut: ['f2'],
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-stock/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-stock/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);
                message = `**${ui.localize(
                    '@sage/x3-stock/pages__mobile_stock_change_by_identifier__notification__creation_error',
                    'An error has occurred',
                )}**\n\n`;
                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-stock/pages_creation_error_connexion_webservice_contact_administrator',
                    'An error has occurred (connection or webservice error). Please contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
        async onClick() {
            if (this.createButton.isHidden) {
                return;
            }
            if (this.miscellaneousReceiptLines.value?.length > 0) {
                this.$.loader.isHidden = false;
                await this.$standardSaveAction.execute(false);
                this.$.loader.isHidden = true;
                await this.$.sound.success();
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-stock/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-stock/pages__mobile_miscellaneous_receipt__notification__document_created',
                        'Document no. {{documentId}} created.',
                        { documentId: this.id.value },
                    ),
                    {
                        fullScreen: true,
                        acceptButton: { text: ui.localize('@sage/x3-stock/button-accept-ok', 'OK') },
                    },
                );
                await this.$.router.emptyPage();
            } else {
                this._notifier.show(
                    ui.localize(
                        '@sage/x3-stock/pages__miscellaneous_receipt__notification__no_products_error',
                        `Enter at least one product.`,
                    ),
                    'error',
                );
            }
        },
    })
    createButton: ui.PageAction;

    @ui.decorators.pageAction<MobileMiscellaneousReceipt>({
        title: 'Add',
        buttonType: 'primary',
        shortcut: ['f3'],
        isHidden: true,
        async onClick() {
            if (this.addButton.isHidden) {
                return;
            }
            if (!validate(this, await this.$.page.isValid, await this._fieldsData(true))) {
                return;
            }
            if (await this._isZeroQuantityValue()) {
                return;
            }
            if (await this._isEmptyLpnAndContainerValues()) {
                return;
            }
            if (await this._isLotProvided()) {
                return;
            }
            this._addMiscLineAndStockDetails();
            this.quantityInPackingUnit.value = null;
            this.transaction.isDisabled = true;
            this.miscellaneousReceiptLinesBlock.isHidden = false;
            this.product.value = null;
            this.createButton.isHidden = false;
            this.createButton.isDisabled = this.miscellaneousReceiptLines.value.length === 0;
            this.addButton.isHidden = true;
            await this._initFieldsDestination();
        },
    })
    addButton: ui.PageAction;

    @ui.decorators.section<MobileMiscellaneousReceipt>({
        isTitleHidden: true,
    })
    mainSection: ui.containers.Section;

    /* step 01*/
    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'small',
        isTitleHidden: true,
    })
    transactionBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'small',
        isTitleHidden: true,
    })
    customizationMiscellaneousReceiptHeaderBlock: ui.containers.Block;

    @ui.decorators.dateField<MobileMiscellaneousReceipt>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Receipt date',
        isMandatory: true,
        width: 'small',
        placeholder: 'Enter...',
        maxDate: DateValue.today().toString(),
    })
    effectiveDate: ui.fields.Date<MobileMiscellaneousReceipt>;

    @ui.decorators.dropdownListField<MobileMiscellaneousReceipt>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Transaction',
        isTransientInput: true,
        onChange() {
            if (this.transaction.value) {
                const transaction = this._transactions.find(trs => trs.code === this.transaction.value);
                if (transaction) {
                    this._selectedTransaction = transaction;
                    this.stockMovementCode.value = this._selectedTransaction.stockMovementCode
                        ? {
                            code: this._selectedTransaction.stockMovementCode.code,
                            _id: this._selectedTransaction.stockMovementCode._id,
                        }
                        : null;
                    this.stockMovementGroup.value = this._selectedTransaction.defaultStockMovementGroup
                        ? {
                            code: this._selectedTransaction.defaultStockMovementGroup.code,
                            _id: this._selectedTransaction.defaultStockMovementGroup._id,
                        }
                        : null;
                    this.stockAutomaticJournal.value = this._selectedTransaction.stockAutomaticJournal
                        ? {
                            code: this._selectedTransaction.stockAutomaticJournal.code,
                            _id: this._selectedTransaction.stockAutomaticJournal._id,
                        }
                        : null;
                }
                this.product.isDisabled = false;
            } else {
                this.product.isDisabled = true;
            }
        },
    })
    transaction: ui.fields.DropdownList<NonNullable<PartialStockEntryTransaction['code']>, MobileMiscellaneousReceipt>;

    @ui.decorators.selectField<MobileMiscellaneousReceipt>({
        isHidden: true,
    })
    entryType: ui.fields.Select<'miscellaneousReceipt', MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        isHidden: true,
    })
    id: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        isHidden: true,
        isTransientInput: true,
    })
    destination: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        canFilter: false,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 14,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementCode: ui.fields.Reference;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        canFilter: false,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 9,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementGroup: ui.fields.Reference;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, AutomaticJournal>({
        parent() {
            return this.defaultBlock;
        },
        node: '@sage/x3-finance-data/AutomaticJournal',
        valueField: 'code',
        isHidden: true,
        canFilter: false,
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockAutomaticJournal: ui.fields.Reference<AutomaticJournal, MobileMiscellaneousReceipt>;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    productBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, Product>({
        parent() {
            return this.productBlock;
        },
        title: 'Product',
        node: '@sage/x3-master-data/Product',
        valueField: 'code',
        helperTextField: 'upc',
        placeholder: 'Scan or select...',
        isTransient: true,
        canFilter: false,
        isAutoSelectEnabled: true,
        isDisabled: true,
        isFullWidth: true,
        shouldSuggestionsIncludeColumns: true,
        filter() {
            return {
                productStatus: { _ne: 'notUsable' },
                _and: [
                    { stockManagementMode: { _ne: 'notManaged' } },
                    {
                        productSites: {
                            _atLeast: 1,
                            ...(this.stockSite.value && { stockSite: { code: this.stockSite.value.code } }),
                            stockManagementMode: { _ne: 'notManaged' },
                        },
                    },
                ],
            };
        },
        async onChange() {
            if (await this.product.value?.code) {
                const response = await this.$.graph
                    .node('@sage/x3-master-data/ProductSite')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                isBeingCounted: true,
                            },
                            {
                                filter: {
                                    product: {
                                        code: this.product.value?.code,
                                    },
                                    ...(this.stockSite.value && {
                                        stockSite: {
                                            code: this.stockSite.value.code,
                                        },
                                    }),
                                },
                            },
                        ),
                    )
                    .execute();

                if (response.edges[0].node.isBeingCounted === true) {
                    if (
                        await dialogConfirmation(
                            this,
                            'warn',
                            ui.localize('@sage/x3-stock/dialog-warning-title', 'Warning'),
                            ui.localize(
                                '@sage/x3-stock/product-blocked-by-count-continue',
                                'Product blocked by count. Do you want to continue?',
                            ),
                            {
                                acceptButton: {
                                    text: ui.localize('@sage/x3-stock/button-accept-yes', 'Yes'),
                                },
                                cancelButton: {
                                    text: ui.localize('@sage/x3-stock/button-cancel-no', 'No'),
                                },
                            },
                        )
                    ) {
                        await this._initFieldsDestination();
                        this.createButton.isHidden = true;
                        this.addButton.isHidden = false;
                    } else {
                        this.product.value = null;
                        this.createButton.isHidden = false;
                        this.createButton.isDisabled = this.miscellaneousReceiptLines.value.length === 0;
                        this.addButton.isHidden = true;
                        await this._initFieldsDestination();
                        this.product.focus();
                    }
                } else {
                    await this._initFieldsDestination();
                    this.createButton.isHidden = true;
                    this.addButton.isHidden = false;
                }
            } else {
                await this._initFieldsDestination();
                this.createButton.isHidden = false;
                this.createButton.isDisabled = this.miscellaneousReceiptLines.value.length === 0;
                this.addButton.isHidden = true;
            }
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: 'localizedDescription1',
            }),
            ui.nestedFields.text({
                bind: 'upc',
                prefix: 'UPC:',
            }),
        ],
    })
    product: ui.fields.Reference<Product, MobileMiscellaneousReceipt>;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    defaultBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationMiscellaneousReceiptLinesBlock: ui.containers.Block;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customizationMiscellaneousReceiptDetailsBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, Warehouse>({
        parent() {
            return this.defaultBlock;
        },
        node: '@sage/x3-stock-data/Warehouse',
        valueField: 'code',
        placeholder: 'Scan or select...',
        isHidden: true,
        canFilter: false,
        isTransient: true,
        filter() {
            return this.stockSite.value
                ? {
                    stockSite: { _id: { _eq: this.stockSite.value._id } },
                }
                : undefined;
        },
    })
    warehouseDestination: ui.fields.Reference<Warehouse, MobileMiscellaneousReceipt>;

    @ui.decorators.dropdownListField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Unit',
        width: 'small',
        options: ['UN'],
        placeholder: 'Scan or select...',
        isTransient: true,
        isMandatory: true,
        async onChange() {
            if (!this.packingUnit.value) return;
            const selectedValue = this.packingUnit.value;
            const packingUnitIndex = this._packingUnits
                .map(packingUnit => packingUnit.node.packingUnit.code)
                .indexOf(selectedValue);
            if (packingUnitIndex !== -1) {
                const selectedUnit = this._packingUnits[packingUnitIndex].node;
                this.packingUnitToStockUnitConversionFactorDestination.value = Number(
                    selectedUnit.packingUnitToStockUnitConversionFactor,
                );
                this.packingUnitToStockUnitConversionFactorDestination.isDisabled =
                    !selectedUnit.isPackingFactorEntryAllowed;
                this.quantityInPackingUnit.scale = selectedUnit.packingUnit.numberOfDecimals;
            } else {
                this.packingUnitToStockUnitConversionFactorDestination.value = 1;
                this.packingUnitToStockUnitConversionFactorDestination.isDisabled = true;
                if (this.packingUnit.value) this._getNumberOfDecimals();
            }
        },
    })
    packingUnit: ui.fields.DropdownList<string, MobileMiscellaneousReceipt>;

    @ui.decorators.numericField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Quantity',
        placeholder: 'Enter...',
        width: 'small',
        validation: /[0-9]{1,}/,
        min: 0,
        isNotZero: true,
        isTransient: true,
        isMandatory: true,
        scale() {
            if (this.packingUnit.value && this._packingUnits) {
                const selectedUnit = this._packingUnits.find(
                    item => item.node.packingUnit.code === this.packingUnit.value,
                )?.node;
                if (selectedUnit) {
                    return selectedUnit.packingUnit.numberOfDecimals;
                } else {
                    return 0;
                }
            } else {
                return 0;
            }
        },
    })
    quantityInPackingUnit: ui.fields.Numeric<MobileMiscellaneousReceipt>;

    @ui.decorators.numericField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Conversion factor',
        isDisabled: true,
        isTransient: true,
        isMandatory: true,
        placeholder: 'Enter...',
    })
    packingUnitToStockUnitConversionFactorDestination: ui.fields.Numeric<MobileMiscellaneousReceipt>;

    @ui.decorators.dropdownListField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Status',
        options: ['A'],
        placeholder: 'Scan or select...',
        isTransient: true,
        isMandatory: true,
        onChange() { },
    })
    status: ui.fields.DropdownList<string, MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, Container>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Container',
        node: '@sage/x3-master-data/Container',
        valueField: 'code',
        isAutoSelectEnabled: true,
        placeholder: 'Scan or select...',
        width: 'large',
        isTransient: true,
        canFilter: false,
        filter() {
            return {
                isInternal: { _eq: true },
                isActive: { _eq: true },
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.text({
                bind: 'containerType',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: 'description',
                title: 'Description',
            }),
        ],
    })
    containerDestination: ui.fields.Reference<Container, MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, LicensePlateNumber>({
        parent() {
            return this.defaultBlock;
        },
        title: 'License Plate Number',
        node: '@sage/x3-stock-data/LicensePlateNumber',
        valueField: 'code',
        isAutoSelectEnabled: true,
        placeholder: 'Scan or select...',
        helperTextField: { container: { code: true } },
        width: 'large',
        canFilter: false,
        isTransient: true,
        filter() {
            let licensePlateNumberFilter: Filter<any> = {
                ...(this.stockSite.value && { stockSite: { code: this.stockSite.value.code } }),
                isActive: { _eq: true },
                _or: [
                    {
                        isSingleProduct: { _eq: true },
                        stock: { _atLeast: 1, product: { product: { code: this.product.value } } },
                    },
                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                    { isSingleProduct: { _in: [false, null] } },
                ],
            };
            if (this.containerDestination.value) {
                licensePlateNumberFilter = {
                    ...licensePlateNumberFilter,
                    container: { code: this.containerDestination.value.code },
                };
            }
            return licensePlateNumberFilter;
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Location',
                bind: 'location',
                valueField: 'code',
                title: 'Location',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Container',
                bind: 'container',
                valueField: 'code',
                title: 'Container',
            }),
            ui.nestedFields.text({
                bind: 'status',
                title: 'Status',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isActive',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isSingleProduct',
                isHidden: true,
            }),
        ],
        async onChange() {
            if (this.licensePlateNumber.value?.code) {
                this._manageLicensePlateNumberLocation();
                if (this.licensePlateNumber.value?.container) {
                    this.containerDestination.value = this.licensePlateNumber.value.container;
                    this.containerDestination.isDisabled = true;
                } else {
                    this.containerDestination.isDisabled = false;
                }
            } else {
                this.location.isHidden = false;
                this.location.isDisabled = false;
                this.location.value = null;
                this.dummyLocation.isHidden = true;
                this.dummyLocation.value = null;
                this.containerDestination.isDisabled = this._newDetail;
            }
            if (this.location.value && this.location.value.code) this._getWarehouseFromLocation();
            await this.$.commitValueAndPropertyChanges();
            this.licensePlateNumber.getNextField(true)?.focus();
        },
    })
    licensePlateNumber: ui.fields.Reference<LicensePlateNumber, MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Suggested location',
        isReadOnly: true,
        width: 'large',
        isTransient: true,
    })
    suggestedLocation: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, Location>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Location',
        node: '@sage/x3-stock-data/Location',
        valueField: 'code',
        placeholder: 'Scan or select...',
        helperTextField: 'type',
        isAutoSelectEnabled: true,
        width: 'large',
        canFilter: false,
        isTransient: true,
        isMandatory() {
            return this._productSite?.isLocationManaged;
        },
        filter() {
            return {
                ...(this.stockSite.value && { stockSite: { code: this.stockSite.value.code } }),
                category: { _nin: ['subcontract', 'customer'] },
            };
        },
        async onChange() {
            if (this.location.value) this._getWarehouseFromLocation();
            await this.$.commitValueAndPropertyChanges();
            if (this.location.value) this.location.getNextField(true)?.focus();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Warehouse',
                bind: 'warehouse',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
    })
    location: ui.fields.Reference<Location, MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Location',
        isHidden: true,
        isDisabled: true,
        isTransient: true,
        width: 'large',
    })
    dummyLocation: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Serial number',
        width: 'large',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        isTransient: true,
        isMandatory() {
            return !(this.serialNumber.isHidden || !!this._productSite.product.serialSequenceNumber);
        },
    })
    serialNumber: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        title: 'Supplier lot',
        width: 'large',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        isTransient: true,
        async onChange() {
            if (this.supplierLot.value) {
                this.supplierLot.value = this.supplierLot.value.toUpperCase();
            }
            if (
                this.supplierLot.value &&
                this._selectedStockManagementRules.lotByDefault === 'supplierLot' &&
                this._selectedStockManagementRules.lotEntry === 'newLot'
            ) {
                if (
                    (await controlLotReceipt(
                        this,
                        this.supplierLot.value,
                        this._productSite.product.code,
                        '19',
                        this.stockSite.value?.code ?? null,
                    )) === false
                ) {
                    this.supplierLot.value = null;
                    this.supplierLot.focus();
                }
            }

            const rules = this._selectedStockManagementRules;
            const shouldUseSupplierLot =
                rules?.lotByDefault === 'supplierLot' &&
                (!this.lot.value || this.lot.value === this._oldSupplierLotValue);
            if (shouldUseSupplierLot) {
                this.lot.value = this.supplierLot.value;
                await this._getLotValues();
            }
            await this.$.commitValueAndPropertyChanges();
        },
        onInputValueChange(inputValue) {
            this._oldSupplierLotValue = this.supplierLot.value ?? '';
        },
    })
    supplierLot: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.filterSelectField<MobileMiscellaneousReceipt, LotsSites>({
        node: '@sage/x3-stock-data/LotsSites',
        title: 'Lot',
        valueField: 'lot',
        isNewEnabled: true,
        placeholder: 'Scan or select...',
        canFilter: false,
        parent() {
            return this.defaultBlock;
        },
        validation: /^$|^[^|]+$/,
        isTransient: true,
        filter() {
            return {
                product: { code: this.product.value?.code },
                ...(this.stockSite.value && { storageSite: { code: this.stockSite.value?.code } }),
                stock: { _atLeast: 1 },
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
        async onChange() {
            if (this.lot.value) {
                this.lot.value = this.lot.value.toUpperCase();
            }
            if (
                this.lot.value &&
                this._selectedStockManagementRules.lotEntry === 'newLot' &&
                (await controlLotReceipt(
                    this,
                    this.lot.value,
                    this._productSite.product.code,
                    '19',
                    this.stockSite.value?.code ?? null,
                )) === false
            ) {
                this.lot.value = null;
                this.lot.focus();
            }
            await this._getLotValues();
        },
    })
    lot: ui.fields.FilterSelect<LotsSites, MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'Sublot',
        validation: /^$|^[^|]+$/,
        isTransient: true,
        isMandatory() {
            return this._productSite?.product?.lotManagementMode === 'lotAndSublot';
        },
        /* sync onInputValueChange(this, rawData: string): Promise<void> {
            this.expirationDate.isDisabled = true;
            this.useByDate.isDisabled = true;
        }, */
        async onChange() {
            if (this.sublot.value) {
                this.sublot.value = this.sublot.value.toUpperCase();
            }
            await this._getLotValues();
            this._getNextField(this.sublot)?.focus();
        },
    })
    sublot: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, MajorVersionStatus>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan or select...',
        title: 'Major version',
        node: '@sage/x3-master-data/MajorVersionStatus',
        valueField: 'code',
        isAutoSelectEnabled: true,
        canFilter: false,
        isTransient: true,
        filter() {
            return {
                product: { code: this.product.value?.code },
                _or: [{ status: 'prototypeVersion' }, { status: 'activeVersion' }, { status: 'stoppedVersion' }],
            };
        },
        async onChange() {
            if (this.minorVersion.isHidden) {
                if (this.majorVersion.value) {
                    await this.$.commitValueAndPropertyChanges();
                    this._getNextField(this.majorVersion)?.focus();
                }
            } else if ((this.minorVersion.isDisabled = !this.majorVersion.value)) {
                this.minorVersion.value = null;
            } else if (this.majorVersion.value?.code) {
                const minorVersions = await this._getMinorVersions(this.majorVersion.value.code);
                this.minorVersion.value = {
                    _id: minorVersions[minorVersions.length - 1]._id,
                    minorVersion: minorVersions[minorVersions.length - 1].minorVersion,
                };
                await this.$.commitValueAndPropertyChanges();
                this._getNextField(this.majorVersion)?.focus();
            }
        },
        isMandatory() {
            return (
                this._productSite?.product?.lotManagementMode !== 'notManaged' &&
                this._productSite?.product?.stockVersionMode &&
                this._productSite?.product?.stockVersionMode !== 'no'
            );
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'code',
                title: 'Major Version',
            }),
            ui.nestedFields.text({
                bind: 'status',
                title: 'Status',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
        ],
    })
    majorVersion: ui.fields.Reference<MajorVersionStatus, MobileMiscellaneousReceipt>;

    @ui.decorators.referenceField<MobileMiscellaneousReceipt, ProductVersion>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan or select...',
        isHidden: true,
        title: 'Minor version',
        node: '@sage/x3-master-data/ProductVersion',
        valueField: 'minorVersion',
        isAutoSelectEnabled: true,
        canFilter: false,
        isTransient: true,
        filter() {
            if (!this.majorVersion.value) {
                return;
            }

            return {
                ...(this.product.value && { product: { code: this.product.value.code } }),
                majorVersion: this.majorVersion.value.code,
                type: 'stock',
                useStatus: 'availableToUse',
            };
        },
        isMandatory() {
            return (
                this._productSite?.product?.lotManagementMode !== 'notManaged' &&
                this._productSite?.product?.stockVersionMode === 'majorAndMinor'
            );
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'minorVersion',
                title: 'Minor version',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'majorVersion',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'useStatus',
                isHidden: true,
            }),
        ],
    })
    minorVersion: ui.fields.Reference<ProductVersion, MobileMiscellaneousReceipt>;

    @ui.decorators.numericField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'potency %',
        isTransient: true,
        scale: 4,
        validation: /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/, // reg ex for any positive numbers (integers or decimals) excluding 0
        min: 0,
        max: 100,
        isMandatory() {
            return Boolean(
                this._productSite?.product?.lotManagementMode !== 'notManaged' &&
                this._selectedTransaction?.isLotPotencyAllowed &&
                this._productSite?.product?.stockManagementMode === 'potencyManaged',
            );
        },
    })
    potencyDestination: ui.fields.Numeric<MobileMiscellaneousReceipt>;

    @ui.decorators.dateField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Enter...',
        title: 'Expiration date',
        isTransient: true,
        isHidden() {
            return this._productSite && this._selectedTransaction
                ? this._productSite.product.lotManagementMode === 'notManaged' ||
                this._productSite.product.expirationManagementMode === 'notManaged' ||
                (!this._selectedTransaction.isLotExpirationDateAllowed &&
                    (this._productSite.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                        this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                        this._productSite.product.expirationManagementMode === 'withoutRounding'))
                : true;
        },
        isMandatory() {
            return this._productSite
                ? this._productSite?.product.lotManagementMode !== 'notManaged' &&
                this._productSite?.product.expirationManagementMode === 'mandatoryEntry'
                : false;
        },
        onChange() {
            this.useByDate.value = useByDateDefaultValue(
                this.expirationDate.value,
                this.effectiveDate.value,
                this._productSite.product.expirationManagementMode,
                Number(this._productSite.product.useByDateCoefficient),
            );
            this.useByDate.maxDate = this.expirationDate.value ?? undefined;
        },
    })
    expirationDate: ui.fields.Date<MobileMiscellaneousReceipt>;

    @ui.decorators.dateField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Enter...',
        title: 'Use-by date',
        isTransient: true,
        isHidden() {
            return this._productSite && this._selectedTransaction
                ? this._productSite.product.lotManagementMode === 'notManaged' ||
                this._productSite.product.expirationManagementMode === 'notManaged' ||
                (!this._selectedTransaction.isLotExpirationDateAllowed &&
                    (this._productSite.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                        this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                        this._productSite.product.expirationManagementMode === 'withoutRounding'))
                : true;
        },
        isMandatory() {
            return this._productSite
                ? this._productSite.product.lotManagementMode !== 'notManaged' &&
                this._productSite.product.expirationManagementMode === 'mandatoryEntry'
                : false;
        },
        isDisabled() {
            return this._productSite
                ? this._productSite?.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                this._productSite?.product.expirationManagementMode === 'roundingMonthEnd' ||
                this._productSite?.product.expirationManagementMode === 'withoutRounding'
                : false;
        },
    })
    useByDate: ui.fields.Date<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom text 1',
        validation: /^$|^[^|]+$/,
        maxLength: 20,
        isTransient: true,
    })
    lotCustomField1: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom text 2',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
    })
    lotCustomField2: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.numericField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Enter...',
        title: 'Lot custom number',
        isTransient: true,
    })
    lotCustomField3: ui.fields.Numeric<MobileMiscellaneousReceipt>;

    @ui.decorators.dateField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Enter...',
        title: 'Lot custom date',
        isTransient: true,
    })
    lotCustomField4: ui.fields.Date<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'Identifier 1',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
    })
    identifier1: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        parent() {
            return this.defaultBlock;
        },
        placeholder: 'Scan...',
        title: 'Identifier 2',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
    })
    identifier2: ui.fields.Text<MobileMiscellaneousReceipt>;

    /* step 02*/

    /*
     *
     *  Technical fields step 02
     *
     */

    @ui.decorators.section<MobileMiscellaneousReceipt>({
        isTitleHidden: true,
        isHidden: true,
    })
    stockHeaderSection: ui.containers.Section;

    @ui.decorators.block<MobileMiscellaneousReceipt>({
        isTitleHidden: true,
        parent() {
            return this.stockHeaderSection;
        },
    })
    stockLineBlock: ui.containers.Block;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    productHeader: ui.fields.Text<MobileMiscellaneousReceipt>;

    @ui.decorators.textField<MobileMiscellaneousReceipt>({
        isTransient: true,
        isFullWidth: true,
        isTitleHidden: true,
    })
    localizedDescription1Header: ui.fields.Text<MobileMiscellaneousReceipt>;

    private async _initFieldsDestination() {
        this.warehouseDestination.value = null;
        this.quantityInPackingUnit.value = null;
        this.licensePlateNumber.value = null;
        this.location.value = null;
        this.dummyLocation.value = null;
        this.serialNumber.value = null;
        this.supplierLot.value = null;
        this.lot.value = null;
        this.sublot.value = null;
        this.majorVersion.value = null;
        this.minorVersion.value = null;
        this.potencyDestination.value = null;
        this.expirationDate.value = null;
        this.useByDate.value = null;
        this.lotCustomField1.value = null;
        this.lotCustomField2.value = null;
        this.lotCustomField3.value = null;
        this.lotCustomField4.value = null;
        this.identifier1.value = null;
        this.identifier2.value = null;
        if (this.product.value?.code && this.stockSite.value?.code) {
            this._productSite = await getProductSite(this, this.product.value.code, this.stockSite.value.code, '');
            this.defaultBlock.isHidden = false;
            this.customizationMiscellaneousReceiptHeaderBlock.isHidden =
                this.customizationMiscellaneousReceiptHeaderBlock.layout.$layout?.$items.length === 0;
            this.customizationMiscellaneousReceiptLinesBlock.isHidden =
                this.customizationMiscellaneousReceiptLinesBlock.layout.$layout?.$items.length === 0;
            this.customizationMiscellaneousReceiptDetailsBlock.isHidden =
                this.customizationMiscellaneousReceiptDetailsBlock.layout.$layout?.$items.length === 0;
            this.quantityInPackingUnit.isHidden = false;
            this._lotManagement();
            this._initPackingUnitFields();
            this.packingUnit.value = this._productSite.product.stockUnit.code;
            this.quantityInPackingUnit.scale = this._productSite.product.stockUnit.numberOfDecimals;
            this._initPackingUnitToStockUnitConversionFactorDestination();
            this._selectedStockManagementRules = await findStockManagementRules(
                this.stockSite.value?.code ?? '',
                this._productSite.product.productCategory.code,
                '1',
                this._selectedTransaction.stockMovementCode?.code ?? null,
                this,
            );
            this._miscellaneousFieldsManagement();
            this.status.isHidden = false;
            this.status.options = await this._getStockStatus();
            this.status.value = this._selectedStockManagementRules.defaultStatus;
            if (!this.suggestedLocation.isHidden) {
                this.suggestedLocation.value = await findDefaultLocation(
                    this._productSite,
                    this._selectedStockManagementRules,
                    this,
                );
                if (!this.suggestedLocation.value) this.suggestedLocation.isHidden = true;
                else if (this._isLocationPreloaded) {
                    this.location.value = (await this._getDefaultLocation(this.suggestedLocation.value)) ?? null;
                }
            }

            if (!this.expirationDate.isHidden) {
                this.expirationDate.value = expirationDateDefaultValue(
                    this._productSite.product.expirationManagementMode,
                    this._productSite.product.expirationLeadTime,
                    this._productSite.product.expirationTimeUnit,
                    this.effectiveDate.value ?? undefined,
                );
            } else {
                this.expirationDate.value = null;
            }

            if (!this.useByDate.isHidden) {
                this.useByDate.value = useByDateDefaultValue(
                    this.expirationDate.value,
                    this.effectiveDate.value,
                    this._productSite.product.expirationManagementMode,
                    Number(this._productSite.product.useByDateCoefficient),
                );
                this.useByDate.maxDate = this.expirationDate.value ?? undefined;
            } else {
                this.useByDate.value = null;
            }
            this.packingUnit.isReadOnly = (this.packingUnit.options ?? []).length <= 1;
        } else {
            this.defaultBlock.isHidden = true;
            this.customizationMiscellaneousReceiptHeaderBlock.isHidden =
                this.customizationMiscellaneousReceiptHeaderBlock.layout.$layout?.$items.length === 0;
            this.customizationMiscellaneousReceiptLinesBlock.isHidden = true;
            this.customizationMiscellaneousReceiptDetailsBlock.isHidden = true;
            this.quantityInPackingUnit.isHidden = true;
            this.lot.isHidden = true;
            this.sublot.isHidden = true;
            this.supplierLot.isHidden = true;
            this.lotCustomField1.isHidden = true;
            this.lotCustomField2.isHidden = true;
            this.lotCustomField3.isHidden = true;
            this.lotCustomField4.isHidden = true;
            this.potencyDestination.isHidden = true;
            // this.expirationDate.isHidden = true;
            // this.useByDate.isHidden = true;
            this.majorVersion.isHidden = true;
            this.minorVersion.isHidden = true;
            this.packingUnitToStockUnitConversionFactorDestination.isHidden = true;
            this.packingUnit.isHidden = true;
            this.location.isHidden = true;
            this.suggestedLocation.isHidden = true;
            this.serialNumber.isHidden = true;
            this.containerDestination.isHidden = true;
            this.licensePlateNumber.isHidden = true;
            this.identifier1.isHidden = true;
            this.identifier2.isHidden = true;
            this.status.isHidden = true;
        }
    }

    private _lotManagement() {
        const lotManagement = this._productSite.product.lotManagementMode;

        const lotNotManaged = lotManagement === 'notManaged';

        this.lot.isHidden = lotNotManaged;
        this.sublot.isHidden = lotNotManaged;
        //user lot area
        this.lotCustomField1.isHidden = lotNotManaged || !this._selectedTransaction.isLotCustomField1Allowed;
        this.lotCustomField2.isHidden = lotNotManaged || !this._selectedTransaction.isLotCustomField2Allowed;
        this.lotCustomField3.isHidden = lotNotManaged || !this._selectedTransaction.isLotCustomField3Allowed;
        this.lotCustomField4.isHidden = lotNotManaged || !this._selectedTransaction.isLotCustomField4Allowed;
        //Supplier Lot
        switch (this._selectedTransaction.supplierLot) {
            case 'entered': {
                this.supplierLot.isHidden = lotNotManaged || false;
                this.supplierLot.isDisabled = false;
                break;
            }
            case 'displayed': {
                this.supplierLot.isHidden = lotNotManaged || false;
                this.supplierLot.isDisabled = true;
                break;
            }
            default: {
                this.supplierLot.isHidden = lotNotManaged || true;
            }
        }

        //Potency management
        const potencyManagement = this._productSite.product.stockManagementMode;
        this.potencyDestination.isHidden =
            lotNotManaged || !this._selectedTransaction.isLotPotencyAllowed || potencyManagement !== 'potencyManaged';

        //Expiration Date
        /* const expirationManagement = this._productSite.product.expirationManagementMode;

        this.expirationDate.isHidden =
            lotNotManaged ||
            expirationManagement === 'notManaged' ||
            (!this._selectedTransaction.isLotExpirationDateAllowed &&
                expirationManagement !== 'manualEntry' &&
                expirationManagement !== 'mandatoryEntry');
        this.useByDate.isHidden = this.expirationDate.isHidden;
        //disable useByDate field
        if (['roundingBeginningMonth1', 'roundingMonthEnd'].includes(expirationManagement))
            this.useByDate.isDisabled = true; */

        //version
        const versionManagementMode = this._productSite.product.stockVersionMode;
        this.majorVersion.isHidden =
            lotNotManaged || !this._productSite.product.stockVersionMode || versionManagementMode === 'no';
        this.minorVersion.isHidden = lotNotManaged || versionManagementMode !== 'majorAndMinor';

        this.sublot.isHidden = lotManagement !== 'lotAndSublot';
    }

    private _initPackingUnitFields() {
        let productPackingList = extractEdges(this._productSite.product.packingUnits.query).filter(productPacking => {
            return !!productPacking.packingUnit?.code;
        });
        this._packingUnits = productPackingList.map(productPacking => {
            return { node: productPacking };
        });
        this._packingUnits.push({
            node: {
                packingUnit: {
                    code: this._productSite.product.stockUnit.code,
                    numberOfDecimals: this._productSite.product.stockUnit.numberOfDecimals,
                    _id: this._productSite.product.stockUnit._id,
                },
                packingUnitToStockUnitConversionFactor: '1',
                isPackingFactorEntryAllowed: false,
                _id: this._productSite.product.stockUnit._id,
            },
        });
        let productPackingUnitSelectValues = productPackingList.map(productPacking => {
            return productPacking.packingUnit.code;
        });

        this.packingUnitToStockUnitConversionFactorDestination.isHidden = false;
        this.packingUnit.isHidden = false;
        this.packingUnit.options = [this._productSite.product.stockUnit.code, ...productPackingUnitSelectValues];
    }

    private _miscellaneousFieldsManagement() {
        //location fields
        this.location.isHidden = !this._productSite.isLocationManaged;
        this.suggestedLocation.isHidden = this.location.isHidden;

        //serial number fields
        const serialNumberManagement = this._productSite.product.serialNumberManagementMode;
        this.serialNumber.isHidden = ['notManaged', 'issued'].includes(serialNumberManagement);
        //lot field: mandatory if no sequence number
        if (this._selectedStockManagementRules.lotEntry === 'no') {
            this.lot.isDisabled = true;
            if (this.sublot.isHidden === false) {
                this.sublot.value = '00001';
                this.sublot.isDisabled = true;
            }
        } else if (this._selectedStockManagementRules.lotEntry === 'newLot') {
            if (this.sublot.isHidden === false) this.sublot.value = '00001';
        } else if (
            this._selectedStockManagementRules.lotEntry === 'free' &&
            this._selectedStockManagementRules.lotByDefault === 'documentNumber'
        ) {
            if (this.sublot.isHidden === false) this.sublot.value = '00001';
        }

        this.lot.isMandatory =
            ['lotAndSublot', 'mandatoryLot'].includes(this._productSite?.product?.lotManagementMode) &&
            !this._productSite?.product?.lotSequenceNumber &&
            // this._selectedStockManagementRules?.lotEntry !== 'no' &&
            this._selectedStockManagementRules?.lotByDefault !== 'documentNumber';

        //containerDestination field
        this.containerDestination.isHidden =
            this._selectedStockManagementRules.licensePlateNumberEntry !== 'mandatory' ||
            !this._productSite.isLicensePlateNumberManaged;
        //license plate number fields
        this.licensePlateNumber.isHidden = !this._productSite.isLicensePlateNumberManaged;
        if (!this.containerDestination.isHidden) {
            this.containerDestination.value = this._productSite.defaultInternalContainer;
        }

        //identifier 1 and identifier 2
        switch (this._selectedTransaction.identifier1Detail) {
            case 'entered': {
                this.identifier1.isHidden = false;
                this.identifier1.isDisabled = false;
                break;
            }
            case 'displayed': {
                this.identifier1.isHidden = false;
                this.identifier1.isDisabled = true;
                break;
            }
            default: {
                this.identifier1.isHidden = true;
            }
        }

        switch (this._selectedTransaction.identifier2Detail) {
            case 'entered': {
                this.identifier2.isHidden = false;
                this.identifier2.isDisabled = false;
                break;
            }
            case 'displayed': {
                this.identifier2.isHidden = false;
                this.identifier2.isDisabled = true;
                break;
            }
            default: {
                this.identifier2.isHidden = true;
            }
        }
    }

    private async _getStockStatus(): Promise<string[]> {
        const selectedStatus: { _regex: string }[] = [];
        this._selectedStockManagementRules.authorizedSubstatus.split(',').forEach(function (status) {
            selectedStatus.push({ _regex: getRegExp(status).source });
        });
        const response = await this.$.graph
            .node('@sage/x3-stock-data/StockStatus')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        code: true,
                    },
                    {
                        filter: {
                            code: { _or: selectedStatus },
                        },
                    },
                ),
            )
            .execute();

        if (!response.edges || response.edges.length === 0) {
            throw new Error(
                ui.localize(
                    '@sage/x3-stock/pages__miscellaneous_receipt_details__notification__invalid_stock_status_error',
                    'No stock status',
                ),
            );
        }

        // transform Stock status response into a string array
        return response.edges.map((stockStatus: any) => {
            return stockStatus.node.code;
        });
    }

    private async _getNumberOfDecimals() {
        try {
            const response = await this.$.graph
                .node('@sage/x3-master-data/UnitOfMeasure')
                .read(
                    {
                        _id: true,
                        numberOfDecimals: true,
                    },
                    `${this.packingUnit.value}`,
                )
                .execute();

            this.quantityInPackingUnit.scale = response.numberOfDecimals;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/error-loading-unit', 'Error loading unit'),
                String(e),
            );
        }
    }

    private async _getWarehouseFromLocation() {
        if (!this.stockSite.value?.code) {
            this.warehouseDestination.value = null;
            return;
        }
        try {
            this._selectedLocation = await this.$.graph
                .node('@sage/x3-stock-data/Location')
                .read(
                    {
                        warehouse: { code: true, _id: true },
                    },
                    `${this.stockSite.value.code}|${this.location.value?.code}`,
                )
                .execute();
            this.warehouseDestination.value = this._selectedLocation.warehouse ?? null;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/error-loading-location', 'Error loading location'),
                String(e),
            );
        }
    }

    private async _getDefaultLocation(codeLocation: string) {
        if (!this.stockSite.value) {
            return undefined;
        }
        try {
            const response = await this.$.graph
                .node('@sage/x3-stock-data/Location')
                .read(
                    {
                        code: true,
                        _id: true,
                        stockSite: {
                            code: true,
                            _id: true,
                        },
                    },
                    `${this.stockSite.value.code}|${codeLocation}`,
                )
                .execute();
            return response;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/error-loading-location', 'Error loading location'),
                String(e),
            );
        }
    }

    private async _getLots(productCode: string, lot: string, sublot: string | null): Promise<ExtractEdges<Lot>[]> {
        try {
            return extractEdges(
                await this.$.graph
                    .node('@sage/x3-stock-data/Lot')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                _id: true,
                                code: true,
                                product: {
                                    code: true,
                                    _id: true,
                                },
                                sublot: true,
                                supplierLot: true,
                                expirationDate: true,
                                lotCustomField1: true,
                                lotCustomField2: true,
                                lotCustomField3: true,
                                lotCustomField4: true,
                                majorVersion: {
                                    code: true,
                                    _id: true,
                                },
                                minorVersion: true,
                                potency: true,
                                useByDate: true,
                            },
                            {
                                filter: {
                                    product: { code: productCode },
                                    code: lot,
                                    ...(sublot && { sublot: sublot }),
                                },
                            },
                        ),
                    )
                    .execute(),
            ) as ExtractEdges<Lot>[];
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                ui.localize('@sage/x3-stock/dialog-error-set-expdate', 'Error while reading the lot information') +
                String(e),
            );
            return [];
        }
    }

    private async _getLotValues(): Promise<void> {
        if (
            this._productSite.product &&
            this.lot.value &&
            (this._productSite.product.lotManagementMode !== 'lotAndSublot' || this.sublot.value)
        ) {
            const result = await this._getLots(this._productSite.product.code, this.lot.value, this.sublot.value);
            if (result.length !== 0) {
                await this._setLotValues(result);
            } else {
                await this._initLotValues();
            }
        } else {
            await this._initLotValues();
        }
    }

    private async _setLotValues(lotValues: ExtractEdges<Lot>[]): Promise<void> {
        if (this._productSite.product.expirationManagementMode !== 'notManaged') {
            this.expirationDate.value = lotValues[0].expirationDate;
        }
        this.useByDate.value = lotValues[0].useByDate;
        this.supplierLot.value = lotValues[0].supplierLot;
        //    this.previousSupplierLot.value = lotValues[0].supplierLot;
        this.lotCustomField1.value = lotValues[0].lotCustomField1;
        this.lotCustomField2.value = lotValues[0].lotCustomField2;
        this.lotCustomField3.value = Number(lotValues[0].lotCustomField3);
        this.lotCustomField4.value = lotValues[0].lotCustomField4;
        this.majorVersion.value = null;
        this.minorVersion.value = null;
        this.majorVersion.value = {
            code: lotValues[0].majorVersion ? lotValues[0].majorVersion.code : undefined,
        };
        this.minorVersion.value = {
            minorVersion: lotValues[0].minorVersion,
        };
        this.potencyDestination.value = Number(lotValues[0].potency);
        this.supplierLot.isDisabled = true;
        this.lotCustomField1.isDisabled = true;
        this.lotCustomField2.isDisabled = true;
        this.lotCustomField3.isDisabled = true;
        this.lotCustomField4.isDisabled = true;
        this.majorVersion.isDisabled = true;
        this.minorVersion.isDisabled = true;
        this.potencyDestination.isDisabled = true;
        this.$.commitValueAndPropertyChanges();
    }

    private async _initLotValues(): Promise<void> {
        const expirationManagement = this._productSite.product.expirationManagementMode;
        if (!this.expirationDate.value) {
            this.expirationDate.value = expirationDateDefaultValue(
                this._productSite.product.expirationManagementMode,
                this._productSite.product.expirationLeadTime,
                this._productSite.product.expirationTimeUnit,
                this.effectiveDate.value ?? undefined,
            );
        }

        this.useByDate.value = useByDateDefaultValue(
            this.expirationDate.value,
            this.effectiveDate.value,
            this._productSite.product.expirationManagementMode,
            Number(this._productSite.product.useByDateCoefficient),
        );

        this.useByDate.maxDate = this.expirationDate.value ?? undefined;

        this.lotCustomField1.value = null;
        this.lotCustomField2.value = null;
        this.lotCustomField3.value = null;
        this.lotCustomField4.value = null;
        this.potencyDestination.value = Number(this._productSite.product.defaultPotencyInPercentage);
        this.supplierLot.isDisabled = false;
        this.lotCustomField1.isDisabled = false;
        this.lotCustomField2.isDisabled = false;
        this.lotCustomField3.isDisabled = false;
        this.lotCustomField4.isDisabled = false;
        this.majorVersion.isDisabled = false;
        this.minorVersion.isDisabled = !this.majorVersion.value;
        this.potencyDestination.isDisabled = false;
    }

    private async _isEmptyLpnAndContainerValues(): Promise<boolean> {
        if (
            !this.containerDestination.isHidden &&
            !this.licensePlateNumber.isHidden &&
            (!this.containerDestination.value || this.containerDestination.value?.code === '') &&
            (!this.licensePlateNumber.value || this.licensePlateNumber.value?.code === '')
        ) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-stock/pages__miscellaneous_receipt_details__notification__container_license_plate_number_mandatory_error',
                    'Container or license plate number is mandatory',
                ),
            );
            return true;
        } else {
            return false;
        }
    }

    private async _isZeroQuantityValue(): Promise<boolean> {
        if (!this.quantityInPackingUnit.value || this.quantityInPackingUnit.value === 0) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-stock/pages__miscellaneous_receipt_details__notification__quantityInPackingUnit__isZero',
                    'Quantity is mandatory',
                ),
            );
            return true;
        } else {
            return false;
        }
    }

    private _getNextField(field: any) {
        let _nextField = field.getNextField(true);
        if (this._productSite.product.expirationManagementMode !== 'manualEntry') {
            while (_nextField === this.expirationDate || _nextField === this.useByDate) {
                _nextField = _nextField.getNextField(true);
            }
        }
        return _nextField;
    }

    private async _isLotProvided(): Promise<boolean> {
        const rules = this._selectedStockManagementRules;
        const shouldUseSupplierLot = rules?.lotByDefault === 'supplierLot';
        if (
            this._productSite?.product?.lotManagementMode !== 'notManaged' &&
            shouldUseSupplierLot &&
            !this.lot.value &&
            !this._productSite?.product?.lotSequenceNumber &&
            rules?.lotEntry === 'no'
        ) {
            this.$.removeToasts();
            this.$.showToast(
                ui.localize(
                    '@sage/x3-stock/pages__utils__notification__invalid_inputs_error',
                    `Check your entry for {{#each fieldNames}}\n - {{this}}{{/each}}`,
                    { fieldNames: [this.lot.title] },
                ),
                { type: 'error', timeout: 5000 },
            );
            return true;
        }
        return false;
    }

    private async _getMinorVersions(majorVersion: string): Promise<ExtractEdges<ProductVersion>[]> {
        return extractEdges(
            await this.$.graph
                .node('@sage/x3-master-data/ProductVersion')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            minorVersion: true,
                        },
                        {
                            filter: {
                                product: {
                                    code: this._productSite.product.code,
                                    _id: this._productSite.product._id,
                                },
                                majorVersion: this.majorVersion.value?.code,
                                type: { _eq: 'stock' },
                                useStatus: { _eq: 'availableToUse' },
                            },
                            orderBy: {
                                minorVersion: 1,
                            },
                        },
                    ),
                )
                .execute(),
        ) as ExtractEdges<ProductVersion>[];
    }

    private _manageLicensePlateNumberLocation() {
        if (!!this.licensePlateNumber.value?.location?.code) {
            this.location.value = this.licensePlateNumber.value.location;
            this.location.isHidden = true;

            this.dummyLocation.value = this.licensePlateNumber.value.location.code;
            this.dummyLocation.isHidden = false;
        } else {
            if (
                !this._miscellaneousReceiptLines?.find(line => {
                    return line.stockDetails?.find(detail => {
                        if (detail.licensePlateNumber === this.licensePlateNumber.value?.code) {
                            this.location.value = { code: detail.location };
                            this.location.isDisabled = true;
                            return true;
                        } else {
                            return false;
                        }
                    });
                })
            ) {
                this.location.value = null;
                this.location.isDisabled = false;
                this.location.isHidden = false;
                this.dummyLocation.value = null;
                this.dummyLocation.isHidden = true;
            }
        }
    }

    private _addMiscLineAndStockDetails() {
        const lineIndex = this.miscellaneousReceiptLines.value.length;
        /*
          This part of the code, which I disabled, initializes the packingUnitToStockUnitConversionFactorDestination
           value by default and replaces the one entered by the user.
        */
        // const selectedValue = this.packingUnit.value;
        // const packingUnitIndex = selectedValue
        //     ? this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.code).indexOf(selectedValue)
        //     : -1;
        // if (packingUnitIndex !== -1) {
        //     const selectedUnit = this._packingUnits[packingUnitIndex].node;
        //     this.packingUnitToStockUnitConversionFactorDestination.value = Number(
        //         selectedUnit.packingUnitToStockUnitConversionFactor,
        //     );
        //     this.packingUnitToStockUnitConversionFactorDestination.isDisabled =
        //         !selectedUnit.isPackingFactorEntryAllowed;
        //     this.quantityInPackingUnit.scale = selectedUnit.packingUnit.numberOfDecimals;
        // } else {
        //     this.packingUnitToStockUnitConversionFactorDestination.value = 1;
        //     this.packingUnitToStockUnitConversionFactorDestination.isDisabled = true;
        //     if (this.packingUnit.value) this._getNumberOfDecimals();
        // }
        const selectedUnit = this._packingUnits.find(unit => unit.node.packingUnit.code === this.packingUnit.value);
        this.miscellaneousReceiptLines.addOrUpdateRecordValue(
            {
                _id: `${lineIndex}`,
                entryType: 'miscellaneousReceipt',
                lineNumber: Number(Math.floor(Math.random() * MAX_INT_32)),
                product: this.product.value
                    ? {
                        code: this.product.value.code,
                        _id: this.product.value._id,
                    }
                    : undefined,
                productDescription: this.product.value?.localizedDescription1,
                packingUnitToStockUnitConversionFactor:
                    this.packingUnitToStockUnitConversionFactorDestination.value?.toString(),
                packingUnit: selectedUnit
                    ? {
                        code: selectedUnit.node.packingUnit.code,
                        numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                        _id: selectedUnit.node.packingUnit._id,
                    }
                    : undefined,
                quantityInPackingUnit: this.quantityInPackingUnit?.value?.toString(),
                container: this.containerDestination?.value?.code,
                ...this.getcustomizationMiscellaneousReceiptLines(),
            },
            0,
            undefined,
        );
        this.miscellaneousReceiptLines.addOrUpdateRecordValue(
            {
                _id: `${lineIndex}`,
                stockSite: this.stockSite.value
                    ? {
                        code: this.stockSite.value.code,
                        _id: this.stockSite.value._id,
                    }
                    : undefined,
                product: this.product.value
                    ? {
                        code: this.product.value.code,
                        _id: this.product.value._id,
                    }
                    : undefined,
                effectiveDate: this.effectiveDate.value ?? undefined,
                lot: this.lot.value ?? undefined,
                sublot: this.sublot.value ?? undefined,
                supplierLot: this.supplierLot.value ?? undefined,
                location: this.location?.value
                    ? {
                        code: this.location.value?.code,
                        _id: this.location.value?._id,
                    }
                    : undefined,
                warehouse: this.warehouseDestination.value
                    ? {
                        code: this.warehouseDestination.value.code,
                        _id: this.warehouseDestination.value._id,
                    }
                    : undefined,
                serialNumber: this.serialNumber?.value ?? undefined,
                transactionType: 'miscellaneousReceipt',
                stockMovementGroup: this._selectedTransaction?.defaultStockMovementGroup
                    ? {
                        code: this._selectedTransaction.defaultStockMovementGroup.code,
                        _id: this._selectedTransaction.defaultStockMovementGroup._id,
                    }
                    : undefined,
                lotCustomField1: this.lotCustomField1?.value ?? undefined,
                lotCustomField2: this.lotCustomField2?.value ?? undefined,
                lotCustomField3: this.lotCustomField3?.value?.toString() ?? undefined,
                lotCustomField4: this.lotCustomField4?.value ?? undefined,
                packingUnit: selectedUnit
                    ? {
                        code: selectedUnit.node.packingUnit.code,
                        numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                        _id: selectedUnit.node.packingUnit._id,
                    }
                    : {
                        code: 'UN',
                        numberOfDecimals: 0,
                        _id: 'UN',
                    },
                quantityInPackingUnit: this.quantityInPackingUnit?.value?.toString(),
                packingUnitToStockUnitConversionFactor:
                    this.packingUnitToStockUnitConversionFactorDestination?.value?.toString(),
                expirationDate: this.expirationDate?.value ?? undefined,
                status: this.status.value
                    ? {
                        code: this.status.value.toString(),
                        _id: this.status.value.toString(),
                    }
                    : undefined,
                potency: this.potencyDestination?.value?.toString(),
                identifier1: this.identifier1?.value?.toString(),
                identifier2: this.identifier2?.value?.toString(),
                useByDate: this.useByDate?.value?.toString(),
                majorVersion: this.majorVersion.value
                    ? {
                        code: this.majorVersion.value.code,
                        _id: this.majorVersion.value._id,
                    }
                    : undefined,
                minorVersion: this.minorVersion.value?.minorVersion,
                licensePlateNumber: this.licensePlateNumber.value
                    ? {
                        code: this.licensePlateNumber.value.code,
                        _id: this.licensePlateNumber.value._id,
                    }
                    : undefined,
                ...this.getcustomizationMiscellaneousReceiptDetails(),
            },
            1,
            `${lineIndex}`,
        );
    }

    /* step 03*/
    @ui.decorators.block<MobileMiscellaneousReceipt>({
        parent() {
            return this.mainSection;
        },
        title: 'Products in receipt',
        width: 'extra-large',
        isHidden: true,
        isTitleHidden: true,
    })
    miscellaneousReceiptLinesBlock: ui.containers.Block;

    @ui.decorators.nestedGridField<MobileMiscellaneousReceipt, [MiscellaneousReceiptLineBinding, StockJournalBinding]>({
        parent() {
            return this.miscellaneousReceiptLinesBlock;
        },
        bind: 'miscellaneousReceiptLines',
        canFilter: true,
        canActivate: true,
        isChangeIndicatorDisabled: true,
        canSelect: false,
        isTitleHidden: true,
        levels: [
            {
                node: '@sage/x3-stock/MiscellaneousReceiptLine',
                childProperty: 'stockDetails',
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'Id',
                        isHiddenDesktop: false,
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lineNumber',
                        title: 'Line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.select({
                        bind: 'entryType',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, MiscellaneousReceiptLineBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: { code: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity in packing unit',
                        isReadOnly: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'productDescription',
                        title: 'Description',
                        isReadOnly: true,
                    }),
                    ui.nestedFields.reference<
                        MobileMiscellaneousReceipt,
                        MiscellaneousReceiptLineBinding,
                        UnitOfMeasure
                    >({
                        bind: 'packingUnit',
                        title: 'Unit packing',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: { code: true },
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId) {
                            if (await this.getSidebarData(rowId)) {
                                if (this._getSidebarMiscellaneousReceiptLine().length === 0) {
                                    this.miscellaneousReceiptLines.openSidebar(rowId, 1);
                                } else {
                                    this.miscellaneousReceiptLines.openSidebar(rowId, 0);
                                }
                            }
                        },
                    },
                    {
                        icon: 'delete',
                        title: 'Delete',
                        onClick(rowId) {
                            this.miscellaneousReceiptLines.removeRecord(rowId, 0);
                            if (!this.miscellaneousReceiptLines.value?.length) {
                                this.$.setPageClean();
                                this._init();
                            }
                        },
                    },
                ],
                sidebar: {
                    layout() {
                        return {
                            mainSection: {
                                title: 'Miscellaneous receipt lines',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarMiscellaneousReceiptLine(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
            {
                node: '@sage/x3-stock-data/StockJournal',
                orderBy: { _id: 1 },
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'Id',
                        isHiddenDesktop: false,
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, Warehouse>({
                        node: '@sage/x3-stock-data/Warehouse',
                        bind: 'warehouse',
                        valueField: 'code',
                        isHidden: true,
                        canFilter: false,
                        isFullWidth: true,
                        filter() {
                            return this.stockSite.value
                                ? {
                                    stockSite: { _id: { _eq: this.stockSite.value.code } },
                                }
                                : undefined;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity',
                        isFullWidth: true,
                        isMandatory: true,
                        isNotZero: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isMandatory: true,
                        placeholder: 'Scan or select...',
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                                isHidden: true,
                            }),
                        ],
                        filter() {
                            return {
                                code: {
                                    _in: this.packingUnit.options,
                                },
                            };
                        },
                        isReadOnly() {
                            return (this.packingUnit.options ?? []).length <= 1;
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (value._id && value.packingUnit?.code) {
                                const packingUnitIndex = this._packingUnits
                                    .map(packingUnit => packingUnit.node.packingUnit.code)
                                    .indexOf(value.packingUnit.code);
                                if (packingUnitIndex !== -1) {
                                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                                    value.packingUnitToStockUnitConversionFactor = String(
                                        selectedUnit.packingUnitToStockUnitConversionFactor,
                                    );
                                } else {
                                    value.packingUnitToStockUnitConversionFactor = '1';
                                }
                                this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInStockUnit',
                        title: 'Quantity in stock unit',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden() {
                            return (this.packingUnit.options ?? []).length <= 1;
                        },
                        isMandatory: true,
                        isFullWidth: true,
                        isDisabled(value, rowValue) {
                            if (rowValue?.packingUnit) {
                                const packingUnitIndex = this._packingUnits
                                    .map(packingUnit => packingUnit.node.packingUnit.code)
                                    .indexOf(rowValue.packingUnit.code);
                                if (packingUnitIndex !== -1) {
                                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                                    return !selectedUnit.isPackingFactorEntryAllowed;
                                } else {
                                    return true;
                                }
                            } else {
                                return true;
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, StockStatus>({
                        bind: 'status',
                        title: 'Status',
                        node: '@sage/x3-stock-data/StockStatus',
                        valueField: { code: true },
                        isMandatory: true,
                        placeholder: 'Scan or select...',
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, Container>({
                        bind: 'container' as any,
                        title: 'Container',
                        node: '@sage/x3-master-data/Container',
                        valueField: 'code',
                        isAutoSelectEnabled: true,
                        placeholder: 'Scan or select...',
                        width: 'large',
                        isTransient: true,
                        canFilter: false,
                        isDisabled(value, rowValue) {
                            return Boolean(rowValue?.licensePlateNumber);
                        },
                        isHidden() {
                            return (
                                this._selectedStockManagementRules.licensePlateNumberEntry !== 'mandatory' ||
                                !this._productSite.isLicensePlateNumberManaged
                            );
                        },
                        filter() {
                            return {
                                isInternal: { _eq: true },
                                isActive: { _eq: true },
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.text({
                                bind: 'containerType',
                                title: 'Type',
                            }),
                            ui.nestedFields.text({
                                bind: 'description',
                                title: 'Description',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, LicensePlateNumber>({
                        bind: 'licensePlateNumber',
                        title: 'License plate number',
                        valueField: 'code',
                        node: '@sage/x3-stock-data/LicensePlateNumber',
                        helperTextField: { container: { code: true } },
                        isFullWidth: true,
                        placeholder: 'Scan or select...',
                        isAutoSelectEnabled: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-stock-data/Location',
                                bind: 'location',
                                valueField: 'code',
                                title: 'Location',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Container',
                                bind: 'container',
                                valueField: 'code',
                                title: 'Container',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-system/Site',
                                bind: 'stockSite',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isActive',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isSingleProduct',
                                isHidden: true,
                            }),
                        ],
                        canFilter: false,
                        isHidden() {
                            return !this._productSite.isLicensePlateNumberManaged;
                        },
                        filter(rowValue) {
                            let licensePlateNumberFilter: Filter<any> = {
                                ...(this.stockSite.value && { stockSite: { code: this.stockSite.value.code } }),
                                isActive: { _eq: true },
                                _or: [
                                    {
                                        isSingleProduct: { _eq: true },
                                        stock: { _atLeast: 1, product: { product: { code: this.product.value } } },
                                    },
                                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                                    { isSingleProduct: { _in: [false, null] } },
                                ],
                            };
                            if ((rowValue as any).container) {
                                licensePlateNumberFilter = {
                                    ...licensePlateNumberFilter,
                                    container: { code: ((rowValue as any).container as Container)?.code },
                                };
                            }
                            return licensePlateNumberFilter;
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.licensePlateNumber) {
                                if (value.licensePlateNumber.location) {
                                    value.location = value.licensePlateNumber.location;
                                } else {
                                    if (
                                        !this.miscellaneousReceiptLines.value.find(line => {
                                            return line.stockDetails?.find(detail => {
                                                if (
                                                    detail.licensePlateNumber?.code === value.licensePlateNumber?.code
                                                ) {
                                                    value.location = { code: detail.location?.code };
                                                    return true;
                                                } else {
                                                    return false;
                                                }
                                            });
                                        })
                                    ) {
                                        value.location = undefined;
                                    }
                                }
                                this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, Location>({
                        bind: 'location',
                        title: 'Location',
                        node: '@sage/x3-stock-data/Location',
                        valueField: { code: true },
                        isFullWidth: true,
                        helperTextField: 'type',
                        placeholder: 'Scan or select...',
                        isHidden() {
                            return !this._productSite.isLocationManaged;
                        },
                        isMandatory() {
                            return this._productSite.isLocationManaged;
                        },
                        isReadOnly(_value, rowValue) {
                            return this._productSite.isLicensePlateNumberManaged && !!rowValue?.licensePlateNumber;
                        },
                        filter() {
                            return {
                                ...(this.stockSite.value?.code && {
                                    stockSite: { code: this.stockSite.value.code },
                                }),
                                category: { _nin: ['subcontract', 'customer'] },
                            };
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.location) {
                                value.warehouse = value.location.warehouse;
                                this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference<MobileMiscellaneousReceipt, Location, Warehouse>({
                                node: '@sage/x3-stock-data/Warehouse',
                                bind: 'warehouse',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'serialNumber',
                        title: 'Serial number',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return ['notManaged', 'issued'].includes(
                                this._productSite.product.serialNumberManagementMode,
                            );
                        },
                        isMandatory() {
                            return (
                                !['notManaged', 'issued'].includes(
                                    this._productSite.product.serialNumberManagementMode,
                                ) && !this._productSite.product.serialSequenceNumber
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'supplierLot',
                        title: 'Supplier lot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                (this._selectedTransaction.supplierLot !== 'entered' &&
                                    this._selectedTransaction.supplierLot !== 'displayed')
                            );
                        },
                        isDisabled() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                this._selectedTransaction.supplierLot !== 'entered'
                            );
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }
                            if (value.supplierLot) {
                                value.supplierLot = value.supplierLot.toUpperCase();
                                if (
                                    this._selectedStockManagementRules.lotByDefault === 'supplierLot' &&
                                    this._selectedStockManagementRules.lotEntry === 'newLot'
                                ) {
                                    if (
                                        !(await controlLotReceipt(
                                            this,
                                            value.supplierLot,
                                            this._productSite.product.code,
                                            '19',
                                            this.stockSite.value?.code ?? null,
                                        ))
                                    ) {
                                        value.supplierLot = undefined;
                                    }
                                }
                            }
                            const rules = this._selectedStockManagementRules;
                            const shouldUseSupplierLot =
                                rules?.lotByDefault === 'supplierLot' &&
                                (!value.lot || value.lot === this._oldSupplierLotValue);
                            if (shouldUseSupplierLot) {
                                value.lot = value.supplierLot;
                                if (
                                    value.lot &&
                                    (this._productSite.product.lotManagementMode !== 'lotAndSublot' || value.sublot)
                                ) {
                                    this._lots = await this._getLots(
                                        this._productSite.product.code,
                                        value.lot,
                                        value.sublot ?? null,
                                    );
                                    if (this._lots.length !== 0) {
                                        value.supplierLot = this._lots[0].supplierLot;
                                        value.lotCustomField1 = this._lots[0].lotCustomField1;
                                        value.lotCustomField2 = this._lots[0].lotCustomField2;
                                        value.lotCustomField3 = this._lots[0].lotCustomField3;
                                        value.lotCustomField4 = this._lots[0].lotCustomField4;
                                        value.majorVersion = this._lots[0].majorVersion
                                            ? {
                                                code: this._lots[0].majorVersion.code,
                                            }
                                            : undefined;
                                        value.minorVersion = this._lots[0].minorVersion
                                            ? this._lots[0].minorVersion
                                            : undefined;
                                        if (this._productSite.product.expirationManagementMode !== 'notManaged') {
                                            value.expirationDate = this._lots[0].expirationDate;
                                        }
                                        value.useByDate = this._lots[0].useByDate;
                                        value.potency = this._lots[0].potency;
                                    } else {
                                        value = this._initStockJournalLotValue(value);
                                    }
                                } else {
                                    value = this._initStockJournalLotValue(value);
                                    this._lots = [];
                                }
                            }
                            this._oldSupplierLotValue = value.supplierLot ?? '';
                            this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'lot',
                        node: '@sage/x3-stock-data/LotsSites',
                        title: 'Lot',
                        valueField: 'lot',
                        placeholder: 'Scan or select...',
                        isFullWidth: true,
                        isNewEnabled: true,
                        filter() {
                            return {
                                product: { code: this._productSite.product.code },
                                ...(this.stockSite.value && { storageSite: { code: this.stockSite.value.code } }),
                                stock: { _atLeast: 1 },
                            };
                        },
                        isMandatory() {
                            return (
                                ['lotAndSublot', 'mandatoryLot'].includes(
                                    this._productSite.product.lotManagementMode,
                                ) &&
                                !this._productSite.product.lotSequenceNumber &&
                                this._selectedStockManagementRules?.lotEntry !== 'no' &&
                                this._selectedStockManagementRules.lotByDefault !== 'documentNumber'
                            );
                        },
                        isReadOnly() {
                            return this._selectedStockManagementRules?.lotEntry === 'no';
                        },
                        async validation(value, rowValue) {
                            const rules = this._selectedStockManagementRules;
                            const shouldUseSupplierLot = rules?.lotByDefault === 'supplierLot';

                            const regex = /^$|^[^|]+$/;
                            if (!regex.test(value)) {
                                return ui.localize('@sage/x3-stock/validate-error-invalid-value', 'Incorrect value');
                            }

                            if (value) {
                                value = value.toUpperCase();
                            }
                            if (
                                value &&
                                rules.lotEntry === 'newLot' &&
                                (await controlLot(
                                    this,
                                    value,
                                    this._productSite.product.code,
                                    '19',
                                    this.stockSite.value?.code ?? null,
                                )) === false
                            ) {
                                return ui.localize(
                                    '@sage/x3-stock/notification-error-receipt-lot',
                                    'The lot number {{ lot }} already exists for this product.',
                                    { lot: value },
                                );
                            }

                            if (
                                shouldUseSupplierLot &&
                                !rowValue.supplierLot &&
                                !this._productSite?.product?.lotSequenceNumber &&
                                rules?.lotEntry === 'no'
                            ) {
                                return ui.localize(
                                    '@sage/x3-stock/pages__utils__notification__invalid_inputs_error',
                                    `Check your entry for {{#each fieldNames}}\n - {{this}}{{/each}}`,
                                    { fieldNames: ['Lot'] },
                                );
                            }
                            return undefined;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'lot',
                                title: 'Lot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'sublot',
                                title: 'Sublot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                        isHidden() {
                            return this._productSite.product.lotManagementMode === 'notManaged';
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }
                            if (
                                value.lot &&
                                (this._productSite.product.lotManagementMode !== 'lotAndSublot' || value.sublot)
                            ) {
                                this._lots = await this._getLots(
                                    this._productSite.product.code,
                                    value.lot,
                                    value.sublot ?? null,
                                );
                                if (this._lots.length !== 0) {
                                    value.supplierLot = this._lots[0].supplierLot;
                                    value.lotCustomField1 = this._lots[0].lotCustomField1;
                                    value.lotCustomField2 = this._lots[0].lotCustomField2;
                                    value.lotCustomField3 = this._lots[0].lotCustomField3;
                                    value.lotCustomField4 = this._lots[0].lotCustomField4;
                                    value.majorVersion = this._lots[0].majorVersion
                                        ? {
                                            code: this._lots[0].majorVersion.code,
                                        }
                                        : undefined;
                                    value.minorVersion = this._lots[0].minorVersion
                                        ? this._lots[0].minorVersion
                                        : undefined;
                                    if (this._productSite.product.expirationManagementMode !== 'notManaged') {
                                        value.expirationDate = this._lots[0].expirationDate;
                                    }
                                    value.useByDate = this._lots[0].useByDate;
                                    value.potency = this._lots[0].potency;
                                } else {
                                    value = this._initStockJournalLotValue(value);
                                }
                            } else {
                                value = this._initStockJournalLotValue(value);
                                this._lots = [];
                            }
                            this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'sublot',
                        title: 'Sublot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return this._productSite.product.lotManagementMode !== 'lotAndSublot';
                        },
                        isMandatory() {
                            return this._productSite.product.lotManagementMode === 'lotAndSublot';
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.sublot) {
                                value.sublot = value.sublot.toUpperCase();
                            }
                            if (value.lot && value.sublot) {
                                this._lots = await this._getLots(
                                    this._productSite.product.code,
                                    value.lot,
                                    value.sublot,
                                );
                                if (this._lots.length !== 0) {
                                    value.supplierLot = this._lots[0].supplierLot;
                                    value.lotCustomField1 = this._lots[0].lotCustomField1;
                                    value.lotCustomField2 = this._lots[0].lotCustomField2;
                                    value.lotCustomField3 = this._lots[0].lotCustomField3;
                                    value.lotCustomField4 = this._lots[0].lotCustomField4;
                                    value.majorVersion = this._lots[0].majorVersion
                                        ? {
                                            code: this._lots[0].majorVersion.code,
                                        }
                                        : undefined;
                                    value.minorVersion = this._lots[0].minorVersion
                                        ? this._lots[0].minorVersion
                                        : undefined;
                                    if (this._productSite.product.expirationManagementMode !== 'notManaged') {
                                        value.expirationDate = this._lots[0].expirationDate;
                                    }
                                    value.useByDate = this._lots[0].useByDate;
                                    value.potency = this._lots[0].potency;
                                } else {
                                    value = this._initStockJournalLotValue(value);
                                }
                            } else {
                                value = this._initStockJournalLotValue(value);
                                this._lots = [];
                            }
                            this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: 'code',
                        isHiddenDesktop: false,
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, MajorVersionStatus>({
                        bind: 'majorVersion',
                        title: 'Major version',
                        node: '@sage/x3-master-data/MajorVersionStatus',
                        valueField: { code: true },
                        isFullWidth: true,
                        placeholder: 'Scan or select...',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._productSite.product.stockVersionMode ||
                                this._productSite.product.stockVersionMode === 'no'
                            );
                        },
                        isMandatory() {
                            return (
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                this._productSite.product.stockVersionMode &&
                                this._productSite.product.stockVersionMode !== 'no'
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                        filter() {
                            return {
                                product: { code: this._productSite.product.code },
                                _or: [
                                    { status: 'prototypeVersion' },
                                    { status: 'activeVersion' },
                                    { status: 'stoppedVersion' },
                                ],
                            };
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.majorVersion?.code) {
                                const minorVersions = await this._getMinorVersions(value.majorVersion.code);
                                value.minorVersion =
                                    minorVersions.length > 0
                                        ? minorVersions[minorVersions.length - 1].minorVersion
                                        : undefined;
                                this._majorVersion = value.majorVersion.code;
                            } else {
                                this._majorVersion = null;
                            }
                            this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Major Version',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'minorVersion',
                        title: 'Minor version',
                        node: '@sage/x3-master-data/ProductVersion',
                        valueField: 'minorVersion',
                        placeholder: 'Scan or select...',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                this._productSite.product.stockVersionMode !== 'majorAndMinor'
                            );
                        },
                        isMandatory() {
                            return (
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                this._productSite.product.stockVersionMode === 'majorAndMinor'
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                        filter(rowValue) {
                            return rowValue.majorVersion
                                ? {
                                    product: { code: this._productSite.product.code },
                                    majorVersion: rowValue.majorVersion.code,
                                    type: 'stock',
                                    useStatus: 'availableToUse',
                                }
                                : {};
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'minorVersion',
                                title: 'Minor version',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'majorVersion',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'type',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'useStatus',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'potency',
                        title: 'potency %',
                        isFullWidth: true,
                        placeholder: 'Scan...',
                        scale: 4,
                        validation: /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/, // reg ex for any positive numbers (integers or decimals) excluding 0
                        min: 0,
                        max: 100,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedTransaction.isLotPotencyAllowed ||
                                this._productSite.product.stockManagementMode !== 'potencyManaged'
                            );
                        },
                        isMandatory() {
                            return Boolean(
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                this._selectedTransaction.isLotPotencyAllowed &&
                                this._productSite.product.stockManagementMode === 'potencyManaged',
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'expirationDate',
                        placeholder: 'Enter...',
                        title: 'Expiration date',
                        isHidden() {
                            return (
                                (this._productSite.product.lotManagementMode === 'notManaged' ||
                                    this._productSite.product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedTransaction.isLotExpirationDateAllowed &&
                                        (this._productSite.product.expirationManagementMode ===
                                            'roundingBeginningMonth1' ||
                                            this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._productSite.product.expirationManagementMode ===
                                            'withoutRounding'))) ??
                                false
                            );
                        },
                        isMandatory() {
                            return (
                                (this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._productSite.product.expirationManagementMode === 'mandatoryEntry') ??
                                false
                            );
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalBinding>) {
                            if (!value._id || !value.expirationDate) {
                                return;
                            }
                            value.useByDate =
                                useByDateDefaultValue(
                                    value.expirationDate,
                                    this.effectiveDate.value,
                                    this._productSite.product.expirationManagementMode,
                                    Number(this._productSite.product.useByDateCoefficient),
                                ) ?? undefined;
                            this.miscellaneousReceiptLines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'useByDate',
                        placeholder: 'Enter...',
                        title: 'Use-by date',
                        isHidden() {
                            return (
                                (this._productSite.product.lotManagementMode === 'notManaged' ||
                                    this._productSite.product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedTransaction.isLotExpirationDateAllowed &&
                                        (this._productSite.product.expirationManagementMode ===
                                            'roundingBeginningMonth1' ||
                                            this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._productSite.product.expirationManagementMode ===
                                            'withoutRounding'))) ??
                                false
                            );
                        },
                        isMandatory() {
                            return (
                                (this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._productSite.product.expirationManagementMode === 'mandatoryEntry') ??
                                false
                            );
                        },
                        isDisabled() {
                            return (
                                this._productSite.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                                this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                                this._productSite.product.expirationManagementMode === 'withoutRounding'
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 1',
                        validation: /^$|^[^|]+$/,
                        maxLength: 20,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedTransaction.isLotCustomField1Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 2',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedTransaction.isLotCustomField2Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lotCustomField3',
                        placeholder: 'Enter...',
                        title: 'Lot custom number',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedTransaction.isLotCustomField3Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'lotCustomField4',
                        placeholder: 'Enter...',
                        title: 'Lot custom date',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedTransaction.isLotCustomField4Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier1',
                        title: 'Identifier 1',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHidden() {
                            return (
                                this._selectedTransaction.identifier1Detail !== 'entered' &&
                                this._selectedTransaction.identifier1Detail !== 'displayed'
                            );
                        },
                        isDisabled() {
                            return this._selectedTransaction.identifier1Detail !== 'entered';
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier2',
                        title: 'Identifier 2',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHiddenDesktop: false,
                        isHidden() {
                            return (
                                this._selectedTransaction.identifier2Detail !== 'entered' &&
                                this._selectedTransaction.identifier1Detail !== 'displayed'
                            );
                        },
                        isDisabled() {
                            return this._selectedTransaction.identifier2Detail !== 'entered';
                        },
                    }),
                    ui.nestedFields.reference<MobileMiscellaneousReceipt, StockJournalBinding, Site>({
                        bind: 'stockSite',
                        title: 'Site',
                        node: '@sage/x3-system/Site',
                        valueField: { code: true },
                        isHiddenDesktop: false,
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.date({
                        bind: 'effectiveDate',
                        title: 'Effective date',
                        isHiddenDesktop: false,
                        isReadOnly: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId) {
                            if (await this.getSidebarData(rowId)) {
                                this.miscellaneousReceiptLines.openSidebar(rowId, 1);
                            }
                        },
                    },
                ],
                sidebar: {
                    title: 'Miscellaneous receipt',
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock details',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarStockJournal(),
                                    },
                                },
                            },
                        };
                    },
                    async onRecordConfirmed(_id: string, recordValue: StockJournalBinding | undefined) {
                        const _miscellaneousReceiptLine = this.miscellaneousReceiptLines.getRecordValue(_id, 0);
                        if (_miscellaneousReceiptLine) {
                            _miscellaneousReceiptLine.packingUnit = recordValue?.packingUnit;
                            _miscellaneousReceiptLine.packingUnitToStockUnitConversionFactor =
                                recordValue?.packingUnitToStockUnitConversionFactor;
                            _miscellaneousReceiptLine.quantityInPackingUnit = recordValue?.quantityInPackingUnit;
                            this.miscellaneousReceiptLines.setRecordValue(_miscellaneousReceiptLine, 0);
                        }
                    },
                },
            },
        ],
    })
    miscellaneousReceiptLines: ui.fields.NestedGrid<[MiscellaneousReceiptLineBinding, StockJournalBinding]>;

    /*
     *
     *  Init functions
     *
     */

    private async _init(): Promise<void> {
        try {
            await this._initSite();
            await this._getIsLocationPreloaded();
            this.effectiveDate.value = DateValue.today().toString();
            this.entryType.value = 'miscellaneousReceipt';
            this.destination.value = this.$.storage.get('mobile-label-destination')?.toString() ?? '';
            this.createButton.isDisabled = true;
            if (this.stockSite.value) {
                await this._initTransaction();
                this._numberOfDecimalList = await getUnitNumberOfDecimalList(this);
                await this._initFieldsDestination();
                this.packingUnit.value = this.packingUnit.options?.[0] ?? null;
                this.packingUnitToStockUnitConversionFactorDestination.value = 1;
            } else {
                this._disablePage();
            }
        } catch (e) {
            ui.console.error(e);
        }
    }

    private async _initSite(): Promise<void> {
        const stockSiteCode = await getSelectedStockSite(
            this,
            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
            ui.localize(
                '@sage/x3-stock/dialog-error-location-inquiry-set-site',
                'Define a default stock site on the user function profile.',
            ),
        );
        if (stockSiteCode) {
            this.stockSite.value = {
                _id: stockSiteCode,
                code: stockSiteCode,
            };
        }
    }

    private async _initTransaction() {
        try {
            this._transactions = extractEdges(
                await this.$.graph
                    .node('@sage/x3-stock/StockEntryTransaction')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                code: true,
                                isActive: true,
                                stockAutomaticJournal: {
                                    code: true,
                                    _id: true,
                                },
                                localizedDescription: true,
                                transactionType: true,
                                isLotCustomField1Allowed: true,
                                isLotCustomField2Allowed: true,
                                isLotCustomField3Allowed: true,
                                isLotCustomField4Allowed: true,
                                isLotPotencyAllowed: true,
                                identifier1Detail: true,
                                identifier2Detail: true,
                                supplierLot: true,
                                isLotExpirationDateAllowed: true,
                                defaultStockMovementGroup: {
                                    code: true,
                                    _id: true,
                                },
                                stockMovementCode: {
                                    code: true,
                                    _id: true,
                                },
                                companyOrSiteGroup: {
                                    group: true,
                                },
                                isUnitChange: true,
                                _id: true,
                            },
                            {
                                filter: {
                                    transactionType: 'miscellaneousReceipt',
                                    isActive: true,
                                },
                            },
                        ),
                    )
                    .execute(),
            );
        } catch (err) {
            ui.console.error(err);
        }

        this.transaction.isHidden = this._transactions.length <= 1;
        this.transaction.isDisabled = false;
        this._transactions.length <= 1 ? (this.transactionBlock.isHidden = true) : null;
        this.transaction.options = this._transactions.map(trs => trs.code).filter(t => !!t) as string[];
        this.transaction.value = this._transactions[0]?.code ?? null;
        this._selectedTransaction = this._transactions[0];
        this.stockMovementCode.value = this._selectedTransaction.stockMovementCode
            ? {
                code: this._selectedTransaction.stockMovementCode.code,
                _id: this._selectedTransaction.stockMovementCode._id,
            }
            : null;
        this.stockMovementGroup.value = this._selectedTransaction.defaultStockMovementGroup
            ? {
                code: this._selectedTransaction.defaultStockMovementGroup.code,
                _id: this._selectedTransaction.defaultStockMovementGroup._id,
            }
            : null;
        this.stockAutomaticJournal.value = this._selectedTransaction.stockAutomaticJournal
            ? {
                code: this._selectedTransaction.stockAutomaticJournal.code,
                _id: this._selectedTransaction.stockAutomaticJournal._id,
            }
            : null;
        this.product.isDisabled = false;
    }

    private async _getIsLocationPreloaded() {
        const responseSite = await this.$.graph
            .node('@sage/x3-master-data/MobileAutomationSetup')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        isLocationPreloaded: true,
                    },
                    {
                        filter: this.stockSite.value
                            ? {
                                site: { code: this.stockSite.value.code },
                            }
                            : undefined,
                    },
                ),
            )
            .execute();
        if (responseSite.edges.length !== 0) {
            responseSite.edges.some(edge => {
                this._isLocationPreloaded = edge.node.isLocationPreloaded;
            });
        } else {
            const response = await this.$.graph
                .node('@sage/x3-master-data/MobileAutomationSetup')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            isLocationPreloaded: true,
                        },
                        {
                            filter: {
                                site: null,
                            },
                        },
                    ),
                )
                .execute();
            if (response.edges.length !== 0) {
                response.edges.some(edge => {
                    this._isLocationPreloaded = edge.node.isLocationPreloaded;
                });
            } else {
                this._isLocationPreloaded = true;
            }
        }
    }

    private _disablePage(): void {
        this.effectiveDate.isDisabled = true;
        this.transaction.isDisabled = true;
        this.product.isDisabled = true;
    }

    private _initStockJournalLotValue(
        value: ExtractEdgesPartial<StockJournalBinding>,
    ): ExtractEdgesPartial<StockJournalBinding> {
        value.lotCustomField1 = '';
        value.lotCustomField2 = '';
        value.lotCustomField3 = '0';
        value.lotCustomField4 = '';
        if (!value.expirationDate) {
            value.expirationDate =
                expirationDateDefaultValue(
                    this._productSite.product.expirationManagementMode,
                    this._productSite.product.expirationLeadTime,
                    this._productSite.product.expirationTimeUnit,
                    this.effectiveDate.value ?? undefined,
                ) ?? undefined;
        }
        value.useByDate =
            useByDateDefaultValue(
                value.expirationDate,
                this.effectiveDate.value,
                this._productSite.product.expirationManagementMode,
                Number(this._productSite.product.useByDateCoefficient),
            ) ?? undefined;
        value.potency = this._productSite.product.defaultPotencyInPercentage;
        return value;
    }

    private async getSidebarData(rowId: string): Promise<boolean> {
        const _miscellaneousLine = this.miscellaneousReceiptLines.getRecordValue(rowId, 0);
        const _stockJournal = this.miscellaneousReceiptLines.getRecordValue(rowId, 1);
        if (_miscellaneousLine && _stockJournal) {
            this._oldSupplierLotValue = _stockJournal.supplierLot ?? '';
            const _productCode = _miscellaneousLine.product?.code;
            if (_productCode && this.stockSite.value?.code) {
                this._productSite = await getProductSite(this, _productCode, this.stockSite.value.code, '');
            }
            this._packingUnits = this._productSite.product.packingUnits?.query.edges.map(packingUnit => {
                return packingUnit;
            });
            this.packingUnit.options = [
                this._productSite.product.stockUnit.code,
                ...this._packingUnits.map(packingUnit => {
                    return packingUnit.node.packingUnit.code;
                }),
            ];
            this._lots =
                _stockJournal.product && _stockJournal.product.code && _stockJournal.lot
                    ? await this._getLots(_stockJournal.product.code, _stockJournal.lot, _stockJournal.sublot ?? null)
                    : [];
            return true;
        } else {
            return false;
        }
    }

    private _getSidebarMiscellaneousReceiptLine(): ui.SidebarFieldDefinition<MiscellaneousReceiptLineBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<MiscellaneousReceiptLineBinding>[] = [];
        this.customizationMiscellaneousReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<MiscellaneousReceiptLineBinding>);
            }
        });
        return _FieldsSidebar;
    }

    private _getSidebarStockJournal(): ui.SidebarFieldDefinition<StockJournalBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockJournalBinding>[] = [];
        _FieldsSidebar.push('packingUnit');
        _FieldsSidebar.push('quantityInPackingUnit');
        if ((this.packingUnit.options ?? []).length > 1) {
            _FieldsSidebar.push('packingUnitToStockUnitConversionFactor');
        }
        _FieldsSidebar.push('status');
        if (this._productSite.isLicensePlateNumberManaged) {
            _FieldsSidebar.push('licensePlateNumber');
        }
        if (this._productSite.isLocationManaged) {
            _FieldsSidebar.push('location');
        }
        if (!['notManaged', 'issued'].includes(this._productSite.product.serialNumberManagementMode)) {
            _FieldsSidebar.push('serialNumber');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            (this._selectedTransaction.supplierLot === 'entered' ||
                this._selectedTransaction.supplierLot === 'displayed')
        ) {
            _FieldsSidebar.push('supplierLot');
        }
        if (this._productSite.product.lotManagementMode !== 'notManaged') {
            _FieldsSidebar.push('lot');
        }
        if (this._productSite.product.lotManagementMode === 'lotAndSublot') {
            _FieldsSidebar.push('sublot');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._productSite.product.stockVersionMode !== 'no'
        ) {
            _FieldsSidebar.push('majorVersion');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._productSite.product.stockVersionMode === 'majorAndMinor'
        ) {
            _FieldsSidebar.push('minorVersion');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedTransaction.isLotPotencyAllowed &&
            this._productSite.product.stockManagementMode === 'potencyManaged'
        ) {
            _FieldsSidebar.push('potency');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            (this._selectedTransaction.isLotExpirationDateAllowed ||
                this._productSite.product.expirationManagementMode === 'manualEntry' ||
                this._productSite.product.expirationManagementMode === 'mandatoryEntry')
        ) {
            _FieldsSidebar.push('expirationDate');
            _FieldsSidebar.push('useByDate');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedTransaction.isLotCustomField1Allowed
        ) {
            _FieldsSidebar.push('lotCustomField1');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedTransaction.isLotCustomField2Allowed
        ) {
            _FieldsSidebar.push('lotCustomField2');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedTransaction.isLotCustomField3Allowed
        ) {
            _FieldsSidebar.push('lotCustomField3');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedTransaction.isLotCustomField4Allowed
        ) {
            _FieldsSidebar.push('lotCustomField4');
        }
        if (
            this._selectedTransaction.identifier1Detail === 'entered' ||
            this._selectedTransaction.identifier1Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier1');
        }
        if (
            this._selectedTransaction.identifier2Detail === 'entered' ||
            this._selectedTransaction.identifier2Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier2');
        }
        this.customizationMiscellaneousReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockJournalBinding>);
            }
        });

        return _FieldsSidebar;
    }

    private getcustomizationMiscellaneousReceiptLines() {
        let response = {};
        this.customizationMiscellaneousReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private getcustomizationMiscellaneousReceiptDetails() {
        let response = {};
        this.customizationMiscellaneousReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private _initPackingUnitToStockUnitConversionFactorDestination() {
        const selectedValue = this.packingUnit.value;
        const packingUnitIndex = selectedValue
            ? this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.code).indexOf(selectedValue)
            : -1;

        if (packingUnitIndex !== -1) {
            const selectedUnit = this._packingUnits[packingUnitIndex].node;
            this.packingUnitToStockUnitConversionFactorDestination.value = Number(
                selectedUnit.packingUnitToStockUnitConversionFactor,
            );
            this.packingUnitToStockUnitConversionFactorDestination.isDisabled =
                !selectedUnit.isPackingFactorEntryAllowed;
            this.quantityInPackingUnit.scale = selectedUnit.packingUnit.numberOfDecimals;
        } else {
            this.packingUnitToStockUnitConversionFactorDestination.value = 1;
            this.packingUnitToStockUnitConversionFactorDestination.isDisabled = true;
            if (this.packingUnit.value) this._getNumberOfDecimals();
        }
    }
}
