import {
    StandardOperations,
    WorkOrderOperationTrackingLineBinding as TimeTrackingLineBinding,
    WorkOrder,
    WorkOrderOperationLine,
    WorkOrderOperationTracking,
} from '@sage/x3-manufacturing-api';
import { RoutingHeader } from '@sage/x3-manufacturing-api-partial';
import { EmployeeId, GraphApi, WorkCenter } from '@sage/x3-manufacturing-data-api';
import { Product } from '@sage/x3-master-data-api';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { Site } from '@sage/x3-system-api-partial';
import { decimal, ExtractEdges, extractEdges } from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import * as ui from '@sage/xtrem-ui';
import { PartialCollectionValueWithIds } from '@sage/xtrem-ui/build/lib/component/types';

import { getBomCode } from '../client-functions/get-bom-code';
import { getEntryTransaction, TransactionList } from '../client-functions/get-entry-transaction';
import { getReleaseUOM } from '../client-functions/get-release-uom';
import { getSelectedStockSite } from '../client-functions/get-selected-stock-site';
import { getTrackingDate } from '../client-functions/get-tracking-date';
import { GetUOMNumberOfDecimalPlaces } from '../client-functions/get-unit-decimal-places';
import { readParameterValue } from '../client-functions/read-parameter';
import { validateWithDetails } from '../client-functions/validation';
import { getWorkOrderOperationsList } from '../client-functions/work-order-operations-list';
import { manufacturing } from '../menu-items/manufacturing';
import { OperationLine, SelectedOperationMethod } from './utils/types';

interface TimeTrackingLine extends TimeTrackingLineBinding {
    enteredSetupTimeHours?: number;
    enteredSetupTimeMinutes?: number;
    enteredRunTimeHours?: number;
    enteredRunTimeMinutes?: number;
    operationDescription?: string;
    expectedQuantity?: number;
    completedQuantity?: number;
}

let paramOperationTrackingControl: string; // Parameter CTLOPEPRE - Expected operation tracking
let showMobileTimeTrackingDetails: boolean;
let operationList: ExtractEdges<WorkOrderOperationLine>[];
let workOrderRoutingList: ExtractEdges<RoutingHeader>[];
let operationMethod: number;

@ui.decorators.page<MobileTimeTracking, WorkOrderOperationTracking>({
    title: 'Time tracking',
    module: 'xtrem-x3-manufacturing',
    mode: 'default',
    isTitleHidden: true,
    menuItem: manufacturing,
    authorizationCode: 'CWSAOT',
    node: '@sage/x3-manufacturing/WorkOrderOperationTracking',
    access: { node: '@sage/x3-manufacturing/WorkOrderOperationTracking' },
    navigationPanel: undefined,
    priority: 100,
    businessActions() {
        return [this.createButton, this.addButton];
    },
    async onLoad() {
        await this.initPage();
        paramOperationTrackingControl = await readParameterValue(
            'CTLOPEPRE',
            '',
            this.productionSite?.value?.code,
            this,
        );
    },
})
export class MobileTimeTracking extends ui.Page<GraphApi> {
    private transactionsList: TransactionList[];

    @ui.decorators.section<MobileTimeTracking>({
        isTitleHidden: true,
    })
    mainSection: ui.containers.Section;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    transactionBlock: ui.containers.Block;

    @ui.decorators.dateField<MobileTimeTracking>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Tracking date',
        isMandatory: true,
        isTransient: true,
        maxDate: DateValue.today().toString(),
        onChange() {
            this.transaction.focus();
        },
    })
    trackingDate: ui.fields.Date;

    @ui.decorators.referenceField<MobileTimeTracking, Site>({
        node: '@sage/x3-system/Site',
        valueField: 'code',
        isHidden: true,
        bind: 'productionSite',
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    productionSite: ui.fields.Reference;

    @ui.decorators.selectField<MobileTimeTracking>({
        parent() {
            return this.transactionBlock;
        },
        title: 'Transaction',
        isTransient: true,
        isMandatory: true,
        async onChange() {
            if (this.transaction.value) {
                this.employeeId.isDisabled = false;
                this.workOrder.isDisabled = false;
                this.employeeId.isMandatory = await this.isEmployeeMandatory(
                    this.transactionsList,
                    this.transaction.value,
                );
                this.employeeId.isHidden = await this.isEmployeeHidden(this.transactionsList, this.transaction.value);
                this.unplannedOperations.isHidden = await this.isUnplannedOperationsHidden(
                    this.transactionsList,
                    this.transaction.value,
                );
                this.isUnplannedOperationsEnabled();
            } else {
                this.employeeId.isDisabled = true;
                this.workOrder.isDisabled = true;
                this.unplannedOperations.isHidden = true;
                this.transaction.focus();
            }
        },
    })
    transaction: ui.fields.Select;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    headerBlock: ui.containers.Block;

    @ui.decorators.textField<MobileTimeTracking>({})
    trackingNumber: ui.fields.Text<MobileTimeTracking>;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customTimeTrackingHeaderBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileTimeTracking, EmployeeId>({
        parent() {
            return this.headerBlock;
        },
        title: 'Employee ID',
        valueField: 'teamId',
        node: '@sage/x3-manufacturing-data/EmployeeId',
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        isDisabled: true,
        isTransient: true,
        isFullWidth: true,
        filter() {
            return {
                isActive: true,
            };
        },
        onChange() {
            if (this.employeeId.isMandatory) {
                if (this.employeeId.value) {
                    this.workOrder.isDisabled = false;
                    this.workOrder.focus();
                } else {
                    this.workOrder.isDisabled = true;
                }
            } else {
                this.workOrder.isDisabled = false;
            }
        },
        columns: [
            ui.nestedFields.numeric({
                bind: 'teamId',
            }),
            ui.nestedFields.text({
                bind: 'description',
                title: 'Description',
                isReadOnly: true,
            }),
        ],
    })
    employeeId: ui.fields.Reference;

    @ui.decorators.referenceField<MobileTimeTracking, WorkOrder>({
        parent() {
            return this.headerBlock;
        },
        title: 'Work order',
        valueField: 'number',
        node: '@sage/x3-manufacturing/WorkOrder',
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        isDisabled: false,
        isFullWidth: true,
        isTransient: true,
        filter() {
            return {
                productionSite: { code: this.productionSite?.value?.code ?? '' },
                orderStatus: { _in: ['firm'] },
                modeType: 'complete',
                workOrderSuspendedFlag: false,
            };
        },
        async onChange() {
            this.setOperationFieldsDisabledStatus();
            if (this.workOrder.value) {
                workOrderRoutingList = await this.getWorkOrderRoutingList();
                this.isUnplannedOperationsEnabled();
                operationList = await getWorkOrderOperationsList(this.workOrder.value.number, this);
                this.releaseUOM.value = await getReleaseUOM(this.workOrder.value.number, this);
                this.releaseTimeUnit.value = await this.getReleaseTimeUnit();
            }
        },
        columns: [
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/WorkOrder',
                bind: 'productionSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'number',
            }),

            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/WorkOrder',
                bind: 'routingCode',
                valueField: 'code',
            }),
            ui.nestedFields.text({
                bind: 'startDate',
                title: 'Start',
                isReadOnly: true,
                prefix: 'Start:',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/WorkOrder',
                bind: 'releasedRouting',
                valueField: { routing: { code: true } },
            }),
        ],
    })
    workOrder: ui.fields.Reference;

    @ui.decorators.textField<MobileTimeTracking>({
        parent() {
            return this.headerBlock;
        },
        title: 'Release unit',
        maxLength: 30,
        isTransient: true,
        isDisabled: true,
        isHidden: true,
    })
    releaseUOM: ui.fields.Text;

    @ui.decorators.textField<MobileTimeTracking>({
        parent() {
            return this.headerBlock;
        },
        title: 'Release time unit',
        maxLength: 30,
        isTransient: true,
        isDisabled: true,
        isHidden: true,
    })
    releaseTimeUnit: ui.fields.Text;

    @ui.decorators.checkboxField<MobileTimeTracking>({
        parent() {
            return this.headerBlock;
        },
        title: 'Unplanned operation tracking',
        isFullWidth: true,
        isHidden: true,
        isDisabled: true,
        isTransient: true,
        async onChange() {
            this.isUnplannedOperationsEnabled();
        },
    })
    unplannedOperations: ui.fields.Checkbox;

    @ui.decorators.referenceField<MobileTimeTracking, StandardOperations>({
        parent() {
            return this.headerBlock;
        },
        title: 'Standard operation',
        valueField: 'code',
        node: '@sage/x3-manufacturing/StandardOperations',
        placeholder: 'Scan or select...',
        isHidden: true,
        isDisabled: true,
        isAutoSelectEnabled: true,
        isFullWidth: false,
        isMandatory: false,
        isTransient: true,
        async onChange() {
            this.stdOperationDescription.value = this.stdOperation.value?.description ?? null;
        },
        filter() {
            return {
                site: {
                    code: this.productionSite.value ? this.productionSite.value.code : '',
                },
            };
        },
        columns: [
            ui.nestedFields.numeric({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: 'description',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/StandardOperations',
                bind: 'operationUom',
                valueField: 'code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/StandardOperations',
                bind: 'site',
                valueField: 'code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/StandardOperations',
                bind: 'mainWorkCenter',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'timeUnit',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'setupTime',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'runTime',
                isHidden: true,
            }),
        ],
    })
    stdOperation: ui.fields.Reference;

    @ui.decorators.referenceField<MobileTimeTracking, WorkOrderOperationLine>({
        parent() {
            return this.headerBlock;
        },
        title: 'Operation',
        valueField: 'operationNumber',
        node: '@sage/x3-manufacturing/WorkOrderOperationLine',
        placeholder: 'Scan or select...',
        isDisabled: true,
        isAutoSelectEnabled: true,
        isFullWidth: true,
        isTransient: true,
        async onChange() {
            if (!this.operationLine.value) {
                return;
            }

            const operationMileStoneType: string = await this.getOperationMilestoneType(
                this.operationLine.value?.operationNumber,
                this.workOrder.value?.number,
            );

            if (operationMileStoneType === 'none') {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_milestone_is_none',
                        'This operation will be tracked with a milestone.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
                return;
            }

            if (!(await validateWithDetails(this))) return;
            operationMethod = SelectedOperationMethod.RouterOperation;
            this.$.setPageClean();

            if (paramOperationTrackingControl === '1') {
                showMobileTimeTrackingDetails = true;
            } else {
                if (
                    (await this.isPreviousOperationTracked(
                        this.operationLine.value?.operationNumber,
                        this.workOrder.value?.number,
                        operationMileStoneType,
                    )) === false
                ) {
                    if (paramOperationTrackingControl === '2') {
                        // "Yes, without block" display warning message
                        this.$.showToast(
                            ui.localize(
                                '@sage/x3-manufacturing/pages__mobile_time_tracking___Previous_operation_is_not_tracked',
                                'The previous operation has not been tracked',
                            ),
                            { type: 'warning', timeout: 5000 },
                        );
                        showMobileTimeTrackingDetails = true;
                    }
                    if (paramOperationTrackingControl === '3') {
                        // "Yes with block" display blocking message
                        this.$.showToast(
                            ui.localize(
                                '@sage/x3-manufacturing/pages__mobile_time_tracking___Previous_operation_is_not_tracked',
                                'The previous operation has not been tracked',
                            ),
                            { type: 'error', timeout: 5000 },
                        );
                        showMobileTimeTrackingDetails = false;
                    }
                } else {
                    showMobileTimeTrackingDetails = true;
                }
            }
            if (showMobileTimeTrackingDetails === true) {
                this.addButton.isHidden = false;
                this.detailBlock.isHidden = false;
                this.setDetailDefaultValues();
                this.setStockTimeUnits();
                this.remainingQuantity.value =
                    (await this.getRemainingQuantity()) < 0 ? 0 : await this.getRemainingQuantity();
            }
        },

        filter() {
            return {
                workOrder: {
                    number: this.workOrder.value?.number,
                },
                expectedWorkCenter: {
                    type: { _ne: 3 },
                },
                workOrderScheduling: {
                    _atLeast: 1,
                    milestone: {
                        _ne: 'none',
                    },
                },
                operationStatus: { _ne: 6 },
            };
        },
        columns: [
            ui.nestedFields.numeric({
                bind: 'operationNumber',
            }),
            ui.nestedFields.text({
                bind: 'operationDescription',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/WorkOrderOperationLine',
                bind: 'expectedWorkCenter',
                valueField: 'code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing/WorkOrderOperationLine',
                bind: 'operationUnit',
                valueField: 'code',
            }),
            ui.nestedFields.numeric({
                bind: 'expectedQuantity',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'completedQuantity',
                title: 'Completed quantity',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'timeUnit',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'setupTime',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'runTime',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'stockOperationConversion',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'operationSplit',
                isHidden: true,
            }),
        ],
    })
    operationLine: ui.fields.Reference;

    @ui.decorators.textField<MobileTimeTracking>({
        parent() {
            return this.headerBlock;
        },
        title: 'Description',
        maxLength: 30,
        isTransient: true,
        isHidden: true,
        isDisabled: true,
        isFullWidth: true,
        placeholder: 'Enter...',
    })
    stdOperationDescription: ui.fields.Text;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.headerBlock;
        },
        title: 'Operation',
        isHidden: true,
        validation: /^\d*[0-9](|\d*[0-9]|,\d*[0-9])?$/,
        isNotZero: true,
        max: 9999,
        isDisabled: true,
        isMandatory: true,
        isTransient: true,
        async onChange() {
            if (!(await validateWithDetails(this))) {
                this.unplannedOperationsLine.value = null;
                return;
            }
            if (operationList.find(ope => ope.operationNumber == this.unplannedOperationsLine.value)) {
                const filteredNumbers = operationList.filter(
                    ope => ope.operationNumber == this.unplannedOperationsLine.value,
                );
                if (filteredNumbers.find(ope => ope.operationStatus === 'excluded')) {
                    this.$.dialog.message(
                        'error',
                        'Field error "Operation"',
                        `${ui.localize(
                            '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_excluded',
                            'Excluded line.',
                        )}`,
                    );
                } else {
                    this.$.dialog.message(
                        'error',
                        'Field error "Operation"',
                        `${ui.localize(
                            '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_exists',
                            'This operation is already planned.',
                        )}`,
                    );
                }
            } else {
                if (this.stdOperation.value === null) {
                    operationMethod = SelectedOperationMethod.UserDefinedOperation;
                } else {
                    operationMethod = SelectedOperationMethod.StandardOperationWithData;
                }
                this.$.setPageClean();
                this.addButton.isHidden = false;
                this.detailBlock.isHidden = false;
                this.setDetailDefaultValues();
                this.setStockTimeUnits();
            }
        },
    })
    unplannedOperationsLine: ui.fields.Numeric;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    detailBlock: ui.containers.Block;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
    })
    customTimeTrackingDetailBlock: ui.containers.Block;

    @ui.decorators.referenceField<MobileTimeTracking, WorkCenter>({
        parent() {
            return this.detailBlock;
        },
        title: 'Actual work center',
        valueField: 'code',
        node: '@sage/x3-manufacturing-data/WorkCenter',
        isTransientInput: false,
        isTransient: true,
        placeholder: 'Scan or select...',
        isAutoSelectEnabled: true,
        isFullWidth: true,
        isMandatory: true,
        filter() {
            return {
                manufacturingSite: { code: this.productionSite.value ? this.productionSite.value.code : '' },
                type: { _ne: 3 },
            };
        },
        onChange() {
            this.remainingQuantity.focus();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'description',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-manufacturing-data/WorkCenterGroup',
                bind: 'workCenterGroup',
                valueField: 'code',
            }),
            ui.nestedFields.text({
                bind: 'type',
                isReadOnly: true,
            }),
        ],
    })
    expectedWorkCenter: ui.fields.Reference;

    /*     @ui.decorators.separatorField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
    })
    blank: ui.fields.Separator; */

    @ui.decorators.textField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Time unit',
        isDisabled: true,
        isTransient: true,
    })
    timeUnit: ui.fields.Text;

    @ui.decorators.textField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Unit',
        isDisabled: true,
        isTransient: true,
    })
    stockUnit: ui.fields.Text;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Quantity',
        isMandatory: true,
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        async onChange() {
            if (this.remainingQuantity.value === undefined || this.remainingQuantity.value === null) {
                this.remainingQuantity.value = 0;
                await validateWithDetails(this);
            }

            if (operationMethod === 0) {
                this.setCloseField();
                if (this.remainingQuantity.value > (await this.getRemainingQuantity())) {
                    this.$.showToast(
                        ui.localize(
                            '@sage/x3-manufacturing/pages__mobile_time_tracking_details___quantity__larger__expected__warning',
                            'The quantity is larger than expected.',
                        ),
                        { type: 'warning', timeout: 5000 },
                    );
                }
            }
            this.trackingLinesBlock.isHidden = false;
        },
    })
    remainingQuantity: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Rejected qty.',
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        async onChange() {
            // if user blanks out value set to zero and re-run screen validation
            if (this.rejectedQuantity.value === undefined || this.rejectedQuantity.value === null) {
                this.rejectedQuantity.value = 0;
                await validateWithDetails(this);
            }
            if (this.remainingQuantity.value === undefined || this.remainingQuantity.value === null) {
                this.remainingQuantity.value = 0;
            }
            if (this.rejectedQuantity.value > this.remainingQuantity.value) {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking_details___rejected_qty_error',
                        'Enter a rejected quantity that matches or is less than the quantity value.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
        },
    })
    rejectedQuantity: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Actual setup time',
        isHidden: true,
        isDisabled: true,
        isTransient: true,
        scale: 4,
    })
    setupTime: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Actual run time',
        isHidden: true,
        isDisabled: true,
        isTransient: true,
        scale: 4,
    })
    runTime: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Setup time hours',
        scale: 4,
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        async onChange() {
            // if user blanks out value set to zero and re-run screen validation
            if (this.enteredSetupTimeHours.value === undefined || this.enteredSetupTimeHours.value === null) {
                this.enteredSetupTimeHours.value = 0;
                await validateWithDetails(this);
            }
        },
    })
    enteredSetupTimeHours: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Setup time minutes',
        scale: 0,
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        async onChange() {
            if (this.enteredSetupTimeMinutes.value === undefined || this.enteredSetupTimeMinutes.value === null) {
                this.enteredSetupTimeMinutes.value = 0;
                await validateWithDetails(this);
            }

            if (this.enteredSetupTimeMinutes.value >= 60) {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking_details___minutes__entered__warning',
                        'Minutes are equal to or greater than 60.',
                    ),
                    { type: 'warning', timeout: 5000 },
                );
            }
        },
    })
    enteredSetupTimeMinutes: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Run time hours',
        scale: 4,
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        async onChange() {
            if (this.enteredRunTimeHours.value === undefined || this.enteredRunTimeHours.value === null) {
                this.enteredRunTimeHours.value = 0;
                await validateWithDetails(this);
            }
        },
    })
    enteredRunTimeHours: ui.fields.Numeric;

    @ui.decorators.numericField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Run time minutes',
        isMandatory: false,
        validation: /^\d*[0-9](|.\d*[0-9]|,\d*[0-9])?$/,
        isTransient: true,
        scale: 0,
        async onChange() {
            if (this.enteredRunTimeMinutes.value === undefined || this.enteredRunTimeMinutes.value === null) {
                this.enteredRunTimeMinutes.value = 0;
                await validateWithDetails(this);
            }

            if (this.enteredRunTimeMinutes.value >= 60) {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking_details___minutes__entered__warning',
                        'Minutes are equal to or greater than 60.',
                    ),
                    { type: 'warning', timeout: 5000 },
                );
            }
        },
    })
    enteredRunTimeMinutes: ui.fields.Numeric;

    @ui.decorators.checkboxField<MobileTimeTracking>({
        parent() {
            return this.detailBlock;
        },
        title: 'Close',
        isTransient: true,
        onChange() {
            if (this.close.value === null || this.close.value === undefined) {
                this.setCloseField();
            }
        },
    })
    close: ui.fields.Checkbox;

    @ui.decorators.block<MobileTimeTracking>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isHidden: true,
    })
    trackingLinesBlock: ui.containers.Block;

    @ui.decorators.tableField<MobileTimeTracking, TimeTrackingLine>({
        parent() {
            return this.trackingLinesBlock;
        },
        title: 'Operations',
        canResizeColumns: true,
        canSelect: false,
        canFilter: false,
        cardView: true,

        isFullWidth: true,
        isReadOnly: true,
        mobileCard: {
            title: ui.nestedFields.text({
                bind: 'workOrderNumber',
                title: 'Work order',
            }),

            titleRight: ui.nestedFields.numeric({
                bind: 'operationNumber',
                title: 'Operation',
            }),
            line2: ui.nestedFields.text({
                bind: 'operationDescription' as any,
                isTransient: true,
            }),
            line2Right: ui.nestedFields.numeric({
                bind: 'totalCompletedQuantity',
                title: 'Quantity',
                postfix(_value, rowValue) {
                    return rowValue?.operationUnit?.toString();
                },
                scale(_value, rowValue) {
                    return rowValue?.operationUnit?.numberOfDecimals ?? 0;
                },
            }),
            line3: ui.nestedFields.numeric({
                bind: 'actualSetupTime',
                title: 'Setup time',
                postfix(_value, rowValue) {
                    return rowValue?.timeUnit;
                },
                scale(_value, rowValue) {
                    return rowValue?.timeUnit === 'minutes' ? 0 : 4;
                },
            }),
            line3Right: ui.nestedFields.numeric({
                bind: 'actualRunTime',
                title: 'Run time',
                postfix(_value, rowValue) {
                    return rowValue?.timeUnit;
                },
                scale(_value, rowValue) {
                    return rowValue?.timeUnit === 'minutes' ? 0 : 4;
                },
            }),
        },
        columns: [
            ui.nestedFields.text({
                bind: 'workOrderNumber',
                title: 'Work order',
            }),
            ui.nestedFields.numeric({
                bind: 'employeeId',
                title: 'Employee ID',
                isHidden: true,
            }),

            ui.nestedFields.numeric({
                bind: 'operationNumber',
                title: 'Operation',
            }),

            ui.nestedFields.text({
                bind: 'operationDescription' as any,
                title: 'Description',
                isTransient: true,
            }),

            ui.nestedFields.numeric({
                bind: 'enteredSetupTimeHours' as any,
                isHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.numeric({
                bind: 'enteredSetupTimeMinutes' as any,
                isHidden: true,
                isTransient: true,
            }),
            ui.nestedFields.numeric({
                bind: 'enteredRunTimeHours' as any,
                isHidden: true,
                isTransient: true,
            }),

            ui.nestedFields.numeric({
                bind: 'expectedQuantity' as any,
                isHidden: true,
                isTransient: true,
            }),

            ui.nestedFields.numeric({
                bind: 'completedQuantity' as any,
                isHidden: true,
                isTransient: true,
            }),

            ui.nestedFields.numeric({
                bind: 'totalCompletedQuantity',
                postfix(_value, rowValue) {
                    return rowValue?.operationUnit?.toString();
                },
                scale(_value, rowValue) {
                    return rowValue?.operationUnit?.numberOfDecimals ?? 0;
                },
            }),

            ui.nestedFields.numeric({
                bind: 'actualRejectedQuantity',
                isHidden: true,
            }),

            ui.nestedFields.numeric({
                bind: 'actualSetupTime',
                postfix(_value, rowValue) {
                    return rowValue?.timeUnit;
                },
                scale: 4,
            }),

            ui.nestedFields.numeric({
                bind: 'actualRunTime',
                postfix(_value, rowValue) {
                    return rowValue?.timeUnit;
                },
                scale: 4,
            }),
            ui.nestedFields.reference<MobileTimeTracking, TimeTrackingLine, RoutingHeader>({
                node: '@sage/x3-manufacturing/RoutingHeader',
                valueField: { routingCode: { code: true } },
                bind: 'routingCode',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileTimeTracking, TimeTrackingLine, Product>({
                node: '@sage/x3-master-data/Product',
                valueField: 'code',
                bind: 'routing',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'bom',
                isHidden: true,
                isTransientInput: true,
            }),
            ui.nestedFields.text({
                bind: 'bomCode',
                isHidden: true,
                isTransientInput: true,
            }),

            ui.nestedFields.reference<MobileTimeTracking, TimeTrackingLine, WorkCenter>({
                node: '@sage/x3-manufacturing-data/WorkCenter',
                valueField: 'code',
                bind: 'actualWorkCenter',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'operationConversion',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'operationSplit',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobileTimeTracking, TimeTrackingLine, Site>({
                node: '@sage/x3-system/Site',
                valueField: 'code',
                bind: 'productionSite',
                isHidden: true,
                columns: [
                    ui.nestedFields.text({
                        bind: 'code',
                    }),
                ],
            }),
            ui.nestedFields.numeric({
                bind: 'balance',
                isHidden: true,
                isTransientInput: true,
            }),
            ui.nestedFields.date({
                bind: 'postingDate',
                isHidden: true,
                isTransientInput: true,
            }),
            ui.nestedFields.text({
                bind: 'transaction',
                isHidden: true,
                isTransientInput: true,
            }),
        ],
        dropdownActions: [
            {
                icon: 'bin',
                title: 'Delete',
                isDestructive: true,
                onClick(rowId: string, rowData: any) {
                    this.workOrderOperationTrackingLines.removeRecord(rowId);

                    if (!this.workOrderOperationTrackingLines.value?.length) {
                        this.$.setPageClean();
                        this.initPage();
                    }
                },
            },
        ],
        inlineActions: [
            {
                icon: 'box_arrow_left',
                title: 'Edit line',

                onClick(rowId: string) {
                    this.workOrderOperationTrackingLines.openSidebar(rowId);
                    this.toggleCreateButton(this.workOrderOperationTrackingLines.value.length);
                    this.addButton.isHidden = true;
                    this.detailBlock.isHidden = true;
                },
            },
        ],
        sidebar: {
            async onRecordOpened(_id, recordValue?: TimeTrackingLine) {
                if (recordValue) {
                    this.remainingQuantity.value = Number(recordValue.totalCompletedQuantity);
                    this.rejectedQuantity.value = Number(recordValue.actualRejectedQuantity);
                    this.stockUnit.value = recordValue.operationUnit.toString();
                    this.enteredSetupTimeHours.value = recordValue.enteredSetupTimeHours ?? null;
                    this.enteredSetupTimeMinutes.value = recordValue.enteredSetupTimeMinutes ?? null;
                    this.enteredRunTimeHours.value = recordValue.enteredRunTimeHours ?? null;
                    this.enteredRunTimeMinutes.value = recordValue.enteredRunTimeMinutes ?? null;
                    this.timeUnit.value = recordValue.timeUnit;
                    this.close.value = recordValue.balance === '2' ? true : false;
                    this.operationLine.value = {
                        operationNumber: recordValue.operationNumber,
                        operationDescription: recordValue.operationDescription,
                        expectedQuantity:
                            (recordValue.expectedQuantity ?? 0) + Number(recordValue.totalCompletedQuantity),
                        completedQuantity: recordValue.completedQuantity,
                    };

                    this.workOrder.value = {
                        number: recordValue.workOrderNumber,
                    };
                    // Get the custom fields to display in the sidebar
                    this.getCustomTimeTrackingLines();
                }

                this.detailBlock.isHidden = true;
                this.addButton.isHidden = true;
                this.createButton.isHidden = false;
            },

            async onRecordDiscarded() {
                this.$.setPageClean();
            },

            async onRecordConfirmed(_, recordValue) {
                if (recordValue) {
                    recordValue.totalCompletedQuantity = this.remainingQuantity.value?.toString() ?? '';
                    recordValue.actualRejectedQuantity = this.rejectedQuantity.value?.toString() ?? '';
                    recordValue.enteredSetupTimeHours = this.enteredSetupTimeHours.value ?? undefined;
                    recordValue.enteredSetupTimeMinutes = this.enteredSetupTimeMinutes.value ?? undefined;
                    recordValue.enteredRunTimeHours = this.enteredRunTimeHours.value ?? undefined;
                    recordValue.enteredRunTimeMinutes = this.enteredRunTimeMinutes.value ?? undefined;
                    this.setActualRunTime();
                    this.setActualSetupTime();
                    (recordValue.actualSetupTime = this.setupTime.value?.toString() ?? ''),
                        (recordValue.actualRunTime = this.runTime.value?.toString() ?? ''),
                        this.setCloseField();
                    recordValue.balance = this.close?.value === true ? '2' : '1';
                    // Check time and quantity values
                    if ((await this.isTimeAndQuantityValid('sideBar')) === false) {
                        this.$.showToast(
                            ui.localize(
                                '@sage/x3-manufacturing/pages__mobile_time_tracking__qty_error',
                                'The operation did not update. Enter a quantity or a time.',
                            ),
                            { type: 'error', timeout: 5000 },
                        );
                        this.close.focus();
                        this.$.setPageClean();
                        return;
                    }
                    // To add the custom detail fields to the table ...this.getCustomTimeTrackingDetails());
                    recordValue = { ...recordValue, ...this.getCustomTimeTrackingDetails() };
                    this.workOrderOperationTrackingLines.addOrUpdateRecordValue(recordValue as any);
                    this.workOrder.isHidden = false;
                    this.workOrder.value = null;
                    this.operationLine.isHidden = false;
                    this.addButton.isHidden = true;
                    this.createButton.isHidden = false;
                    this.toggleCreateButton(this.workOrderOperationTrackingLines.value.length);
                    this.detailBlock.isHidden = true;
                    this.clearFields();
                    this.isUnplannedOperationsEnabled();
                    this.$.setPageClean();
                }
            },
            layout() {
                return {
                    mainSection: {
                        title: 'Operation',
                        blocks: {
                            operationBlock: {
                                fields: this.getCustomSidebarTimeTrackingLine(),
                            },
                        },
                    },
                };
            },
        },
    })
    workOrderOperationTrackingLines: ui.fields.Table<TimeTrackingLine>;

    @ui.decorators.pageAction<MobileTimeTracking>({
        title: 'Create',
        buttonType: 'primary',
        shortcut: ['f2'],
        isDisabled: true,
        async onClick() {
            if (this.createButton.isHidden) {
                return;
            }
            if (this.workOrderOperationTrackingLines.value?.length > 0) {
                this.$.loader.isHidden = false;
                await this.$standardSaveAction.execute(false);
                this.$.loader.isHidden = true;
                await this.$.sound.success();
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-manufacturing/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking__notification__document_created',
                        'Time tracking created: {{documentId}}.',
                        { documentId: this.trackingNumber.value },
                    ),
                    {
                        fullScreen: true,
                        acceptButton: { text: ui.localize('@sage/x3-manufacturing/button-accept-ok', 'OK') },
                    },
                );
                await this.$.router.emptyPage();
            } else {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking__no_operation_error',
                        'Enter at least one operation.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
        },
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-manufacturing/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-manufacturing/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);

                message = `**${ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking__notification__creation_error',
                    'An error occurred.',
                )}**\n\n`;

                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-manufacturing/pages_creation_error_connexion_webservice_contact_administrator',
                    'An error has occurred (connection or webservice error). Please contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-manufacturing/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
    })
    createButton: ui.PageAction;

    @ui.decorators.pageAction<MobileTimeTracking>({
        title: 'Add',
        buttonType: 'primary',
        shortcut: ['f3'],
        isHidden: true,
        async onClick() {
            if (this.addButton.isHidden) {
                return;
            }
            if (!(await validateWithDetails(this))) return;

            if (this.workOrder.value === null) {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking__work_order_error',
                        'Scan or select a work order.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
                return;
            }

            if (this.unplannedOperations.value === false && this.operationLine.value === null) {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_error',
                        'Scan or select an operation.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
                return;
            }
            if (this.unplannedOperations.value === true && this.unplannedOperationsLine.value === null) {
                return;
            }
            // Calculate the actual run time and setup time
            await this.setActualRunTime();
            await this.setActualSetupTime();

            // Check time and quantity values
            if ((await this.isTimeAndQuantityValid('addButton')) === false) {
                return;
            }
            // Get the BOM values, code and product for tracking line
            const bomValues = await getBomCode(this.workOrder.value?.number, this);
            const lineIndex = this.workOrderOperationTrackingLines.value.length;
            let trackingLine: Partial<TimeTrackingLine> = {
                _id: `${lineIndex}`,
                workOrderNumber: this.workOrder.value?.number,
                operationNumber: this.operationLine.value?.operationNumber ?? this.unplannedOperationsLine.value,
                operationDescription:
                    this.operationLine.value?.operationDescription ?? this.stdOperationDescription.value,
                employeeId: this.employeeId?.value?.teamId,
                postingDate: this.trackingDate.value ?? undefined,
                transaction: this.transaction.value ?? undefined,

                expectedQuantity: this.operationLine.value?.expectedQuantity,
                completedQuantity: this.operationLine.value?.completedQuantity,
                productionSite: this.productionSite?.value?.code,
                totalCompletedQuantity: this.remainingQuantity.value?.toString() ?? undefined,
                actualRejectedQuantity: this.rejectedQuantity.value?.toString() ?? undefined,
                operationUnit: this.stockUnit.value ?? this.operationLine.value?.operationUnit.code,
                timeUnit: this.timeUnit.value ? this.timeUnit.value : '',
                actualSetupTime: this.setupTime.value?.toString() ?? undefined,
                actualRunTime: this.runTime?.value !== undefined ? this.runTime?.value?.toString() : undefined,
                enteredSetupTimeHours: this.enteredSetupTimeHours.value ?? undefined,
                enteredSetupTimeMinutes: this.enteredSetupTimeMinutes.value ?? undefined,
                enteredRunTimeHours: this.enteredRunTimeHours.value ?? undefined,
                enteredRunTimeMinutes: this.enteredRunTimeMinutes.value ?? undefined,
                routingCode: this.workOrder.value?.routingCode.code,
                routing: this.workOrder.value?.releasedRouting?.routing.code,
                bom: bomValues.product.code,
                bomCode: bomValues.bomCode.code.toString(),
                actualWorkCenter: this.expectedWorkCenter.value?.code,
                operationConversion: this.operationLine.value?.stockOperationConversion,
                balance: this.close.value === true ? '2' : '1',
                operationSplit: this.operationLine.value?.operationSplit,
            };

            // Add the custom detail fields to the tracking line
            trackingLine = { ...trackingLine, ...this.getCustomTimeTrackingDetails() };
            this.workOrderOperationTrackingLines.addOrUpdateRecordValue(trackingLine as any);
            this.trackingLinesBlock.isHidden = false;
            this.transaction.isDisabled = true;
            this.employeeId.isDisabled = true;
            this.workOrder.isDirty = !this.workOrder.isDirty;
            this.addButton.isHidden = true;
            this.createButton.isHidden = false;
            this.toggleCreateButton(this.workOrderOperationTrackingLines.value.length);
            this.detailBlock.isHidden = true;
            this.clearFields();
            this.isUnplannedOperationsEnabled();
        },
    })
    addButton: ui.PageAction;

    /** Functions Validation and Fields Management */

    async initPage() {
        this.transactionsList = await getEntryTransaction(this, 60);
        this.transactionsList.length === 1 ? (this.transaction.isHidden = true) : (this.transaction.isHidden = false);
        await this.setMandatoryFields(this.transactionsList);
        this.detailBlock.isHidden = true;
        this.trackingLinesBlock.isHidden = true;
        this.addButton.isHidden = true;

        //custom blocks display
        this.customTimeTrackingHeaderBlock.isHidden =
            this.customTimeTrackingHeaderBlock.layout.$layout?.$items.length === 0;
        this.customTimeTrackingDetailBlock.isHidden =
            this.customTimeTrackingDetailBlock.layout.$layout?.$items.length === 0;
    }

    async setMandatoryFields(transactions: any[]): Promise<boolean | undefined> {
        let transactionIsDisabled: boolean;

        const selectedStockSite = await getSelectedStockSite(this);
        if (selectedStockSite) {
            this.productionSite.value = {
                code: selectedStockSite,
            };
        }
        if (this.productionSite.value) {
            this.trackingDate.value = await getTrackingDate();
            if (!transactions || transactions.length === 0) {
                this.transactionBlock.isDisabled = true;
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking___no_transaction_error',
                        'No transactions were found.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
                return (this.transactionBlock.isDisabled = true);
            }
            transactionIsDisabled = transactions.length < 1;

            if (transactionIsDisabled) {
                this.setEntryTransaction(transactions, transactionIsDisabled, true);
                return;
            } else {
                this.setEntryTransaction(transactions, transactionIsDisabled, false);
            }

            this.employeeId.isMandatory = await this.isEmployeeMandatory(transactions, this.transaction?.value ?? '');
            this.employeeId.isHidden = await this.isEmployeeHidden(transactions, this.transaction?.value ?? '');
            this.unplannedOperations.isHidden = await this.isUnplannedOperationsHidden(
                transactions,
                this.transaction?.value ?? '',
            );
        } else {
            this.transactionBlock.isDisabled = true;
            this.$.dialog.message(
                'error',
                'Error',
                ui.localize(
                    '@sage/x3-manufacturing/pages__default_stock_site_error',
                    'Define a default stock site on the user function profile.',
                ),
            );
            return (this.transactionBlock.isDisabled = true);
        }
    }

    setEntryTransaction(transaction: any[], transactionIsDisabled: boolean, allFieldsDisabled: boolean) {
        const transactionOptions: string[] = transaction.map(transactions => transactions.transaction);
        this.transaction.options = transactionOptions;
        this.transaction.value = transaction[0].transaction;
        this.transaction.isDisabled = transactionIsDisabled;
        this.employeeId.isDisabled = allFieldsDisabled;
    }

    /**
     * Is the employee id mandatory
     * @param transactions
     * @param currentTransaction
     * @returns boolean
     */

    async isEmployeeMandatory(transactions: TransactionList[], currentTransaction: string): Promise<boolean> {
        const transactionIndex = transactions.findIndex(transaction => transaction.transaction === currentTransaction);

        switch (transactions[transactionIndex].employeeEntry) {
            case 'mandatory':
                this.employeeId.value = { teamId: await this.setDefaultEmployee() };
                this.workOrder.isDisabled = !this.employeeId.value;
                !this.employeeId.value ? (this.workOrder.isDisabled = true) : (this.workOrder.isDisabled = false);
                return true;

            case 'optional':
                this.employeeId.value = { teamId: await this.setDefaultEmployee() };

                return false;

            case 'prohibited':
                return false;

            default:
                return false;
        }
    }

    /**
     * Is the employee id field to be hidden
     * @param transactions
     * @param currentTransaction
     * @returns boolean
     */
    async isEmployeeHidden(transactions: TransactionList[], currentTransaction: string): Promise<boolean> {
        const transaction = transactions.find(transaction => transaction.transaction === currentTransaction);

        if (transaction?.employeeEntry === 'prohibited') {
            this.employeeId.value = null;
            return true;
        }

        return false;
    }

    /** Function returns whether unplanned operations checkbox is hidden based on unplanned operation flag
        in the entry transaction - True | False */
    async isUnplannedOperationsHidden(transactions: TransactionList[], currentTransaction: string): Promise<boolean> {
        const transactionIndex: number = transactions.findIndex(
            transaction => transaction.transaction === currentTransaction,
        );
        let isUnplannedOperationHiddenFlag: boolean = true;

        if (transactions[transactionIndex].unplannedOperations === true) {
            isUnplannedOperationHiddenFlag = false;
        } else {
            this.unplannedOperations.value = false;
        }

        return isUnplannedOperationHiddenFlag;
    }

    setOperationFieldsDisabledStatus(): void {
        if (this.workOrder.value) {
            this.unplannedOperations.isDisabled = false;
            this.operationLine.isDisabled = false;
            this.unplannedOperationsLine.isDisabled = false;
            this.stdOperation.isDisabled = false;
            this.stdOperationDescription.isDisabled = false;
            this.unplannedOperations.isHidden === true ? this.operationLine.focus() : this.unplannedOperations.focus();
        } else {
            this.unplannedOperations.isDisabled = true;
            this.unplannedOperations.value = false;
            this.operationLine.isDisabled = true;
            this.operationLine.value = null;
            this.unplannedOperationsLine.isDisabled = true;
            this.unplannedOperationsLine.value = null;
            this.stdOperation.isDisabled = true;
            this.stdOperation.value = null;
            this.stdOperationDescription.isDisabled = true;
            this.stdOperationDescription.value = null;
        }
    }

    isUnplannedOperationsEnabled(): void {
        if (this.unplannedOperations.value === true) {
            this.operationLine.isHidden = true;
            this.operationLine.value = null;
            this.unplannedOperationsLine.isHidden = false;
            this.unplannedOperationsLine.isDisabled = false;
            this.stdOperation.isHidden = false;
            this.stdOperation.isDisabled = false;
            this.stdOperationDescription.isHidden = false;
            this.stdOperationDescription.isDisabled = false;
        } else if (this.unplannedOperations.value === false) {
            this.operationLine.isHidden = false;
            this.operationLine.isDisabled = false;
            this.unplannedOperationsLine.isHidden = true;
            this.unplannedOperationsLine.value = null;
            this.stdOperation.isHidden = true;
            this.stdOperation.value = null;
            this.stdOperationDescription.isHidden = true;
            this.stdOperationDescription.value = null;
        }
    }

    /**
     * Function to disable create button if no lines are present in the table
     * @param linesLength
     * @returns boolean
     */

    toggleCreateButton(linesLength: number): boolean {
        if (linesLength > 0) {
            return (this.createButton.isDisabled = false);
        } else {
            return (this.createButton.isDisabled = true);
        }
    }

    /*
     * set current employee if not already set
     */

    async setCurrentEmployee() {
        if (this.employeeId.value === null) {
            this.employeeId.value = { teamId: await this.setDefaultEmployee() };
        }
    }

    async getDate(): Promise<string> {
        return DateValue.today().toString();
    }

    /** Function to set the default employee for the user */
    async setDefaultEmployee(): Promise<number | undefined> {
        try {
            const employeeIDList = await this.$.graph
                .node('@sage/x3-manufacturing-data/EmployeeId')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            teamId: true,
                        },
                        {
                            filter: {
                                isActive: true,
                                userCode: this.$.userCode,
                            },
                        },
                    ),
                )
                .execute();

            return employeeIDList.edges[0]?.node.teamId;
        } catch (err) {
            return undefined;
        }
    }

    async getWorkOrderRoutingList(): Promise<ExtractEdges<RoutingHeader>[]> {
        try {
            const woRoutingList = await this.$.graph
                .node('@sage/x3-manufacturing/RoutingHeader')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            timeUnit: true,
                        },
                        {
                            filter: {
                                routing: { code: this.workOrder.value?.releasedRouting.routing.code },
                                routingCode: { code: this.workOrder.value?.routingCode.code },
                                site: this.productionSite.value,
                            },
                            first: 1000,
                        },
                    ),
                )
                .execute();

            return extractEdges(woRoutingList.edges[0].node.timeUnit);
        } catch (err) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___cannot_retrieve_work_order_routing',
                    'Error while loading the work order routing.',
                ),
                { type: 'error', timeout: 5000 },
            );
            return [];
        }
    }

    /** function to test if operation can be tracked - tracking is only permitted if previous operation
     *  has already been tracked and exists X3 or has been added to the grid pending creation */
    async isPreviousOperationTracked(
        operationNumber: number,
        workOrder: any,
        operationMileStoneType: string | undefined,
    ): Promise<boolean> {
        // If this is the first operation allow user to track as no previous operation will exist
        if (operationMileStoneType === 'normalTracking') {
            if ((await this.getPreviousOperation(operationNumber, workOrder)) === undefined) {
                return true;
            }
        }

        if (operationMileStoneType === 'range') {
            // If this is the first operation allow user to track as no previous operation will exist
            if ((await this.getPreviousOperation(operationNumber, workOrder)) === undefined) {
                return true;
            }

            let startFromOperation: number | undefined = operationNumber;
            let previousOperation: number | undefined;
            let result: string | undefined;

            do {
                previousOperation = await this.getPreviousOperation(startFromOperation, workOrder);
                // If the first operation in the routing has been reached and no
                // previous operation exists that can be tracked
                if (previousOperation === undefined) {
                    previousOperation = startFromOperation;
                    result = await this.getOperationMilestoneType(previousOperation, workOrder);
                    break;
                }
                result = await this.getOperationMilestoneType(previousOperation, workOrder);
                startFromOperation = previousOperation;
            } while (result === 'none');

            operationNumber = previousOperation;
            // When the first operation in the routing has been reached and type is = none
            if (result === 'none') {
                return true;
            }
        }

        // Test if previous operation has already been already been tracked and exists in X3
        const allowTracking = await this.isPreviousOperationTrackedDB(operationNumber, operationList);
        if (allowTracking === true) {
            return true;
        }

        // if operation not in X3 then test if operation has been added to mobile trackuing grid waiting to be processed
        if (allowTracking === false) {
            if (operationMileStoneType === 'normalTracking') {
                return await this.isOperationOnTrackingGrid(
                    await this.getPreviousOperation(operationNumber, workOrder),
                    workOrder,
                );
            }
            if (operationMileStoneType === 'range') {
                return await this.isOperationOnTrackingGrid(operationNumber, workOrder);
            }
        }
        return false;
    }

    /** function to test if an operation number exists on the mobile page 1 tracking grid */

    async isOperationOnTrackingGrid(operationNumber: number | undefined, workOrderNumber: number): Promise<boolean> {
        const trackingLines: PartialCollectionValueWithIds<TimeTrackingLine>[] =
            this.workOrderOperationTrackingLines.value ?? [];

        if (trackingLines === undefined || trackingLines === null) {
            return false;
        }

        if (
            trackingLines.find(
                ope => ope.operationNumber === operationNumber && ope.workOrderNumber === String(workOrderNumber),
            )
        ) {
            return true;
        }

        return false;
    }

    /** function to test if the previous operation for the current operation has been tracking in X3 */
    async isPreviousOperationTrackedDB(
        opNumber: number,
        opList: ExtractEdges<WorkOrderOperationLine>[],
    ): Promise<boolean | undefined> {
        try {
            let result: boolean = false;
            for (let i = 0; i <= opList.length - 1; i++) {
                if (opList[i]?.operationNumber === opNumber) {
                    if (opList[i]?.operationStatus !== 'pending') {
                        result = true;
                        break;
                    }
                    if (opList[i]?.operationStatus === 'pending' && i !== 0) {
                        break;
                    }
                }
            }
            return result;
        } catch (err) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_tracking_status_error',
                    'Unable to check operation tracking status',
                ),
                { type: 'error', timeout: 5000 },
            );
            return undefined;
        }
    }

    infoDialogOption(): ui.dialogs.DialogOptions {
        const options = {
            acceptButton: {
                text: 'Yes',
            },
            cancelButton: {
                text: 'No',
            },
        };
        return options;
    }

    successDialogOption(): ui.dialogs.DialogOptions {
        const options = {
            acceptButton: {
                text: 'OK',
            },
        };
        return options;
    }

    errorDialogOption(): ui.dialogs.DialogOptions {
        const options = {
            acceptButton: {
                text: 'Back',
            },
            cancelButton: {
                text: 'Reset',
            },
        };
        return options;
    }

    /** function to return the milestone type for an operation */
    async getOperationMilestoneType(operation: number, woNumber: string): Promise<string> {
        try {
            const milestoneType = await this.$.graph
                .node('@sage/x3-manufacturing/WorkOrderScheduling')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            milestone: true,
                        },
                        {
                            filter: {
                                operationNumber: operation,
                                workOrder: woNumber,
                            },
                        },
                    ),
                )
                .execute();

            return milestoneType.edges[0].node.milestone;
        } catch (err) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___operation_milestone_type_error',
                    'Error while checking the operation milestone type.',
                ),
                { type: 'error', timeout: 5000 },
            );
            return '';
        }
    }

    /** Function to get release time unit for selected work order **/
    async getReleaseTimeUnit(): Promise<string> {
        try {
            const result = await this.$.graph
                .node('@sage/x3-manufacturing/RoutingHeader')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            timeUnit: true,
                        },
                        {
                            filter: {
                                routing: { code: this.workOrder.value?.releasedRouting.routing.code },
                                routingCode: { code: this.workOrder.value?.routingCode.code },
                                site: this.productionSite.value,
                            },
                            first: 1000,
                        },
                    ),
                )
                .execute();

            return result.edges[0].node.timeUnit;
        } catch (err) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___loading_rel_time_unit_code_error',
                    'Error while loading the release time unit.',
                ),
                { type: 'error', timeout: 5000 },
            );
            return '';
        }
    }

    /** Function to get release unit for selected work order **/

    async getReleaseUOM(): Promise<string> {
        try {
            const result = await this.$.graph
                .node('@sage/x3-manufacturing/WorkOrderProductLine')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            releaseUnit: { code: true },
                        },
                        {
                            filter: {
                                workOrder: this.workOrder.value?.number,
                            },
                        },
                    ),
                )

                .execute();

            return result.edges[0].node.releaseUnit.code.toString();
        } catch (err) {
            this.$.dialog.message(
                'error',
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_tracking___loading_rel_BOM_code_error',
                    'Error loading the release unit.',
                ),
                String(err),
            );
            return '';
        }
    }

    /** Function to return the previous operation number */
    async getPreviousOperation(operation: number, woNumber: string): Promise<number | undefined> {
        try {
            const previousOperation = await this.$.graph
                .node('@sage/x3-manufacturing/WorkOrderScheduling')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            operationNumber: true,
                        },
                        {
                            filter: {
                                downstreamOperationNumber: operation,
                                workOrder: woNumber,
                            },
                        },
                    ),
                )
                .execute();

            return previousOperation.edges[0].node.operationNumber;
        } catch (err) {
            return undefined;
        }
    }

    /** function verify if the time and quantity values enter are permitted - True | False
     * @param calledFrom
     * @returns boolean
     */
    async isTimeAndQuantityValid(calledFrom: string): Promise<boolean> {
        let result: boolean = true;
        const totalTime = (this.runTime.value ?? 0) + (this.setupTime.value ?? 0);
        const totalQuantity = (this.remainingQuantity.value ?? 0) - (this.rejectedQuantity.value ?? 0);

        if (totalQuantity < 0) {
            if (calledFrom !== 'sideBar') {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking___rejected_qty_error',
                        'Enter a rejected quantity that matches or is less than the quantity value.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
            result = false;
        }

        if ((this.remainingQuantity.value ?? 0) + totalTime === 0) {
            if (calledFrom !== 'sideBar') {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-manufacturing/pages__mobile_time_tracking___invalid__quantity__time___values__error',
                        'Enter a quantity or a time.',
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
            result = false;
        }

        if (totalTime === 0 && totalQuantity > 0) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___no__time__entered__warning',
                    'No time entered.',
                ),
                { type: 'warning', timeout: 5000 },
            );
            result = true;
        }

        return result;
    }

    /** function to set the Close field - yes/no based on the quantity being tracked */
    async setCloseField(): Promise<void> {
        const operationLine: Partial<OperationLine> = this.operationLine.value ?? {};
        const workOrderOperationTrackingLines: PartialCollectionValueWithIds<TimeTrackingLine>[] =
            this.workOrderOperationTrackingLines.value ?? [];

        const workOrderNumber = this.workOrder.value?.number;
        const remainingQuantity = this.remainingQuantity.value ?? 0;
        const completedQuantity = operationLine.completedQuantity ?? 0;
        const expectedQuantity = operationLine.expectedQuantity ?? 0;

        let automaticClosingPercentage: decimal = 100;
        let completedQuantityPercentage: decimal = 0;

        // Total quantity = quantity of tracked lines + quantity to track + completed quantity
        const totalQuantity =
            (await this.qtyTrackedOnDeviceOnly(
                (workOrderOperationTrackingLines as unknown as PartialCollectionValueWithIds<TimeTrackingLine>[]) ?? [],
                workOrderNumber,
                operationLine.operationNumber ?? 0,
            )) +
            remainingQuantity +
            completedQuantity;

        if (totalQuantity > 0 && expectedQuantity > 0) {
            completedQuantityPercentage = (totalQuantity / (expectedQuantity ?? 1)) * 100;
        }

        if (this.expectedWorkCenter.value) {
            automaticClosingPercentage = await this.getAutomaticClosingPercent(
                this.expectedWorkCenter.value.code,
                (await getSelectedStockSite(this)) ?? '',
            );
        }

        if (automaticClosingPercentage != 0) {
            if (completedQuantityPercentage >= automaticClosingPercentage) {
                this.close.value = true;
            } else {
                this.close.value = false;
            }
        } else {
            this.close.value = false;
        }
    }

    /**
     * clearFields()
     */

    clearFields() {
        this.expectedWorkCenter.value = null;
        this.timeUnit.value = null;
        this.stockUnit.value = null;
        this.operationLine.value = null;
        this.unplannedOperationsLine.value = null;
        this.stdOperation.value = null;
        this.stdOperationDescription.value = null;
        this.remainingQuantity.value = null;
        this.rejectedQuantity.value = null;
        this.setupTime.value = null;
        this.runTime.value = null;
        this.enteredSetupTimeHours.value = null;
        this.enteredSetupTimeMinutes.value = null;
        this.enteredRunTimeHours.value = null;
        this.enteredRunTimeMinutes.value = null;
        this.close.value = null;
        this.unplannedOperations.value = false;
    }

    /** Initialize-set detail block field values */
    async setDetailDefaultValues() {
        this.rejectedQuantity.value = 0;
        this.enteredSetupTimeMinutes.value = 0;
        this.enteredSetupTimeHours.value = 0;
        this.enteredRunTimeMinutes.value = 0;
        this.enteredRunTimeHours.value = 0;
        if (operationMethod === 0) {
            await this.setWorkCenter(this.operationLine?.value?.expectedWorkCenter.code);
            this.remainingQuantity.value =
                (await this.getRemainingQuantity()) < 0 ? 0 : await this.getRemainingQuantity();
            this.runTime.value = this.operationLine?.value?.runTime;
            this.setupTime.value = this.operationLine?.value?.setupTime;
            this.close.isDisabled = false;
            this.setCloseField();
        }
        if (operationMethod !== 0) {
            if (operationMethod === 1) {
                await this.setWorkCenter(this.stdOperation?.value?.mainWorkCenter.code);
            }
            this.setupTime.value = Number(this.stdOperation?.value?.setupTime);
            this.runTime.value = Number(this.stdOperation?.value?.runTime);
            this.close.value = true;
            this.close.isDisabled = true;
            this.remainingQuantity.value = 1;
        }
    }

    /**
     * Set the stock and time units based on the operation method
     *
     */
    async setStockTimeUnits() {
        //   RouterOperation is 0
        //   StandardOperationWithData is 1
        //   StandardOperationNoData is 2
        //   UserDefinedOperation is 3
        if (operationMethod === 0) {
            this.stockUnit.value = this.operationLine?.value?.operationUnit.code;
            this.timeUnit.value = this.operationLine?.value?.timeUnit;
        }
        if (operationMethod === 1) {
            try {
                this.stockUnit.value = this.stdOperation?.value?.operationUom.code;
            } catch (error) {
                this.stockUnit.value = this.releaseUOM.value;
            }
            this.timeUnit.value = this.stdOperation?.value?.timeUnit;
        }
        if (operationMethod === 2) {
            this.stockUnit.value = this.releaseUOM.value;
            this.timeUnit.value = this.releaseTimeUnit.value;
            this.remainingQuantity.value = 1;
            this.setScaleForNumericFields(this.stockUnit?.value ?? '');
        }
    }

    /** Set the scale for numeric field based on their unit of measure */
    async setScaleForNumericFields(stockUnit: string) {
        const numberOfDecimalPlaces = await GetUOMNumberOfDecimalPlaces(this, stockUnit);
        this.remainingQuantity.scale = numberOfDecimalPlaces;
        this.rejectedQuantity.scale = numberOfDecimalPlaces;
    }

    /** Set the work centre */
    async setWorkCenter(workCenter: string) {
        try {
            this.expectedWorkCenter.value = extractEdges(
                await this.$.graph
                    .node('@sage/x3-manufacturing-data/WorkCenter')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                code: true,
                            },
                            {
                                filter: {
                                    code: workCenter,
                                },
                            },
                        ),
                    )
                    .execute(),
            );

            if (this.expectedWorkCenter.value) {
                this.expectedWorkCenter.value.code = workCenter;
            }
        } catch (error) {
            this.$.dialog.message(
                'error',
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_completed_quantity_details___loading_work_centre_error',
                    'Error loading work center.',
                ),
                String(error),
            );
        }
    }

    /** set the remaining quantity value. Expected qty minus the completed qty
     *  if completed exceeds expected then set qty to zero */
    public async getRemainingQuantity(): Promise<number> {
        const operationLine: Partial<OperationLine> = this.operationLine.value ?? {};
        const workOrderOperationTrackingLines: PartialCollectionValueWithIds<TimeTrackingLine>[] =
            this.workOrderOperationTrackingLines.value ?? [];
        const workOrder: string = this.workOrder?.value?.number ?? '';

        let remainingQuantity: number = 0;
        if (operationLine !== null && operationLine !== undefined) {
            remainingQuantity =
                (await this.remainingQtyInX3Database(
                    operationLine?.expectedQuantity ?? 0,
                    operationLine?.completedQuantity ?? 0,
                )) -
                (await this.qtyTrackedOnDeviceOnly(
                    workOrderOperationTrackingLines as PartialCollectionValueWithIds<TimeTrackingLine>[],
                    workOrder,
                    operationLine.operationNumber ?? 0,
                ));
        }
        return remainingQuantity;
    }

    /**Calculate the remaining quantity in X3 database
     * @param expectedQuantity
     * @param completedQuantity
     * @returns number
     */
    async remainingQtyInX3Database(expectedQuantity: number, completedQuantity: number): Promise<number> {
        const remainingQtyInX3Database = expectedQuantity - completedQuantity;
        return remainingQtyInX3Database;
    }

    /** Calculate the total quantities tracked for the work order - operation
     * @param trackingLines
     * @param workOrder
     * @param operationNumber
     * @returns number
     */

    async qtyTrackedOnDeviceOnly(
        trackedLines: PartialCollectionValueWithIds<TimeTrackingLine>[],
        workOrderNumber: string,
        operationNumber: number,
    ): Promise<number> {
        const qtyTrackedOnDeviceOnly: number = trackedLines
            .filter(line => {
                return line.workOrderNumber === workOrderNumber && line.operationNumber === operationNumber;
            }) // filter the lines for the work order and operation
            .reduce((total, line) => total + Number(line.totalCompletedQuantity ?? 0), 0); // sum the total completed quantity
        return qtyTrackedOnDeviceOnly;
    }

    /** Function coverts the setup time entered by the user to operation time unit
     *  updates runtime page field with new value return: true | false was calculation successfully
     */
    async setActualRunTime(): Promise<boolean> {
        if ((this.enteredRunTimeHours?.value ?? 0) + (this.enteredRunTimeMinutes?.value ?? 0) === 0) {
            this.runTime.value = 0;
            return true;
        }

        try {
            if (this.timeUnit.value === 'hours') {
                const time = Math.floor((this.enteredRunTimeHours?.value ?? 0) * 60);
                this.runTime.value = ((this.enteredRunTimeMinutes?.value ?? 0) + time) / 60;
            }
            if (this.timeUnit.value === 'minutes') {
                const time = Math.floor((this.enteredRunTimeHours?.value ?? 0) * 60);
                this.runTime.value = (this.enteredRunTimeMinutes?.value ?? 0) + time;
            }
            return true;
        } catch (error) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___update__runtime__error',
                    'The actual run time calculation failed.',
                ),
                { type: 'error', timeout: 5000 },
            );
            return false;
        }
    }
    /** Function coverts the setup time entered by the user to operation time unit
     *  updates setuptime page field with new value return: true | fales was calculation succefully
     */
    async setActualSetupTime(): Promise<boolean> {
        if ((this.enteredSetupTimeHours?.value ?? 0) + (this.enteredSetupTimeMinutes?.value ?? 0) === 0) {
            this.setupTime.value = 0;
            return true;
        }
        try {
            if (this.timeUnit.value === 'hours') {
                const time = Math.floor((this.enteredSetupTimeHours?.value ?? 0) * 60);
                this.setupTime.value = ((this.enteredSetupTimeMinutes?.value ?? 0) + time) / 60;
            }
            if (this.timeUnit.value === 'minutes') {
                const time = Math.floor((this.enteredSetupTimeHours?.value ?? 0) * 60);
                this.setupTime.value = (this.enteredSetupTimeMinutes?.value ?? 0) + time;
            }
            return true;
        } catch (error) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-manufacturing/pages__mobile_time_tracking___update__setup__time__error',
                    'The actual setup time calculation failed.',
                ),
                { type: 'error', timeout: 5000 },
            );
            return false;
        }
    }

    /**
     * Automatic closing percentage for expected work center & site
     * @param workCenter
     * @param site
     * @returns
     */
    async getAutomaticClosingPercent(workCenter: string, site: string): Promise<decimal> {
        try {
            const result = await this.$.graph
                .node('@sage/x3-manufacturing-data/WorkCenter')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            automaticClosingPercentage: true,
                        },
                        {
                            filter: {
                                code: workCenter,
                                manufacturingSite: { code: site },
                            },
                        },
                    ),
                )
                .execute();

            return parseFloat(result.edges[0].node.automaticClosingPercentage);
        } catch (err) {
            return 0;
        }
    }

    /** Function to add the custom block to the time tracking lines table
     * @returns custom block details item:value
     */
    private getCustomTimeTrackingDetails() {
        let response = {};
        this.customTimeTrackingDetailBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    /** Function to get the custom detail fields from the time tracking lines table
     *
     * @returns
     */
    private getCustomTimeTrackingLines() {
        let response = {};
        this.customTimeTrackingDetailBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private getCustomSidebarTimeTrackingLine(): ui.SidebarFieldDefinition<TimeTrackingLine>[] {
        const fieldsSidebar: ui.SidebarFieldDefinition<TimeTrackingLine>[] = [];
        fieldsSidebar.push(this.stockUnit);
        fieldsSidebar.push(this.timeUnit);
        fieldsSidebar.push(this.remainingQuantity);
        fieldsSidebar.push(this.rejectedQuantity);
        fieldsSidebar.push(this.setupTime);
        fieldsSidebar.push(this.runTime);
        fieldsSidebar.push(this.enteredSetupTimeHours);
        fieldsSidebar.push(this.enteredSetupTimeMinutes);
        fieldsSidebar.push(this.enteredRunTimeHours);
        fieldsSidebar.push(this.enteredRunTimeMinutes);
        fieldsSidebar.push(this.close);

        this.customTimeTrackingDetailBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                fieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<TimeTrackingLine>);
            }
        });
        return fieldsSidebar;
    }
}
