import { ProductSite } from '@sage/x3-master-data-api';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import { GraphApi, Stock } from '@sage/x3-stock-data-api';
import { stockControl } from '@sage/x3-stock-data/build/lib/menu-items/stock-control';
import {
    aggregateEdgesSelector,
    AggregateEdgesSelector,
    AggregateGroupSelector,
    AggregateQueryOptions,
    AggregateQuerySelector,
    AggregateValuesSelector,
    Edges,
    extractEdges,
} from '@sage/xtrem-client';
import * as ui from '@sage/xtrem-ui';

// TODO: use lot management enum. See Jira ticket X3-162570.
const lotNotManaged = 'notManaged';

@ui.decorators.page<MobileViewStockByProductSite, ProductSite>({
    isTitleHidden: true,
    title: 'View stock by product-site',
    menuItem: stockControl,
    priority: 500,
    node: '@sage/x3-master-data/ProductSite',
    authorizationCode: 'INQSTOPRO',
    skipDirtyCheck: true,
    mode: 'default',
    async onLoad() {
        this.siteField.value = await getSelectedStockSite(
            this,
            ui.localize('@sage/x3-stock/dialog-error-title', 'Error'),
            ui.localize(
                '@sage/x3-stock/dialog-error-location-inquiry-set-site',
                'Define a default stock site on the user function profile.',
            ),
        );
    },
    headerCard() {
        return {
            title: this.siteField,
        };
    },
    navigationPanel: {
        canFilter: false,
        isHeaderHidden: true,
        isAutoSelectEnabled: true,
        listItem: {
            title: ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
            }),
            titleRight: ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'upc',
            }),
            line2: ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'localizedDescription1',
            }),
            line4: ui.nestedFields.numeric({
                bind: '_id',
                canFilter: false,
                isHidden: true,
            }),
            line5: ui.nestedFields.checkbox({
                bind: 'isLocationManaged',
                canFilter: false,
                isHidden: true,
            }),
            line6: ui.nestedFields.checkbox({
                bind: 'isLicensePlateNumberManaged',
                canFilter: false,
                isHidden: true,
            }),
            line7: ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'lotManagementMode',
                isHidden: true,
            }),
        },
        optionsMenu: [
            {
                title: 'Site: ', //+ this.$.storage.get('adc-selected-site').toString(),
                graphQLFilter: storage => ({ stockSite: { code: String(storage.get('mobile-selected-stock-site')) } }),
            },
        ],
        onSelect(listItemValue: any) {
            if (!listItemValue.isLocationManaged) {
                this.checkStockByProduct(this.siteField.value as string, listItemValue.product.code, '', '');
            } else {
                this.checkStockDataByProductAndSite(
                    listItemValue._id,
                    listItemValue.product.code,
                    this.siteField.value as string,
                    String(listItemValue.isLicensePlateNumberManaged),
                    listItemValue.product.lotManagementMode,
                );
            }
            return true;
        },
    },
})
export class MobileViewStockByProductSite extends ui.Page<GraphApi> {
    @ui.decorators.section<MobileViewStockByProductSite>({
        isTitleHidden: true,
        isHidden: true,
    })
    mainSection: ui.containers.Section;

    @ui.decorators.block<MobileViewStockByProductSite>({
        parent() {
            return this.mainSection;
        },
    })
    mainBlock: ui.containers.Block;

    @ui.decorators.labelField<MobileViewStockByProductSite>({
        parent() {
            return this.mainBlock;
        },
        isTransient: true,
    })
    stockUnit: ui.fields.Label;

    @ui.decorators.labelField<MobileViewStockByProductSite>({
        parent() {
            return this.mainBlock;
        },
        isTransient: true,
    })
    siteField: ui.fields.Label;

    private generateAggregateStockRequest<T extends ui.ClientNode>(
        selector: AggregateQuerySelector<T, AggregateGroupSelector<T>, AggregateValuesSelector<T>>,
        filter: AggregateQueryOptions<T>,
        numberOfRecords = 500,
    ) {
        const tempAggregateRequest: AggregateEdgesSelector<
            T,
            AggregateGroupSelector<T>,
            AggregateValuesSelector<T>
        > = aggregateEdgesSelector<T, AggregateGroupSelector<T>, AggregateValuesSelector<T>>(selector, filter);
        tempAggregateRequest.__args.first = numberOfRecords; // (X3-197381) TODO: Have to set some sort of hard limit. To be superseded in non-transient way
        return this.$.graph.node('@sage/x3-stock-data/Stock').aggregate.query(tempAggregateRequest);
    }

    private async batchAggregateRequests(stockSiteCode: string, productCode: string) {
        const groupByLocation: AggregateGroupSelector<Stock> = {
            location: {
                code: {
                    _by: 'value',
                },
            },
            product: {
                product: {
                    stockUnit: {
                        code: {
                            _by: 'value',
                        },
                    },
                },
            },
        };
        const filterByProduct = {
            product: { product: { code: productCode } },
            stockSite: stockSiteCode,
        };

        // Separate aggregate calls in order to exclude blank lots from distinct count
        const requests = {
            aggregateSumStockQuantity: this.generateAggregateStockRequest<Stock>(
                {
                    group: groupByLocation,
                    values: {
                        quantityInStockUnit: { sum: true },
                    },
                },
                {
                    filter: filterByProduct,
                },
            ),
            aggregateDistinctLots: this.generateAggregateStockRequest<Stock>(
                {
                    group: groupByLocation,
                    values: {
                        lot: { distinctCount: true },
                    },
                },
                {
                    filter: {
                        ...filterByProduct,
                        lot: { _ne: ' ' },
                    },
                },
            ),
            aggregateDistinctLpn: this.generateAggregateStockRequest<Stock>(
                {
                    group: groupByLocation,
                    values: {
                        licensePlateNumber: { code: { distinctCount: true } },
                    },
                },
                {
                    filter: {
                        ...filterByProduct,
                        licensePlateNumber: { code: { _ne: ' ' } },
                    },
                },
            ),
        };

        return await new ui.queryUtils.BatchRequest(requests).execute();
    }

    private async isStockDataAvailable(stockSiteCode: string, productCode: string): Promise<boolean> {
        type LocationAggregations = {
            stockUnit: string;
            stockQuantity: number;
        };

        const mess: string = ui.localize('@sage/x3-stock/product-required', 'Selected product is required');
        if (!productCode) {
            this.$.showToast(mess, { type: 'warning' });
            return false;
        }

        const aggregatedResults = this.batchAggregateRequests(stockSiteCode, productCode);

        // Check if the selected product returns no records
        if (!(await aggregatedResults).aggregateSumStockQuantity.edges) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-stock/notification-warning-product-inquiry-zero-products',
                    'Product {{code}} has zero locations(s)',
                    { code: productCode },
                ),
                { type: 'error' },
            );
            return false;
        }

        // Initialize the data structure map to consolidate and store aggregated values at the corresponding location
        const aggregationsLocationMap = new Map<string, LocationAggregations>();

        // Populate the map with all existing locations (including stock units of the products) within the stock node, based on the selected site and product
        extractEdges((await aggregatedResults).aggregateSumStockQuantity as Edges<any>).forEach(aggregation => {
            aggregationsLocationMap.set(aggregation.group.location.code, {
                stockUnit: aggregation.group.product.product.stockUnit.code,
                stockQuantity: Number(aggregation.values.quantityInStockUnit.sum),
            });
        });

        const locationsValue = new Array<{
            _id: string;
            location: string;
            stockQuantity: string;
            stockUnit: string;
        }>();

        aggregationsLocationMap.forEach((value: LocationAggregations, key: string) => {
            locationsValue.push({
                _id: key, // this is required, otherwise during onRowClick, rowItem will be blank
                location: key,
                stockQuantity: `${value.stockQuantity.toString()} ${value.stockUnit}`,
                stockUnit: value.stockUnit,
            });
        });

        if (locationsValue.length === 0) {
            setTimeout(
                () =>
                    this.$.showToast(
                        ui.localize(
                            '@sage/x3-stock/selected-product-no-results',
                            'Product {{code}} has no stock records.',
                            {
                                code: productCode,
                            },
                        ),
                        { type: 'info' },
                    ),
                10,
            );
            return false;
        }

        return true;
    }

    private async checkStockDataByProductAndSite(
        id: string,
        productCode: string,
        stockSiteCode: string,
        _LpnManage: string,
        _lotManagementMode: string,
    ): Promise<void> {
        try {
            const isAvaible = await this.isStockDataAvailable(stockSiteCode, productCode);
            if (isAvaible) {
                this.$.router.goTo(`@sage/x3-stock/MobileViewStockByProductSiteSelectLocation`, {
                    _id: id,
                    product: productCode,
                    site: stockSiteCode,
                    isLicensePlateNumberManaged: _LpnManage,
                    lotManagementMode: _lotManagementMode,
                });
            }
        } catch (error) {
            console.error('Error fetching data:', error);
        }
    }

    private async isStockDataAvailableForProduct(
        siteCode: string,
        productCode: string,
        locationCode: string,
        lpnCode: string,
    ): Promise<boolean> {
        const mess: string = ui.localize('@sage/x3-stock/product-required', 'Selected product is required');
        if (!productCode) {
            this.$.showToast(mess, { type: 'warning' });
            return false;
        }

        const response = extractEdges(
            await this.$.graph
                .node('@sage/x3-stock-data/Stock')
                .query(
                    ui.queryUtils.edgesSelector<Stock>(
                        {
                            _id: true,
                            lot: true,
                            sublot: true,
                            quantityInPackingUnit: true,
                            packingUnit: {
                                code: true,
                            },
                            quantityInStockUnit: true,
                            product: {
                                product: {
                                    stockUnit: {
                                        code: true,
                                    },
                                    serialNumberManagementMode: true,
                                    lotManagementMode: true,
                                    code: true,
                                    localizedDescription1: true,
                                    expirationManagementMode: true,
                                },
                            },
                            allocatedQuantity: true,
                            serialNumber: true,
                            status: {
                                code: true,
                            },
                            identifier1: true,
                            identifier2: true,
                            qualityAnalysisRequestId: true,
                            owner: true,
                            isBeingCounted: true,
                            packingUnitToStockUnitConversionFactor: true,
                            stockId: true,
                            lotReference: {
                                expirationDate: true,
                                useByDate: true,
                                lotCustomField1: true,
                                lotCustomField2: true,
                                majorVersion: { code: true },
                            },
                        },
                        {
                            first: 500,
                            filter: {
                                stockSite: { code: siteCode },
                                location: locationCode ? { code: locationCode } : undefined,
                                product: { product: { code: productCode } },
                                licensePlateNumber: lpnCode ? lpnCode : '',
                            },
                            orderBy: {
                                lot: 1,
                                serialNumber: 1,
                            },
                        },
                    ),
                )
                .execute(),
        );

        // Should never happen
        if (response.length === 0) {
            setTimeout(
                () =>
                    this.$.showToast(
                        ui.localize(
                            '@sage/x3-stock/selected-product-no-results',
                            'Product {{code}} has no stock records.',
                            {
                                code: productCode,
                            },
                        ),
                        { type: 'info' },
                    ),
                10,
            );
            return false;
        }

        return true;
    }

    private async checkStockByProduct(
        siteCode: string,
        productCode: string,
        locationCode: string,
        lpnCode: string,
    ): Promise<void> {
        try {
            const isAvaible = await this.isStockDataAvailableForProduct(siteCode, productCode, locationCode, lpnCode);
            if (isAvaible) {
                this.$.router.goTo(`@sage/x3-stock/MobileViewStockByProductSiteProductDetails`, {
                    site: siteCode,
                    location: locationCode,
                    product: productCode,
                    licensePlateNumber: lpnCode,
                });
            }
        } catch (error) {
            const mess: string = ui.localize('@sage/x3-stock/error-loading-stock', 'Error loading stock');
            this.$.showToast(mess, { type: 'error' });
        }
    }
}
