import { AutomaticJournal } from '@sage/x3-finance-data-api';
import {
    Container,
    MajorVersionStatus,
    Product,
    ProductSearchCriteria,
    ProductSite,
    ProductVersion,
    Supplier,
    SupplierProduct,
    UnitOfMeasure,
} from '@sage/x3-master-data-api';
import { dialogConfirmation, dialogMessage } from '@sage/x3-master-data/lib/client-functions/dialogs';
import { getSelectedStockSite } from '@sage/x3-master-data/lib/client-functions/get-selected-stock-site';
import {
    GraphApi,
    ProductSupplierView,
    PurchaseEntryTransaction,
    PurchaseOrder,
    PurchaseOrderLine,
    PurchaseOrderQuantityLinesSupplierView,
    PurchaseReceiptLineBinding,
    PurchaseShipment,
    PurchaseShipmentLine,
} from '@sage/x3-purchasing-api';
import {
    LicensePlateNumber,
    Location,
    Lot,
    LotsSites,
    StockJournalBinding,
    StockManagementRules,
    StockStatus,
    Warehouse,
} from '@sage/x3-stock-data-api';
import { receipt } from '@sage/x3-stock-data/build/lib/menu-items/receipt';
import { MiscellaneousTable, Site } from '@sage/x3-system-api';
import { BarcodeTransactionFlow } from '@sage/x3-system/lib/client-functions/barcode-transaction-flow';
import {
    AsyncCompositeAllowed,
    AsyncVoidFunction,
    DictionaryDataGeneric,
    DictionaryDataGs1,
    DictionaryFieldSupported,
} from '@sage/x3-system/lib/client-functions/screen-management';
import { SupportManagementService } from '@sage/x3-system/lib/client-functions/support-management-service';
import { TypeOfServiceBarcode } from '@sage/x3-system/lib/client-functions/type-of-barcode-management';
import { DataTitle } from '@sage/x3-system/lib/shared-functions/parsed-element';
import { getRegExp } from '@sage/x3-system/lib/shared-functions/pat-converter';
import { ExtractEdges, ExtractEdgesPartial, Filter, extractEdges } from '@sage/xtrem-client';
import { DateValue } from '@sage/xtrem-date-time';
import { MAX_INT_32 } from '@sage/xtrem-shared';
import * as ui from '@sage/xtrem-ui';
import { expirationDateDefaultValue, useByDateDefaultValue } from '../client-functions/defaultValue';
import { controlLotReceipt } from '../client-functions/purchase-receipt-details-control';
import { readParameterValue } from '../client-functions/read-parameter';
import { findDefaultLocation, findStockManagementRules } from '../client-functions/stock-management-rules';

export type DeepPartial<T> = T extends Object ? { [K in keyof T]?: DeepPartial<T[K]> } : T;
type PartialPurchaseEntryTransaction = DeepPartial<PurchaseEntryTransaction>;

// Key to use with Composite Data for this application
export const mobileApplicationKey = 'MobilePurchaseReceiptKey';
const QTY_DECIMAL_PLACES: number = 8;
const CONVERSION_RATE_DECIMAL_PLACES: number = 6;

/**
 * Composite data management for this application.
 * @enum PurchaseReceiptDictionaryFieldKey
 *
 * This enum provides per field the associated access key in the data dictionary.
 * It must be maintained in accordance with the fields of this dictionary.
 * Never translate..
 */

enum PurchaseReceiptDictionaryFieldKey {
    closePoLine = 'closePoLine',
    expirationDate = 'expirationDate',
    identifier1 = 'identifier1',
    identifier2 = 'identifier2',
    licensePlateNumber = 'licensePlateNumber',
    location = 'location',
    lot = 'lot',
    lotCustomField1 = 'lotCustomField1',
    lotCustomField2 = 'lotCustomField2',
    lotCustomField3 = 'lotCustomField3',
    lotCustomField4 = 'lotCustomField4',
    majorVersion = 'majorVersion',
    minorVersion = 'minorVersion',
    packingUnitToStockUnitConversionFactor = 'packingUnitToStockUnitConversionFactor',
    product = 'product',
    productFromPurchaseOrder = 'productFromPurchaseOrder',
    purchaseOrder = 'purchaseOrder',
    quantityReceived = 'quantityReceived',
    receiptUnit = 'receiptUnit',
    serialNumber = 'serialNumber',
    status = 'status',
    sublot = 'sublot',
    supplierLot = 'supplierLot',
    useByDate = 'useByDate',
}

enum Section {
    main = 'main',
    destination = 'destination',
    shipment = 'shipment',
    receipt = 'receipt',
}

type packingUnit = {
    node: {
        packingUnit: {
            code: string;
            numberOfDecimals: number;
            name: string;
            _id: string;
        };
        packingUnitToStockUnitConversionFactor: string;
        isPackingFactorEntryAllowed: boolean;
        _id: string;
        isPackingUnit: boolean;
    };
};

interface StockJournalClosePoLine extends StockJournalBinding {
    closePoLine: string;
    remainingQuantityInStockUnit: number;
    containerDetail: string;
}

const balanceOptions: string[] = [
    ui.localize('@sage/x3-purchasing/Yes', 'Yes'),
    ui.localize('@sage/x3-purchasing/No', 'No'),
];

function formatUnitForDisplay(code: string, conversionRate: string, stockUnit: string): string {
    return `${code} ${'='} ${conversionRate} ${stockUnit}`;
}
@ui.decorators.page<MobilePurchaseReceipt>({
    title: 'Purchase receipt',
    module: 'x3-purchasing',
    mode: 'default',
    menuItem: receipt,
    priority: 100,
    node: '@sage/x3-purchasing/PurchaseReceipt',
    isTitleHidden: true,
    authorizationCode: 'CWSPTH',
    navigationPanel: undefined,
    access: { node: '@sage/x3-purchasing/PurchaseReceipt' },
    async onLoad() {
        if (!(await this._init())) {
            this._disablePage();
        }
    },
    businessActions() {
        return [this.createButton, this.addButton];
    },
})
export class MobilePurchaseReceipt extends SupportManagementService<GraphApi> {
    /*
     *
     *  Technical properties
     *
     */
    private _entryTransactionMap: Map<string, PartialPurchaseEntryTransaction>;
    private _selectedEntryTransaction: PartialPurchaseEntryTransaction | undefined;
    private _productSearchCriteria1: ProductSearchCriteria;
    private _productSearchCriteria2: ProductSearchCriteria;
    private _purchaseShipmentId: string | undefined;
    private _purchaseShipmentLineNumber: string | undefined;
    private _shipmentLines: any[] = [];
    private _purchaseShipmentIds: string[] = [];
    private _isLocationPreloaded: boolean;
    private _productSite: ProductSite;
    private _packingUnits: packingUnit[];
    private _isOverReceiptAllowed: boolean | undefined;
    private _purchaseStockManagementRules: StockManagementRules;
    private _purchaseOrderLine: PurchaseOrderLine | undefined;
    private _supplierProduct: SupplierProduct | undefined;
    private _purchaseShipmentLine: PurchaseShipmentLine | undefined;
    private _effectiveDate = DateValue.today().toString();
    private _quantityToReceiveNoRounded: number;
    private _quantityToReceiveRounded: number;
    private _remainingQuantityNoRoundedInPackingUnit: number;
    private _remainingQuantityRoundedInPackingUnit: number;
    private _remainingQuantityInitialInStockUnit: number;
    private _remainingQuantityInitialInOrderUnit: number;
    private _locationDetail: ExtractEdgesPartial<Location> | undefined;
    private _containerDetail: string | undefined;
    private _lots: ExtractEdges<Lot>[];
    private _directReceiptAllowed?: boolean;

    /**
     *
     * Technical composite data properties
     *
     */

    /**
     * This value is initialized only when dispatching
     * and reset when user change lot or expiration date.
     */
    /* @internal */
    private _batchLot: string | null = null;

    /**
     * This value is initialized only when dispatching
     * and reset when user change lot or expiration date.
     */
    /* @internal */
    private _expiryDate: string | null = null;

    @ui.decorators.referenceField<MobilePurchaseReceipt, Site>({
        node: '@sage/x3-system/Site',
        valueField: 'code',
        isHidden: true,
        bind: 'receiptSite',
        helperTextField: 'name',
        columns: [
            ui.nestedFields.text({
                bind: 'code',
            }),
            ui.nestedFields.text({
                bind: '_id',
            }),
        ],
    })
    receiptSite: ui.fields.Reference<Site, MobilePurchaseReceipt>;

    /*
     *
     *  Page Actions
     *
     */

    @ui.decorators.pageAction<MobilePurchaseReceipt>({
        title: 'Create',
        buttonType: 'primary',
        isDisabled: true,
        shortcut: ['f2'],
        async onError(e) {
            this.$.loader.isHidden = true;
            const options: ui.dialogs.DialogOptions = {
                acceptButton: {
                    text: ui.localize('@sage/x3-purchasing/button-goback', 'Go back'),
                },
                cancelButton: {
                    text: ui.localize('@sage/x3-purchasing/button-cancel', 'Cancel'),
                },
                size: 'small',
                mdContent: true,
            };

            let message = '';
            if (
                e.errors &&
                e.errors.length &&
                e.errors[0].extensions &&
                e.errors[0].extensions.diagnoses &&
                e.errors[0].extensions.diagnoses.length
            ) {
                const _messages = <string[]>[];
                e.errors[0].extensions.diagnoses
                    .filter((d: { severity: number; message: any }) => d.severity > 2 && d.message)
                    .forEach((d: { message: any }) => {
                        const _message = d.message.split(`\n`);
                        _messages.push(..._message);
                    });
                const _result = _messages.length ? <string[]>_messages : <string[]>e.message.split(`\n`);
                message = `**${ui.localize(
                    '@sage/x3-purchasing/dialog-error-purchase-receipt-creation',
                    'An error has occurred',
                )}**\n\n`;
                if (_result.length === 1) {
                    message += `${_result[0]}`;
                } else {
                    message += _result.map(item => `* ${item}`).join('\n');
                }
            } else {
                message = `${ui.localize(
                    '@sage/x3-purchasing/pages_creation_error_network_contact_administrator',
                    'A network error occurred. Contact your administrator.',
                )}`;
            }
            await this.$.sound.error();
            if (
                !(await dialogConfirmation(
                    this,
                    'error',
                    ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                    message,
                    options,
                ))
            ) {
                await this.$.router.emptyPage();
            }
        },
        async onClick() {
            const validateError = (await this.receiptDate.validate()) || (await this.supplierPackingSlip.validate());
            if (validateError) {
                this.$.removeToasts();
                this.$.showToast(
                    `${ui.localize(
                        '@sage/x3-purchasing/notification-error-purchase-receipt-error',
                        'Error',
                    )}: ${validateError}`,
                    { type: 'error', timeout: 30000 },
                );
            } else if (this.lines.value?.length > 0) {
                this.$.loader.isHidden = false;
                await this.$standardSaveAction.execute(false);
                this.$.loader.isHidden = true;
                await this.$.sound.success();
                await dialogMessage(
                    this,
                    'success',
                    ui.localize('@sage/x3-purchasing/dialog-success-title', 'Success'),
                    ui.localize(
                        '@sage/x3-purchasing/dialog-success-purchase-receipt-creation',
                        'Document no. {{receiptId}} created.',
                        { receiptId: this.id.value },
                    ),
                );
                await this.$.router.emptyPage();
            } else {
                this.$.showToast(
                    ui.localize(
                        '@sage/x3-purchasing/notification-error-purchase-receipt-no-products',
                        `Enter at least one product.`,
                    ),
                    { type: 'error', timeout: 5000 },
                );
            }
        },
    })
    createButton: ui.PageAction;

    @ui.decorators.pageAction<MobilePurchaseReceipt>({
        title: 'Add',
        buttonType: 'primary',
        shortcut: ['f3'],
        isHidden: true,
        async onClick() {
            if (await this._isZeroQuantityValue()) {
                return;
            }
            if (await this._isEmptyLpnAndContainerValues()) {
                return;
            }
            if (!(await this._validate())) {
                return;
            }
            this._addLineAndStockDetails();
            this.quantityReceived.value = null;
            this.previousQuantityToReceive.value = null;
            this.transaction.isDisabled = true;

            this._disableDestinationFields();

            await this._setActiveSection(Section.receipt);
        },
    })
    addButton: ui.PageAction;

    /*
     *
     *  Sections
     *
     */

    @ui.decorators.section<MobilePurchaseReceipt>({
        isTitleHidden: true,
    })
    mainSection: ui.containers.Section;

    /*
     *
     *  Blocks
     *
     */

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
    })
    productBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'small',
        isTitleHidden: true,
    })
    customizationPurchaseReceiptHeaderBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
        isHidden: true,
    })
    destinationBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    customizationPurchaseReceiptLinesBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        width: 'extra-large',
        isTitleHidden: true,
        isHidden: true,
    })
    customizationPurchaseReceiptDetailsBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
        isHidden: true,
    })
    shipmentBlock: ui.containers.Block;

    @ui.decorators.block<MobilePurchaseReceipt>({
        parent() {
            return this.mainSection;
        },
        isTitleHidden: true,
        isHidden: true,
    })
    receiptBlock: ui.containers.Block;

    /*
     *
     *  Fields
     *
     */

    @ui.decorators.textField<MobilePurchaseReceipt>({
        isHidden: true,
    })
    id: ui.fields.Text<MobilePurchaseReceipt>;

    @ui.decorators.dateField<MobilePurchaseReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Receipt date',
        isMandatory: true,
        maxDate: DateValue.today().toString(), // restricts to current date or earlier
    })
    receiptDate: ui.fields.Date;

    @ui.decorators.dropdownListField<MobilePurchaseReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Transaction',
        isMandatory: true,
        isTransientInput: true,
        onChange() {
            if (this.transaction.value) {
                this._selectedEntryTransaction = this._entryTransactionMap.get(this.transaction.value);
                this.stockMovementCode.value = this._selectedEntryTransaction?.stockMovementCode
                    ? {
                          code: this._selectedEntryTransaction.stockMovementCode.code,
                          _id: this._selectedEntryTransaction.stockMovementCode._id,
                      }
                    : null;
                this.stockAutomaticJournal.value = this._selectedEntryTransaction?.stockAutomaticJournal
                    ? {
                          code: this._selectedEntryTransaction.stockAutomaticJournal.code,
                          _id: this._selectedEntryTransaction.stockAutomaticJournal._id,
                      }
                    : null;
            }
        },
    })
    transaction: ui.fields.DropdownList;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransientInput: true,
    })
    destination: ui.fields.Text<MobilePurchaseReceipt>;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransientInput: true,
    })
    document: ui.fields.Text<MobilePurchaseReceipt>;

    @ui.decorators.referenceField<MobilePurchaseReceipt, MiscellaneousTable>({
        node: '@sage/x3-system/MiscellaneousTable',
        valueField: 'code',
        canFilter: false,
        isHidden: true,
        filter: {
            glossaryId: {
                glossaryId: 14,
            },
        },
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockMovementCode: ui.fields.Reference;

    @ui.decorators.referenceField<MobilePurchaseReceipt, AutomaticJournal>({
        node: '@sage/x3-finance-data/AutomaticJournal',
        valueField: 'code',
        isHidden: true,
        canFilter: false,
        columns: [
            ui.nestedFields.text({
                bind: '_id',
            }),
            ui.nestedFields.text({
                bind: 'code',
            }),
        ],
    })
    stockAutomaticJournal: ui.fields.Reference<AutomaticJournal, MobilePurchaseReceipt>;

    @ui.decorators.labelField<MobilePurchaseReceipt>({
        isTransient: true,
        isHidden: true,
    })
    legalCompany: ui.fields.Label;

    @ui.decorators.referenceField<MobilePurchaseReceipt, PurchaseOrder>({
        parent() {
            return this.productBlock;
        },
        title: 'Purchase order',
        placeholder: 'Scan or select...',
        node: '@sage/x3-purchasing/PurchaseOrder',
        valueField: 'id',
        helperTextField: { orderFromSupplier: { code: { code: true } } },
        isHelperTextHidden: true,
        canFilter: false,
        isTransient: true,
        filter() {
            const filter: Filter<PurchaseOrder> = {
                isClosed: { _eq: false },
                isIntersite: { _eq: false },
                signatureStatus: { _in: ['inFull', 'notManaged', 'automatic'] },
                receiptStatus: { _in: ['no', 'inPart'] },
                purchaseSite: { legalCompany: { code: this.legalCompany.value ?? undefined } },
                purchaseOrderLines: {
                    _atLeast: 1,
                    isClosed: { _eq: false },
                    ...(this.receiptSite.value && { receiptSite: { code: this.receiptSite.value.code } }),
                    sourceRequest: { _ne: 'directOrder' },
                    productType: { _ne: 'subcontract' },
                    lineType: { _eq: 'normal' },
                },
            };

            if (this.supplier.value) {
                filter.orderFromSupplier = { code: this.supplier.value.code };
            }

            return filter;
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.purchaseOrder, rawData);
        },
        async onChange() {
            await this._onChange_purchaseOrder();
        },
        isFullWidth: true,
        isAutoSelectEnabled: true,
        columns: [
            ui.nestedFields.text({
                bind: 'id',
                title: 'Order no.',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'purchaseSite',
                valueField: 'code',
                title: 'Order site',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Supplier',
                bind: 'orderFromSupplier',
                valueField: { code: { code: true } },
                title: 'Supplier',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'internalOrderReference',
                title: 'Internal or order reference',
                isReadOnly: true,
            }),
            // (X3-227347) TODO: Obsolete: Having to specify & hide fields used in filter that don't need to be displayed
            ui.nestedFields.checkbox({
                bind: 'isClosed',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'signatureStatus',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'receiptStatus',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'receiptSite',
                valueField: '_id',
                isHidden: true,
            }),
            ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseOrder, Site>({
                node: '@sage/x3-system/Site',
                bind: 'purchaseSite',
                valueField: { legalCompany: { code: true } },
                isHidden: true,
            }),
        ],
    })
    purchaseOrder: ui.fields.Reference;

    /**
     * OnChange purchaseOrder readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_purchaseOrder: AsyncVoidFunction = async () => {
        await this._setActiveSection(Section.main, !this.purchaseOrder.value);
        // perform another screen initialization for current workFlow enabled
        // TODO: This operation deletes the composite data, see if you want to transfer them here
        if (!(await this._updateControlManager())) {
            // TODO: What action should be taken in the event of a fatal error ?
        }
    };

    @ui.decorators.referenceField<MobilePurchaseReceipt, Supplier>({
        parent() {
            return this.productBlock;
        },
        title: 'Supplier',
        placeholder: 'Scan or select...',
        node: '@sage/x3-master-data/Supplier',
        valueField: { code: { code: true } },
        canFilter: false,
        filter: {
            isActive: { _eq: true },
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.supplier, rawData);
        },
        async onChange() {
            await this._onChange_supplier();
        },
        isFullWidth: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        columns: [
            ui.nestedFields.reference({
                bind: 'code',
                title: 'Supplier',
                isReadOnly: true,
                node: '@sage/x3-master-data/BusinessPartner',
                valueField: 'code',
            }),
            ui.nestedFields.text({
                bind: 'shortCompanyName',
                title: 'Short description',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'companyName1',
                title: 'Company name',
                isReadOnly: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isActive',
                isHidden: true,
            }),
        ],
    })
    supplier: ui.fields.Reference;

    /**
     * OnChange supplier readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_supplier: AsyncVoidFunction = async () => {
        await this._setActiveSection(Section.main, !this.supplier.value);
    };

    @ui.decorators.linkField<MobilePurchaseReceipt>({
        parent() {
            return this.productBlock;
        },
        isTransient: true,
        isTitleHidden: true,
        size: 'small',
        isFullWidth: true,
        width: 'small',
        map() {
            return ui.localize('@sage/x3-purchasing/select-supplier-packing-slip', 'Add a supplier packing slip');
        },
        async onClick() {
            this.supplierPackingSlip.isHidden = false;
            this.supplierPackingSlipLink.isHidden = true;
            this.supplierPackingSlip.focus();
        },
    })
    supplierPackingSlipLink: ui.fields.Link;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.productBlock;
        },
        title: 'Supplier packing slip',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        maxLength: 20,
        isHidden: true,
        async onChange() {
            await this._onChange_supplierPackingSlip();
        },
    })
    supplierPackingSlip: ui.fields.Text;

    /**
     * OnChange readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_supplierPackingSlip: AsyncVoidFunction = async () => {
        if (this.supplierPackingSlip.value) {
            this.supplierPackingSlip.getNextField(true)?.focus();
        }
    };

    @ui.decorators.referenceField<MobilePurchaseReceipt, ProductSupplierView>({
        parent() {
            return this.productBlock;
        },
        title: 'Product',
        placeholder: 'Scan or select...',
        node: '@sage/x3-purchasing/ProductSupplierView',
        valueField: { product: true },
        helperTextField: 'upc',
        canFilter: false,
        filter() {
            return {
                ...(this.receiptSite.value && { stockSite: this.receiptSite.value.code }),
                supplier: this.supplier.value?.code.code ?? undefined,
                productStatus: 'active',
                isReceived: { _eq: true },
                _or: [
                    { isPurchased: { _eq: true } },
                    { isSubcontracted: { _eq: true } },
                    { isService: { _eq: true } },
                    { isPhantom: { _eq: true } },
                    { isDeliverable: { _eq: true } },
                    { isSold: { _eq: true } },
                ],
            };
        },
        onError(error: any, originScreenId: string, originElementId: string) {
            ui.console.warn(`Error on ${originScreenId} ${originElementId}: ${error.message || error}`);
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.product, rawData);
        },
        async onChange() {
            await this._onChange_product();
        },
        isTransient: true,
        isFullWidth: true,
        isDisabled: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        columns: [
            ui.nestedFields.text({
                bind: 'product',
                title: 'Product',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'upc',
                title: 'UPC',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'globalTradeItemNumber',
                title: 'GTIN',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'description1',
                title: 'Description',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'supplierUpc',
                title: 'Supplier UPC',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'supplierProduct',
                title: 'Supplier product',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'supplierProductDescription',
                title: 'Supplier product description',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'productCategory',
                title: 'Product category',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'productStatus',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isReceived',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isPurchased',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isSubcontracted',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isService',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isPhantom',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isDeliverable',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isSold',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'supplier',
                isHidden: true,
            }),
        ],
    })
    product: ui.fields.Reference;

    @ui.decorators.referenceField<MobilePurchaseReceipt, PurchaseOrderQuantityLinesSupplierView>({
        parent() {
            return this.productBlock;
        },
        title: 'Product',
        placeholder: 'Scan or select...',
        node: '@sage/x3-purchasing/PurchaseOrderQuantityLinesSupplierView',
        valueField: { product: true },
        helperTextField: { upc: true },
        canFilter: false,
        filter() {
            return {
                productStatus: 'active',
                isReceived: { _eq: true },
                _or: [
                    { isIntersite: { _eq: true } },
                    { isIntersite: { _eq: false }, isManufactured: { _eq: false } },
                    { isIntersite: { _eq: false }, isPhantom: { _eq: false } },
                ],
                isClosed: { _eq: false },
                sourceRequest: { _ne: 'directOrder' },
                productType: { _ne: 'subcontract' },
                lineType: 'normal',
                ...(this.receiptSite.value && { receiptSite: this.receiptSite.value.code }),
                purchaseOrder: this.purchaseOrder.value?.id,
            };
        },
        onError(error: any, originScreenId: string, originElementId: string) {
            console.warn(`Error on ${originScreenId} ${originElementId}: ${error.message || error}`);
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.productFromPurchaseOrder, rawData);
        },
        async onChange() {
            await this._onChange_productFromPurchaseOrder();
        },
        isTransient: true,
        isFullWidth: true,
        isHidden: true,
        isDisabled: true,
        isAutoSelectEnabled: true,
        shouldSuggestionsIncludeColumns: true,
        columns: [
            ui.nestedFields.text({
                bind: 'product',
                title: 'Product',
                isReadOnly: true,
            }),
            ui.nestedFields.numeric({
                bind: 'lineNumber',
                title: 'Order line',
                prefix: 'Line no.',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'supplierUpc',
                title: 'Supplier upc',
                isReadOnly: true,
                canFilter: true,
                isHidden() {
                    return this._productSearchCriteria1 !== 'supplierUpcCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'supplierProduct',
                title: 'Supplier product',
                isReadOnly: true,
                canFilter: true,
                isHidden() {
                    return this._productSearchCriteria1 !== 'supplierProductCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'upc',
                title: 'UPC',
                isReadOnly: true,
                isHidden() {
                    return this._productSearchCriteria1 !== 'x3UpcCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'globalTradeItemNumber',
                title: 'GTIN',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'description1',
                title: 'Description',
                isReadOnly: true,
                isHidden() {
                    return this._productSearchCriteria1 !== 'x3ProductDescription';
                },
            }),
            ui.nestedFields.text({
                bind: 'supplierUpc',
                title: 'Supplier upc',
                isReadOnly: true,
                canFilter: true,
                isHidden() {
                    return this._productSearchCriteria2 !== 'supplierUpcCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'supplierProduct',
                title: 'Supplier product',
                isReadOnly: true,
                canFilter: true,
                isHidden() {
                    return this._productSearchCriteria2 !== 'supplierProductCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'upc',
                title: 'UPC',
                isReadOnly: true,
                isHidden() {
                    return this._productSearchCriteria2 !== 'x3UpcCode';
                },
            }),
            ui.nestedFields.text({
                bind: 'description1',
                title: 'Description',
                isReadOnly: true,
                isHidden() {
                    return this._productSearchCriteria2 !== 'x3ProductDescription';
                },
            }),
            ui.nestedFields.numeric({
                bind: 'purchaseUnitToStockUnitConversionFactor',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPurchaseUnitOrdered',
                title: 'Ordered qty',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPurchaseUnitReceived',
                title: 'Received qty',
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.label({
                title: 'Ordered',
                prefix: 'Ordered',
                postfix(_value, rowData) {
                    return rowData?.purchaseUnit;
                },
                isTitleHidden: false,
                isTransient: true,
                bind: '_id',
                map(_value, rowData) {
                    return rowData.quantityInPurchaseUnitOrdered.toString();
                },
            }),
            ui.nestedFields.label({
                title: 'Expected',
                prefix: 'Expected',
                postfix(_value, rowData) {
                    return rowData?.purchaseUnit;
                },
                isTitleHidden: false,
                isTransient: true,
                bind: '_id',
                map(_value, rowData) {
                    let quantityInPurchaseUnitReceived = 0;
                    let quantityInPurchaseUnitToReceive = Math.max(
                        Number(rowData.quantityInPurchaseUnitOrdered) - Number(rowData.quantityInPurchaseUnitReceived),
                        0,
                    );

                    this.lines.value
                        .filter(
                            line =>
                                line.purchaseOrder?.id === rowData.purchaseOrder &&
                                Number(line.purchaseOrderLineNumber) === Number(rowData.lineNumber) &&
                                line.product?.code === rowData.product,
                        )
                        .map(
                            line =>
                                (quantityInPurchaseUnitReceived +=
                                    (Number(line.quantityInReceiptUnitReceived) *
                                        Number(line.receiptUnitToStockUnitConversionFactor)) /
                                    rowData.purchaseUnitToStockUnitConversionFactor),
                        );
                    if (quantityInPurchaseUnitReceived > 0) {
                        quantityInPurchaseUnitToReceive = Math.max(
                            Number(rowData.quantityInPurchaseUnitOrdered) -
                                Number(rowData.quantityInPurchaseUnitReceived) -
                                Number(quantityInPurchaseUnitReceived),
                            0,
                        );
                    }

                    return Number.isInteger(Number(quantityInPurchaseUnitToReceive))
                        ? quantityInPurchaseUnitToReceive.toString()
                        : quantityInPurchaseUnitToReceive.toFixed(QTY_DECIMAL_PLACES);
                },
            }),
            ui.nestedFields.numeric({
                bind: 'sequenceNumber',
                isTitleHidden: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'purchaseUnit',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'productStatus',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isReceived',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isIntersite',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isManufactured',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isPhantom',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isClosed',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'sourceRequest',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'workInProgressStatus',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'productType',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'lineType',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'receiptSite',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'purchaseOrder',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'supplier',
                isHidden: true,
            }),
        ],
    })
    productFromPurchaseOrder: ui.fields.Reference;

    @ui.decorators.dropdownListField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Unit',
        placeholder: 'Enter...',
        options: ['UN'],
        isMandatory() {
            return this._isMandatoryDestination();
        },
        isTransient: true,
        async onChange() {
            await this._onChange_receiptUnit();
        },
    })
    receiptUnit: ui.fields.DropdownList;

    /**
     * OnChange receiptUnit readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_receiptUnit: AsyncVoidFunction = async () => {
        if (this.receiptUnit.value) {
            const selectedUnit = this._packingUnits.find(
                item => item.node.packingUnit.name === this.receiptUnit.value,
            )?.node;
            if (selectedUnit) {
                const previousUnit = this._packingUnits.find(
                    item => item.node.packingUnit.code === this.previousReceiptUnitCode.value,
                )?.node;

                if (this.productFromPurchaseOrder.value) {
                    if (this._remainingQuantityInitialInStockUnit) {
                        // calcul packUnitTostockUnittoStockUnitConversionFactor : QTYSTU/QTYUOM
                        if (this._remainingQuantityInitialInOrderUnit) {
                            this.packingUnitToStockUnitConversionFactor.value =
                                this._remainingQuantityInitialInStockUnit / this._remainingQuantityInitialInOrderUnit;
                            if (selectedUnit.packingUnit.code === this.productFromPurchaseOrder.value.orderUnit)
                                selectedUnit.packingUnitToStockUnitConversionFactor =
                                    this.packingUnitToStockUnitConversionFactor.value.toString();
                            if (
                                previousUnit &&
                                previousUnit.packingUnit.code === this.productFromPurchaseOrder.value.orderUnit
                            )
                                previousUnit.packingUnitToStockUnitConversionFactor =
                                    this.packingUnitToStockUnitConversionFactor.value.toString();
                        }
                    }
                }

                if (previousUnit) {
                    let selectedUnitConversionFactor: string;
                    if (Number(selectedUnit.packingUnitToStockUnitConversionFactor) !== 0) {
                        selectedUnitConversionFactor = (
                            Number(previousUnit.packingUnitToStockUnitConversionFactor) /
                            Number(selectedUnit.packingUnitToStockUnitConversionFactor)
                        ).toString();
                    } else {
                        selectedUnitConversionFactor = Number(
                            previousUnit.packingUnitToStockUnitConversionFactor,
                        ).toString();
                    }

                    const tmp = Math.pow(10, selectedUnit?.packingUnit.numberOfDecimals);
                    if (this.quantityReceived.value) {
                        this._quantityToReceiveNoRounded *= Number(selectedUnitConversionFactor);
                        if (tmp !== 0) {
                            this._quantityToReceiveRounded = Math.round(this._quantityToReceiveNoRounded * tmp) / tmp;
                        }
                        this.quantityReceived.value = this._quantityToReceiveRounded;
                        this.previousQuantityToReceive.value = this.quantityReceived.value;
                    }
                    if (this.productFromPurchaseOrder.value) {
                        if (
                            this._remainingQuantityInitialInStockUnit &&
                            this.packingUnitToStockUnitConversionFactor.value !== 0
                        )
                            this._remainingQuantityNoRoundedInPackingUnit = Number(
                                this._remainingQuantityInitialInStockUnit /
                                    Number(this.packingUnitToStockUnitConversionFactor.value),
                            );
                        if (tmp !== 0) {
                            this._remainingQuantityRoundedInPackingUnit =
                                Math.round(this._remainingQuantityNoRoundedInPackingUnit * tmp) / tmp;
                        } else {
                            this._remainingQuantityNoRoundedInPackingUnit *= Number(
                                this.packingUnitToStockUnitConversionFactor.value,
                            );
                            if (tmp !== 0) {
                                this._remainingQuantityRoundedInPackingUnit =
                                    Math.round(this._quantityToReceiveNoRounded * tmp) / tmp;
                            }
                        }
                        this.destinationRemainingQuantityInPackingUnit.value =
                            this._remainingQuantityRoundedInPackingUnit;
                    }
                    this.quantityReceived.focus(); // this moves cursor to quantityToReceive and auto-round & update # of decimal places allowed in its current value
                    this.previousReceiptUnit.value = this.receiptUnit.value;
                    this.previousReceiptUnitCode.value = selectedUnit.packingUnit.code;

                    this.receiptUnitCode.value = selectedUnit.packingUnit.code;
                    this.packingUnitToStockUnitConversionFactor.value = Number(
                        selectedUnit.packingUnitToStockUnitConversionFactor,
                    );
                    this.packingUnitToStockUnitConversionFactor.isDisabled = !selectedUnit.isPackingFactorEntryAllowed; // TODO: Obsolete isPackingFactorEntryAllowed because packingUnitToStockUnitConversionFactor is now a hidden field
                    this.quantityReceived.scale = selectedUnit.packingUnit.numberOfDecimals;
                }
            }
        }
    };

    @ui.decorators.labelField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransient: true,
    })
    receiptUnitCode: ui.fields.Label;

    @ui.decorators.labelField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransient: true,
    })
    previousReceiptUnit: ui.fields.Label;

    @ui.decorators.labelField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransient: true,
    })
    previousReceiptUnitCode: ui.fields.Label;

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Quantity',
        placeholder: 'Enter...',
        isMandatory() {
            return this._isMandatoryDestination();
        },
        isTransient: true,
        min: 0,
        isNotZero: true,
        validation(value: number) {
            if (this._isMandatoryDestination()) {
                const regex = /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/; // reg ex for any positive numbers (integers or decimals) excluding 0
                if ((value.toString().match(regex)?.length ?? 0) === 0) {
                    return ui.localize('@sage/x3-purchasing/validate-error-invalid-value', 'invalid value');
                }
                if (
                    this.productFromPurchaseOrder.value &&
                    Number(this.quantityReceived.value) * Number(this.packingUnitToStockUnitConversionFactor.value) >
                        Number(this.destinationRemainingQuantityInStockUnit.value) &&
                    !this._isOverReceiptAllowed
                ) {
                    return ui.localize(
                        '@sage/x3-purchasing/validation-error-purchase-receipt-details-quantity-received-is-greater-than-the-ordered-quantity',
                        'The quantity received is greater than the ordered quantity.',
                    );
                }
            }
            return undefined;
        },
        scale() {
            if (this.receiptUnit.value) {
                const selectedUnit = this._packingUnits.find(
                    item => item.node.packingUnit.name === this.receiptUnit.value,
                )?.node;
                if (selectedUnit) {
                    return selectedUnit.packingUnit.numberOfDecimals;
                } else {
                    return 0;
                }
            } else {
                return 0;
            }
        },
        async onChange() {
            await this._onChange_quantityReceived();
        },
    })
    quantityReceived: ui.fields.Numeric<MobilePurchaseReceipt>;

    /**
     * OnChange receiptUnit readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_quantityReceived: AsyncVoidFunction = async () => {
        if (this.productFromPurchaseOrder.value) {
            const tmp = Math.pow(10, this.quantityReceived.scale ?? 1);
            let quantity: number =
                Number(this.quantityReceived.value) * Number(this.packingUnitToStockUnitConversionFactor.value);
            if (tmp !== 0) {
                quantity = Math.round(quantity * tmp) / tmp;
            }
            if (quantity > (this.destinationRemainingQuantityInStockUnit.value ?? 0)) {
                if (!this._isOverReceiptAllowed) {
                    await dialogMessage(
                        this,
                        'error',
                        ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                        ui.localize(
                            '@sage/x3-purchasing/dialog-error-purchase-receipt-details-quantity-received-is-greater-than-the-ordered-quantity',
                            `The quantity received is greater than the ordered quantity.`,
                        ),
                        {
                            fullScreen: false,
                            acceptButton: {
                                text: ui.localize('@sage/x3-purchasing/button-accept-ok', 'OK'),
                            },
                        },
                    );
                    this.quantityReceived.focus();
                    return;
                } else {
                    if (
                        !(await dialogConfirmation(
                            this,
                            'warn',
                            ui.localize('@sage/x3-purchasing/dialog-warning-title', 'Warning'),
                            ui.localize(
                                '@sage/x3-purchasing/dialog-warn-purchase-receipt-details-quantity-message',
                                'Quantity larger than expected. Validate ?',
                            ),
                            {
                                fullScreen: false,
                                acceptButton: {
                                    text: ui.localize('@sage/x3-purchasing/button-accept-yes', 'Yes'),
                                },
                                cancelButton: {
                                    text: ui.localize('@sage/x3-purchasing/button-cancel-no', 'No'),
                                },
                            }, //
                        ))
                    ) {
                        this.quantityReceived.value = this.previousQuantityToReceive.value;
                        await this.$.commitValueAndPropertyChanges();
                        this.quantityReceived.focus();
                    }
                }
            }

            if (
                Number(this.quantityReceived.value) * Number(this.packingUnitToStockUnitConversionFactor.value) ===
                this.destinationRemainingQuantityInStockUnit.value
            ) {
                this.closePoLine.value = balanceOptions[0]; //Yes
            } else if (
                Number(this.quantityReceived.value) * Number(this.packingUnitToStockUnitConversionFactor.value) <
                Number(this.destinationRemainingQuantityInStockUnit.value)
            ) {
                this.closePoLine.value = balanceOptions[1]; //No
            }
        }
        if (!this.quantityReceived.value) {
            this.quantityReceived.value = 0;
        }
        if (this.quantityReceived.value !== this._quantityToReceiveRounded) {
            this._quantityToReceiveNoRounded = this.quantityReceived.value;
        }
        this._quantityToReceiveRounded = this.quantityReceived.value;
        this.previousQuantityToReceive.value = this.quantityReceived.value;
    };

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransient: true,
    })
    previousQuantityToReceive: ui.fields.Numeric;

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        isReadOnly: true,
        isHidden: true,
        isTransient: true,
    })
    destinationRemainingQuantityInPackingUnit: ui.fields.Numeric;

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        isReadOnly: true,
        isHidden: true,
        isTransient: true,
    })
    destinationRemainingQuantityInStockUnit: ui.fields.Numeric;

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Conversion factor',
        isDisabled: true,
        isTransient: true,
        scale: 6,
        async onChange() {
            await this._onChange_packingUnitToStockUnitConversionFactor();
        },
    })
    packingUnitToStockUnitConversionFactor: ui.fields.Numeric;

    /**
     * OnChange packingUnitToStockUnitConversionFactor readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_packingUnitToStockUnitConversionFactor: AsyncVoidFunction = async () => {
        if (this._purchaseOrderLine) {
            const lineAlreadyEnter = this.lines.value.find(
                line => line.purchaseOrderLine?._id == this._purchaseShipmentLine?._id,
            );
            if (lineAlreadyEnter) {
                lineAlreadyEnter.stockDetails?.forEach(detail => {
                    if (
                        this.packingUnitToStockUnitConversionFactor.value !==
                        Number(detail.packingUnitToStockUnitConversionFactor)
                    ) {
                        this.packingUnitToStockUnitConversionFactor.value = Number(
                            detail.packingUnitToStockUnitConversionFactor,
                        );
                        this.packingUnitToStockUnitConversionFactor.focus();
                        this.$.showToast(
                            `${ui.localize(
                                '@sage/x3-purchasing/notification-The-conversion-factor-can-t-be-different-as-the-one-entered-previously-on-the-same-stock-line',
                                "The conversion factor can't be different as the one entered previously on the same stock line.",
                            )}`,
                            { type: 'error', timeout: 5000 },
                        );
                    }
                });
            }
        }
    };

    @ui.decorators.dropdownListField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Status',
        placeholder: 'Scan or select...',
        isMandatory() {
            return this._isMandatoryDestination();
        },
        isTransient: true,
        async onChange() {
            await this._onChange_status();
        },
    })
    status: ui.fields.DropdownList;

    /**
     * OnChange status readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_status: AsyncVoidFunction = async () => {
        this.status.getNextField(true)?.focus();
    };

    @ui.decorators.referenceField<MobilePurchaseReceipt, Container>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Container',
        node: '@sage/x3-master-data/Container',
        valueField: 'code',
        isAutoSelectEnabled: true,
        placeholder: 'Scan or select...',
        width: 'large',
        isTransient: true,
        canFilter: false,
        filter() {
            return {
                isInternal: { _eq: true },
                isActive: { _eq: true },
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
            }),
            ui.nestedFields.text({
                bind: 'containerType',
                title: 'Type',
            }),
            ui.nestedFields.text({
                bind: 'description',
                title: 'Description',
            }),
        ],
    })
    destinationContainer: ui.fields.Reference;

    @ui.decorators.referenceField<MobilePurchaseReceipt, LicensePlateNumber>({
        parent() {
            return this.destinationBlock;
        },
        title: 'License plate number',
        placeholder: 'Scan or select...',
        isFullWidth: true,
        node: '@sage/x3-stock-data/LicensePlateNumber',
        helperTextField: { container: { code: true } },
        isAutoSelectEnabled: true,
        valueField: 'code',
        canFilter: false,
        filter() {
            let licensePlateNumberFilter: Filter<LicensePlateNumber> = {
                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
                isActive: { _eq: true },
                _or: [
                    {
                        isSingleProduct: { _eq: true },
                        stock: { _atLeast: 1, product: { product: { code: this._productSite.product.code } } },
                    },
                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                    { isSingleProduct: { _eq: false } },
                ],
            };

            if (this.location.value?.code) {
                licensePlateNumberFilter._or = [
                    {
                        location: { code: this.location.value.code },
                    },
                    {
                        // to also include entries without location that have 'free' status
                        _and: [{ location: { code: undefined } }, { status: 'free' }],
                    },
                ];
            }

            if (this.destinationContainer.value) {
                licensePlateNumberFilter = {
                    ...licensePlateNumberFilter,
                    container: { code: (this.destinationContainer.value as Container)?.code },
                };
            }

            return licensePlateNumberFilter;
        },
        async onChange() {
            await this._onChange_licensePlateNumber();
        },
        isTransient: true,
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'License Plate Number',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Location',
                bind: 'location',
                valueField: 'code',
                title: 'Location',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Container',
                bind: 'container',
                valueField: 'code',
                title: 'Container',
                isReadOnly: true,
            }),
            ui.nestedFields.label({
                bind: 'status',
                title: 'Status',
                map(value: any) {
                    switch (value) {
                        case 'free':
                            return 'Free';
                        case 'inStock':
                            return 'In Stock';
                        default:
                            return '';
                    }
                },
                borderColor: ui.tokens.colorsYang100,
                optionType: '@sage/x3-stock-data/ContainerStatus',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isActive',
                isHidden: true,
            }),
            ui.nestedFields.checkbox({
                bind: 'isSingleProduct',
                isHidden: true,
            }),
        ],
    })
    licensePlateNumber: ui.fields.Reference;

    /**
     * OnChange licensePlateNumber readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_licensePlateNumber: AsyncVoidFunction = async () => {
        if (!this.licensePlateNumber.value) {
            this.location.isDisabled = false;
            this.location.value = null;
            await this.$.commitValueAndPropertyChanges(); // without this, when you clear out LPN and then, without tabbing out, click Location's lookup button directly, nothing will happen
            return;
        }

        // LPN can be associated with or without a location
        // if location is NOT popualted manually by the user

        if (!!this.licensePlateNumber.value?.location?.code) {
            this.location.isDisabled = true;
            this.location.value = this.licensePlateNumber.value.location;
            if ((this.licensePlateNumber.value as LicensePlateNumber).container) {
                this.destinationContainer.value = (this.licensePlateNumber.value as LicensePlateNumber).container;
                this.destinationContainer.isDisabled = true;
            } else {
                this.destinationContainer.isDisabled = false;
            }
        } else {
            if (this.location.isDisabled) {
                this.location.value = null;
                this.location.isDisabled = false;
            }
        }

        await this.$.commitValueAndPropertyChanges();
        this.licensePlateNumber.getNextField(true)?.focus();
    };

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Suggested Location',
        isReadOnly: true,
        width: 'large',
        isTransient: true,
    })
    suggestedLocation: ui.fields.Text;

    @ui.decorators.referenceField<MobilePurchaseReceipt, Location>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Location',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/Location',
        isAutoSelectEnabled: true,
        valueField: 'code',
        canFilter: false,
        isTransient: true,
        helperTextField: 'type',
        filter() {
            return {
                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
            };
        },
        async onChange() {
            await this._onChange_location();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Code',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                title: 'Type',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-system/Site',
                bind: 'stockSite',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-stock-data/Warehouse',
                bind: 'warehouse',
                valueField: 'code',
                isHidden: true,
            }),
        ],
        minLookupCharacters: 1,
        isFullWidth: true,
        isMandatory: true,
    })
    location: ui.fields.Reference;

    /**
     * OnChange location readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_location: AsyncVoidFunction = async () => {
        if (this.location.value) this.location.getNextField(true)?.focus();
        this.warehouse.value = !this.location.value?.warehouse ? '' : this.location.value.warehouse.code;
    };

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        isHidden: true,
        isReadOnly: true,
        isTransient: true,
    })
    warehouse: ui.fields.Text;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Serial number',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isTransient: true,
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.serialNumber, rawData);
        },
        async onChange() {
            await this._onChange_serialNumber();
        },
    })
    serialNumber: ui.fields.Text;

    // TODO: Make reference field after X3-193258 is resolved to add custom values on reference fields
    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Supplier lot',
        placeholder: 'Scan...',
        isFullWidth: true,
        validation: /^$|^[^|]+$/,
        isTransient: true,
        async onChange() {
            await this._onChange_supplierLot();
        },
    })
    supplierLot: ui.fields.Text;

    /**
     * OnChange supplierLot readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_supplierLot: AsyncVoidFunction = async () => {
        const supplierLot = this.supplierLot.value?.replace(/\s/g, '')?.toUpperCase() ?? '';
        if (supplierLot !== this.supplierLot.value) {
            this.supplierLot.value = supplierLot;
            await this.$.commitValueAndPropertyChanges();
        }

        if (this.supplierLot.value && this._purchaseStockManagementRules) {
            if (this._purchaseStockManagementRules.lotByDefault === 'supplierLot') {
                if (this._purchaseStockManagementRules.lotEntry === 'newLot') {
                    if (
                        (await controlLotReceipt(
                            this,
                            this.supplierLot.value,
                            this._productSite.product.code,
                            '6',
                            this.receiptSite.value?.code ?? '',
                        )) === false
                    ) {
                        this.supplierLot.value = null;
                        this.supplierLot.focus();
                    }
                }

                /**
                 *  Only when previous lot value has not same and has editable :
                 * - Assign new lot value.
                 * - If composite control block become invalid, reset it.
                 * - Commit page change.
                 * - If lot has editable, perform explicit lot change event (no _getLotValues) and stay in this field
                 * - Overwise next focus to next field focusable
                 */
                if (
                    this.supplierLot.value &&
                    this.supplierLot.value !== this.lot.value &&
                    this.previousSupplierLot.value === this.lot.value
                ) {
                    this.lot.value = this.supplierLot.value;
                    if (this._batchLot && this.lot.value !== this._batchLot) {
                        // this._resetCompositeDataControl();
                    }
                    await this.$.commitValueAndPropertyChanges();
                    if (!this.lot.isHidden && !this.lot.isDisabled && !this.lot.isReadOnly) {
                        await this._onChange_lot();
                        this.lot.focus();
                    } else {
                        this.lot.getNextField(true)?.focus();
                    }
                } else {
                    await this.$.commitValueAndPropertyChanges();
                }
            }
            this.previousSupplierLot.value = this.supplierLot.value;
        }
    };

    @ui.decorators.labelField<MobilePurchaseReceipt>({
        isHidden: true,
        isTransient: true,
    })
    previousSupplierLot: ui.fields.Label;

    @ui.decorators.filterSelectField<MobilePurchaseReceipt, LotsSites>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Lot',
        placeholder: 'Scan or select...',
        node: '@sage/x3-stock-data/LotsSites',
        valueField: 'lot',
        validation: /^$|^[^|]+$/,
        isTransient: true,
        isNewEnabled: true,
        minLookupCharacters: 1,
        canFilter: false,
        filter() {
            return {
                product: { code: this._productSite.product.code },
                ...(this.receiptSite.value && { storageSite: { code: this.receiptSite.value.code } }),
                stock: { _atLeast: 1 },
            };
        },
        columns: [
            ui.nestedFields.text({
                bind: 'lot',
                title: 'Lot',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'sublot',
                title: 'Sublot',
                isReadOnly: true,
            }),
        ],
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.lot, rawData);
        },
        async onChange() {
            await this._onChange_lot();
        },
    })
    lot: ui.fields.FilterSelect<Lot>;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        isHidden: true,
        isTransient: true,
    })
    svgLot: ui.fields.Text;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Sublot',
        placeholder: 'Scan...',
        isMandatory: true,
        isTransient: true,
        validation: /^$|^[^|]+$/,
        async onChange() {
            await this._onChange_sublot();
        },
    })
    sublot: ui.fields.Text;

    /**
     * OnChange sublot readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_sublot: AsyncVoidFunction = async () => {
        this.sublot.value = this.sublot.value?.toUpperCase() ?? '';
        await this._getLotValues();
        this._getNextField(this.sublot)?.focus();
    };

    @ui.decorators.referenceField<MobilePurchaseReceipt, MajorVersionStatus>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Major version',
        placeholder: 'Scan or select...',
        node: '@sage/x3-master-data/MajorVersionStatus',
        isAutoSelectEnabled: true,
        valueField: 'code',
        canFilter: false,
        isTransient: true,
        filter() {
            const filter: Filter<MajorVersionStatus> = {
                product: { code: this._productSite.product.code },
                status: { _in: ['prototypeVersion', 'activeVersion', 'stoppedVersion'] },
            };

            // if majorAndMinor managed, filter down further for major version with available minor versions
            if (!this.minorVersion.isHidden) {
                filter.minorVersions = {
                    _atLeast: 1,
                    type: { _eq: 'stock' },
                    useStatus: { _eq: 'availableToUse' },
                };
            }

            return filter;
        },
        async onChange() {
            await this._onChange_majorVersion();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'code',
                title: 'Major version',
                isReadOnly: true,
            }),
            ui.nestedFields.text({
                bind: 'status',
                title: 'Status',
            }),
            // (X3-227347) TODO: Obsolete: Having to specify & hide fields used in filter that don't need to be displayed
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
                isHidden: true,
            }),
        ],
        isMandatory: true,
    })
    majorVersion: ui.fields.Reference;

    /**
     * OnChange majorVersion readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_majorVersion: AsyncVoidFunction = async () => {
        // if not major & minor version managed
        if (this.minorVersion.isHidden) {
            if (this.majorVersion.value) {
                this.majorVersion.getNextField(true)?.focus();
            }
        }
        // if major version is cleared out, clear out minor version if any as well
        else if ((this.minorVersion.isDisabled = !this.majorVersion.value)) {
            this.minorVersion.value = null;
        } else {
            // Auto-populate minor version based on the last minor version available
            const minorVersions = await this._getMinorVersions(this.majorVersion.value.code);
            this.minorVersion.value = minorVersions[minorVersions.length - 1];
            this._getNextField(this.majorVersion)?.focus();
        }
    };

    @ui.decorators.referenceField<MobilePurchaseReceipt, ProductVersion>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Minor version',
        placeholder: 'Scan or select...',
        node: '@sage/x3-master-data/ProductVersion',
        isAutoSelectEnabled: true,
        valueField: 'minorVersion',
        canFilter: false,
        isTransient: true,
        filter() {
            return {
                product: { code: this._productSite.product.code },
                majorVersion: this.majorVersion.value?.code,
                type: 'stock',
                useStatus: 'availableToUse',
            };
        },
        async onChange() {
            await this._onChange_minorVersion();
        },
        columns: [
            ui.nestedFields.text({
                bind: 'minorVersion',
                title: 'Minor version',
                isReadOnly: true,
            }),
            ui.nestedFields.reference({
                bind: 'product',
                node: '@sage/x3-master-data/Product',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'majorVersion',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'type',
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: 'useStatus',
                isHidden: true,
            }),
        ],
        isMandatory: true,
    })
    minorVersion: ui.fields.Reference;

    /**
     * OnChange minorVersion readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_minorVersion: AsyncVoidFunction = async () => {
        if (this.minorVersion.value) this._getNextField(this.minorVersion)?.focus();
    };

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Potency %',
        isMandatory: true,
        scale: 4,
        isTransient: true,
        validation: /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/, // reg ex for any positive numbers (integers or decimals) excluding 0
        min: 0,
        max: 100,
        async onChange() {
            await this._onChange_potency();
        },
    })
    potency: ui.fields.Numeric;

    /**
     * OnChange destinationPotency readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_potency: AsyncVoidFunction = async () => {};

    @ui.decorators.dateField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Expiration date',
        placeholder: 'Enter...',
        isTransient: true,
        isHidden() {
            return this._productSite && this._selectedEntryTransaction
                ? this._productSite.product.lotManagementMode === 'notManaged' ||
                      this._productSite.product.expirationManagementMode === 'notManaged' ||
                      (!this._selectedEntryTransaction.isLotExpirationDateAllowed &&
                          (this._productSite.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                              this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                              this._productSite.product.expirationManagementMode === 'withoutRounding'))
                : true;
        },
        isMandatory() {
            return this._isMandatoryDestination() && !!this._productSite
                ? this._productSite?.product.lotManagementMode !== 'notManaged' &&
                      this._productSite?.product.expirationManagementMode === 'mandatoryEntry'
                : false;
        },
        async onInputValueChange(this, rawData: string): Promise<void> {
            await this.scan(this.expirationDate, rawData);
        },
        async onChange() {
            await this._onChange_expirationDate();
        },
    })
    expirationDate: ui.fields.Date;

    @ui.decorators.dateField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Enter...',
        title: 'Use-by date',
        isTransient: true,
        isHidden() {
            return this._productSite && this._selectedEntryTransaction
                ? this._productSite.product.lotManagementMode === 'notManaged' ||
                      this._productSite.product.expirationManagementMode === 'notManaged' ||
                      (!this._selectedEntryTransaction.isLotExpirationDateAllowed &&
                          (this._productSite.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                              this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                              this._productSite.product.expirationManagementMode === 'withoutRounding'))
                : true;
        },
        isMandatory() {
            return this._isMandatoryDestination() && !!this._productSite
                ? this._productSite.product.lotManagementMode !== 'notManaged' &&
                      this._productSite.product.expirationManagementMode === 'mandatoryEntry'
                : false;
        },
        isDisabled() {
            return this._productSite
                ? this._productSite?.product.expirationManagementMode === 'roundingBeginningMonth1' ||
                      this._productSite?.product.expirationManagementMode === 'roundingMonthEnd' ||
                      this._productSite?.product.expirationManagementMode === 'withoutRounding'
                : false;
        },
        maxDate() {
            return this.expirationDate.value ?? undefined;
        },
        async onChange() {
            await this._onChange_useByDate();
        },
    })
    useByDate: ui.fields.Date;

    /**
     * OnChange useByDate readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_useByDate: AsyncVoidFunction = async () => {};

    @ui.decorators.dateField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Expiration date',
        isDisabled: true,
        isHidden: true,
        isTransient: true,
    })
    calculatedExpirationDate: ui.fields.Date;

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom field 1',
        validation: /^$|^[^|]+$/,
        maxLength: 20,
        isTransient: true,
        async onChange() {
            await this._onChange_lotCustomField1();
        },
    })
    lotCustomField1: ui.fields.Text;

    /**
     * OnChange lotCustomField1 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_lotCustomField1: AsyncVoidFunction = async () => {};

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom field 2',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        async onChange() {
            await this._onChange_lotCustomField2();
        },
    })
    lotCustomField2: ui.fields.Text;

    /**
     * OnChange lotCustomField2 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_lotCustomField2: AsyncVoidFunction = async () => {};

    @ui.decorators.numericField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Scan...',
        title: 'Lot custom field 3',
        isTransient: true,
        async onChange() {
            await this._onChange_lotCustomField3();
        },
    })
    lotCustomField3: ui.fields.Numeric;

    /**
     * OnChange lotCustomField3 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_lotCustomField3: AsyncVoidFunction = async () => {};

    @ui.decorators.dateField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        placeholder: 'Enter...',
        title: 'Lot custom field 4',
        isTransient: true,
        async onChange() {
            await this._onChange_lotCustomField4();
        },
    })
    lotCustomField4: ui.fields.Date;

    /**
     * OnChange lotCustomField4 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_lotCustomField4: AsyncVoidFunction = async () => {};

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Identifier 1',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        async onChange() {
            await this._onChange_identifier1();
        },
    })
    identifier1: ui.fields.Text;

    /**
     * OnChange identifier1 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_identifier1: AsyncVoidFunction = async () => {};

    @ui.decorators.textField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Identifier 2',
        placeholder: 'Scan...',
        validation: /^$|^[^|]+$/,
        maxLength: 10,
        isTransient: true,
        async onChange() {
            await this._onChange_identifier2();
        },
    })
    identifier2: ui.fields.Text;

    /**
     * OnChange identifier2 readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_identifier2: AsyncVoidFunction = async () => {};

    @ui.decorators.dropdownListField<MobilePurchaseReceipt>({
        parent() {
            return this.destinationBlock;
        },
        title: 'Close PO line',
        placeholder: 'Select...',
        isFullWidth: true,
        isTransient: true,
        async onChange() {
            await this._onChange_closePoLine();
        },
    })
    closePoLine: ui.fields.DropdownList;

    /**
     * OnChange closePoLine readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_closePoLine: AsyncVoidFunction = async () => {};

    @ui.decorators.nestedGridField<MobilePurchaseReceipt, [PurchaseReceiptLineBinding, StockJournalClosePoLine]>({
        parent() {
            return this.receiptBlock;
        },
        bind: 'lines',
        canFilter: true,
        canActivate: true,
        isChangeIndicatorDisabled: true,
        canSelect: false,
        isTitleHidden: true,
        levels: [
            {
                node: '@sage/x3-purchasing/PurchaseReceiptLine',
                childProperty: 'stockDetails',
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'Id',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lineNumber',
                        title: 'Line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: { code: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInReceiptUnitReceived',
                        title: 'Quantity in receipt unit',
                        isReadOnly: true,
                        postfix(_value, rowValue) {
                            return rowValue?.receiptUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.receiptUnit?.numberOfDecimals ?? 0;
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, UnitOfMeasure>({
                        bind: 'receiptUnit',
                        title: 'Receipt unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: { code: true },
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, PurchaseOrder>({
                        bind: 'purchaseOrder',
                        title: 'Purchase order',
                        node: '@sage/x3-purchasing/PurchaseOrder',
                        valueField: { id: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'id',
                                title: 'Order no.',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'purchaseOrderLineNumber',
                        title: 'Purchase order line number',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, PurchaseOrderLine>({
                        bind: 'purchaseOrderLine',
                        title: 'Purchase order line',
                        node: '@sage/x3-purchasing/PurchaseOrderLine',
                        valueField: { lineNumber: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'lineNumber',
                                title: 'Number line',
                            }),
                            ui.nestedFields.text({
                                bind: 'sequenceNumber',
                                title: 'Sequence number',
                            }),
                        ],
                        filter(rowValue) {
                            return rowValue.purchaseOrder
                                ? {
                                      purchaseOrder: { _id: rowValue.purchaseOrder._id },
                                  }
                                : undefined;
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, Site>({
                        bind: 'receiptSite',
                        title: 'Receipt Site',
                        node: '@sage/x3-system/Site',
                        valueField: { code: true },
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, MiscellaneousTable>({
                        bind: 'stockMovementGroup',
                        title: 'Stock movement group',
                        node: '@sage/x3-system/MiscellaneousTable',
                        valueField: { code: true },
                        isHidden: true,
                        filter: {
                            glossaryId: {
                                glossaryId: 9,
                            },
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, PurchaseShipment>({
                        bind: 'purchaseShipment',
                        title: 'Purchase shipment',
                        node: '@sage/x3-purchasing/PurchaseShipment',
                        valueField: { id: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'id',
                                title: 'Order no.',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, PurchaseShipmentLine>({
                        bind: 'purchaseShipmentLine',
                        title: 'Purchase shipment line',
                        node: '@sage/x3-purchasing/PurchaseShipmentLine',
                        valueField: { lineNumber: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'lineNumber',
                                title: 'Number line',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, PurchaseReceiptLineBinding, Container>({
                        bind: 'container',
                        title: 'Container',
                        node: '@sage/x3-master-data/Container',
                        valueField: 'code',
                        isAutoSelectEnabled: true,
                        placeholder: 'Scan or select...',
                        width: 'large',
                        isTransient: true,
                        canFilter: false,
                        isDisabled(value, rowValue) {
                            return Boolean(rowValue?.licensePlateNumber);
                        },
                        isHidden() {
                            return (
                                this._purchaseStockManagementRules?.licensePlateNumberEntry !== 'mandatory' ||
                                !this._productSite?.isLicensePlateNumberManaged
                            );
                        },
                        filter() {
                            return {
                                isInternal: { _eq: true },
                                isActive: { _eq: true },
                            };
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.text({
                                bind: 'containerType',
                                title: 'Type',
                            }),
                            ui.nestedFields.text({
                                bind: 'description',
                                title: 'Description',
                            }),
                        ],
                    }),
                    ui.nestedFields.dropdownList<MobilePurchaseReceipt>({
                        title: 'Close PO line',
                        placeholder: 'Select...',
                        bind: 'balance',
                        isTransientInput: true,
                        isHidden() {
                            return !this._purchaseOrderLine;
                        },
                        options() {
                            return balanceOptions;
                        },
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId) {
                            if (await this.getSidebarData(rowId)) {
                                if (this._getSidebarPurchaseReceiptLine().length === 0) {
                                    this.lines.openSidebar(rowId, 1);
                                } else {
                                    this.lines.openSidebar(rowId, 0);
                                }
                            }
                        },
                    },
                    {
                        icon: 'delete',
                        title: 'Delete',
                        isDestructive: true,
                        onClick(rowId) {
                            this.lines.removeRecord(rowId, 0);
                            if (!this.lines.value?.length) {
                                this.$.setPageClean();
                                this._init()
                                    .then(() => {
                                        this._setActiveSection(Section.main, true).then(() => {});
                                    })
                                    .catch(error => {
                                        ui.console.error(`Error during initialization:\n${error}`);
                                    });
                            }
                        },
                    },
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        isHidden() {
                            return this._getSidebarPurchaseReceiptLine().length === 0;
                        },
                        async onClick(rowId) {
                            const _productCode = this.lines.getRecordValue(rowId, 0)?.product?.code;
                            if (_productCode && this.receiptSite.value) {
                                this._productSite = await this.getProductSite(
                                    _productCode,
                                    this.receiptSite.value?.code ?? '',
                                );
                            }
                            this._purchaseStockManagementRules = await findStockManagementRules(
                                this.receiptSite.value?.code ?? '',
                                this._productSite.product.productCategory.code,
                                '3',
                                this._selectedEntryTransaction?.stockMovementCode?.code ?? null,
                                this,
                            );
                            this._initPackingUnitFields();
                            this.receiptUnit.options = this._packingUnits.map(
                                packingUnit => packingUnit.node.packingUnit.name,
                            );
                            this.lines.openSidebar(rowId, 0);
                        },
                    },
                ],
                sidebar: {
                    layout() {
                        return {
                            mainSection: {
                                title: 'Purchase receipt line',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarPurchaseReceiptLine(),
                                    },
                                },
                            },
                        };
                    },
                },
            },
            {
                node: '@sage/x3-stock-data/StockJournal',
                orderBy: { _id: 1 },
                columns: [
                    ui.nestedFields.text({
                        bind: '_id',
                        title: 'Id',
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, Warehouse>({
                        node: '@sage/x3-stock-data/Warehouse',
                        bind: 'warehouse',
                        valueField: 'code',
                        isHidden: true,
                        canFilter: false,
                        isFullWidth: true,
                        filter() {
                            return this.receiptSite.value
                                ? {
                                      stockSite: { _id: { _eq: this.receiptSite.value.code } },
                                  }
                                : undefined;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInPackingUnit',
                        title: 'Quantity',
                        min: 0,
                        isNotZero: true,
                        isMandatory: true,
                        isFullWidth: true,
                        postfix(_value, rowValue) {
                            return rowValue?.packingUnit?.code;
                        },
                        scale(_value, rowValue) {
                            return rowValue?.packingUnit?.numberOfDecimals ?? 0;
                        },
                        validation(value: number) {
                            const regex = /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/; // reg ex for any positive numbers (integers or decimals) excluding 0
                            if ((value.toString().match(regex)?.length ?? 0) === 0) {
                                return ui.localize('@sage/x3-purchasing/validate-error-invalid-value', 'invalid value');
                            }
                            return undefined;
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }
                            if (this._purchaseOrderLine) {
                                const tmp = Math.pow(10, value.packingUnit?.numberOfDecimals ?? 1);
                                let quantity: number =
                                    Number(value.quantityInPackingUnit) *
                                    Number(value.packingUnitToStockUnitConversionFactor);
                                if (tmp !== 0) {
                                    quantity = Math.round(quantity * tmp) / tmp;
                                }
                                if (quantity > Number(value.remainingQuantityInStockUnit)) {
                                    if (!this._isOverReceiptAllowed) {
                                        await dialogMessage(
                                            this,
                                            'error',
                                            ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                                            ui.localize(
                                                '@sage/x3-purchasing/dialog-error-purchase-receipt-details-quantity-received-is-greater-than-the-ordered-quantity',
                                                `The quantity received is greater than the ordered quantity.`,
                                            ),
                                            {
                                                fullScreen: false,
                                                acceptButton: {
                                                    text: ui.localize('@sage/x3-purchasing/button-accept-ok', 'OK'),
                                                },
                                            },
                                        );
                                        value.quantityInPackingUnit = this.previousQuantityToReceive.value?.toString();
                                    } else {
                                        if (
                                            !(await dialogConfirmation(
                                                this,
                                                'warn',
                                                ui.localize('@sage/x3-purchasing/dialog-warning-title', 'Warning'),
                                                ui.localize(
                                                    '@sage/x3-purchasing/dialog-warn-purchase-receipt-details-quantity-message',
                                                    'Quantity larger than expected. Validate ?',
                                                ),
                                                {
                                                    fullScreen: false,
                                                    acceptButton: {
                                                        text: ui.localize(
                                                            '@sage/x3-purchasing/button-accept-yes',
                                                            'Yes',
                                                        ),
                                                    },
                                                    cancelButton: {
                                                        text: ui.localize('@sage/x3-purchasing/button-cancel-no', 'No'),
                                                    },
                                                },
                                            ))
                                        ) {
                                            value.quantityInPackingUnit =
                                                this.previousQuantityToReceive.value?.toString();
                                        }
                                    }
                                }
                            }
                            if (!value.quantityInPackingUnit) {
                                value.quantityInPackingUnit = '0';
                            }
                            if (Number(value.quantityInPackingUnit) !== this._quantityToReceiveRounded) {
                                this._quantityToReceiveNoRounded = Number(value.quantityInPackingUnit);
                            }
                            this._quantityToReceiveRounded = Number(value.quantityInPackingUnit);
                            this.previousQuantityToReceive.value = Number(value.quantityInPackingUnit);
                            if (this._purchaseOrderLine) {
                                if (
                                    Number(value.quantityInPackingUnit) *
                                        Number(value.packingUnitToStockUnitConversionFactor) ===
                                    Number(value.remainingQuantityInStockUnit)
                                ) {
                                    value.closePoLine = balanceOptions[0]; //Yes
                                } else if (
                                    Number(value.quantityInPackingUnit) *
                                        Number(value.packingUnitToStockUnitConversionFactor) <
                                    Number(value.remainingQuantityInStockUnit)
                                ) {
                                    value.closePoLine = balanceOptions[1]; //No
                                }
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, UnitOfMeasure>({
                        bind: 'packingUnit',
                        title: 'Unit',
                        node: '@sage/x3-master-data/UnitOfMeasure',
                        valueField: 'code',
                        isMandatory: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.numeric({
                                bind: 'numberOfDecimals',
                                title: 'Decimals',
                                isHidden: true,
                            }),
                        ],
                        filter() {
                            return {
                                code: {
                                    _in: this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.code),
                                },
                            };
                        },
                        isReadOnly() {
                            return (this.receiptUnit.options ?? []).length <= 1;
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }
                            if (value.packingUnit?.code) {
                                const selectedUnit = this._packingUnits.find(
                                    item => item.node.packingUnit.code === value.packingUnit?.code,
                                )?.node;
                                const previousUnit = this._packingUnits.find(
                                    item => item.node.packingUnit.code === this.previousReceiptUnitCode.value,
                                )?.node;
                                if (selectedUnit && previousUnit) {
                                    if (this._purchaseOrderLine) {
                                        if (this._remainingQuantityInitialInStockUnit) {
                                            // calcul packUnitTostockUnittoStockUnitConveriosnFactor : QTYSTU/QTYUOM
                                            if (this._remainingQuantityInitialInOrderUnit) {
                                                value.packingUnitToStockUnitConversionFactor = (
                                                    Number(this._remainingQuantityInitialInStockUnit) /
                                                    Number(this._remainingQuantityInitialInOrderUnit)
                                                ).toString();
                                                if (
                                                    selectedUnit.packingUnit.code ===
                                                    this._purchaseOrderLine.orderUnit.code
                                                )
                                                    selectedUnit.packingUnitToStockUnitConversionFactor =
                                                        value.packingUnitToStockUnitConversionFactor;
                                                if (
                                                    previousUnit.packingUnit.code ===
                                                    this._purchaseOrderLine.orderUnit.code
                                                )
                                                    previousUnit.packingUnitToStockUnitConversionFactor =
                                                        value.packingUnitToStockUnitConversionFactor;
                                            }
                                        }
                                    }
                                    let selectedUnitConversionFactor: string;
                                    if (Number(selectedUnit.packingUnitToStockUnitConversionFactor) !== 0) {
                                        selectedUnitConversionFactor = (
                                            Number(previousUnit.packingUnitToStockUnitConversionFactor) /
                                            Number(selectedUnit.packingUnitToStockUnitConversionFactor)
                                        ).toString();
                                    } else {
                                        selectedUnitConversionFactor = Number(
                                            previousUnit.packingUnitToStockUnitConversionFactor,
                                        ).toString();
                                    }
                                    const tmp = Math.pow(10, selectedUnit?.packingUnit.numberOfDecimals);
                                    if (value.quantityInPackingUnit) {
                                        this._quantityToReceiveNoRounded *= Number(selectedUnitConversionFactor);
                                        if (tmp !== 0) {
                                            this._quantityToReceiveRounded =
                                                Math.round(this._quantityToReceiveNoRounded * tmp) / tmp;
                                        }
                                        value.quantityInPackingUnit = this._quantityToReceiveRounded.toString();
                                    }
                                    if (this._purchaseOrderLine) {
                                        if (
                                            this._remainingQuantityInitialInStockUnit &&
                                            Number(value.packingUnitToStockUnitConversionFactor) !== 0
                                        )
                                            this._remainingQuantityNoRoundedInPackingUnit = Number(
                                                this._remainingQuantityInitialInStockUnit /
                                                    Number(value.packingUnitToStockUnitConversionFactor),
                                            );
                                        if (tmp !== 0) {
                                            this._remainingQuantityRoundedInPackingUnit =
                                                Math.round(this._remainingQuantityNoRoundedInPackingUnit * tmp) / tmp;
                                        } else {
                                            this._remainingQuantityNoRoundedInPackingUnit *= Number(
                                                value.packingUnitToStockUnitConversionFactor,
                                            );
                                            if (tmp !== 0) {
                                                this._remainingQuantityRoundedInPackingUnit =
                                                    Math.round(Number(value.quantityInPackingUnit) * tmp) / tmp;
                                            }
                                        }
                                        this.destinationRemainingQuantityInPackingUnit.value =
                                            this._remainingQuantityRoundedInPackingUnit;
                                    }
                                    this.previousQuantityToReceive.value = Number(value.quantityInPackingUnit);
                                    this.previousReceiptUnit.value = selectedUnit.packingUnit.name;
                                    this.previousReceiptUnitCode.value = selectedUnit.packingUnit.code;
                                    value.packingUnitToStockUnitConversionFactor =
                                        selectedUnit.packingUnitToStockUnitConversionFactor;
                                    this.lines.addOrUpdateRecordValue(value, 1, value._id);
                                }
                            }
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'quantityInStockUnit',
                        title: 'Quantity in stock unit',
                        isHidden: true,
                    }),
                    ui.nestedFields.numeric({
                        bind: 'packingUnitToStockUnitConversionFactor',
                        title: 'Packing unit to stock unit conversion factor',
                        isHidden() {
                            return (this.receiptUnit.options ?? []).length <= 1;
                        },
                        isMandatory: true,
                        isFullWidth: true,
                        scale: 6,
                        isDisabled(value, rowValue) {
                            if (rowValue?.packingUnit) {
                                const packingUnitIndex = this._packingUnits
                                    .map(packingUnit => packingUnit.node.packingUnit.code)
                                    .indexOf(rowValue.packingUnit.code);
                                if (packingUnitIndex !== -1) {
                                    const selectedUnit = this._packingUnits[packingUnitIndex].node;
                                    return !selectedUnit.isPackingFactorEntryAllowed;
                                } else {
                                    return true;
                                }
                            } else {
                                return true;
                            }
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (this._purchaseOrderLine) {
                                const lineAlreadyEnter = this.lines.value.find(
                                    line =>
                                        line._id !== value._id &&
                                        line.purchaseOrderLine?._id === this._purchaseShipmentLine?._id,
                                );
                                if (lineAlreadyEnter) {
                                    lineAlreadyEnter.stockDetails?.forEach(detail => {
                                        if (
                                            Number(value.packingUnitToStockUnitConversionFactor) !==
                                            Number(detail.packingUnitToStockUnitConversionFactor)
                                        ) {
                                            value.packingUnitToStockUnitConversionFactor =
                                                detail.packingUnitToStockUnitConversionFactor;
                                            this.$.showToast(
                                                `${ui.localize(
                                                    '@sage/x3-purchasing/notification-The-conversion-factor-can-t-be-different-as-the-one-entered-previously-on-the-same-stock-line',
                                                    "The conversion factor can't be different as the one entered previously on the same stock line.",
                                                )}`,
                                                { type: 'error', timeout: 5000 },
                                            );
                                        }
                                    });
                                }
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, StockStatus>({
                        bind: 'status',
                        title: 'Status',
                        node: '@sage/x3-stock-data/StockStatus',
                        valueField: { code: true },
                        isMandatory: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'containerDetail',
                        node: '@sage/x3-master-data/Container',
                        title: 'Container',
                        valueField: 'code',
                        placeholder: 'Scan or select...',
                        isFullWidth: true,
                        isTransient: true,
                        filter() {
                            return {
                                isInternal: { _eq: true },
                                isActive: { _eq: true },
                            };
                        },
                        isReadOnly(_value, rowValue) {
                            return !!rowValue?.licensePlateNumber?.container;
                        },
                        isHidden() {
                            return (
                                this._purchaseStockManagementRules.licensePlateNumberEntry !== 'mandatory' ||
                                !this._productSite.isLicensePlateNumberManaged
                            );
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            this._containerDetail = value.containerDetail;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.text({
                                bind: 'containerType',
                                title: 'Type',
                            }),
                            ui.nestedFields.text({
                                bind: 'description',
                                title: 'Description',
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, LicensePlateNumber>({
                        bind: 'licensePlateNumber',
                        title: 'License plate number',
                        valueField: 'code',
                        node: '@sage/x3-stock-data/LicensePlateNumber',
                        isFullWidth: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-stock-data/Location',
                                bind: 'location',
                                valueField: 'code',
                                title: 'Location',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Container',
                                bind: 'container',
                                valueField: 'code',
                                title: 'Container',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-system/Site',
                                bind: 'stockSite',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isActive',
                                isHidden: true,
                            }),
                            ui.nestedFields.checkbox({
                                bind: 'isSingleProduct',
                                isHidden: true,
                            }),
                        ],
                        canFilter: false,
                        isHidden() {
                            return !this._productSite.isLicensePlateNumberManaged;
                        },
                        filter() {
                            let licensePlateNumberFilter: Filter<LicensePlateNumber> = {
                                ...(this.receiptSite.value && { stockSite: { code: this.receiptSite.value.code } }),
                                isActive: { _eq: true },
                                _or: [
                                    {
                                        isSingleProduct: { _eq: true },
                                        stock: {
                                            _atLeast: 1,
                                            product: { product: { code: this._productSite.product.code } },
                                        },
                                    },
                                    { isSingleProduct: { _eq: true }, stock: { _none: true } },
                                    { isSingleProduct: { _eq: false } },
                                ],
                            };
                            if (this._locationDetail) {
                                licensePlateNumberFilter._or = [
                                    {
                                        location: { code: this._locationDetail.code },
                                    },
                                    {
                                        // to also include entries without location that have 'free' status
                                        _and: [{ location: { code: undefined } }, { status: 'free' }],
                                    },
                                ];
                            }
                            if (this._containerDetail) {
                                licensePlateNumberFilter = {
                                    ...licensePlateNumberFilter,
                                    container: { code: this._containerDetail },
                                };
                            }
                            return licensePlateNumberFilter;
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.licensePlateNumber) {
                                if (value.licensePlateNumber.location) {
                                    value.location = value.licensePlateNumber.location;
                                    this._locationDetail = value.licensePlateNumber.location;
                                }
                                if (value.licensePlateNumber.container) {
                                    value.containerDetail = value.licensePlateNumber.container.code;
                                    this._containerDetail = value.licensePlateNumber.container.code;
                                }
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, Location>({
                        bind: 'location',
                        title: 'Location',
                        node: '@sage/x3-stock-data/Location',
                        valueField: { code: true },
                        isFullWidth: true,
                        isHidden() {
                            return !this._productSite.isLocationManaged;
                        },
                        isMandatory() {
                            return this._productSite.isLocationManaged;
                        },
                        isReadOnly(_value, rowValue) {
                            return this._productSite.isLicensePlateNumberManaged && !!rowValue?.licensePlateNumber;
                        },
                        filter() {
                            return {
                                ...(this.receiptSite.value && {
                                    stockSite: { code: this.receiptSite.value.code },
                                }),
                                category: { _nin: ['subcontract', 'customer'] },
                            };
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.location) {
                                value.warehouse = value.location.warehouse;
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                            this._locationDetail = value.location;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                            ui.nestedFields.reference<MobilePurchaseReceipt, Location, Warehouse>({
                                node: '@sage/x3-stock-data/Warehouse',
                                bind: 'warehouse',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.text({
                        bind: 'serialNumber',
                        title: 'Serial number',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return ['notManaged', 'issued'].includes(
                                this._productSite.product.serialNumberManagementMode,
                            );
                        },
                        isMandatory() {
                            return (
                                !['notManaged', 'issued'].includes(
                                    this._productSite.product.serialNumberManagementMode,
                                ) && !this._productSite.product.serialSequenceNumber
                            );
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'supplierLot',
                        title: 'Supplier lot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return this._productSite.product.lotManagementMode === 'notManaged';
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.supplierLot) {
                                value.supplierLot = value.supplierLot.toUpperCase();
                                if (this._purchaseStockManagementRules.lotByDefault === 'supplierLot') {
                                    if (this._purchaseStockManagementRules.lotEntry === 'newLot') {
                                        if (
                                            !(await controlLotReceipt(
                                                this,
                                                value.supplierLot,
                                                this._productSite.product.code,
                                                '6',
                                                this.receiptSite.value?.code ?? '',
                                            ))
                                        ) {
                                            value.supplierLot = undefined;
                                        }
                                    }
                                    if (
                                        value.supplierLot &&
                                        value.supplierLot != value.lot &&
                                        this.previousSupplierLot.value === value.lot
                                    ) {
                                        if (
                                            value.sublot &&
                                            this._productSite.product.lotManagementMode === 'lotAndSublot' &&
                                            value.lot != value.supplierLot
                                        ) {
                                            value.sublot = undefined;
                                        }
                                        value.lot = value.supplierLot;
                                        if (
                                            value.lot &&
                                            (this._productSite.product.lotManagementMode !== 'lotAndSublot' ||
                                                value.sublot)
                                        ) {
                                            this._lots = await this._getLots(
                                                this._productSite.product.code,
                                                value.lot,
                                                value.sublot ?? null,
                                            );
                                            if (this._lots.length !== 0) {
                                                value.supplierLot = this._lots[0].supplierLot;
                                                value.lotCustomField1 = this._lots[0].lotCustomField1;
                                                value.lotCustomField2 = this._lots[0].lotCustomField2;
                                                value.lotCustomField3 = this._lots[0].lotCustomField3;
                                                value.lotCustomField4 = this._lots[0].lotCustomField4;
                                                value.majorVersion = this._lots[0].majorVersion
                                                    ? {
                                                          code: this._lots[0].majorVersion.code,
                                                      }
                                                    : undefined;
                                                value.minorVersion = this._lots[0].minorVersion
                                                    ? this._lots[0].minorVersion
                                                    : undefined;
                                            } else {
                                                value.lotCustomField1 = undefined;
                                                value.lotCustomField2 = undefined;
                                                value.lotCustomField3 = undefined;
                                                value.lotCustomField4 = undefined;
                                            }
                                        } else {
                                            value.lotCustomField1 = undefined;
                                            value.lotCustomField2 = undefined;
                                            value.lotCustomField3 = undefined;
                                            value.lotCustomField4 = undefined;
                                            this._lots = [];
                                        }
                                    }
                                }
                                this.previousSupplierLot.value = value.supplierLot ?? null;
                                this.lines.addOrUpdateRecordValue(value, 1, value._id);
                            }
                        },
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'lot',
                        node: '@sage/x3-stock-data/LotsSites',
                        title: 'Lot',
                        valueField: 'lot',
                        placeholder: 'Scan or select...',
                        isFullWidth: true,
                        isNewEnabled: true,
                        validation: /^$|^[^|]+$/,
                        filter() {
                            return {
                                product: { code: this._productSite.product.code },
                                ...(this.receiptSite.value && { storageSite: { code: this.receiptSite.value.code } }),
                                stock: { _atLeast: 1 },
                            };
                        },
                        isMandatory() {
                            return (
                                ['lotAndSublot', 'mandatoryLot'].includes(
                                    this._productSite.product.lotManagementMode,
                                ) &&
                                !this._productSite.product.lotSequenceNumber &&
                                this._purchaseStockManagementRules.lotByDefault !== 'documentNumber'
                            );
                        },
                        isDisabled() {
                            return this._purchaseStockManagementRules.lotEntry === 'no';
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'lot',
                                title: 'Lot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'sublot',
                                title: 'Sublot',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                        isHidden() {
                            return this._productSite.product.lotManagementMode === 'notManaged';
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value.lot || !value._id) {
                                return;
                            }

                            value.lot = value.lot.toUpperCase();
                            if (
                                value.lot &&
                                this._purchaseStockManagementRules.lotEntry === 'newLot' &&
                                (await controlLotReceipt(
                                    this,
                                    value.lot,
                                    this._productSite.product.code,
                                    '6',
                                    this.receiptSite.value?.code ?? '',
                                )) === false
                            ) {
                                value.lot = undefined;
                            }
                            if (
                                value.lot &&
                                (this._productSite.product.lotManagementMode !== 'lotAndSublot' || value.sublot)
                            ) {
                                this._lots = await this._getLots(
                                    this._productSite.product.code,
                                    value.lot,
                                    value.sublot ?? null,
                                );
                                if (this._lots.length !== 0) {
                                    value.supplierLot = this._lots[0].supplierLot;
                                    value.lotCustomField1 = this._lots[0].lotCustomField1;
                                    value.lotCustomField2 = this._lots[0].lotCustomField2;
                                    value.lotCustomField3 = this._lots[0].lotCustomField3;
                                    value.lotCustomField4 = this._lots[0].lotCustomField4;
                                    value.majorVersion = this._lots[0].majorVersion
                                        ? {
                                              code: this._lots[0].majorVersion.code,
                                          }
                                        : undefined;
                                    value.minorVersion = this._lots[0].minorVersion
                                        ? this._lots[0].minorVersion
                                        : undefined;
                                } else {
                                    value.lotCustomField1 = undefined;
                                    value.lotCustomField2 = undefined;
                                    value.lotCustomField3 = undefined;
                                    value.lotCustomField4 = undefined;
                                }
                            } else {
                                value.lotCustomField1 = undefined;
                                value.lotCustomField2 = undefined;
                                value.lotCustomField3 = undefined;
                                value.lotCustomField4 = undefined;
                                this._lots = [];
                            }
                            this.lines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'sublot',
                        title: 'Sublot',
                        width: 'large',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        isFullWidth: true,
                        isHidden() {
                            return this._productSite.product.lotManagementMode !== 'lotAndSublot';
                        },
                        isMandatory() {
                            return this._productSite.product.lotManagementMode === 'lotAndSublot';
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.sublot) {
                                value.sublot = value.sublot.toUpperCase();
                            }
                            if (value.lot && value.sublot) {
                                this._lots = await this._getLots(
                                    this._productSite.product.code,
                                    value.lot,
                                    value.sublot,
                                );
                                if (this._lots.length !== 0) {
                                    value.supplierLot = this._lots[0].supplierLot;
                                    value.lotCustomField1 = this._lots[0].lotCustomField1;
                                    value.lotCustomField2 = this._lots[0].lotCustomField2;
                                    value.lotCustomField3 = this._lots[0].lotCustomField3;
                                    value.lotCustomField4 = this._lots[0].lotCustomField4;
                                    value.majorVersion = this._lots[0].majorVersion
                                        ? {
                                              code: this._lots[0].majorVersion.code,
                                          }
                                        : undefined;
                                    value.minorVersion = this._lots[0].minorVersion
                                        ? this._lots[0].minorVersion
                                        : undefined;
                                } else {
                                    value.lotCustomField1 = undefined;
                                    value.lotCustomField2 = undefined;
                                    value.lotCustomField3 = undefined;
                                    value.lotCustomField4 = undefined;
                                }
                            } else {
                                value.lotCustomField1 = undefined;
                                value.lotCustomField2 = undefined;
                                value.lotCustomField3 = undefined;
                                value.lotCustomField4 = undefined;
                                this._lots = [];
                            }
                            this.lines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, Product>({
                        bind: 'product',
                        title: 'Product',
                        node: '@sage/x3-master-data/Product',
                        valueField: 'code',
                        isReadOnly: true,
                        isHidden: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, MajorVersionStatus>({
                        bind: 'majorVersion',
                        title: 'Major version',
                        node: '@sage/x3-master-data/MajorVersionStatus',
                        valueField: { code: true },
                        isFullWidth: true,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._productSite.product.stockVersionMode ||
                                this._productSite.product.stockVersionMode === 'no'
                            );
                        },
                        isMandatory() {
                            return (
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                this._productSite.product.stockVersionMode &&
                                this._productSite.product.stockVersionMode !== 'no'
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                        filter() {
                            return {
                                product: { code: this._productSite.product.code },
                                status: { _in: ['prototypeVersion', 'activeVersion', 'stoppedVersion'] },
                                ...(this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._productSite.product.stockVersionMode === 'majorAndMinor' && {
                                        minorVersions: {
                                            _atLeast: 1,
                                            type: { _eq: 'stock' },
                                            useStatus: { _eq: 'availableToUse' },
                                        },
                                    }),
                            };
                        },
                        async onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value._id) {
                                return;
                            }

                            if (value.majorVersion && value.majorVersion.code) {
                                const minorVersions = await this._getMinorVersions(value.majorVersion.code);
                                value.minorVersion =
                                    minorVersions.length > 0
                                        ? minorVersions[minorVersions.length - 1].minorVersion
                                        : undefined;
                            }
                            this.lines.addOrUpdateRecordValue(value, 1, value._id);
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Major version',
                            }),
                            ui.nestedFields.text({
                                bind: 'status',
                                title: 'Status',
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.filterSelect({
                        bind: 'minorVersion',
                        title: 'Minor version',
                        node: '@sage/x3-master-data/ProductVersion',
                        valueField: 'minorVersion',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                this._productSite.product.stockVersionMode !== 'majorAndMinor'
                            );
                        },
                        isMandatory() {
                            return (
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                this._productSite.product.stockVersionMode === 'majorAndMinor'
                            );
                        },
                        filter(rowValue) {
                            return rowValue.majorVersion
                                ? {
                                      product: { code: this._productSite.product.code },
                                      majorVersion: rowValue.majorVersion.code,
                                      type: 'stock',
                                      useStatus: 'availableToUse',
                                  }
                                : {};
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                        columns: [
                            ui.nestedFields.text({
                                bind: 'minorVersion',
                                title: 'Minor version',
                                isReadOnly: true,
                            }),
                            ui.nestedFields.reference({
                                node: '@sage/x3-master-data/Product',
                                bind: 'product',
                                valueField: 'code',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'majorVersion',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'type',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'useStatus',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                        ],
                    }),
                    ui.nestedFields.numeric({
                        bind: 'potency',
                        title: 'Potency %',
                        isFullWidth: true,
                        placeholder: 'Scan...',
                        scale: 4,
                        validation: /^([1-9][0-9]*(\.[0-9]+)?|[0]+\.[0-9]*[1-9][0-9]*)$/, // reg ex for any positive numbers (integers or decimals) excluding 0
                        min: 0,
                        max: 100,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotPotencyAllowed ||
                                this._productSite.product.stockManagementMode !== 'potencyManaged'
                            );
                        },
                        isMandatory() {
                            return Boolean(
                                this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._selectedEntryTransaction?.isLotPotencyAllowed &&
                                    this._productSite.product.stockManagementMode === 'potencyManaged',
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'expirationDate',
                        placeholder: 'Enter...',
                        title: 'Expiration date',
                        isHidden() {
                            return (
                                (this._productSite.product.lotManagementMode === 'notManaged' ||
                                    this._productSite.product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedEntryTransaction?.isLotExpirationDateAllowed &&
                                        (this._productSite.product.expirationManagementMode ===
                                            'roundingBeginningMonth1' ||
                                            this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._productSite.product.expirationManagementMode ===
                                                'withoutRounding'))) ??
                                false
                            );
                        },
                        isMandatory() {
                            return (
                                (this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._productSite.product.expirationManagementMode === 'mandatoryEntry') ??
                                false
                            );
                        },
                        onChange(_id, value: ExtractEdgesPartial<StockJournalClosePoLine>) {
                            if (!value.expirationDate) {
                                return;
                            }
                            value.useByDate =
                                useByDateDefaultValue(
                                    value.expirationDate,
                                    this._effectiveDate,
                                    this._productSite.product.expirationManagementMode,
                                    Number(this._productSite.product.useByDateCoefficient),
                                ) ?? undefined;
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'useByDate',
                        placeholder: 'Enter...',
                        title: 'Use-by date',
                        isHidden() {
                            return (
                                (this._productSite.product.lotManagementMode === 'notManaged' ||
                                    this._productSite.product.expirationManagementMode === 'notManaged' ||
                                    (!this._selectedEntryTransaction?.isLotExpirationDateAllowed &&
                                        (this._productSite.product.expirationManagementMode ===
                                            'roundingBeginningMonth1' ||
                                            this._productSite.product.expirationManagementMode === 'roundingMonthEnd' ||
                                            this._productSite.product.expirationManagementMode ===
                                                'withoutRounding'))) ??
                                false
                            );
                        },
                        isMandatory() {
                            return (
                                (this._productSite.product.lotManagementMode !== 'notManaged' &&
                                    this._productSite.product.expirationManagementMode === 'mandatoryEntry') ??
                                false
                            );
                        },
                        isDisabled() {
                            return ['roundingBeginningMonth1', 'roundingMonthEnd', 'withoutRounding'].includes(
                                this._productSite.product.expirationManagementMode,
                            );
                        },
                        maxDate(rowValue) {
                            return rowValue?.expirationDate;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField1',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 1',
                        validation: /^$|^[^|]+$/,
                        maxLength: 20,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField1Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'lotCustomField2',
                        placeholder: 'Scan...',
                        title: 'Lot custom text 2',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField2Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'lotCustomField3',
                        placeholder: 'Enter...',
                        title: 'Lot custom number',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField3Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.date({
                        bind: 'lotCustomField4',
                        placeholder: 'Enter...',
                        title: 'Lot custom date',
                        isHidden() {
                            return (
                                this._productSite.product.lotManagementMode === 'notManaged' ||
                                !this._selectedEntryTransaction?.isLotCustomField4Allowed
                            );
                        },
                        isDisabled() {
                            return this._lots && this._lots.length > 0;
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier1',
                        title: 'Identifier 1',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHidden() {
                            return this._selectedEntryTransaction?.identifier1Detail === 'hidden';
                        },
                        isDisabled() {
                            return this._selectedEntryTransaction?.identifier1Detail !== 'entered';
                        },
                    }),
                    ui.nestedFields.text({
                        bind: 'identifier2',
                        title: 'Identifier 2',
                        placeholder: 'Scan...',
                        validation: /^$|^[^|]+$/,
                        maxLength: 10,
                        isHidden() {
                            return (
                                this._selectedEntryTransaction?.identifier2Detail !== 'entered' &&
                                this._selectedEntryTransaction?.identifier1Detail !== 'displayed'
                            );
                        },
                        isDisabled() {
                            return this._selectedEntryTransaction?.identifier2Detail !== 'entered';
                        },
                    }),
                    ui.nestedFields.dropdownList<MobilePurchaseReceipt>({
                        title: 'Close PO line',
                        placeholder: 'Select...',
                        bind: 'closePoLine',
                        isTransient: true,
                        isHidden() {
                            return !this._purchaseOrderLine;
                        },
                        options() {
                            return balanceOptions;
                        },
                    }),
                    ui.nestedFields.numeric({
                        bind: 'remainingQuantityInStockUnit',
                        isTransient: true,
                        isHidden: true,
                    }),
                    ui.nestedFields.reference<MobilePurchaseReceipt, StockJournalClosePoLine, Site>({
                        bind: 'stockSite',
                        title: 'Site',
                        node: '@sage/x3-system/Site',
                        valueField: { code: true },
                        isReadOnly: true,
                        columns: [
                            ui.nestedFields.text({
                                bind: '_id',
                                isHidden: true,
                            }),
                            ui.nestedFields.text({
                                bind: 'code',
                                title: 'Code',
                            }),
                        ],
                    }),
                    ui.nestedFields.date({
                        bind: 'effectiveDate',
                        title: 'Effective date',
                        isReadOnly: true,
                    }),
                ],
                dropdownActions: [
                    {
                        icon: 'box_arrow_left',
                        title: 'Edit line',
                        isDisabled() {
                            return false;
                        },
                        async onClick(rowId) {
                            if (await this.getSidebarData(rowId)) {
                                this.lines.openSidebar(rowId, 1);
                            }
                        },
                    },
                ],
                sidebar: {
                    title: 'Purchase receipt',
                    layout() {
                        return {
                            mainSection: {
                                title: 'Stock details',
                                blocks: {
                                    mainBlock: {
                                        fields: this._getSidebarStockJournal(),
                                    },
                                },
                            },
                        };
                    },
                    async onRecordConfirmed(_id: string, recordValue: StockJournalClosePoLine | undefined) {
                        const _purchaseReceiptLine = this.lines.getRecordValue(_id, 0);
                        if (_purchaseReceiptLine) {
                            _purchaseReceiptLine.quantityInReceiptUnitReceived = recordValue?.quantityInPackingUnit;
                            _purchaseReceiptLine.receiptUnit = recordValue?.packingUnit;
                            _purchaseReceiptLine.receiptUnitToPurchaseUnitConversionFactor =
                                recordValue?.packingUnitToStockUnitConversionFactor;
                            _purchaseReceiptLine.receiptUnitToStockUnitConversionFactor =
                                recordValue?.packingUnitToStockUnitConversionFactor;
                            _purchaseReceiptLine.balance = recordValue?.closePoLine;
                            _purchaseReceiptLine.container = recordValue?.containerDetail;
                            this.lines.setRecordValue(_purchaseReceiptLine, 0);
                        }
                    },
                },
            },
        ],
    })
    lines: ui.fields.NestedGrid<[PurchaseReceiptLineBinding, StockJournalClosePoLine]>;

    @ui.decorators.tableField<MobilePurchaseReceipt>({
        parent() {
            return this.shipmentBlock;
        },
        node: '@sage/x3-purchasing/PurchaseShipmentLine',
        title: 'Shipments',
        isChangeIndicatorDisabled: false,
        canFilter: false,
        canSelect: false,
        canExport: false,
        canResizeColumns: false,
        canUserHideColumns: false,
        isTransient: true,
        isFullWidth: true,
        isDisabled: false,
        hasSearchBoxMobile: false,
        cardView: true,
        displayMode: ui.fields.TableDisplayMode.compact,
        mobileCard: undefined,
        columns: [
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'code',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/Product',
                bind: 'product',
                valueField: 'localizedDescription1',
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-purchasing/PurchaseShipment',
                bind: 'purchaseShipment',
                valueField: 'id',
            }),
            ui.nestedFields.text({
                bind: 'lineNumber',
            }),
            ui.nestedFields.label({
                title: 'Shipped qty',
                prefix: 'Shipped qty.',
                postfix(_value, rowData) {
                    return rowData?.purchaseUnit?.code;
                },
                isTitleHidden: false,
                isTransient: true,
                bind: '_id',
                map(_value, rowData) {
                    return Number.isInteger(Number(rowData.quantityInPurchaseUnitShipped))
                        ? rowData.quantityInPurchaseUnitShipped.toString()
                        : rowData.quantityInPurchaseUnitShipped.toFixed(QTY_DECIMAL_PLACES);
                },
            }),
            ui.nestedFields.label({
                title: 'Expected qty',
                prefix: 'Expected qty.',
                postfix(_value, rowData) {
                    return rowData?.purchaseUnit?.code;
                },
                isTransient: true,
                bind: '_id',
                map(_value, rowData) {
                    let quantityInStockUnitReceived = 0;
                    let quantityInPurchaseUnitExpected = rowData.quantityInPurchaseUnitExpected;

                    if (this.lines.value.length > 0) {
                        let quantityInStockUnitExpected = Number(rowData.quantityInStockUnitShipped);
                        this.lines.value
                            .filter(line => line.purchaseShipmentLine?._id === rowData._id)
                            .map(
                                line =>
                                    (quantityInStockUnitReceived +=
                                        Number(line.quantityInReceiptUnitReceived) *
                                        Number(line.receiptUnitToStockUnitConversionFactor)),
                            );

                        if (quantityInStockUnitReceived > 0) {
                            quantityInPurchaseUnitExpected = Math.max(
                                (Number(quantityInStockUnitExpected) - Number(quantityInStockUnitReceived)) /
                                    Number(rowData.product.purchaseUnitToStockUnitConversionFactor),
                                0,
                            );
                        }
                    }

                    return Number.isInteger(Number(quantityInPurchaseUnitExpected))
                        ? quantityInPurchaseUnitExpected.toString()
                        : quantityInPurchaseUnitExpected.toFixed(QTY_DECIMAL_PLACES);
                },
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPurchaseUnitShipped',
                title: 'Shipped qty',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInPurchaseUnitExpected',
                title: 'Expected qty',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'purchaseUnit',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInOrderUnitShipped',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'quantityInStockUnitShipped',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: 'orderUnitToPurchaseUnitConversionFactor',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'stockUnit',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.reference({
                node: '@sage/x3-master-data/UnitOfMeasure',
                bind: 'orderUnit',
                valueField: 'code',
                isHidden: true,
            }),
            ui.nestedFields.numeric({
                bind: { product: { purchaseUnitToStockUnitConversionFactor: true } },
                isReadOnly: true,
                isHidden: true,
            }),
            ui.nestedFields.text({
                bind: '_id',
                isHidden: true,
            }),
        ],
        async onRowClick(recordId: string, rowItem: PurchaseShipmentLine) {
            this._purchaseShipmentId = rowItem.purchaseShipment?.id;
            this._purchaseShipmentLineNumber = rowItem.lineNumber?.toString();
            if (this._purchaseShipmentId && this._purchaseShipmentLineNumber) {
                this._purchaseShipmentLine = await this.getPurchaseShipmentLine(
                    this._purchaseShipmentId,
                    Number(this._purchaseShipmentLineNumber),
                );
                await this._setActiveSection(Section.destination);
            }
        },
    })
    shipments: ui.fields.Table<any>;

    /**
     *
     * Methods section
     *
     *
     */

    private async _init(): Promise<boolean> {
        this.receiptDate.value = DateValue.today().toString();
        const purchaseOrderId = this.$.queryParameters.purchaseOrderId;
        const supplierCode = this.$.queryParameters.supplierCode;
        try {
            const receiptSiteCode = await getSelectedStockSite(
                this,
                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-purchasing/dialog-error-location-inquiry-set-site',
                    'Define a default stock site on the user function profile.',
                ),
            );

            if (!receiptSiteCode) {
                return false;
            }

            this.receiptSite.value = {
                _id: receiptSiteCode,
                code: receiptSiteCode,
            };

            // perform screen initialization for current workFlow enabled
            if (!(await this._initControlManager(receiptSiteCode))) {
                return false;
            }

            await this._getProductSearch();
            await this._initTransaction();
            this.closePoLine.options = balanceOptions;
        } catch (error) {
            this.$.showToast(error.message, { type: 'error' });
            return false;
        }

        // Preloading purchaseOrderId and supplierCode into the purchaseOrder and supplier code fields
        if (purchaseOrderId && supplierCode) {
            this.purchaseOrder.value = {
                id: purchaseOrderId,
                _id: purchaseOrderId,
                orderFromSupplier: {
                    code: { code: supplierCode },
                    _id: supplierCode,
                },
            };
            await this.$.commitValueAndPropertyChanges();
        }

        this.supplierPackingSlip.value = null;
        this.supplierPackingSlip.isHidden = true;
        this.supplierPackingSlipLink.isHidden = false;
        this.destination.value = this.$.storage.get('mobile-label-destination')?.toString() ?? '';
        this.document.value = this.$.storage.get('mobile-document-destination')?.toString() ?? '';
        this.legalCompany.value = this.receiptSite.value
            ? await this._getLegalCompany(this.receiptSite.value.code ?? '')
            : null;

        this._isOverReceiptAllowed =
            (await readParameterValue('RCPORD', this.$.userCode ?? '', this.receiptSite.value.code ?? '', this)) === 2;

        // Disallow direct receipts based on PTHDIR setting. Storing value in local storage so only need to read once.
        this._directReceiptAllowed =
            (await readParameterValue('PTHDIR', this.$.userCode ?? '', this.receiptSite.value.code ?? '', this)) == 2;

        await this._setActiveSection(Section.main, true);
        return true;
    }

    /**
     * Initialize ControlManager
     * @returns true when ControlManager is usable
     */
    private async _initControlManager(site: string): Promise<boolean> {
        const isPOWorkflow = !!this.purchaseOrder.value;
        return await this.createAndInitServiceManagement(
            site,
            mobileApplicationKey,
            BarcodeTransactionFlow.purchaseReceipt,
            this._getDictionaryFieldSupported(),
            undefined,
            this._checkCompositeDataAllowed,
        );
    }

    /**
     * Update fields mappings
     * @returns true when ControlManager is usable
     */
    /** @internal */
    private async _updateControlManager(): Promise<boolean> {
        const isPOWorkflow = !!this.purchaseOrder.value;
        const _ = this._getDictionaryFieldSupported();
        return await this.setScreenFieldSupported(this._getDictionaryFieldSupported());
    }

    /**
     * Get dictionary field supported
     *  @returns DictionaryFieldSupported
     */
    private _getDictionaryFieldSupported(): DictionaryFieldSupported {
        const isPOWorkflow = !!this.purchaseOrder.value;
        return <DictionaryFieldSupported>{
            [DataTitle.gtin]: {
                mainField: isPOWorkflow ? this.productFromPurchaseOrder : this.product,
            },
            [DataTitle.serial]: {
                mainField: this.serialNumber,
            },
            [DataTitle.batchLot]: {
                mainField: this.lot,
            },
            [DataTitle.expiryDate]: {
                mainField: this.expirationDate,
            },
        };
    }

    private async _getLegalCompany(defaultSite: string): Promise<string> {
        const response = await this.$.graph
            .node('@sage/x3-system/Site')
            .read(
                {
                    _id: true, // TODO: Remove unneeded property
                    legalCompany: { code: true },
                },
                defaultSite,
            )
            .execute();

        if (!response) {
            // This should never occur as long as the user function profile has a valid site set up
            throw new Error(
                ui.localize(
                    '@sage/x3-purchasing/error-purchase-receipt-no-legal-company',
                    'The invalid site {{site}} has no legal company.',
                    {
                        site: defaultSite,
                    },
                ),
            );
        }

        return response.legalCompany.code;
    }

    private async _getEntryTransactionOptions(): Promise<string[]> {
        const response = extractEdges(
            await this.$.graph
                // with 'provides' property defined in accessCode of this node, should automatically return only transactions that are accessible for the current user
                .node('@sage/x3-purchasing/PurchaseEntryTransaction')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            code: true,
                            stockMovementCode: {
                                code: true,
                                _id: true,
                            },
                            defaultStockMovementGroup: {
                                code: true,
                                _id: true,
                            },
                            stockAutomaticJournal: {
                                code: true,
                                _id: true,
                            },
                            isLotCustomField1Allowed: true,
                            isLotCustomField2Allowed: true,
                            isLotCustomField3Allowed: true,
                            isLotCustomField4Allowed: true,
                            identifier1Entry: true,
                            identifier1Detail: true,
                            identifier2Entry: true,
                            identifier2Detail: true,
                            isLotPotencyAllowed: true,
                            isLotExpirationDateAllowed: true,
                        },
                        {
                            filter: {
                                entryTransactionType: 'receipt',
                                isActive: true,
                            },
                        },
                    ),
                )
                .execute(),
        ) as PurchaseEntryTransaction[];

        this._entryTransactionMap = new Map<string, PurchaseEntryTransaction>(); // TODO: Put this in a better place

        // transform Entry Transaction response into a string array
        return response.map((entryTransaction: PurchaseEntryTransaction) => {
            this._entryTransactionMap.set(entryTransaction.code, {
                code: entryTransaction.code,
                _id: entryTransaction._id,
                stockAutomaticJournal: entryTransaction.stockAutomaticJournal,
                stockMovementCode: entryTransaction.stockMovementCode,
                defaultStockMovementGroup: entryTransaction.defaultStockMovementGroup,
                isLotCustomField1Allowed: entryTransaction.isLotCustomField1Allowed,
                isLotCustomField2Allowed: entryTransaction.isLotCustomField2Allowed,
                isLotCustomField3Allowed: entryTransaction.isLotCustomField3Allowed,
                isLotCustomField4Allowed: entryTransaction.isLotCustomField4Allowed,
                identifier1Entry: entryTransaction.identifier1Entry,
                identifier1Detail: entryTransaction.identifier1Detail,
                identifier2Entry: entryTransaction.identifier2Entry,
                identifier2Detail: entryTransaction.identifier2Detail,
                isLotPotencyAllowed: entryTransaction.isLotPotencyAllowed,
                isLotExpirationDateAllowed: entryTransaction.isLotExpirationDateAllowed,
            });
            return entryTransaction.code;
        });
    }

    private _disablePage(): void {
        this.purchaseOrder.isDisabled = true;
        this.supplier.isDisabled = true;
        this.receiptDate.isDisabled = true;
        this.transaction.isDisabled = true;
        this.supplierPackingSlip.isDisabled = true;
    }

    private async _getProductSearch() {
        const responseSite = await this.$.graph
            .node('@sage/x3-master-data/MobileAutomationSetup')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        criteria1: true,
                        criteria2: true,
                        isLocationPreloaded: true,
                        // _id: true,
                    },
                    {
                        filter: this.receiptSite.value
                            ? {
                                  site: { code: this.receiptSite.value.code },
                              }
                            : undefined,
                    },
                ),
            )
            .execute();
        if (responseSite.edges.length !== 0) {
            responseSite.edges.some(edge => {
                this._productSearchCriteria1 = edge.node.criteria1;
                this._productSearchCriteria2 = edge.node.criteria2;
                this._isLocationPreloaded = edge.node.isLocationPreloaded;
            });
        } else {
            const response = await this.$.graph
                .node('@sage/x3-master-data/MobileAutomationSetup')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            criteria1: true,
                            criteria2: true,
                            isLocationPreloaded: true,
                            // _id: true,
                        },
                        {
                            filter: {
                                site: null,
                            },
                        },
                    ),
                )
                .execute();
            if (response.edges.length !== 0) {
                response.edges.some(edge => {
                    this._productSearchCriteria1 = edge.node.criteria1;
                    this._productSearchCriteria2 = edge.node.criteria2;
                    this._isLocationPreloaded = edge.node.isLocationPreloaded;
                });
            } else {
                this._productSearchCriteria1 = 'none';
                this._productSearchCriteria2 = 'none';
                this._isLocationPreloaded = true;
            }
        }
    }

    private async _initTransaction() {
        this.transaction.options = await this._getEntryTransactionOptions();
        switch (this.transaction.options.length) {
            case 0:
                this._disablePage();
                throw new Error('No transaction, cannot continue');
            case 1:
                this.transaction.value = this.transaction.options[0];
                this.transaction.isHidden = true;
                break;
            default:
                this.transaction.value = this.transaction.options[0];
                this.transaction.isDisabled = this.lines.value.length > 0;
                break;
        }
        if (this.transaction.value) {
            this._selectedEntryTransaction = this._entryTransactionMap.get(this.transaction.value);
            this.stockMovementCode.value = this._selectedEntryTransaction?.stockMovementCode
                ? {
                      code: this._selectedEntryTransaction.stockMovementCode.code,
                      _id: this._selectedEntryTransaction.stockMovementCode._id,
                  }
                : null;
            this.stockAutomaticJournal.value = this._selectedEntryTransaction?.stockAutomaticJournal
                ? {
                      code: this._selectedEntryTransaction.stockAutomaticJournal.code,
                      _id: this._selectedEntryTransaction.stockAutomaticJournal._id,
                  }
                : null;
        }
    }

    /**
     * OnChange readonly process
     *
     * Used both decorator and bar code manager.
     * @returns Promise<void>
     */

    /** @internal */
    private readonly _onChange_product: AsyncVoidFunction = async () => {
        /* When processing composite codes, this event will be called systematically,
         * because currently, the service does not know if this field is part of a
         * displayed section or not.
         * So we need to check if there is a need to process the event.
         */
        if (this.controlManager.isDispatchInProgress && this._getActiveSection() !== Section.main) {
            return;
        }

        if (!this.product.value || !!(await this.product.validate())) {
            await this._setActiveSection(Section.main, true);
            return;
        }

        this._purchaseShipmentId = undefined;
        this._purchaseShipmentLineNumber = undefined;
        this._purchaseOrderLine = undefined;
        this._supplierProduct = undefined;
        this._purchaseShipmentLine = undefined;

        this._productSite = await this.getProductSite(this.product.value.product, this.receiptSite.value?.code ?? '');
        if (this._productSite.isBeingCounted === true) {
            if (
                !(await dialogConfirmation(
                    this,
                    'warn',
                    ui.localize('@sage/x3-purchasing/dialog-warning-title', 'Warning'),
                    ui.localize(
                        '@sage/x3-purchasing/product-blocked-by-count-continue',
                        'Product blocked by count. Do you want to continue?',
                    ),
                    {
                        acceptButton: {
                            text: ui.localize('@sage/x3-purchasing/button-accept-yes', 'Yes'),
                        },
                        cancelButton: {
                            text: ui.localize('@sage/x3-purchasing/button-cancel-no', 'No'),
                        },
                    },
                ))
            ) {
                //cancel button clicked - promise rejected
                await this._setActiveSection(Section.main);
                return;
            }
        }

        this._supplierProduct = await this.getSupplierProduct(this.product.value.product, this.product.value.supplier);

        await this._setActiveSection(Section.destination);
    };

    /** @internal */
    private readonly _onChange_productFromPurchaseOrder: AsyncVoidFunction = async () => {
        this._purchaseShipmentId = undefined;
        this._purchaseShipmentLineNumber = undefined;
        this._purchaseOrderLine = undefined;
        this._supplierProduct = undefined;
        this._purchaseShipmentLine = undefined;

        /* When processing composite codes, this event will be called systematically,
         * because currently, the service does not know if this field is part of a
         * displayed section or not.
         * So we need to check if there is a need to process the event.
         */
        if (this.controlManager.isDispatchInProgress && this._getActiveSection() !== Section.main) {
            return;
        }

        if (!this.productFromPurchaseOrder.value || !!(await this.productFromPurchaseOrder.validate())) {
            await this._setActiveSection(Section.main, true);
            return;
        }

        this._shipmentLines = await this._readShipmentLines();
        this._cleanShipmentLines();

        if (this._shipmentLines?.length === 1) {
            this._purchaseShipmentId = this._shipmentLines[0].purchaseShipment?.id;
            this._purchaseShipmentLineNumber = this._shipmentLines[0].lineNumber?.toString();
        }

        this._productSite = await this.getProductSite(
            this.productFromPurchaseOrder.value.product,
            this.receiptSite.value?.code ?? '',
        );

        if (this._productSite.isBeingCounted === true) {
            if (
                !(await dialogConfirmation(
                    this,
                    'warn',
                    ui.localize('@sage/x3-purchasing/dialog-warning-title', 'Warning'),
                    ui.localize(
                        '@sage/x3-purchasing/product-blocked-by-count-continue',
                        'Product blocked by count. Do you want to continue?',
                    ),
                    {
                        acceptButton: {
                            text: ui.localize('@sage/x3-purchasing/button-accept-yes', 'Yes'),
                        },
                        cancelButton: {
                            text: ui.localize('@sage/x3-purchasing/button-cancel-no', 'No'),
                        },
                    },
                ))
            ) {
                //cancel button clicked - promise rejected
                await this._setActiveSection(Section.main);
                return;
            }
        }
        this.supplier.value = this.purchaseOrder.value?.orderFromSupplier;
        this._supplierProduct = await this.getSupplierProduct(
            this.productFromPurchaseOrder.value.product,
            this.productFromPurchaseOrder.value.supplier,
        );

        if (this.productFromPurchaseOrder.value) {
            this._purchaseOrderLine = await this.getPurchaseOrderLine(
                this.productFromPurchaseOrder.value.purchaseOrder,
                this.productFromPurchaseOrder.value.lineNumber,
                this.productFromPurchaseOrder.value.sequenceNumber,
            );
        }

        if (this._shipmentLines?.length > 1) {
            await this._setActiveSection(Section.shipment);
            return;
        }

        if (this._purchaseShipmentId && this._purchaseShipmentLineNumber) {
            this._purchaseShipmentLine = await this.getPurchaseShipmentLine(
                this._purchaseShipmentId,
                Number(this._purchaseShipmentLineNumber),
            );
        }

        await this._setActiveSection(Section.destination);
    };

    private async _clearPage() {
        this._resetCompositeDataControl();
        this._disableDestinationFields();
        this.closePoLine.isHidden = true;
        this.expirationDate.isHidden = true;
        this.useByDate.isHidden = true;
    }

    private _disableDestinationFields() {
        this.quantityReceived.isHidden = true;
        this.lot.isHidden = true;
        this.sublot.isHidden = true;
        this.supplierLot.isHidden = true;
        this.lotCustomField1.isHidden = true;
        this.lotCustomField2.isHidden = true;
        this.lotCustomField3.isHidden = true;
        this.lotCustomField4.isHidden = true;
        this.potency.isHidden = true;
        this.majorVersion.isHidden = true;
        this.minorVersion.isHidden = true;
        this.packingUnitToStockUnitConversionFactor.isHidden = true;
        this.receiptUnit.isHidden = true;
        this.location.isHidden = true;
        this.suggestedLocation.isHidden = true;
        this.serialNumber.isHidden = true;
        this.destinationContainer.isHidden = true;
        this.licensePlateNumber.isHidden = true;
        this.identifier1.isHidden = true;
        this.identifier2.isHidden = true;
        this.status.isHidden = true;
    }

    private async _initializePage() {
        this._resetCompositeDataControl();
        this.warehouse.value = null;
        this.quantityReceived.value = null;
        this.previousQuantityToReceive.value = null;
        this.licensePlateNumber.value = null;
        this.location.value = null;
        this.serialNumber.value = null;
        this.supplierLot.value = null;
        this.previousSupplierLot.value = null;
        this.lot.value = null;
        this.sublot.value = null;
        this.majorVersion.value = null;
        this.minorVersion.value = null;
        this.lotCustomField1.value = null;
        this.lotCustomField2.value = null;
        this.lotCustomField3.value = null;
        this.lotCustomField4.value = null;
        this.identifier1.value = null;
        this.identifier2.value = null;

        if (this._productSite) {
            this._purchaseStockManagementRules = await findStockManagementRules(
                this.receiptSite.value?.code ?? '',
                this._productSite.product.productCategory.code,
                '3',
                this._selectedEntryTransaction?.stockMovementCode?.code ?? null,
                this,
            );

            this.status.isHidden = false;
            this.status.options = await this._getStockStatus();
            this.status.value = this._purchaseStockManagementRules?.defaultStatus ?? null;

            this._initLotFields();
            this._initPackingUnitFields();
            this._miscellaneousFields();

            if (!this.suggestedLocation.isHidden) {
                this.suggestedLocation.value = await findDefaultLocation(
                    this._productSite,
                    this._purchaseStockManagementRules,
                    this,
                );
                if (!this.suggestedLocation.value) this.suggestedLocation.isHidden = true;
                else if (this._isLocationPreloaded) {
                    this.location.value = (await this._getDefaultLocation(this.suggestedLocation.value)) ?? null;
                }
            }

            this.closePoLine.options = balanceOptions;
            this.closePoLine.value = balanceOptions[0];
            this._getDefaultValue();
        }
    }

    private _initPackingUnitFields() {
        this._packingUnits = [];

        this._packingUnits.push({
            node: {
                packingUnit: {
                    code: this._productSite.product.stockUnit.code,
                    numberOfDecimals: this._productSite.product.stockUnit.numberOfDecimals,
                    _id: this._productSite.product.stockUnit._id,
                    name: this._productSite.product.stockUnit.code,
                },
                packingUnitToStockUnitConversionFactor: '1',
                isPackingFactorEntryAllowed: false,
                _id: this._productSite.product.stockUnit._id,
                isPackingUnit: false,
            },
        });

        if (this._productSite.product.stockUnit.code !== this._productSite.product.purchaseUnit.code) {
            this._packingUnits.push({
                node: {
                    packingUnit: {
                        code: this._productSite.product.purchaseUnit.code,
                        numberOfDecimals: this._productSite.product.purchaseUnit.numberOfDecimals,
                        _id: this._productSite.product.purchaseUnit._id,
                        name: formatUnitForDisplay(
                            this._productSite.product.purchaseUnit.code,
                            this._productSite.product.purchaseUnitToStockUnitConversionFactor,
                            this._productSite.product.stockUnit.code,
                        ),
                    },
                    packingUnitToStockUnitConversionFactor:
                        this._productSite.product.purchaseUnitToStockUnitConversionFactor.toString(),
                    isPackingFactorEntryAllowed: false,
                    _id: this._productSite.product.purchaseUnit._id,
                    isPackingUnit: false,
                },
            });
        }

        let productPackingList = extractEdges(this._productSite.product.packingUnits.query).filter(productPacking => {
            return !!productPacking.packingUnit?.code;
        });
        let duplicateUnit = false;
        if (productPackingList.length > 0) {
            for (const productPacking of productPackingList) {
                duplicateUnit = false;
                for (const receiptUnit of this._packingUnits) {
                    if (
                        productPacking.packingUnit.code === receiptUnit.node.packingUnit.code &&
                        productPacking.packingUnitToStockUnitConversionFactor ===
                            receiptUnit.node.packingUnitToStockUnitConversionFactor
                    ) {
                        duplicateUnit = true;
                        receiptUnit.node.isPackingUnit = true;
                        break;
                    }
                }
                if (!duplicateUnit) {
                    this._packingUnits.push({
                        node: {
                            packingUnit: {
                                code: productPacking.packingUnit?.code,
                                numberOfDecimals: productPacking.packingUnit?.numberOfDecimals,
                                _id: productPacking.packingUnit?._id,
                                name: formatUnitForDisplay(
                                    productPacking.packingUnit?.code,
                                    productPacking.packingUnitToStockUnitConversionFactor,
                                    this._productSite.product.stockUnit.code,
                                ),
                            },
                            packingUnitToStockUnitConversionFactor:
                                productPacking.packingUnitToStockUnitConversionFactor,
                            isPackingFactorEntryAllowed: productPacking.isPackingFactorEntryAllowed,
                            _id: productPacking._id,
                            isPackingUnit: true,
                        },
                    });
                }
            }
        }
        if (this._supplierProduct?.purchaseUnit) {
            duplicateUnit = false;
            for (const receiptUnit of this._packingUnits) {
                if (
                    this._supplierProduct.purchaseUnit.code === receiptUnit.node.packingUnit.code &&
                    this._supplierProduct.purchaseUnitToStockUnitConversionFactor ===
                        receiptUnit.node.packingUnitToStockUnitConversionFactor
                ) {
                    duplicateUnit = true;
                    break;
                    1;
                }
            }
            if (!duplicateUnit) {
                this._packingUnits.push({
                    node: {
                        packingUnit: {
                            code: this._supplierProduct.purchaseUnit.code,
                            numberOfDecimals: this._supplierProduct.purchaseUnit.numberOfDecimals,
                            _id: this._supplierProduct.purchaseUnit._id,
                            name: formatUnitForDisplay(
                                this._supplierProduct.purchaseUnit.code,
                                this._supplierProduct.purchaseUnitToStockUnitConversionFactor,
                                this._productSite.product.stockUnit.code,
                            ),
                        },
                        packingUnitToStockUnitConversionFactor:
                            this._supplierProduct.purchaseUnitToStockUnitConversionFactor,
                        isPackingFactorEntryAllowed: false,
                        _id: this._supplierProduct.purchaseUnit._id,
                        isPackingUnit: false,
                    },
                });
            }
        }
        if (this._supplierProduct?.packingUnit) {
            duplicateUnit = false;
            const unitConversion: string = Number.parseFloat(
                (
                    Number(this._supplierProduct.packingUnitToPurchaseUnitConversionFactor) *
                    Number(this._supplierProduct.purchaseUnitToStockUnitConversionFactor)
                ).toFixed(CONVERSION_RATE_DECIMAL_PLACES),
            ).toString();
            for (const receiptUnit of this._packingUnits) {
                if (
                    this._supplierProduct.packingUnit.code === receiptUnit.node.packingUnit.code &&
                    unitConversion.toString() === receiptUnit.node.packingUnitToStockUnitConversionFactor
                ) {
                    duplicateUnit = true;
                    break;
                }
            }
            if (!duplicateUnit) {
                this._packingUnits.push({
                    node: {
                        packingUnit: {
                            code: this._supplierProduct.packingUnit.code,
                            numberOfDecimals: this._supplierProduct.packingUnit.numberOfDecimals,
                            _id: this._supplierProduct.packingUnit._id,
                            name: formatUnitForDisplay(
                                this._supplierProduct.packingUnit.code,
                                unitConversion,
                                this._productSite.product.stockUnit.code,
                            ),
                        },
                        packingUnitToStockUnitConversionFactor: unitConversion,
                        isPackingFactorEntryAllowed: false,
                        _id: this._supplierProduct.packingUnit._id,
                        isPackingUnit: false,
                    },
                });
            }
        }

        this.packingUnitToStockUnitConversionFactor.isHidden = false;
        this.receiptUnit.isHidden = false;
        this.receiptUnit.options = this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.name);
        this.receiptUnit.isReadOnly = this.receiptUnit.options.length <= 1;
        this.receiptUnit.isDisabled = false;
    }

    private _initLotFields() {
        const lotManagement = this._productSite.product.lotManagementMode;
        const lotManagementRule = this._purchaseStockManagementRules.lotEntry;
        const lotByDefault = this._purchaseStockManagementRules.lotByDefault;
        const lotNotManaged = lotManagement === 'notManaged';
        this.lot.isHidden = lotNotManaged;
        this.supplierLot.isHidden = lotNotManaged;
        this.sublot.isHidden = lotManagement !== 'lotAndSublot';
        this.sublot.isMandatory = !this.sublot.isHidden;
        if (!this.lot.isHidden) {
            if (lotManagementRule === 'no') {
                this.lot.isDisabled = true;
                if (this.sublot.isHidden === false) {
                    this.sublot.value = '00001';
                    this.sublot.isDisabled = true;
                }
            } else if (lotManagementRule === 'newLot') {
                if (this.sublot.isHidden === false) this.sublot.value = '00001';
            } else if (lotManagementRule === 'free') {
                if (this.sublot.isHidden === false) this.sublot.value = '00001';
            }
        }

        this.lotCustomField1.isHidden = !this._selectedEntryTransaction?.isLotCustomField1Allowed || lotNotManaged;
        this.lotCustomField2.isHidden = !this._selectedEntryTransaction?.isLotCustomField2Allowed || lotNotManaged;
        this.lotCustomField3.isHidden = !this._selectedEntryTransaction?.isLotCustomField3Allowed || lotNotManaged;
        this.lotCustomField4.isHidden = !this._selectedEntryTransaction?.isLotCustomField4Allowed || lotNotManaged;

        this.lot.isMandatory =
            !this.lot.isHidden &&
            !this._productSite.product.lotSequenceNumber &&
            lotManagement !== 'optionalLot' &&
            lotByDefault !== 'documentNumber';
    }

    private _initVersionFields() {
        const versionManagementMode = this._productSite.product.stockVersionMode;

        this.majorVersion.isHidden = !['major', 'majorAndMinor'].includes(versionManagementMode);
        this.minorVersion.isHidden = versionManagementMode !== 'majorAndMinor';
        this.minorVersion.isDisabled = true;

        if (
            this._purchaseOrderLine &&
            !this.majorVersion.isHidden &&
            this._purchaseOrderLine.majorProductVersion &&
            this._purchaseOrderLine.majorProductVersion.code
        ) {
            this.majorVersion.value = { code: this._purchaseOrderLine.majorProductVersion.code };

            // check for minor version
            if (!this.minorVersion.isHidden && this._purchaseOrderLine.minorProductVersion) {
                this.minorVersion.value = { minorVersion: this._purchaseOrderLine.minorProductVersion };
                this.minorVersion.isDisabled = false;
            }
        }
    }

    private _miscellaneousFields() {
        const potencyManagement = this._productSite.product.stockManagementMode;
        const expirationManagement = this._productSite.product.expirationManagementMode;

        this.potency.isHidden =
            this.lot.isHidden ||
            !this._selectedEntryTransaction?.isLotPotencyAllowed ||
            potencyManagement !== 'potencyManaged';
        if (!this.potency.isHidden) {
            this.potency.value = Number(this._productSite.product.defaultPotencyInPercentage);
        }

        //serial number fields
        const serialNumberManagement = this._productSite.product.serialNumberManagementMode;
        this.serialNumber.isHidden = ['notManaged', 'issued'].includes(serialNumberManagement);
        this.serialNumber.isMandatory = !this.serialNumber.isHidden && !this._productSite.product.serialSequenceNumber;

        this.licensePlateNumber.isHidden = !this._productSite.isLicensePlateNumberManaged;

        //location fields
        this.location.isHidden = !this._productSite.isLocationManaged;
        this.location.isMandatory = !this.location.isHidden;
        this.location.isDisabled = false;
        this.suggestedLocation.isHidden = this.location.isHidden;
        this.expirationDate.isHidden = expirationManagement === 'notManaged';
        this.useByDate.isHidden = expirationManagement === 'notManaged';

        if (expirationManagement !== 'notManaged' && expirationManagement !== 'manualEntry') {
            this.expirationDate.value =
                expirationDateDefaultValue(
                    expirationManagement,
                    this._productSite.product.expirationLeadTime,
                    this._productSite.product.expirationTimeUnit,
                    this._effectiveDate,
                ) ?? '';

            this.useByDate.value =
                useByDateDefaultValue(
                    this.expirationDate.value,
                    this._effectiveDate,
                    expirationManagement,
                    Number(this._productSite.product.useByDateCoefficient),
                ) ?? '';
            this.useByDate.maxDate = this.expirationDate.value;
        }

        //identifier 1 and identifier 2
        switch (this._selectedEntryTransaction?.identifier1Detail) {
            case 'entered': {
                this.identifier1.isHidden = false;
                this.identifier1.isDisabled = false;
                break;
            }
            case 'displayed': {
                this.identifier1.isHidden = false;
                this.identifier1.isDisabled = true;
                break;
            }
            default: {
                this.identifier1.isHidden = true;
            }
        }

        switch (this._selectedEntryTransaction?.identifier2Detail) {
            case 'entered': {
                this.identifier2.isHidden = false;
                this.identifier2.isDisabled = false;
                break;
            }
            case 'displayed': {
                this.identifier2.isHidden = false;
                this.identifier2.isDisabled = true;
                break;
            }
            default: {
                this.identifier2.isHidden = true;
            }
        }

        this.destinationContainer.isHidden =
            this._purchaseStockManagementRules.licensePlateNumberEntry !== 'mandatory' ||
            !this._productSite.isLicensePlateNumberManaged;
        if (!this.destinationContainer.isHidden) {
            this.destinationContainer.value = this._productSite.defaultInternalContainer;
        }
    }

    private _getDefaultValue() {
        let orderUnitToStockUnitConversionFactor: string;
        let defaultReceiptUnit: string;
        let defaultReceiptUnitCode: string;
        if (this._purchaseOrderLine) {
            this.packingUnitToStockUnitConversionFactor.value = Number(
                this._purchaseOrderLine.orderUnitToPurchaseUnitConversionFactor,
            );
            orderUnitToStockUnitConversionFactor = Number.parseFloat(
                (
                    Number(this._purchaseOrderLine.quantityInStockUnitOrdered) /
                    Number(this._purchaseOrderLine.quantityInOrderUnitOrdered)
                ).toFixed(CONVERSION_RATE_DECIMAL_PLACES),
            ).toString();
            if (
                this._purchaseOrderLine.orderUnit.code === this._productSite.product.stockUnit.code &&
                orderUnitToStockUnitConversionFactor === '1'
            ) {
                defaultReceiptUnit = this._purchaseOrderLine.orderUnit.code;
                defaultReceiptUnitCode = this._purchaseOrderLine.orderUnit.code;
            } else {
                defaultReceiptUnitCode = this._purchaseOrderLine.orderUnit.code;
                defaultReceiptUnit = formatUnitForDisplay(
                    this._purchaseOrderLine.orderUnit.code,
                    orderUnitToStockUnitConversionFactor,
                    this._productSite.product.stockUnit.code,
                );
            }
            // Add order unit calculated from supplier purchase unit when used on PO
            const findUnit = this._packingUnits.find(
                packingUnit => packingUnit.node.packingUnit.code === defaultReceiptUnitCode,
            ); // TODO Issue: Rounding issue on orderUnitToStockUnitConversionFactor can cause an existing match to not be found and thus, create a 'duplicate' option in receiptUnit
            if (!findUnit) {
                this._packingUnits.push({
                    node: {
                        packingUnit: {
                            _id: this._purchaseOrderLine.orderUnit._id,
                            code: this._purchaseOrderLine.orderUnit.code,
                            numberOfDecimals:
                                this._packingUnits.find(
                                    packingUnit =>
                                        packingUnit.node.packingUnit.code === this._purchaseOrderLine?.orderUnit.code,
                                )?.node.packingUnit.numberOfDecimals ?? 0,
                            name: defaultReceiptUnit,
                        },
                        packingUnitToStockUnitConversionFactor: orderUnitToStockUnitConversionFactor.toString(),
                        isPackingFactorEntryAllowed: false,
                        isPackingUnit: true,
                        _id: this._purchaseOrderLine.orderUnit._id,
                    },
                });
                //Update available units
                this.receiptUnit.options?.push(defaultReceiptUnit);
            }
        } else {
            //If supplier product purchase unit exists, use it as default.  Otherwise use the product's purchase unit.
            let defaultOrderUnit: string;
            if (this._supplierProduct && this._supplierProduct?.purchaseUnit?.code) {
                defaultOrderUnit = this._supplierProduct?.purchaseUnit.code;
                orderUnitToStockUnitConversionFactor = Number.parseFloat(
                    this._supplierProduct?.purchaseUnitToStockUnitConversionFactor,
                ).toString();
            } else {
                defaultOrderUnit = this._productSite.product.purchaseUnit?.code;
                orderUnitToStockUnitConversionFactor = Number.parseFloat(
                    this._productSite.product.purchaseUnitToStockUnitConversionFactor,
                ).toString();
            }
            if (
                defaultOrderUnit === this._productSite.product.stockUnit.code &&
                orderUnitToStockUnitConversionFactor === '1'
            ) {
                defaultReceiptUnit = defaultOrderUnit;
                defaultReceiptUnitCode = defaultOrderUnit;
            } else {
                defaultReceiptUnitCode = defaultOrderUnit;
                defaultReceiptUnit = formatUnitForDisplay(
                    defaultOrderUnit,
                    orderUnitToStockUnitConversionFactor,
                    this._productSite.product.stockUnit.code,
                );
            }
        }

        const selectedUnit = this._packingUnits.find(
            packingUnit => packingUnit.node.packingUnit.code === defaultReceiptUnitCode,
        );
        if (!selectedUnit) return false;
        this.packingUnitToStockUnitConversionFactor.value = Number(orderUnitToStockUnitConversionFactor);
        this.quantityReceived.scale = selectedUnit.node.packingUnit.numberOfDecimals; // set # of decimal places restriction based on selected unit of measure
        if (
            selectedUnit.node.packingUnit.code !== this._productSite.product.stockUnit.code &&
            orderUnitToStockUnitConversionFactor !== '1'
        ) {
            this.receiptUnit.value = formatUnitForDisplay(
                selectedUnit.node.packingUnit.code,
                selectedUnit.node.packingUnitToStockUnitConversionFactor,
                this._productSite.product.stockUnit.code,
            );
        } else {
            this.receiptUnit.value = defaultReceiptUnit;
        }
        this.receiptUnitCode.value = selectedUnit.node.packingUnit.code;
        this.previousReceiptUnit.value = this.receiptUnit.value;
        this.previousReceiptUnitCode.value = selectedUnit.node.packingUnit.code;
        this.packingUnitToStockUnitConversionFactor.isDisabled = !selectedUnit.node.isPackingFactorEntryAllowed;

        // Calculating remaining quantity only applicable for PO Line workflow
        if (this._purchaseOrderLine) {
            this._initVersionFields();

            const remainingQuantity: number[] = [0, 0, 0];
            this.calculateRemainingQuantity(remainingQuantity);
            this.destinationRemainingQuantityInStockUnit.value = remainingQuantity[0];
            //the packing unit is the order unit, so I can put remainingQuantityInOrderUnit in the remainingQuantityInPackingUnit field
            this.destinationRemainingQuantityInPackingUnit.value = remainingQuantity[1];
            const tmp = Math.pow(10, selectedUnit?.node.packingUnit.numberOfDecimals);
            if (tmp !== 0) {
                this.destinationRemainingQuantityInStockUnit.value =
                    Math.round(this.destinationRemainingQuantityInStockUnit.value * tmp) / tmp;
            }
            this.quantityReceived.value = this.destinationRemainingQuantityInPackingUnit.value; // should auto-round value appropriately based on scale property that was set for this field
            this._quantityToReceiveNoRounded = this.quantityReceived.value;
            this._quantityToReceiveRounded = this.quantityReceived.value;
            this.previousQuantityToReceive.value = this.quantityReceived.value;
            this.closePoLine.value = balanceOptions[0]; //Yes
            this._remainingQuantityInitialInStockUnit = this._purchaseShipmentLine
                ? Number(this._purchaseShipmentLine.quantityInStockUnitShipped)
                : Number(this._purchaseOrderLine.quantityInStockUnitOrdered);
            this._remainingQuantityInitialInOrderUnit = this._purchaseShipmentLine
                ? Number(this._purchaseShipmentLine.quantityInOrderUnitShipped)
                : Number(this._purchaseOrderLine.quantityInOrderUnitOrdered);
            this.closePoLine.isHidden = false;
        } else {
            this._initVersionFields();
            //Quantity values not used for direct receipts
            this.destinationRemainingQuantityInPackingUnit.value = 0;
            this.destinationRemainingQuantityInPackingUnit.isHidden = true;
            this.closePoLine.isHidden = true;
            this._remainingQuantityInitialInStockUnit = 0;
            this._remainingQuantityInitialInOrderUnit = 0;
        }
    }

    private async _getShipments() {
        const response = await this.$.graph
            .node('@sage/x3-purchasing/PurchaseShipmentLine')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        product: {
                            code: true,
                            purchaseUnitToStockUnitConversionFactor: true,
                            _id: true,
                        },
                        purchaseShipment: {
                            id: true,
                            _id: true,
                        },
                        lineNumber: true,
                        quantityInPurchaseUnitShipped: true,
                        quantityInPurchaseUnitExpected: true,
                        purchaseUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        quantityInOrderUnitShipped: true,
                        quantityInStockUnitShipped: true,
                        orderUnitToPurchaseUnitConversionFactor: true,
                        stockUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        orderUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                    },
                    {
                        filter: {
                            purchaseOrderNumber: { id: this._purchaseOrderLine?.purchaseOrder.id },
                            purchaseOrderLineNumber: Number(this._purchaseOrderLine?.lineNumber),
                            purchaseShipment: {
                                id: { _in: this._purchaseShipmentIds },
                            },
                        },
                    },
                ),
            )
            .execute();
        this.shipments.value = response.edges.map(line => {
            return {
                _id: line.node._id,
                product: line.node.product,
                purchaseShipment: line.node.purchaseShipment,
                lineNumber: line.node.lineNumber,
                quantityInPurchaseUnitShipped: line.node.quantityInPurchaseUnitShipped,
                quantityInPurchaseUnitExpected: line.node.quantityInPurchaseUnitExpected,
                purchaseUnit: line.node.purchaseUnit,
                quantityInOrderUnitShipped: line.node.quantityInOrderUnitShipped,
                quantityInStockUnitShipped: line.node.quantityInStockUnitShipped,
                orderUnitToPurchaseUnitConversionFactor: line.node.orderUnitToPurchaseUnitConversionFactor,
                stockUnit: line.node.stockUnit,
                orderUnit: line.node.orderUnit,
            };
        });
    }

    private async _readShipmentLines() {
        return extractEdges(
            await this.$.graph
                .node('@sage/x3-purchasing/PurchaseShipmentLine')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            purchaseShipment: { id: true, _id: true },
                            lineNumber: true,
                            product: { code: true, _id: true },
                            purchaseOrderNumber: { id: true, _id: true },
                            purchaseOrderLineNumber: true,
                            quantityInPurchaseUnitShipped: true,
                            quantityInPurchaseUnitReceived: true,
                            quantityInPurchaseUnitPreReceived: true,
                            quantityInOrderUnitShipped: true,
                            quantityInStockUnitShipped: true,
                            isClosed: true,
                            isPreReceived: true,
                        },
                        {
                            filter: {
                                purchaseOrderNumber: { id: this.purchaseOrder.value?.id },
                                purchaseOrderLineNumber: this.productFromPurchaseOrder?.value?.lineNumber,
                                purchaseShipment: {
                                    shipmentStatus: {
                                        _nin: ['received', 'closed'],
                                    },
                                },
                                isClosed: { _eq: false },
                            },
                        },
                    ),
                )
                .execute(),
        ) as PurchaseShipmentLine[];
    }

    private _cleanShipmentLines() {
        this._shipmentLines.forEach((element: any) => {
            if (
                (element.isPreReceived &&
                    Number(element.quantityInPurchaseUnitPreReceived) >
                        Number(element.quantityInPurchaseUnitReceived)) ||
                (!element.isPreReceived &&
                    Number(element.quantityInPurchaseUnitShipped) > Number(element.quantityInPurchaseUnitReceived))
            ) {
                this._purchaseShipmentIds.push(element?.purchaseShipment?.id);
            } else {
                const resultIndex = this._shipmentLines.indexOf(element);
                this._shipmentLines.splice(resultIndex, 1);
            }
        });

        if (this.lines.value.length !== 0) {
            this._shipmentLines.forEach(lineShip => {
                if (
                    lineShip.purchaseOrderNumber === this.purchaseOrder.value?.id &&
                    lineShip.purchaseOrderLineNumber === this.productFromPurchaseOrder?.value?.lineNumber
                ) {
                    let QuantityInStockUnitReceived = 0;
                    let quantityInStockUnitExpected = Number(lineShip.quantityInStockUnitShipped);

                    this.lines.value.forEach(line => {
                        if (line.purchaseShipmentLine && line.purchaseShipmentLine._id === lineShip._id) {
                            QuantityInStockUnitReceived +=
                                Number(line.quantityInReceiptUnitReceived) *
                                Number(line.receiptUnitToStockUnitConversionFactor);
                        }
                    });

                    if (Number(quantityInStockUnitExpected) - Number(QuantityInStockUnitReceived) <= 0) {
                        let resultIndex = this._purchaseShipmentIds.indexOf(lineShip.purchaseShipment?.id);
                        this._purchaseShipmentIds.splice(resultIndex, 1);
                        resultIndex = this._shipmentLines.indexOf(lineShip);
                        this._shipmentLines.splice(resultIndex, 1);
                    }
                }
            });
        }
    }

    private async getProductSite(productCode: string, siteCode: string) {
        const productSiteToReceive = await this.$.graph
            .node('@sage/x3-master-data/ProductSite')
            .read(
                {
                    _id: true,
                    stockSite: { code: true },
                    isLocationManaged: true,
                    isLicensePlateNumberManaged: true,
                    defaultInternalContainer: {
                        code: true,
                        _id: true,
                    },
                    isBeingCounted: true,
                    defaultLocations: {
                        query: {
                            edges: {
                                node: {
                                    defaultLocation: true,
                                    locationNumber: true,
                                    defaultLocationType: true,
                                    _id: true,
                                },
                            },
                        },
                    },
                    product: {
                        _id: true,
                        code: true,
                        localizedDescription1: true,
                        upc: true,
                        globalTradeItemNumber: true,
                        productCategory: {
                            code: true,
                            _id: true,
                        },
                        lotManagementMode: true,
                        serialNumberManagementMode: true,
                        stockVersionMode: true,
                        expirationManagementMode: true,
                        serialSequenceNumber: true,
                        lotSequenceNumber: true,
                        stockManagementMode: true,
                        defaultPotencyInPercentage: true,
                        expirationLeadTime: true,
                        expirationTimeUnit: true,
                        useByDateCoefficient: true,
                        stockUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        purchaseUnit: {
                            code: true,
                            numberOfDecimals: true,
                            _id: true,
                        },
                        purchaseUnitToStockUnitConversionFactor: true,
                        packingUnits: {
                            query: {
                                edges: {
                                    node: {
                                        packingUnit: {
                                            code: true,
                                            numberOfDecimals: true,
                                            _id: true,
                                        },
                                        packingUnitToStockUnitConversionFactor: true,
                                        isPackingFactorEntryAllowed: true,
                                        _id: true,
                                    },
                                },
                            },
                        },
                    },
                },
                `${productCode}|${siteCode}`,
            )
            .execute();

        if (!productSiteToReceive) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing/pages__purchase_receipt__notification__invalid_product_site_error',
                    `Could not retrieve your product {{ productCode }} for the site {{ siteCode }}`,
                    {
                        productCode: productCode,
                        siteCode: siteCode,
                    },
                ),
                {
                    type: 'error',
                    timeout: 5000,
                },
            );
            return null;
        }

        this._purchaseStockManagementRules = await findStockManagementRules(
            this.receiptSite.value?.code ?? '',
            productSiteToReceive.product.productCategory.code,
            '3',
            this._selectedEntryTransaction?.stockMovementCode?.code ?? null,
            this,
        );

        return productSiteToReceive as any;
    }

    private async getPurchaseOrderLine(
        purchaseOrderNumber: string,
        purchaseOrderLineNumber: number,
        purchaseOrderLine: number,
    ) {
        const purchaseOrderLineToReceive = await this.$.graph
            .node('@sage/x3-purchasing/PurchaseOrderLine')
            .read(
                {
                    _id: true,
                    purchaseOrder: {
                        _id: true,
                        id: true,
                    },
                    lineNumber: true,
                    sequenceNumber: true,
                    orderUnit: {
                        code: true,
                        _id: true,
                    },
                    purchaseUnit: {
                        code: true,
                        numberOfDecimals: true,
                        _id: true,
                    },
                    quantityInStockUnitOrdered: true,
                    quantityInStockUnitReceived: true,
                    quantityInOrderUnitOrdered: true,
                    quantityInPurchaseUnitOrdered: true,
                    quantityInPurchaseUnitReceived: true,
                    orderUnitToPurchaseUnitConversionFactor: true,
                    majorProductVersion: { code: true },
                    minorProductVersion: true,
                },
                `${purchaseOrderNumber}|${purchaseOrderLineNumber}|${purchaseOrderLine}`,
            )
            .execute();

        if (!purchaseOrderLineToReceive) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing/pages__purchase_receipt__notification__invalid_purchase_order_line_error',
                    `Could not retrieve your purchase order line {{ purchaseOrderNumber }} {{ purchaseOrderLineNumber }} {{ purchaseOrderLine }}`,
                    {
                        purchaseOrderNumber: purchaseOrderNumber,
                        purchaseOrderLineNumber: purchaseOrderLineNumber,
                        purchaseOrderLine: purchaseOrderLine,
                    },
                ),
                {
                    type: 'error',
                    timeout: 5000,
                },
            );
            return null;
        }
        return purchaseOrderLineToReceive as any;
    }

    private async getSupplierProduct(product: string, supplier: string) {
        const supplierProductToReceive = await this.$.graph
            .node('@sage/x3-master-data/SupplierProduct')
            .read(
                {
                    _id: true,
                    purchaseUnit: {
                        code: true,
                        numberOfDecimals: true,
                        _id: true,
                    },
                    purchaseUnitToStockUnitConversionFactor: true,
                    packingUnit: {
                        code: true,
                        numberOfDecimals: true,
                        _id: true,
                    },
                    packingUnitToPurchaseUnitConversionFactor: true,
                },
                `${product}|${supplier}`,
            )
            .execute();

        if (!supplierProductToReceive) {
            return undefined;
        }

        return supplierProductToReceive as any;
    }

    private async getPurchaseShipmentLine(id: string, lineNumber: number) {
        const shipmentLineToReceive = await this.$.graph
            .node('@sage/x3-purchasing/PurchaseShipmentLine')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        purchaseShipment: { id: true, _id: true },
                        lineNumber: true,
                        product: { code: true, _id: true },
                        quantityInPurchaseUnitExpected: true,
                        quantityInOrderUnitShipped: true,
                        quantityInPurchaseUnitShipped: true,
                        quantityInStockUnitShipped: true,
                        orderUnitToPurchaseUnitConversionFactor: true,
                        orderUnit: { code: true, _id: true },
                        stockUnit: { code: true, _id: true },
                        _id: true,
                    },
                    {
                        filter: {
                            purchaseShipment: {
                                id: id,
                            },
                            lineNumber: lineNumber,
                        },
                    },
                ),
            )
            .execute();

        if (!shipmentLineToReceive) {
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing/error-loading-purchase-shipment-line',
                    'Error loading purchase shipment line',
                ),
                {
                    type: 'error',
                    timeout: 5000,
                },
            );
            return null;
        }

        if (shipmentLineToReceive.edges.length === 1) {
            return shipmentLineToReceive.edges[0].node as any;
        } else {
            return null;
        }
    }

    private _addLineAndStockDetails() {
        const lineIndex = this.lines.value.length;
        const selectedUnit = this._packingUnits.find(unit => unit.node.packingUnit.name === this.receiptUnit.value);
        this.lines.addOrUpdateRecordValue(
            {
                _id: `${lineIndex}`,
                lineNumber: Number(Math.floor(Math.random() * MAX_INT_32)),
                product: {
                    code: this._productSite.product.code,
                    _id: this._productSite.product._id,
                },
                purchaseOrder: this.purchaseOrder.value ?? undefined,
                ...(this.productFromPurchaseOrder.value && {
                    purchaseOrderLineNumber: this.productFromPurchaseOrder.value.lineNumber,
                    purchaseOrderLine: {
                        _id: `${this.purchaseOrder.value?.id}|${this.productFromPurchaseOrder.value.lineNumber}|${this.productFromPurchaseOrder.value.sequenceNumber}`,
                        purchaseOrder: this.purchaseOrder.value ?? undefined,
                        lineNumber: this.productFromPurchaseOrder.value.lineNumber,
                        sequenceNumber: this.productFromPurchaseOrder.value.sequenceNumber,
                    },
                }),
                receiptSite: this.receiptSite.value ?? undefined,
                quantityInReceiptUnitReceived: this.quantityReceived.value?.toString(),
                receiptUnit: selectedUnit
                    ? {
                          code: selectedUnit.node.packingUnit.code,
                          numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                          _id: selectedUnit.node.packingUnit._id,
                      }
                    : this._productSite
                      ? this._productSite.product.stockUnit
                      : undefined,
                receiptUnitToPurchaseUnitConversionFactor:
                    this.packingUnitToStockUnitConversionFactor.value?.toString(),
                receiptUnitToStockUnitConversionFactor: this.packingUnitToStockUnitConversionFactor.value?.toString(),
                ...(this._selectedEntryTransaction?.defaultStockMovementGroup && {
                    stockMovementGroup: {
                        code: this._selectedEntryTransaction?.defaultStockMovementGroup.code,
                        _id: this._selectedEntryTransaction?.defaultStockMovementGroup._id,
                    },
                }),
                ...(this._purchaseShipmentLine && {
                    purchaseShipment: {
                        _id: this._purchaseShipmentLine.purchaseShipment._id,
                        id: this._purchaseShipmentLine.purchaseShipment.id,
                    },
                    purchaseShipmentLine: {
                        _id: this._purchaseShipmentLine._id,
                        lineNumber: this._purchaseShipmentLine.lineNumber,
                    },
                }),
                container: this.destinationContainer.value?.code,
                balance: this.closePoLine.value?.toString(),
                ...this.getCustomizationPurchaseReceiptLines(),
            },
            0,
            undefined,
        );
        this.lines.addOrUpdateRecordValue(
            {
                _id: `${lineIndex}`,
                stockSite: this.receiptSite.value
                    ? {
                          code: this.receiptSite.value.code,
                          _id: this.receiptSite.value._id,
                      }
                    : undefined,
                product: {
                    code: this._productSite.product.code,
                    _id: this._productSite.product._id,
                },
                effectiveDate: this.receiptDate.value ?? undefined,
                lot: this.lot.value ?? undefined,
                sublot: this.sublot.value ?? undefined,
                supplierLot: this.supplierLot.value ?? undefined,
                location: this.location.value
                    ? {
                          code: this.location.value.code,
                          _id: this.location.value._id,
                      }
                    : undefined,
                /*  warehouse: this.warehouseDestination.value
                    ? {
                          code: this.warehouseDestination.value.code,
                          _id: this.warehouseDestination.value._id,
                      }
                    : undefined, */
                serialNumber: this.serialNumber.value ?? undefined,
                lotCustomField1: this.lotCustomField1.value ?? undefined,
                lotCustomField2: this.lotCustomField2.value ?? undefined,
                lotCustomField3: this.lotCustomField3.value?.toString() ?? undefined,
                lotCustomField4: this.lotCustomField4.value ?? undefined,
                packingUnit: selectedUnit
                    ? {
                          code: selectedUnit.node.packingUnit.code,
                          numberOfDecimals: selectedUnit.node.packingUnit.numberOfDecimals,
                          _id: selectedUnit.node.packingUnit._id,
                      }
                    : this._productSite
                      ? this._productSite.product.stockUnit
                      : undefined,
                quantityInPackingUnit: this.quantityReceived.value?.toString(),
                packingUnitToStockUnitConversionFactor: this.packingUnitToStockUnitConversionFactor.value?.toString(),
                expirationDate: this.expirationDate?.value ?? undefined,
                status: this.status.value
                    ? {
                          code: this.status.value.toString(),
                          _id: this.status.value.toString(),
                      }
                    : undefined,
                potency: this.potency.value?.toString(),
                identifier1: this.identifier1.value?.toString(),
                identifier2: this.identifier2.value?.toString(),
                useByDate: this.useByDate.value?.toString(),
                majorVersion: this.majorVersion.value
                    ? {
                          code: this.majorVersion.value.code,
                          _id: this.majorVersion.value._id,
                      }
                    : undefined,
                minorVersion: this.minorVersion.value?.minorVersion,
                licensePlateNumber: this.licensePlateNumber.value
                    ? {
                          code: this.licensePlateNumber.value.code,
                          _id: this.licensePlateNumber.value._id,
                          container: {
                              code: this.licensePlateNumber.value.code,
                              _id: this.licensePlateNumber.value._id,
                          },
                      }
                    : undefined,
                closePoLine: this.closePoLine.value?.toString(),
                containerDetail: this.destinationContainer.value?.code,
                remainingQuantityInStockUnit: this.destinationRemainingQuantityInStockUnit.value ?? 0,
                ...this.getCustomizationPurchaseReceiptDetails(),
            },
            1,
            `${lineIndex}`,
        );
    }

    private readonly _onChange_serialNumber: AsyncVoidFunction = async () => {};

    /** @internal */
    private readonly _onChange_lot: AsyncVoidFunction = async () => {
        this.lot.value ? (this.lot.value = this.lot.value.toUpperCase()) : '';
        if (this.lot.value && this._purchaseStockManagementRules?.lotEntry === 'newLot') {
            if (
                !(await controlLotReceipt(
                    this,
                    this.lot.value,
                    this._productSite.product.code,
                    '6',
                    this.receiptSite.value?.code ?? '',
                ))
            ) {
                this.lot.value = null;
                this.lot.focus();
            }
        }

        /**
         * This bloc initialize current batch / lot during scanning,
         * any change after invalidate the entire control data block
         * */
        if (this.controlManager.isDispatchInProgress) {
            this._batchLot = this.lot.value;
        } else if (this._batchLot !== this.lot.value) {
            this._resetCompositeDataControl();
        }

        if (this.sublot.value && !this.sublot.isHidden && this.svgLot.value !== this.lot.value) {
            this.sublot.value = null;
        }

        await this._getLotValues();

        this.svgLot.value = this.lot.value;

        if (this.lot.value) {
            this._getNextField(this.lot)?.focus();
        }
    };

    /** @internal */
    private readonly _onChange_expirationDate: AsyncVoidFunction = async () => {
        /**
         * This bloc initialize current batch / lot during scanning,
         * any change after invalidate the entire control data block
         * */
        if (this.controlManager.isDispatchInProgress) {
            this._expiryDate = this.expirationDate.value;
        } else if (this._expiryDate !== this.expirationDate.value) {
            this._resetCompositeDataControl();
        }

        this.useByDate.value = useByDateDefaultValue(
            this.expirationDate.value ?? '',
            this._effectiveDate,
            this._productSite.product.expirationManagementMode,
            Number(this._productSite.product.useByDateCoefficient),
        );
        this.useByDate.maxDate = this.expirationDate.value ?? undefined;
    };

    /**
     * Reinitialize all control data block
     */
    /** @internal */
    private _resetCompositeDataControl(): void {
        this._batchLot = null;
        this._expiryDate = null;
    }

    private _getNextField(field: any) {
        let _nextField = field.getNextField(true);
        if (this._productSite.product.expirationManagementMode !== 'manualEntry') {
            while (_nextField === this.expirationDate || _nextField === this.useByDate) {
                _nextField = _nextField.getNextField(true);
            }
        }
        return _nextField;
    }

    /**
     * This asynchronous readonly function check if current composite data is valid or not
     * @param dictionaryDataComposite : dictionary of composite data block.
     * @param scannedFieldId : scanned field id.
     * @returns false when data must be discarded
     */
    /** @internal */
    private readonly _checkCompositeDataAllowed: AsyncCompositeAllowed = async (
        dictionaryDataComposite: DictionaryDataGs1 | DictionaryDataGeneric,
        scannedFieldId?: string,
    ) => {
        const _section = this._getActiveSection();
        const _productCode = <string | undefined>this._productSite?.product?.code;
        if (_productCode) {
            if ([Section.destination, Section.receipt].includes(_section)) {
                switch (this.typeOfServiceBarcode) {
                    case <TypeOfServiceBarcode>'gs1':
                        const _productGs1 = dictionaryDataComposite[DataTitle.gtin]?.data;
                        const _globalTradeItemNumber = <string | undefined>(
                            this._productSite?.product?.globalTradeItemNumber
                        );

                        if (
                            (!_productGs1 && Object.keys(dictionaryDataComposite).length) ||
                            (_productGs1 && _globalTradeItemNumber !== _productGs1)
                        ) {
                            await this.$.sound.error();
                            await dialogMessage(
                                this,
                                'error',
                                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                                ui.localize(
                                    '@sage/x3-purchasing/dialog-error-gs-1-data-is-not-related-to-product',
                                    'The GS1 data is not related to the product: {{ product }}.',
                                    { product: _productCode },
                                ),
                                {
                                    fullScreen: false,
                                },
                            );
                            return false;
                        }
                        break;

                    case <TypeOfServiceBarcode>'composite':
                        const _genericDictionary = <DictionaryDataGeneric>(<unknown>dictionaryDataComposite);
                        const _productComposite =
                            _genericDictionary[
                                !this.purchaseOrder.value
                                    ? PurchaseReceiptDictionaryFieldKey.product
                                    : PurchaseReceiptDictionaryFieldKey.productFromPurchaseOrder
                            ]?.data;

                        // We check that we have not provided a GTIN, UPC or other code to identify the product...
                        if (_productComposite && _productCode !== _productComposite) {
                            try {
                                if (
                                    (!this.purchaseOrder.value &&
                                        (await this.product.fetchSuggestions(<string>_productComposite))[0]?.code ===
                                            _productCode) ||
                                    (this.purchaseOrder.value &&
                                        (
                                            await this.productFromPurchaseOrder.fetchSuggestions(
                                                <string>_productComposite,
                                            )
                                        )[0]?.product === _productCode)
                                ) {
                                    break;
                                }
                            } catch (error) {}
                        }
                        if (
                            (!_productComposite && Object.keys(dictionaryDataComposite).length) ||
                            (_productComposite && _productCode !== _productComposite)
                        ) {
                            await this.$.sound.error();
                            await dialogMessage(
                                this,
                                'error',
                                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                                ui.localize(
                                    '@sage/x3-purchasing/dialog-error-composite-data-is-not-related-to-product',
                                    'The composite data is not related to the product: {{ product }}.',
                                    { product: _productCode },
                                ),
                                {
                                    fullScreen: false,
                                },
                            );
                            return false;
                        }
                        break;

                    default:
                        break;
                }
            }
        }

        return true;
    };

    private async _getLots(productCode: string, lot: string, sublot: string | null): Promise<ExtractEdges<Lot>[]> {
        try {
            return extractEdges(
                await this.$.graph
                    .node('@sage/x3-stock-data/Lot')
                    .query(
                        ui.queryUtils.edgesSelector(
                            {
                                _id: true,
                                code: true,
                                product: {
                                    code: true,
                                    _id: true,
                                },
                                sublot: true,
                                supplierLot: true,
                                expirationDate: true,
                                lotCustomField1: true,
                                lotCustomField2: true,
                                lotCustomField3: true,
                                lotCustomField4: true,
                                majorVersion: {
                                    code: true,
                                    _id: true,
                                },
                                minorVersion: true,
                                potency: true,
                                useByDate: true,
                            },
                            {
                                filter: {
                                    product: { code: productCode },
                                    code: lot,
                                    ...(sublot && { sublot: sublot }),
                                },
                            },
                        ),
                    )
                    .execute(),
            ) as ExtractEdges<Lot>[];
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                ui.localize('@sage/x3-purchasing/dialog-error-set-expdate', 'Error while reading the lot information') +
                    String(e),
            );
            return [];
        }
    }

    private async _getLotValues(): Promise<void> {
        if (this._productSite.product && this.lot.value && (this.sublot.isHidden || this.sublot.value)) {
            const result = await this._getLots(this._productSite.product.code, this.lot.value, this.sublot.value);
            if (result.length !== 0) {
                await this._setLotValues(result);
            } else {
                await this._initLotValues();
            }
        } else {
            await this._initLotValues();
        }
    }

    private async _setLotValues(lotValues: ExtractEdges<Lot>[]): Promise<void> {
        if (this._productSite.product.expirationManagementMode !== 'notManaged') {
            this.expirationDate.value = lotValues[0].expirationDate;
        }
        this.useByDate.value = lotValues[0].useByDate;
        this.supplierLot.value = lotValues[0].supplierLot;
        this.previousSupplierLot.value = lotValues[0].supplierLot;
        this.lotCustomField1.value = lotValues[0].lotCustomField1;
        this.lotCustomField2.value = lotValues[0].lotCustomField2;
        this.lotCustomField3.value = Number(lotValues[0].lotCustomField3);
        this.lotCustomField4.value = lotValues[0].lotCustomField4;
        this.majorVersion.value = null;
        this.minorVersion.value = null;
        this.majorVersion.value = {
            code: lotValues[0].majorVersion ? lotValues[0].majorVersion.code : undefined,
        };
        this.minorVersion.value = {
            minorVersion: lotValues[0].minorVersion,
        };
        this.potency.value = Number(lotValues[0].potency);
        this.supplierLot.isDisabled = true;
        this.lotCustomField1.isDisabled = true;
        this.lotCustomField2.isDisabled = true;
        this.lotCustomField3.isDisabled = true;
        this.lotCustomField4.isDisabled = true;
        this.majorVersion.isDisabled = true;
        this.minorVersion.isDisabled = true;
        this.potency.isDisabled = true;
        this.$.commitValueAndPropertyChanges();
    }

    private async _initLotValues(): Promise<void> {
        /* const expirationManagement = this._productSite.product.expirationManagementMode;
          if (this.isExpirationManaged) {
            this.expirationDate.value = this.useByDate.value = null;
        } */

        /**
         * When dispatching has finished, lot and expiry date must have data :
         * in this case, our may using saved date.
         * Otherwise, user have make some change OR composite information have missing,
         * and all controls data must be cleared before to continue with standard value.
         */
        if (!this.controlManager.isDispatchInProgress) {
            if (this._batchLot !== null && this._expiryDate !== null) {
                this.expirationDate.value = this._expiryDate;
            } else {
                this._resetCompositeDataControl();
            }
        }

        // Initialize only expiration date when is not received before
        if (!this.expirationDate.value) {
            this.expirationDate.value = expirationDateDefaultValue(
                this._productSite.product.expirationManagementMode,
                this._productSite.product.expirationLeadTime,
                this._productSite.product.expirationTimeUnit,
                this._effectiveDate,
            );
        }

        this.useByDate.value = useByDateDefaultValue(
            this.expirationDate.value,
            this._effectiveDate,
            this._productSite.product.expirationManagementMode,
            Number(this._productSite.product.useByDateCoefficient),
        );

        this.useByDate.maxDate = this.expirationDate.value ?? undefined;

        this.lotCustomField1.value = null;
        this.lotCustomField2.value = null;
        this.lotCustomField3.value = null;
        this.lotCustomField4.value = null;
        this.potency.value = Number(this._productSite.product.defaultPotencyInPercentage);
        this.supplierLot.isDisabled = false;
        this.lotCustomField1.isDisabled = false;
        this.lotCustomField2.isDisabled = false;
        this.lotCustomField3.isDisabled = false;
        this.lotCustomField4.isDisabled = false;
        this.majorVersion.isDisabled = false;
        this.minorVersion.isDisabled = !this.majorVersion.value;
        this.potency.isDisabled = false;
    }

    private async _validate(): Promise<boolean> {
        const fieldErrors = new Array<string>();
        if (await this._isFieldInvalid(this.receiptUnit)) fieldErrors.push(this.receiptUnit.title ?? '');
        if (await this._isFieldInvalid(this.quantityReceived)) fieldErrors.push(this.quantityReceived.title ?? '');
        if (await this._isFieldInvalid(this.status)) fieldErrors.push(this.status.title ?? '');
        if (await this._isFieldInvalid(this.licensePlateNumber)) fieldErrors.push(this.licensePlateNumber.title ?? '');
        if (await this._isFieldInvalid(this.location)) fieldErrors.push(this.location.title ?? '');
        if (await this._isFieldInvalid(this.serialNumber)) fieldErrors.push(this.serialNumber.title ?? '');
        if (await this._isFieldInvalid(this.supplierLot)) fieldErrors.push(this.supplierLot.title ?? '');
        if (await this._isFieldInvalid(this.lot)) fieldErrors.push(this.lot.title ?? '');
        if (await this._isFieldInvalid(this.sublot)) fieldErrors.push(this.sublot.title ?? '');
        if (await this._isFieldInvalid(this.majorVersion)) fieldErrors.push(this.majorVersion.title ?? '');
        if (await this._isFieldInvalid(this.minorVersion)) fieldErrors.push(this.minorVersion.title ?? '');
        if (await this._isFieldInvalid(this.potency)) fieldErrors.push(this.potency.title ?? '');
        if (await this._isFieldInvalid(this.expirationDate)) fieldErrors.push(this.expirationDate.title ?? '');
        if (await this._isFieldInvalid(this.useByDate)) fieldErrors.push(this.useByDate.title ?? '');
        if (await this._isFieldInvalid(this.calculatedExpirationDate))
            fieldErrors.push(this.calculatedExpirationDate.title ?? '');
        if (await this._isFieldInvalid(this.lotCustomField1)) fieldErrors.push(this.lotCustomField1.title ?? '');
        if (await this._isFieldInvalid(this.lotCustomField2)) fieldErrors.push(this.lotCustomField2.title ?? '');
        if (await this._isFieldInvalid(this.lotCustomField3)) fieldErrors.push(this.lotCustomField3.title ?? '');
        if (await this._isFieldInvalid(this.lotCustomField4)) fieldErrors.push(this.lotCustomField4.title ?? '');
        if (await this._isFieldInvalid(this.identifier1)) fieldErrors.push(this.identifier1.title ?? '');
        if (await this._isFieldInvalid(this.identifier2)) fieldErrors.push(this.identifier2.title ?? '');
        if (await this._isFieldInvalid(this.closePoLine)) fieldErrors.push(this.closePoLine.title ?? '');
        if (await this._isFieldInvalid(this.destinationContainer))
            fieldErrors.push(this.destinationContainer.title ?? '');
        if (fieldErrors.length > 0) {
            this.$.removeToasts();
            this.$.showToast(
                ui.localize(
                    '@sage/x3-purchasing/pages__utils__notification__invalid_entry_error',
                    `Please check your entry for {{#each fieldNames}}\n - {{this}}{{/each}}`,
                    { fieldNames: fieldErrors },
                ),
                { type: 'error', timeout: 5000 },
            );
            await this.$.page.validate();
        }
        return !fieldErrors.length;
    }

    private async _isFieldInvalid(
        field:
            | ui.fields.FilterSelect
            | ui.fields.DropdownList
            | ui.fields.Text
            | ui.fields.Reference
            | ui.fields.Numeric
            | ui.fields.Select
            | ui.fields.Date,
    ): Promise<boolean> {
        return !field.isHidden && ((field.isMandatory && !field.value) || !!(await field.validate()));
    }

    private async _isEmptyLpnAndContainerValues(): Promise<boolean> {
        if (
            !this.destinationContainer.isHidden &&
            !this.licensePlateNumber.isHidden &&
            (!this.destinationContainer.value || this.destinationContainer.value?.code === '') &&
            (!this.licensePlateNumber.value || this.licensePlateNumber.value?.code === '')
        ) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-purchasing/pages__purchase_receipt__notification__container_license_plate_number_mandatory_error',
                    'Container or license plate number is mandatory',
                ),
            );
            return true;
        } else {
            return false;
        }
    }

    private async _isZeroQuantityValue(): Promise<boolean> {
        if (!this.quantityReceived.value || this.quantityReceived.value === 0) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing/dialog-error-title', 'Error'),
                ui.localize(
                    '@sage/x3-purchasing/pages__purchase_receipt__notification__quantityReceived__isZero',
                    'Quantity is mandatory',
                ),
            );
            return true;
        } else {
            return false;
        }
    }

    private async _getMinorVersions(majorVersion: string): Promise<ExtractEdges<ProductVersion>[]> {
        return extractEdges(
            await this.$.graph
                .node('@sage/x3-master-data/ProductVersion')
                .query(
                    ui.queryUtils.edgesSelector(
                        {
                            _id: true,
                            minorVersion: true,
                        },
                        {
                            filter: {
                                product: { code: this._productSite.product.code },
                                majorVersion: majorVersion,
                                type: { _eq: 'stock' },
                                useStatus: { _eq: 'availableToUse' },
                            },
                            orderBy: {
                                minorVersion: 1,
                            },
                        },
                    ),
                )
                .execute(),
        ) as ExtractEdges<ProductVersion>[];
    }

    private async _getStockStatus(): Promise<string[]> {
        const selectedStatus: { _regex: string }[] = [];
        this._purchaseStockManagementRules.authorizedSubstatus.split(',').forEach(function (status) {
            selectedStatus.push({ _regex: getRegExp(status).source });
        });
        const response = await this.$.graph
            .node('@sage/x3-stock-data/StockStatus')
            .query(
                ui.queryUtils.edgesSelector(
                    {
                        _id: true,
                        code: true,
                    },
                    {
                        filter: {
                            code: { _or: selectedStatus },
                        },
                    },
                ),
            )
            .execute();

        if (!response.edges || response.edges.length === 0) {
            throw new Error(
                ui.localize(
                    '@sage/x3-purchasing/pages__purchase_receipt__notification__invalid_stock_status_error',
                    'No stock status',
                ),
            );
        }

        return response.edges.map((stockStatus: any) => {
            return stockStatus.node.code;
        });
    }

    private async _getDefaultLocation(codeLocation: string) {
        if (!this.receiptSite.value) {
            return undefined;
        }
        try {
            const response = await this.$.graph
                .node('@sage/x3-stock-data/Location')
                .read(
                    {
                        code: true,
                        _id: true,
                        stockSite: {
                            code: true,
                            _id: true,
                        },
                    },
                    `${this.receiptSite.value.code}|${codeLocation}`,
                )
                .execute();
            return response;
        } catch (e) {
            await dialogMessage(
                this,
                'error',
                ui.localize('@sage/x3-purchasing/error-loading-location', 'Error loading location'),
                String(e),
            );
        }
    }

    private calculateRemainingQuantity(remainingQuantity: number[]) {
        if (!this._purchaseOrderLine) {
            return;
        }
        // calculate the conversion rate from stock unit to order unit
        // TODO: Unfortunately there appears to be cases in which toFixed() doesn't round correctly
        // TODO: Verify if there is a better way of calculating conversion rate. This is based on 4GL ENVADC.src
        let conversionRatePurchaseUnitToStockUnit = 0;
        let conversionRateOrderUnitToStockUnit = 0;
        let pendingQuantityInStockUnitReceived = 0;
        let quantityInStockUnitExpected = 0;
        let remainingQuantityInStockUnit = 0;
        let remainingQuantityInOrderUnit = 0;
        let remainingQuantityInPurchaseUnit = 0;

        if (this._purchaseShipmentLine) {
            conversionRatePurchaseUnitToStockUnit = Number.parseFloat(
                (
                    Number(this._purchaseShipmentLine.quantityInStockUnitShipped) /
                    Number(this._purchaseShipmentLine.quantityInPurchaseUnitShipped)
                ).toFixed(QTY_DECIMAL_PLACES),
            );
            conversionRateOrderUnitToStockUnit = Number.parseFloat(
                (
                    Number(this._purchaseOrderLine.quantityInStockUnitOrdered) /
                    Number(this._purchaseOrderLine.quantityInOrderUnitOrdered)
                ).toFixed(QTY_DECIMAL_PLACES),
            ); // using Number.parseFloat() to remove any trailing decimal zeros due to toFixed()
        } else {
            conversionRateOrderUnitToStockUnit = Number.parseFloat(
                (
                    Number(this._purchaseOrderLine.quantityInStockUnitOrdered) /
                    Number(this._purchaseOrderLine.quantityInOrderUnitOrdered)
                ).toFixed(QTY_DECIMAL_PLACES),
            ); // using Number.parseFloat() to remove any trailing decimal zeros due to toFixed()
            conversionRatePurchaseUnitToStockUnit = Number.parseFloat(
                (
                    Number(this._purchaseOrderLine.quantityInStockUnitOrdered) /
                    Number(this._purchaseOrderLine.quantityInPurchaseUnitOrdered)
                ).toFixed(QTY_DECIMAL_PLACES),
            ); // using Number.parseFloat() to remove any trailing decimal zeros due to toFixed()
        }

        try {
            if (this.lines.value.length > 0) {
                // iterate through receipt lines user has added so far for the same PO Line & product as this selected PO line
                if (this._purchaseShipmentLine?.purchaseShipment) {
                    pendingQuantityInStockUnitReceived = this.calculCurrentQuantityInStockUnitReceived();
                } else {
                    this.lines.value
                        .filter(line => line.purchaseOrderLine?._id === this._purchaseOrderLine?._id)
                        .map(
                            line =>
                                (pendingQuantityInStockUnitReceived +=
                                    Number(line.quantityInReceiptUnitReceived) *
                                    Number(line.receiptUnitToStockUnitConversionFactor)),
                        );
                }
            }
        } catch (error) {
            ui.console.warn(`Error loading saved purchase receipt session data: ${error}`);
        }
        // quantity is initially calculated based on remaining stock quantity and then converted to order unit
        if (this._purchaseShipmentLine) {
            quantityInStockUnitExpected =
                Number(this._purchaseShipmentLine.quantityInPurchaseUnitExpected) *
                conversionRatePurchaseUnitToStockUnit;
            remainingQuantityInStockUnit = Math.max(
                0,
                Number(quantityInStockUnitExpected) - pendingQuantityInStockUnitReceived,
            );
            if (conversionRateOrderUnitToStockUnit !== 0) {
                remainingQuantityInOrderUnit = remainingQuantityInStockUnit / conversionRateOrderUnitToStockUnit;
            }
            if (conversionRatePurchaseUnitToStockUnit !== 0) {
                const tmp = Math.pow(10, this._purchaseOrderLine.purchaseUnit.numberOfDecimals);
                if (tmp !== 0) {
                    remainingQuantityInPurchaseUnit =
                        Math.round((remainingQuantityInStockUnit / conversionRatePurchaseUnitToStockUnit) * tmp) / tmp;
                } else {
                    remainingQuantityInPurchaseUnit =
                        remainingQuantityInStockUnit / conversionRatePurchaseUnitToStockUnit;
                }
            }
        } else {
            remainingQuantityInStockUnit = Math.max(
                0,
                Number(this._purchaseOrderLine.quantityInStockUnitOrdered) -
                    Number(this._purchaseOrderLine.quantityInStockUnitReceived) -
                    pendingQuantityInStockUnitReceived,
            );
            if (conversionRateOrderUnitToStockUnit !== 0) {
                remainingQuantityInOrderUnit = remainingQuantityInStockUnit / conversionRateOrderUnitToStockUnit;
            }
            if (conversionRatePurchaseUnitToStockUnit !== 0) {
                const tmp = Math.pow(10, this._purchaseOrderLine.purchaseUnit.numberOfDecimals);
                if (tmp !== 0) {
                    remainingQuantityInPurchaseUnit =
                        Math.round((remainingQuantityInStockUnit / conversionRatePurchaseUnitToStockUnit) * tmp) / tmp;
                } else {
                    remainingQuantityInPurchaseUnit =
                        remainingQuantityInStockUnit / conversionRatePurchaseUnitToStockUnit;
                }
            }
        }
        remainingQuantity[0] = remainingQuantityInStockUnit;
        remainingQuantity[1] = remainingQuantityInOrderUnit;
        remainingQuantity[2] = remainingQuantityInPurchaseUnit;
    }

    private calculCurrentQuantityInStockUnitReceived(): number {
        let currentQuantityInStockUnitReceived = 0;
        this.lines.value
            ?.filter(line => line.purchaseShipmentLine?._id === this._purchaseShipmentLine?._id)
            .map(
                line =>
                    (currentQuantityInStockUnitReceived +=
                        Number(line.quantityInReceiptUnitReceived) *
                        Number(line.receiptUnitToStockUnitConversionFactor)),
            );
        return currentQuantityInStockUnitReceived;
    }

    private async getSidebarData(rowId: string): Promise<boolean> {
        const _purchaseReceiptLine = this.lines.getRecordValue(rowId, 0);
        const _stockJournal = this.lines.getRecordValue(rowId, 1);
        if (_purchaseReceiptLine && _stockJournal) {
            const _productCode = _purchaseReceiptLine.product?.code;
            if (_productCode && this.receiptSite.value) {
                this._productSite = await this.getProductSite(_productCode, this.receiptSite.value?.code ?? '');
            }
            this._purchaseStockManagementRules = await findStockManagementRules(
                this.receiptSite.value?.code ?? '',
                this._productSite.product.productCategory.code,
                '3',
                this._selectedEntryTransaction?.stockMovementCode?.code ?? null,
                this,
            );
            this._initPackingUnitFields();
            this.receiptUnit.options = this._packingUnits.map(packingUnit => packingUnit.node.packingUnit.name);

            this.previousReceiptUnit.value =
                _stockJournal.packingUnit && _stockJournal.packingUnit.code !== this._productSite.product.stockUnit.code
                    ? formatUnitForDisplay(
                          _stockJournal.packingUnit.code ?? '',
                          _stockJournal.packingUnitToStockUnitConversionFactor ?? '',
                          this._productSite.product.stockUnit.code,
                      )
                    : this._productSite.product.stockUnit.code;
            this.previousReceiptUnitCode.value =
                _stockJournal.packingUnit && _stockJournal.packingUnit.code
                    ? _stockJournal.packingUnit.code
                    : this._productSite.product.stockUnit.code;
            this._purchaseShipmentLine =
                _purchaseReceiptLine.purchaseShipment && _purchaseReceiptLine.purchaseShipmentLine
                    ? await this.getPurchaseShipmentLine(
                          _purchaseReceiptLine.purchaseShipment.id ?? '',
                          Number(_purchaseReceiptLine.purchaseShipmentLine.lineNumber),
                      )
                    : undefined;
            this._purchaseOrderLine =
                _purchaseReceiptLine.purchaseOrder && _purchaseReceiptLine.purchaseOrderLine
                    ? await this.getPurchaseOrderLine(
                          _purchaseReceiptLine.purchaseOrder.id ?? '',
                          Number(_purchaseReceiptLine.purchaseOrderLine.lineNumber),
                          Number(_purchaseReceiptLine.purchaseOrderLine.sequenceNumber),
                      )
                    : undefined;
            this._remainingQuantityInitialInStockUnit = this._purchaseShipmentLine
                ? Number(this._purchaseShipmentLine.quantityInStockUnitShipped)
                : this._purchaseOrderLine
                  ? Number(this._purchaseOrderLine.quantityInStockUnitOrdered)
                  : 0;
            this._remainingQuantityInitialInOrderUnit = this._purchaseShipmentLine
                ? Number(this._purchaseShipmentLine.quantityInOrderUnitShipped)
                : this._purchaseOrderLine
                  ? Number(this._purchaseOrderLine.quantityInOrderUnitOrdered)
                  : 0;

            const remainingQuantity: number[] = [0, 0, 0];
            this.calculateRemainingQuantity(remainingQuantity);
            const tmp = Math.pow(10, _stockJournal.packingUnit?.numberOfDecimals ?? 1);
            let quantity: number =
                Number(_stockJournal.quantityInPackingUnit) *
                Number(_stockJournal.packingUnitToStockUnitConversionFactor);
            if (tmp !== 0) {
                quantity = Math.round(quantity * tmp) / tmp;
            }
            _stockJournal.remainingQuantityInStockUnit = remainingQuantity[0] + quantity;
            this.lines.setRecordValue(_stockJournal, 1);

            this._quantityToReceiveNoRounded = Number(_stockJournal.quantityInPackingUnit);
            this._quantityToReceiveRounded = Number(_stockJournal.quantityInPackingUnit);
            this.previousQuantityToReceive.value = Number(_stockJournal.quantityInPackingUnit);
            this.previousSupplierLot.value = _stockJournal.supplierLot ?? null;
            this._lots =
                _stockJournal.product && _stockJournal.product.code && _stockJournal.lot
                    ? await this._getLots(_stockJournal.product.code, _stockJournal.lot, _stockJournal.sublot ?? null)
                    : [];
            this._locationDetail = _stockJournal.location;
            this._containerDetail = _stockJournal.containerDetail;
            return true;
        } else {
            return false;
        }
    }

    private _getSidebarPurchaseReceiptLine(): ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>[] = [];
        this.customizationPurchaseReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<PurchaseReceiptLineBinding>);
            }
        });
        return _FieldsSidebar;
    }

    private _getSidebarStockJournal(): ui.SidebarFieldDefinition<StockJournalClosePoLine>[] {
        const _FieldsSidebar: ui.SidebarFieldDefinition<StockJournalClosePoLine>[] = [];
        _FieldsSidebar.push('packingUnit');
        _FieldsSidebar.push('quantityInPackingUnit');
        if ((this.receiptUnit.options ?? []).length > 1) {
            _FieldsSidebar.push('packingUnitToStockUnitConversionFactor');
        }
        _FieldsSidebar.push('status');
        if (
            this._purchaseStockManagementRules.licensePlateNumberEntry === 'mandatory' &&
            this._productSite.isLicensePlateNumberManaged
        ) {
            _FieldsSidebar.push('containerDetail');
        }
        if (this._productSite.isLicensePlateNumberManaged) {
            _FieldsSidebar.push('licensePlateNumber');
        }
        if (this._productSite.isLocationManaged) {
            _FieldsSidebar.push('location');
        }
        if (!['notManaged', 'issued'].includes(this._productSite.product.serialNumberManagementMode)) {
            _FieldsSidebar.push('serialNumber');
        }
        if (this._productSite.product.lotManagementMode !== 'notManaged') {
            _FieldsSidebar.push('supplierLot');
        }
        if (this._productSite.product.lotManagementMode !== 'notManaged') {
            _FieldsSidebar.push('lot');
        }
        if (this._productSite.product.lotManagementMode === 'lotAndSublot') {
            _FieldsSidebar.push('sublot');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._productSite.product.stockVersionMode !== 'no'
        ) {
            _FieldsSidebar.push('majorVersion');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._productSite.product.stockVersionMode === 'majorAndMinor'
        ) {
            _FieldsSidebar.push('minorVersion');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotPotencyAllowed &&
            this._productSite.product.stockManagementMode === 'potencyManaged'
        ) {
            _FieldsSidebar.push('potency');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            (this._selectedEntryTransaction?.isLotExpirationDateAllowed ||
                this._productSite.product.expirationManagementMode === 'manualEntry' ||
                this._productSite.product.expirationManagementMode === 'mandatoryEntry')
        ) {
            _FieldsSidebar.push('expirationDate');
            _FieldsSidebar.push('useByDate');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField1Allowed
        ) {
            _FieldsSidebar.push('lotCustomField1');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField2Allowed
        ) {
            _FieldsSidebar.push('lotCustomField2');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField3Allowed
        ) {
            _FieldsSidebar.push('lotCustomField3');
        }
        if (
            this._productSite.product.lotManagementMode !== 'notManaged' &&
            this._selectedEntryTransaction?.isLotCustomField4Allowed
        ) {
            _FieldsSidebar.push('lotCustomField4');
        }
        if (
            this._selectedEntryTransaction?.identifier1Detail === 'entered' ||
            this._selectedEntryTransaction?.identifier1Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier1');
        }
        if (
            this._selectedEntryTransaction?.identifier2Detail === 'entered' ||
            this._selectedEntryTransaction?.identifier2Detail === 'displayed'
        ) {
            _FieldsSidebar.push('identifier2');
        }
        _FieldsSidebar.push('closePoLine');
        this.customizationPurchaseReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                _FieldsSidebar.push(item.$bind as ui.SidebarFieldDefinition<StockJournalClosePoLine>);
            }
        });

        return _FieldsSidebar;
    }

    private getCustomizationPurchaseReceiptLines() {
        let response = {};
        this.customizationPurchaseReceiptLinesBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    private getCustomizationPurchaseReceiptDetails() {
        let response = {};
        this.customizationPurchaseReceiptDetailsBlock.layout.$layout?.$items.forEach(item => {
            if (item.$bind) {
                response = {
                    ...response,
                    ...{
                        [item.$bind]: (this as any)[item.$bind].value,
                    },
                };
            }
        });
        return response;
    }

    /**
     *
     * Technical screen management methods
     *
     */

    /**
     *  @returns the active section
     */
    private _getActiveSection(): Section {
        // Section.receipt is never returned (virtual section)
        if (!this.destinationBlock.isHidden) {
            return Section.destination;
        } else if (!this.productBlock.isHidden) {
            return Section.main;
        } else if (!this.shipmentBlock.isHidden) {
            return Section.shipment;
        } else {
            return Section.main;
        }
    }

    /**
     *
     * @param newSection
     * @returns  the old active section
     */
    private async _setActiveSection(newSection: Section, isMainReset?: Boolean): Promise<Section> {
        const _previousSection = this._getActiveSection();
        let _newSection = newSection;
        // validate the new section
        switch (_newSection) {
            case Section.main:
                if (isMainReset) {
                    if (!this.purchaseOrder.value && !this.lines.value.length) {
                        this.supplier.value = null;
                    }
                    await this._clearPage();
                    if (_previousSection !== Section.main) {
                        await this._initializePage();
                    }
                }
                this.product.value = null;
                this.productFromPurchaseOrder.value = null;
                break;

            case Section.destination:
                await this._initializePage();
                this.quantityReceived.isHidden = false;
                this.customizationPurchaseReceiptLinesBlock.isHidden =
                    !this.customizationPurchaseReceiptLinesBlock.layout.$layout?.$items.length;
                this.customizationPurchaseReceiptDetailsBlock.isHidden =
                    !this.customizationPurchaseReceiptDetailsBlock.layout.$layout?.$items.length;
                break;

            case Section.receipt:
                await this._initializePage();
                this.product.value = null;
                this.productFromPurchaseOrder.value = null;
                this.supplier.isDisabled = true;
                break;

            case Section.shipment:
                this.customizationPurchaseReceiptLinesBlock.isHidden = true;
                this.customizationPurchaseReceiptDetailsBlock.isHidden = true;
                await this._getShipments();
                break;

            default:
                _newSection = Section.main;
                await this._clearPage();
                await this._initializePage();
                this.product.value = null;
                this.productFromPurchaseOrder.value = null;
                break;
        }

        // hide/show buttons
        this.createButton.isHidden = ![Section.main, Section.destination, Section.receipt].includes(_newSection);
        this.addButton.isHidden = _newSection !== Section.destination;

        this.createButton.isDisabled = this.lines.value.length === 0 || this.createButton.isHidden;

        /**
         * To avoid display effects, we prefer to fill in the fields before showing the sections.
         */

        this.productBlock.isHidden = _newSection === Section.shipment;

        this.destinationBlock.isHidden = _newSection !== Section.destination;

        this.customizationPurchaseReceiptHeaderBlock.isHidden =
            this.customizationPurchaseReceiptHeaderBlock.layout.$layout?.$items.length === 0;

        this.shipmentBlock.isHidden = _newSection !== Section.shipment;

        this.receiptBlock.isHidden = _newSection === Section.shipment || !this.lines.value?.length;

        // Action to do after change
        switch (_newSection) {
            case Section.main:
                if (!this.purchaseOrder.value) {
                    this.productFromPurchaseOrder.value = null;
                    this.productFromPurchaseOrder.isHidden = true;
                    this.productFromPurchaseOrder.isDisabled = true;
                    this.supplier.isDisabled = !this._directReceiptAllowed;
                    this.product.isHidden = false;

                    if (!this.supplier.value) {
                        this.product.isDisabled = true;
                        this.supplier.focus();
                    } else {
                        this.product.isDisabled = false;
                        this.product.focus();
                    }
                } else {
                    this.supplier.value = this.purchaseOrder.value?.orderFromSupplier;
                    this.supplier.isDisabled = true;
                    this.product.isHidden = true;
                    this.product.isDisabled = true;
                    this.productFromPurchaseOrder.isHidden = false;
                    this.productFromPurchaseOrder.isDisabled = false;
                    this.productFromPurchaseOrder.focus();
                }
                break;
            case Section.destination:
                this.supplier.isDisabled = true;
                break;
            case Section.shipment:
                break;
            case Section.receipt:
                break;

            default:
                break;
        }

        await this.$.commitValueAndPropertyChanges();

        return _previousSection;
    }

    /**
     * @returns true if the active section is the destination section
     */
    private _isMandatoryDestination(): boolean {
        return this._getActiveSection() === Section.main;
    }
}
